/*
 * calculator.l
 * To compile: make calculator
 *  yacc -d calculator.y
 *  flex calculator.l
 *  cc -x c++ -o calculator y.tab.c lex.yy.c -ll -lm
 * NOTE: variables cannot exceed 7 letters
 */
%{
#include "y.tab.h"
#include <math.h>
#include <string.h>
extern int lineno;
struct symtab* symlook(char *s);
%}
%%
([0-9]+|([0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?) {
     yylval.dval = atof(yytext);
     return NUMBER;
     }
[ \t\n{}]                       /* ignore whitespace     */
">=" { return GE; }
"<=" { return LE; }
"==" { return EQ; }
if   { return IF; }
then { return THEN; }
else { return ELSE; }
"&&" { return LOGICAL_AND; }
"||" { return LOGICAL_OR; }
"+=" { return PLUS_EQUALS; }
"-=" { return MINUS_EQUALS; }
"*=" { return TIMES_EQUALS; }
"/=" { return DIVIDE_EQUALS; }
"++" { return INCREMENT; }
"--" { return DECREMENT; }
"image" { return IMAGE; }
"red" { return RED; }
"green" { return GREEN; }
"blue" { return BLUE; }
"real" { return YYREAL; }
"imag" { return YYIMAG; }
"wave" { return YYWAVE; }
[A-Za-z][A-Za-z0-9]* {        /* return symbol pointer */
    struct symtab *sp = symlook(yytext);
    yylval.symp = sp;
    return NAME;
    }
\"[^\"]*\" { 
    struct symtab *sp = symlook(yytext);
    yylval.symp = sp;
    return YYSTRING; }
"$" {return 0; }              /* end of input          */
;  { lineno++; return yytext[0]; }/* end of line       */
.  return yytext[0];          /* anything else         */
%%
