/*=================================================================================  
 * calculator.y
 * Latest revision: 09-19-1998                                              
 * Copyright (C) 1998 by Thomas J. Nelson                                  
 * All rights reserved.                                                     
 * 
 * Each rule must start in column 1
 * To compile: 
 *     rm y*.o l*.o
 *     make calculator
 *     yacc -d calculator.y     Bison doesn't work - make sure autoconf doesn't use it. 
 *     flex calculator.l        Must be run manually after changing calculator.l
 *     cc -o calculator y.tab.c lex.yy.c -ll -lm
 *=================================================================================
 * After running flex, it is necessary to edit lex.yy.c:
 *     remove static from    static char yy_hold_char;              (line 211)
 *     remove static from    static char yy_n_chars;                (line 213)
 *     remove static from    static int yyinput YY_PROTO(( void )); (line 455)
 *     remove static from    static int input YY_PROTO(( void ));   (line 457)
 *     remove static from    static int yyinput()                   (line 1201)
 *     remove static from    static int input()                     (line 1203)
 *     At the macro starting at line 513, i.e. from:
 *
 *          #ifndef YY_INPUT
 *          #define YY_INPUT(buf,result,max_size) \
 *          . . .
 *          #endif
 * 
 *    the entire macro must be replaced with:
 *
 *          #ifndef YY_INPUT
 *          #define YY_INPUT(buf,result,max_size) \
 *              strncpy(buf,text,max_size); \
 *              result = strlen(buf);
 *          #endif
 *
 *    Add these lines at the top of the file after line 20:
 *          extern char *text;
 *          int n; 
 *          int input(void);
 *          struct symtab* symlook(char *s);
 *    Also, remove any exit()'s from line 1566 in the function:
 *        static void yy_fatal_error( yyconst char msg[] )
 *          exit( YY_EXIT_FAILURE );
 * 
 *
 * If the following messages don't appear, the compilation didn't work:
 *        y.tab.c: In function `int yyparse()':
 *        y.tab.c:774: warning: suggest parentheses around assignment used as truth value
 *        y.tab.c:822: warning: label `yyerrlab' defined but not used
 *        y.tab.c:817: warning: label `yynewerror' defined but not used
 *        y.tab.c: At top level:
 *        y.tab.c:2: warning: `char yysccsid[36]' defined but not used
 * 
 *
 *=================================================================================
 * For false condition, characters are thrown away until desired stop point:
 *          while(yy_hold_char != ';') yyinput();
 *          yyinput();
 *     For true condition, set if_true to 1.
 *
 * yy_n_chars = no of chars in current evaluated string          
 * yyleng = no of chars in current token                         
 * yytext = current token                                        
 * Use input() for C, yyinput() for C++                          
 *   throws away next token & prepares for more input             
 * 
 * These 3 lines would throw away the rest of the current string 
 *   which is not what we want. (The yy_buffer_state typedef struct must also
 *   be included for the following line to work.)
 *          len = yy_n_chars - strlen(yy_current_buffer->yy_ch_buf) - 1;
 *          yy_c_buf_p += len;                                          
 *          yyinput();                                                    
 *
 * yy_hold_char is current character being input.
 *
 * The input string must end with \n\255 to parse correctly.  
 *================================================================================
 */

%{
#ifndef DIGITAL
#include<stdio.h>
#include<string.h>
#include<stdlib.h> 
#include<math.h>
#include "calculator.h"
#include "xmtnimage.h"
#endif

void addfunc(char *name, double(*func)(double));
void addfunc2(char *name, double(*func)(double,double));
void addfunc3(char *name, double(*func)(char *));
double drand(double a); 
void skip_to_end(void);
double dmax(double a, double b);
double dmin(double a, double b);
struct symtab* symlook(char *s);
int input(void);
int yyinput(void);
#ifndef MIPS
extern int yylex(void);
#endif
void yyerror(char *s);
void yyrestart ( FILE *input_file );

extern Globals     g;
extern Images     *z;
extern int         ci;
extern char yy_hold_char, yy_n_chars;
extern int ii,rr,gg,bb; 
extern char *text;
extern char *yytext;

extern char *yy_c_buf_p;
extern int yy_init;		/* whether we need to initialize */
extern int yy_start;	/* start state number */


int if_true=0;
int rr,gg,bb,v,x,y,frame,bpp,ino,len;
int lineno;
int position=0;
static char tempstring[128];
%}

%union {  double dval;
          struct symtab *symp;
       }

/*  tokens  */ 
%token <symp> NAME
%token <dval> NUMBER
%token <symp> YYSTRING
%token IF THEN ELSE
%token EQ LE GE
%token PLUS_EQUALS,MINUS_EQUALS,TIMES_EQUALS,DIVIDE_EQUALS
%token LOGICAL_AND,LOGICAL_OR,INCREMENT,DECREMENT
%token IMAGE RED GREEN BLUE YYREAL YYIMAG YYWAVE

/* left-associative operatore - Also assigns precedence, last %left statement 
    has the highest precedence.
*/

%left LOGICAL_AND LOGICAL_OR
%left '&' '|' '^' 
%left '<' '>' EQ GE LE
%left '-' '+'
%left '*' '/' 
%left INCREMENT DECREMENT
%nonassoc UMINUS '!' '~'
%type <dval> expression

%%

statement_list: 
        statement 
        |   statement_list statement
        |   '{' statement_list '}' statement
        ;
        
statement: 
        NAME '=' expression { $1->value = $3; } ';'
        |   NAME PLUS_EQUALS expression { $1->value += $3; } ';'
        |   NAME MINUS_EQUALS expression { $1->value -= $3; }';'
        |   NAME TIMES_EQUALS expression { $1->value *= $3; }';'
        |   NAME DIVIDE_EQUALS expression
                        { if ($3 == 0.0)
                              yyerror("Division by zero");
                          else
                              $1->value /= $3;
                        }  ';'
        |   NAME INCREMENT { $1->value++; } ';'
        |   NAME DECREMENT { $1->value--; } ';'
        |   expression   { add_variable("temp", $1); } ';'
        |   '\n'
        |   IF '(' expression ')'
                {
                   if( $3 == 1)  if_true=1;
                               
                   else        {  if_true=0;
                                  skip_to_end();
                               }
                }
        |   ELSE
                {  
                   if(if_true) skip_to_end();
                } 
                
        /* Repeat above statement definitions terminated with \n */
                
        |   NAME '=' expression { $1->value = $3; } '\n'
        |   NAME PLUS_EQUALS expression { $1->value += $3; } '\n'
        |   NAME MINUS_EQUALS expression { $1->value -= $3; }'\n'
        |   NAME TIMES_EQUALS expression { $1->value *= $3; }'\n'
        |   NAME DIVIDE_EQUALS expression
                        { if ($3 == 0.0)
                              yyerror("Division by zero");
                          else
                              $1->value /= $3;
                        }  '\n'
        |   NAME INCREMENT { $1->value++; } '\n'
        |   NAME DECREMENT { $1->value--; } '\n'
        |   expression   { add_variable("temp", $1); } '\n'
        ;




expression: 
        expression '+' expression { $$ = $1 + $3; }
        |   expression '-' expression { $$ = $1 - $3; }
        |   expression '*' expression { $$ = $1 * $3; }
        |   expression '/' expression 
                        { if ($3 == 0.0)
                              yyerror("Division by zero");
                          else
                              $$ = $1 / $3; 
                        }
        |   '-' expression %prec UMINUS { $$ = -$2; }
        |   '!' expression %prec '!' { $$ = !$2; }
        |   '~' expression %prec '~' { $$ = (double)(~(int)$2); }
        |   '(' expression ')'          { $$ = $2; }
        |   expression '<' expression { $$ = $1 < $3; }
        |   expression '>' expression { $$ = $1 > $3; }
        |   expression LE expression { $$ = $1 <= $3; }
        |   expression GE expression { $$ = $1 >= $3; }
        |   expression EQ expression { $$ = $1 == $3; }
        |   expression LOGICAL_AND expression { $$ = $1 && $3; }
        |   expression LOGICAL_OR expression { $$ = $1 || $3; }
        |   expression '&' expression { $$ = (double)((int)$1 & (int)$3); }
        |   expression '|' expression { $$ = (double)((int)$1 | (int)$3); }
        |   expression '^' expression { $$ = (double)((int)$1 ^ (int)$3); }
        |   NUMBER                   
        |   NAME {  $$ = $1->value; }
        |   YYSTRING  
                 {   len = strlen($1->name)-2;
                     strncpy(tempstring,$1->name+1,len);
                     tempstring[len]=0;
                     $$ = double_image_number(tempstring);
                 }
        |   NAME '(' expression ')' 
                 {  
                    if($1->funcptr1)
                        $$ = ($1->funcptr1)($3);
                    else
                    {   sprintf(tempstring, "%s is not a 1-argument function\n", $1->name);
                        yyerror(tempstring);
                        $$ = 0.0;
                    }
                 }
        |   NAME '(' expression ',' expression ')' 
                 {  
                    if($1->funcptr2)
                        $$ = ($1->funcptr2)($3,$5);
                    else
                    {   sprintf(tempstring, "%s is not a 2-argument function\n", $1->name);
                        yyerror(tempstring);
                        $$ = 0.0;
                    }
                 }
        |   IMAGE '[' expression ']' '['expression ']' '['expression ']' '['expression ']'
                {    ino   = (int) $3;
                     frame = (int) $6;
                     x     = (int) $9;
                     y     = (int) $12;
                     if(between(ino, 0, g.image_count) &&
                        between(frame, 0, z[ino].frames-1) &&
                        between(x, 0, z[ino].xsize-1) &&
                        between(y, 0, z[ino].ysize-1))
                     {  bpp   = z[ino].bpp;
                        $$ = (double)(pixelat(z[ino].image[frame][y] + x*g.off[bpp], bpp)); 
                     }else yyerror("Array out of bounds");
                }
        |   RED '[' expression ']' '['expression ']' '['expression ']' '['expression ']'
                {    ino   = (int) $3;
                     frame = (int) $6;
                     x     = (int) $9;
                     y     = (int) $12;
                     if(between(ino, 0, g.image_count) &&
                        between(frame, 0, z[ino].frames-1) &&
                        between(x, 0, z[ino].xsize-1) &&
                        between(y, 0, z[ino].ysize-1))
                     {  bpp   = z[ino].bpp;
                        v = pixelat(z[ino].image[frame][y] + x*g.off[bpp], bpp); 
                        valuetoRGB(v,rr,gg,bb,bpp);
                        $$ = (double)rr;
                     }else yyerror("Array out of bounds");
                }
        |   GREEN '[' expression ']' '['expression ']' '['expression ']' '['expression ']'
                {    ino   = (int) $3;
                     frame = (int) $6;
                     x     = (int) $9;
                     y     = (int) $12;
                     if(between(ino, 0, g.image_count) &&
                        between(frame, 0, z[ino].frames-1) &&
                        between(x, 0, z[ino].xsize-1) &&
                        between(y, 0, z[ino].ysize-1))
                     {  bpp   = z[ino].bpp;
                        v = pixelat(z[ino].image[frame][y] + x*g.off[bpp], bpp); 
                        valuetoRGB(v,rr,gg,bb,bpp);
                        $$ = (double)gg;
                     }else yyerror("Array out of bounds");
                }
        |   BLUE '[' expression ']' '['expression ']' '['expression ']' '['expression ']'
                {    ino   = (int) $3;
                     frame = (int) $6;
                     x     = (int) $9;
                     y     = (int) $12;
                     if(between(ino, 0, g.image_count) &&
                        between(frame, 0, z[ino].frames-1) &&
                        between(x, 0, z[ino].xsize-1) &&
                        between(y, 0, z[ino].ysize-1))
                     {  bpp   = z[ino].bpp;
                        v = pixelat(z[ino].image[frame][y] + x*g.off[bpp], bpp); 
                        valuetoRGB(v,rr,gg,bb,bpp);
                        $$ = (double)bb;
                     }else yyerror("Array out of bounds");
                }
        |   YYREAL '[' expression ']' '['expression ']' '['expression ']'
                {    ino   = (int) $3;
                     x     = (int) $6;
                     y     = (int) $9;
                     if(between(ino, 0, g.image_count-1) &&
                        between(x, 0, z[ino].xsize-1) &&
                        between(y, 0, z[ino].ysize-1) &&
                        z[ino].floatexists)
                          $$ = z[ino].fft[y][x].real();
                     else yyerror("Array out of bounds");
                }
        |   YYIMAG '[' expression ']' '['expression ']' '['expression ']'
                {    ino   = (int) $3;
                     x     = (int) $6;
                     y     = (int) $9;
                     if(between(ino, 0, g.image_count-1) &&
                        between(x, 0, z[ino].xsize-1) &&
                        between(y, 0, z[ino].ysize-1) &&
                        z[ino].floatexists)
                          $$ = z[ino].fft[y][x].imag();
                     else yyerror("Array out of bounds");
                }
        |   YYWAVE '[' expression ']' '['expression ']' '['expression ']'
                {    ino   = (int) $3;
                     x     = (int) $6;
                     y     = (int) $9;
                     if(between(ino, 0, g.image_count-1) &&
                        between(x, 0, z[ino].xsize-1) &&
                        between(y, 0, z[ino].ysize-1) &&
                        z[ino].waveletexists)
                          $$ = z[ino].wavelet[y][x];
                     else yyerror("Array out of bounds");
                }
        ;
%%


void initialize_evaluator(void)
{   
   
   /* 1-parameter functions - new functions can be added here */
   addfunc("abs",  fabs); 
   addfunc("acos", acos); 
   addfunc("asin", asin); 
   addfunc("atan", atan); 
   addfunc("cos",  cos); 
   addfunc("cosh", cosh); 
   addfunc("exp",  exp);
   addfunc("log",  log); 
   addfunc("log10",log10); 
   addfunc("rand", drand); 
   addfunc("sin",  sin); 
   addfunc("sinh", sinh); 
   addfunc("sqrt", sqrt);
   addfunc("tan",  tan); 
   addfunc("tanh", tanh); 
   addfunc("erf",  erf); 

#ifdef HAVE_ACOSH
   addfunc("acosh",acosh); 
#endif
#ifdef HAVE_CBRT
   addfunc("cbrt", cbrt); 
#endif

   /* 2-parameter functions - new functions can be added here */
   addfunc2("pow", pow); 
   addfunc2("min", dmin); 
   addfunc2("max", dmax); 

   /* Functions that take a string as argument */
   /* To use this, replace the YYSTRING entry with the following:
        |   NAME '(' YYSTRING ')' 
                 {
                    if($1->funcptr3)
                    {   len = strlen($3->name)-2;
                        strncpy(tempstring,$3->name+1,len);
                        tempstring[len]=0;
                        $$ = ($1->funcptr3)(tempstring);
                    }else
                    {   sprintf(tempstring, "Argument to %s is not a string\n", $1->name);
                        yyerror(tempstring);
                        $$ = 0.0;
                    }                    
     Can't have both because it causes a shift/reduce conflict.
   */
   /* addfunc3("ino", double_image_number); */

   lineno=1;
   yy_n_chars = 0;
}

void yyerror(char *s)
{  
   formula_error(s); 
   yyrestart(NULL);
}

void add_variable(char *name, double value)
{
   struct symtab *sp = symlook(name);
   sp->funcptr1 = NULL;
   sp->funcptr2 = NULL;
   sp->funcptr3 = NULL;
   sp->value = value;
}

void addfunc(char *name, double(*func)(double))
{
   struct symtab *sp = symlook(name);
   sp->funcptr1 = func;
   sp->funcptr2 = NULL;
   sp->funcptr3 = NULL;
}

void addfunc2(char *name, double(*func)(double, double))
{
   struct symtab *sp = symlook(name);
   sp->funcptr1 = NULL;
   sp->funcptr2 = func;
   sp->funcptr3 = NULL;
}

void addfunc3(char *name, double(*func)(char *))
{
   struct symtab *sp = symlook(name);
   sp->funcptr1 = NULL;
   sp->funcptr2 = NULL;
   sp->funcptr3 = func;
}


double read_variable(char *name)
{
   struct symtab *sp = symlook(name);
   return sp->value;
}


struct symtab* symlook(char *s)
{
   struct symtab *sp;
   for(sp=symtab; sp< &symtab[NSYMS]; sp++)
   {  
       if(sp->name && !strcmp(sp->name, s)) return sp;  /* is it already here */
       if(!sp->name){ sp->name=strdup(s); return sp; }  /* is it free         */
   }
   yyerror("Too many symbols, ask a wizard to enlarge me");
   return NULL;
}



void skip_to_end(void)
{
   while(strchr("\n ",yy_hold_char)!=NULL) yyinput(); 
   int got_left_brace = yy_hold_char=='{';
   if(got_left_brace)
   {   while(yy_hold_char != '}') yyinput();
   }else
   {   while(strchr(";}\n",yy_hold_char)==NULL) yyinput(); 
   }
   yyinput();
   while(yy_hold_char=='\n') yyinput(); 
}

double dmax(double a, double b){ if(a>b) return a; return b; }
double dmin(double a, double b){ if(a<b) return a; return b; }
double drand(double a){ a=a; return rand(); }
