//--------------------------------------------------------------------------//
// xmtnimage14.cc                                                           //
// Filtering routines                                                       //
// Latest revision: 05-22-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
//--------------------------------------------------------------------------//
 
#include "xmtnimage.h"


extern int         ci;                 // current image
extern Images     *z;
extern Globals     g;

//--------------------------------------------------------------------------//
// filter                                                                   //
// 2-d filter the image                                                     //
// Uses a separate routine for 8-bit mode vs color modes.                   //
//--------------------------------------------------------------------------//
void filter(int noofargs, char **arg)
{  
  static int filter_type=1;
  static int ksize=3;  
  static int range=7,kmult=1;
  static int max_bkg = 0;
  static int sharpfac=50;

  if(memorylessthan(16384)){ message(g.nomemory,ERROR); return; } // minimal memory
  int knumber=1;
  int entireimage=0;
  
  //-------------------------------------------------------------------------//
  //   Filter type              Center value    Surround                     //
  //                             3   5   9  15                               //
  //   -----------------------  --------------  --------                     //
  //     0 low pass                    1           1                         //
  //     1 high pass              ksize*ksize     -1                         //
  //     2 laplace               -ksize*ksize      1                         //
  //     3 background subtract        n/a         n/a                        //
  //     4 median filter noise filter n/a         n/a                        //
  //   >=5 arbitrary                  n/a        arbitrary                   //
  //-------------------------------------------------------------------------//
  //                                                                         //
  //  User defined filters                                                   //
  //  To add a new filter:    (1) increase 1st dimension of arrays three,    //
  //                              five, nine, and fifteen                    //
  //                          (2) add new element in menu                    //
  //                          (3) increase value of dialog.radiono[0]        //
  //                          (4) add filters below - make sure total value  //
  //                                of elements in array is 0, otherwise     //
  //                                filter will also make image darker or    //
  //                                lighter.                                 //
  //                          (5) Increase value of `nooffilters'.           //
  //                                                                         //
  //        1st dim.= filter type - 5                                        //
  //                                        // Filters                       //
  //-------------------------------------------------------------------------//

 if(noofargs==0)
     get_filter_parameters(filter_type, ksize, range, kmult, max_bkg, sharpfac);
 else
 {  
      if(noofargs>0) filter_type = atoi(arg[1]);
      if(noofargs>1) knumber     = atoi(arg[2]);
      if(knumber==1) ksize=3; 
      if(knumber==2) ksize=5; 
      if(knumber==3) ksize=9; 
      if(knumber==4) ksize=15; 
      if(noofargs>2) sharpfac    = atoi(arg[3]);
      if(noofargs>3) range       = atoi(arg[4]);
      if(noofargs>4) kmult       = atoi(arg[5]);
      if(noofargs>5) max_bkg     = atoi(arg[6]);
 }
 drawselectbox(OFF);
 g.state = BUSY;
 printstatus(g.state);
 if(g.selectedimage>-1 && ci>-1) entireimage=1; else entireimage=0;
 if(ci>=0 && z[ci].is_zoomed) entireimage=0;

 if(g.getout==0)
 switch(filter_type)
 {
     case 13: 
        remove_low_frequencies(ksize, kmult, g.ulx, g.uly, g.lrx, g.lry);
        break;
     default:
        filter_region(g.ulx, g.uly, g.lrx, g.lry, filter_type, 
            ksize, sharpfac, range, kmult, max_bkg, entireimage, 1);
 }
 g.getout=0;
 g.state = NORMAL;
}


//--------------------------------------------------------------------------//
// get_filter_parameters                                                    //
//--------------------------------------------------------------------------//
void get_filter_parameters(int &filter_type, int &ksize, int &range, int &kmult, 
       int &max_bkg, int &sharpfac)
{
   Dialog dialog;
   dialog.needrepeat=0; 
   int j,k;
   int nooffilters=14;   // No.of different filters in menu

   strcpy(dialog.title,"Filter");
   dialog.boxmenutitle[1] = new char[40];
   dialog.boxmenu[1] = new char*[nooffilters];
   for(k=0;k<nooffilters;k++) dialog.boxmenu[1][k] = new char[40];
   dialog.boxmenuselection[1]=filter_type;
   strcpy(dialog.boxmenutitle[1],"Filter Type");
   strcpy(dialog.boxmenu[1][0],"Low-pass");
   strcpy(dialog.boxmenu[1][1],"High-pass");
   strcpy(dialog.boxmenu[1][2],"Laplace edge enhan.");
   strcpy(dialog.boxmenu[1][3],"Background subtract");
   strcpy(dialog.boxmenu[1][4],"Background flatten ");
   strcpy(dialog.boxmenu[1][5],"Noise(median) filter");
   strcpy(dialog.boxmenu[1][6],"Sharpen (/)");
   strcpy(dialog.boxmenu[1][7],"Sharpen (|)");
   strcpy(dialog.boxmenu[1][8],"Sharpen (-)");
   strcpy(dialog.boxmenu[1][9],"Sharpen (\\)");
   strcpy(dialog.boxmenu[1][10],"Edge detect (-)");
   strcpy(dialog.boxmenu[1][11],"Edge detect (|)");
   strcpy(dialog.boxmenu[1][12],"Sobel edge detect.");
   strcpy(dialog.boxmenu[1][13],"Remove low frequencies");

   strcpy(dialog.radio[0][0],"Kernel");
   strcpy(dialog.radio[0][1]," 3 x  3");
   strcpy(dialog.radio[0][2]," 5 x  5");
   strcpy(dialog.radio[0][3]," 9 x  9");
   strcpy(dialog.radio[0][4],"15 x 15");
 
   if(ksize==3) dialog.radioset[0]=1;
   if(ksize==5) dialog.radioset[0]=2;
   if(ksize==9) dialog.radioset[0]=3;
   if(ksize==15) dialog.radioset[0]=4;

   strcpy(dialog.radio[1][0],"Background level");
   strcpy(dialog.radio[1][1],"Black");
   strcpy(dialog.radio[1][2],"White");

   if(max_bkg==0) dialog.radioset[1]=1;
   else  dialog.radioset[1]=2;

   strcpy(dialog.boxes[0],"Parameters");
   strcpy(dialog.boxes[1],"Filter type");
   strcpy(dialog.boxes[2],"Range(noise filter)");
   strcpy(dialog.boxes[3],"Kernel multiplier");
   strcpy(dialog.boxes[4],"Amount of filtering");
   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=NON_EDIT_LIST;
   dialog.boxtype[2]=INTCLICKBOX;
   dialog.boxtype[3]=INTCLICKBOX;
   dialog.boxtype[4]=INTCLICKBOX;
 
   dialog.boxmin[1]=1; dialog.boxmax[1]=nooffilters;
   dialog.boxmin[2]=1; dialog.boxmax[2]=200;
   dialog.boxmin[3]=1; dialog.boxmax[3]=50;
   dialog.boxmin[4]=1; dialog.boxmax[4]=100;
 
   sprintf(dialog.answer[2],"%d",range);
   sprintf(dialog.answer[3],"%d",kmult);
   sprintf(dialog.answer[4],"%d",sharpfac);
 
   dialog.radiono[0]=5;
   dialog.radiono[1]=3;
   dialog.radiono[2]=0;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
 
   dialog.noofradios=2;
   dialog.noofboxes=5;
   dialog.helptopic=7;
   dialog.keep=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   dialog.transient = 1;
   strcpy(dialog.path,".");
   g.getout=0;
 
   ////-------------////
   dialogbox(dialog);
   ////-------------////

   if(dialog.radioset[0]==1) ksize=3; 
   if(dialog.radioset[0]==2) ksize=5; 
   if(dialog.radioset[0]==3) ksize=9; 
   if(dialog.radioset[0]==4) ksize=15;
 
   for(k=0;k<nooffilters;k++) delete[] dialog.boxmenu[1][k];
   delete[] dialog.boxmenu[1];
   delete[] dialog.boxmenutitle[1];
 
   filter_type = dialog.boxmenuselection[1];
   if(dialog.radioset[1]==1) max_bkg=0; else max_bkg=1;
   range = atoi(dialog.answer[2]);
   kmult = atoi(dialog.answer[3]); if(kmult<1)kmult=1;
   sharpfac = atoi(dialog.answer[4]);
  
   if((filter_type<=2)&&(kmult>1))
   {    if(g.want_messages > 1) message("Invalid parameters - \nSetting kernel multiplier to 1",WARNING);
        kmult=1;
   }
}


//--------------------------------------------------------------------------//
// filter_region                                                            //
//--------------------------------------------------------------------------//
int filter_region(int x1, int y1, int x2, int y2, 
         int filter_type, int ksize, int sharpfac, 
         int range, int kmult, int max_bkg, int entireimage, int want_progress_bar)
{
  Widget www, scrollbar;
  int status=OK,f,h=0,bad=0,i,j,k,l,a,index,xindex,ii,jj,asource,
      dx,dy,k2,l2,v,hit=0, skip=0,rr,gg,bb,noscreen=0,
      count,maxbpp,bestvalue=0,sum,rsum,gsum,bsum,
      rvalue=0,gvalue=0,bvalue=0,bpp,obpp=0,
      centerfac=1,bufindex,iii,jjj,lll,convert24=0,
      s1,s2,s3,s4,kk,ll,      // For Sobel
      value=0,value2,v1,v2,v3,v4,rv1=0,rv2=0,rv3=0,rv4=0,xsize,
      gv1=0,gv2=0,gv3=0,gv4=0,bv1=0,bv2=0,bv3=0,bv4=0,
      orr,ogg,obb,
      rv,rs1,rs2,rs3,rs4,gv,gs1,gs2,gs3,gs4,bv,bs1,bs2,bs3,bs4;
 
  int *cpix;

  double average=0,fx,fy,dv,dvx,dvy,fx1,fy1=0,fy2=0,rfy1=0,gfy1=0,bfy1=0,
         rdv,rdvx,rdvy,gdv,gdvx,gdvy,bdv,bdvx,bdvy,v21=0,v43=0,v31=0,v42=0,
         shold, shnew;

  const int three[6][3][3] = {
                             { {   0, -1,  1 },
                               {  -1,  1,  1 },
                               {  -1,  1,  0 }  },

                             { {  -1,  0,  1 },
                               {  -1,  1,  1 },
                               {  -1,  0,  1 }  },

                             { {  -1, -1, -1 },
                               {   0,  1,  0 },
                               {   1,  1,  1 }  },

                             { {   1,  1,  0 },
                               {   1,  1, -1 },
                               {   0, -1, -1 }  },

                             { {  -1, -1, -1 },
                               {   1,  1,  1 },
                               {   0,  0,  0 }  },

                             { {  -1,  1,  0 },
                               {  -1,  1,  0 },
                               {  -1,  1,  0 }  } };

                                                            
  const int five[6][5][5] = {{ {   0,  0, -1, -1,  1 },
                               {   0, -1, -1,  1,  1 },
                               {  -1, -1,  1,  1,  1 },
                               {  -1, -1,  1,  1,  0 },
                               {  -1,  1,  1,  0,  0 }  },

                             { {  -1, -1,  0,  1,  1 },
                               {  -1, -1,  0,  1,  1 },
                               {  -1, -1,  1,  1,  1 },
                               {  -1, -1,  0,  1,  1 },
                               {  -1, -1,  0,  1,  1 }  },

                             { {  -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1 },
                               {   1,  1,  1, -1, -1 },
                               {   1,  1,  1,  1,  1 },
                               {   1,  1,  1,  1,  1 }  },

                             { {   1,  1,  1,  0,  0 },
                               {   1,  1,  1,  1,  0 },
                               {  -1, -1, -1,  1,  1 },
                               {   0, -1, -1, -1,  1 },
                               {   0,  0, -1, -1, -1 }  },

                             { {  -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1 },
                               {   2,  2,  2,  2,  2 },
                               {  -0, -0, -0, -0, -0 },
                               {  -0, -0, -0, -0, -0 }  },
 
                             { {  -1, -1,  2,  0,  0 },
                               {  -1, -1,  2,  0,  0 },
                               {  -1, -1,  2,  0,  0 },
                               {  -1, -1,  2,  0,  0 },
                               {  -1, -1,  2,  0,  0 }  } };
  

  const int nine[6][9][9] ={ { {   0,  0,  0, -1, -1, -1, -1, -1,  1 },
                               {   0,  0, -1, -1, -1, -1, -1,  1,  1 },
                               {   0, -1, -1, -1, -1, -1,  1,  1,  1 },
                               {  -1, -1, -1, -1, -1,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  1,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  1,  1,  1,  1,  1 },
                               {  -1, -1, -1,  1,  1,  1,  1,  1,  0 },
                               {  -1, -1,  1,  1,  1,  1,  1,  0,  0 },
                               {  -1,  1,  1,  1,  1,  1,  0,  0,  0 }  },

                             { {  -1, -1, -1, -1, -0,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1, -0,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1, -0,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1, -0,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  1,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  0,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  0,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  0,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  0,  1,  1,  1,  1 }  },

                             { {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {   0,  0,  0,  0,  1,  0,  0,  0,  0 },
                               {   1,  1,  1,  1,  1,  1,  1,  1,  1 },
                               {   1,  1,  1,  1,  1,  1,  1,  1,  1 },
                               {   1,  1,  1,  1,  1,  1,  1,  1,  1 },
                               {   1,  1,  1,  1,  1,  1,  1,  1,  1 }  },

                             { {  -1,  1,  1,  1,  1,  1,  0,  0,  0 },
                               {  -1, -1,  1,  1,  1,  1,  1,  0,  0 },
                               {  -1, -1, -1,  1,  1,  1,  1,  1,  0 },
                               {  -1, -1, -1, -1,  1,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1,  1,  1,  1,  1,  1 },
                               {  -1, -1, -1, -1, -1,  1,  1,  1,  1 },
                               {   0, -1, -1, -1, -1, -1,  1,  1,  1 },
                               {   0,  0, -1, -1, -1, -1, -1,  1,  1 },
                               {   0,  0,  0, -1, -1, -1, -1, -1,  1 }  },

                             { {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {  -1, -1, -1, -1, -1, -1, -1, -1, -1 },
                               {   4,  4,  4,  4,  4,  4,  4,  4,  4 },
                               {   0,  0,  0,  0,  0,  0,  0,  0,  0 },
                               {   0,  0,  0,  0,  0,  0,  0,  0,  0 },
                               {   0,  0,  0,  0,  0,  0,  0,  0,  0 },
                               {   0,  0,  0,  0,  0,  0,  0,  0,  0 }  },

                             { {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 },
                               {  -1, -1, -1, -1,  4,  0,  0,  0,  0 }  } };
                                      


  const int fifteen[6][15][15] = {

        { {    0,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {    0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1 },
          {    0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1 },
          {    0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1 },
          {    0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0 },
          {   -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0 },
          {   -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0 },
          {   -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0 }  },

        { {   -1, -1, -1, -1, -1, -1, -1,  0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1,  0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1,  0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1,  0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1,  0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1,  0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -0,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -0,  1,  1,  1,  1,  1,  1,  1 }  },

        { {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {    0,  0,  0,  0,  0,  0, -1,  2, -1,  0,  0,  0,  0,  0,  0 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 }  },


        { {    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0 },
          {   -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0 },
          {   -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0 },
          {   -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0 },
          {   -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0 },
          {   -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1,  1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1,  1 },
          {    0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1,  1 },
          {    0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1,  1 },
          {    0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1,  1 },
          {    0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  1 },
          {    0,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 }  },

        { {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
          {    7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7 },
          {    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
          {    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
          {    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
          {    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
          {    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
          {    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
          {    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 }  },

        { {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 },
          {   -1, -1, -1, -1, -1, -1, -1,  7,  0,  0,  0,  0,  0,  0,  0 }  } };


  shnew = ((double)sharpfac)/100.0;      // Fraction of new pixel to use
  shold = 1.0-shnew;                     // Fraction of old pixel to keep

  a=ksize/2;            //              kernel     a
                        //                3        1 
                        //                5        2 
                        //                9        4
                        //               15        7
  asource=a*kmult;      // 1/2 the kernel size * kernel multiplier, used for
                        // gathering pixels farther away.
                        // asource is max. distance from the center pixel
                        // kmult is no. of source pixels to skip
  centerfac = ksize*ksize;   // Factor to multiply center pixel by (for 
                             // high-pass), or factor to divide surrounding
                             // summed pixels by (low-pass).
  count = ksize*ksize;
                             // Pick which nth lowest pixel to use if back-
                             // ground or median filtering
  if(filter_type==3) bestvalue = ksize/2; 
  if(filter_type==5) bestvalue = count/2;
  xsize = x2 - x1;
                 // Array for storing one line of filtered pixels
                 // Make it big enough for biggest possible image (32 bpp)
  array<int> saveline(xsize+3,18*kmult);      
  array<int> bkbuf(xsize+2,36);        
  for(k=0;k<=g.image_count;k++) z[k].hit=0;
                 // Array for storing each center rgb for 1 row of pixels
  cpix     = new int[xsize+2*asource+4];
                 // Array for storing each column of summed source pixels
  array<int> colbuf(xsize+2*asource+4,3); 
  array<int> sorted(xsize+2*asource+2,3);    
  if(!sorted.allocated) status=NOMEM;
  maxbpp = g.bitsperpixel;
                 // Filtering speed can be doubled if we know all the pixels
                 // are on the same image.
  if(ci>=0)
  {   if(z[ci].shell && entireimage) switchto(ci);
      if(z[ci].colortype==INDEXED) 
      {    status=change_image_depth(ci,24,0); 
           convert24=1; 
      }
      bpp=z[ci].bpp;
  }else bpp=g.bitsperpixel;

  if(status!=OK)
  {   message("Insufficient memory buffers,\naborting",ERROR);
      if(cpix    !=NULL) delete[] cpix;
      return -1;   
  }

  //// Start of filtering - This section tests each pixel to determine    
  //// whether the image at that point wants to be color. If so, the pixel
  //// is treated as color and its r,g,and b are filtered separately      
  //// then combined.                                                     
  //// The destination pixel (pixel being changed) is i,j.                
  //// Leave a boundary of asource pixels around the outside to avoid     
  //// boundary effects.                                                  
  //// For background flattening, get 4 corners for gradient              

  g.want_switching = 0;     //// Make sure they dont click around on other images
  if(filter_type==4)                       // background flattening
  {    dx=(x2-x1)/3; dy=(y2-y1)/3;         // Increments
       get_average_value(x1   ,y1   ,x1+dx,y1+dy,v1,rv1,gv1,bv1);
       get_average_value(x2-dx,y1   ,x2   ,y1+dy,v2,rv2,gv2,bv2);
       get_average_value(x1   ,y2-dy,x1+dx,y2   ,v3,rv3,gv3,bv3);
       get_average_value(x2-dx,y2-dy,x2   ,y2   ,v4,rv4,gv4,bv4);
       average = (v1+v2+v3+v4)/4;
       v21=(double)(v2-v1);
       v43=(double)(v4-v3);
       v31=(double)(v3-v1);
       v42=(double)(v4-v2);
  }  

  ii = ci;
  if(want_progress_bar) progress_bar_open(www, scrollbar);
  for(j=y1+asource; j<y2-asource; j++)        
  { 
    if(want_progress_bar) progress_bar_update(scrollbar, 100*(j-y1-asource)/(y2-y1-2*asource));
    if(keyhit()) if(getcharacter()==27){ skip = 1; break; }
    if(j>0)
       index = j%(asource+1); 
    else
       index = asource + (j - asource) % (asource+1);

    switch(filter_type)
    { case 0:       // low pass
      case 1:       // high pass
      case 2:       // laplace
      {
         for(i=x1; i<x2;i++)     
         {  hit=1;         
                    // ii is image no. at pixel i,j.
                    // This line takes 1/4 the processing time
                    // so speed it up by reading directly from image buffer.
            if(entireimage && ii>=0 && z[ii].bpp==8)
            {
               jjj = j-z[ii].ypos;
               iii = i-z[ii].xpos; 
               if(iii<0 || jjj<0)    // Sanity check
               {   fprintf(stderr,"Internal error at line %d in %s\n",__LINE__,__FILE__);
                   bad=1;
                   break;
               }
               if(jjj>=0 && jjj<z[ii].ysize && iii>=0 && iii<z[ii].xsize)
                   cpix[i-x1] = z[ii].image[z[ii].cf][jjj][iii];
            }else if(entireimage && ii>=0 && z[ii].bpp>8)
            { 
               bpp = z[ii].bpp;
               jjj = j-z[ii].ypos;
               iii = i-z[ii].xpos; 
               if(iii<0 || jjj<0)    // Sanity check
               {   fprintf(stderr,"Internal error at line %d in %s\n",__LINE__,__FILE__);
                   bad=1;
                   break;
               }
               if(jjj>=0 && jjj<z[ii].ysize && iii>=0 && iii<z[ii].xsize)
                   cpix[i-x1] = pixelat(z[ii].image[z[ii].cf][jjj]+iii*g.off[bpp],bpp); 
            }else
            {
               cpix[i-x1] = (int)readpixelonimage(i,j,bpp,ii);
            }
            if(bpp!=obpp) 
            { 
               if(ii>=0 && bpp==8)
               {  obpp=bpp; 
                  switch_palette(ii);
               }
            }   
            if(ii>=0){ z[ii].hit=1; noscreen=1; } else noscreen=0;
            if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
            maxbpp = max(bpp,maxbpp);
            if((bpp==8)||(ii>=0 && z[ii].hitgray))
            {   
                colbuf.p[0][i-x1]=0; // Collect colums of pixels
                   // Sum the values of all needed source pixels in each column.
                   // This loop takes about 40% the processing time

                if(entireimage)
                {   for(l=j-asource;l<=j+asource;l+=kmult)    
                    {  lll = l-z[ii].ypos;
                       iii = i-z[ii].xpos; 
                       if(lll>=0 && lll<z[ii].ysize && iii>=0 && iii<z[ii].xsize)
                           colbuf.p[0][i-x1] +=
                               pixelat(z[ii].image[z[ii].cf][lll] + iii*g.off[bpp], bpp); 
                    }
                }else
                { 
                     for(l=j-asource;l<=j+asource;l+=kmult)    
                        colbuf.p[0][i-x1]+=(int)readpixelonimage(i,l,bpp,ii);
                }
            }else
            {  
                colbuf.p[0][i-x1]=0; // Collect colums of pixels
                colbuf.p[1][i-x1]=0;
                colbuf.p[2][i-x1]=0;

                   // Sum the values of all needed source pixels in each column.
                   // This loop takes 40% of the processing time. If all
                   //  pixels are on the same image, we can speed it up 2x
                   //  by using pixelat().

                if(entireimage && ii>=0)
                {   bpp=z[ii].bpp;
                    for(l=j-asource;l<=j+asource;l+=kmult)    
                    {  lll = l-z[ii].ypos;
                       iii = i-z[ii].xpos; 
                       if(lll>=0 && lll<z[ii].ysize && iii>=0 && iii<z[ii].xsize)
                           value = pixelat(z[ii].image[z[ii].cf][lll]+iii*g.off[bpp],bpp); 
                       valuetoRGB(value,rr,gg,bb,bpp);
                       colbuf.p[0][i-x1] += rr;
                       colbuf.p[1][i-x1] += gg;
                       colbuf.p[2][i-x1] += bb;
                    }
                }else
                { 
                    for(l=j-asource;l<=j+asource;l+=kmult)    
                    {  value = (int)readpixelonimage(i,l,bpp,ii);
                       valuetoRGB(value,rr,gg,bb,bpp);
                       colbuf.p[0][i-x1] += rr;
                       colbuf.p[1][i-x1] += gg;
                       colbuf.p[2][i-x1] += bb;
                    }
                }  
            }
            
         } 
         if(bad) break;  
                   // Iterate through colbuf, add up the appropriate columns,
                   // and add the center pixel.
                   // This loop takes about 1/2 the processing time.
         for(i=x1+asource; i<x2-asource; i++)     
         { 
           ii=whichimage(i,j,bpp);
           if(ii>=0){ z[ii].hit=1; noscreen=1; } else noscreen=0;
           if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
           if(bpp!=obpp) 
           {  if(ii>=0 && bpp==8)
              {  obpp=bpp; 
                 switch_palette(ii);
              }
           }   
           maxbpp = max(bpp,maxbpp);

           if(bpp==8 || (ii>=0 && z[ii].hitgray))
           {  sum = 0;
              for(l=i-asource;l<=i+asource;l+=kmult)
                  sum += colbuf.p[0][l-x1];   
              switch(filter_type)
              {   case 0: value = sum / centerfac;            
                          value = (int)(shnew*value + shold*cpix[i-x1]); 
                          break;
                  case 1: value =-sum + centerfac*cpix[i-x1]; 
                          value = cpix[i-x1] + (int)(shnew*value);
                          break;
                  case 2: value = sum - centerfac*cpix[i-x1]; 
                          value = cpix[i-x1] + (int)(shnew*value);
                          break;
              }   

                   //  Next set the pixel to its new value 
              if(value<0) value=0;
              if(value>(int)g.maxvalue[bpp]) value=(int)g.maxvalue[bpp];
              xindex = i - (x1 + asource);

              if(j-asource > y1+asource)
              {    //  This speeds it up 2x here but it only speeds it up
                   //  for 8 bpp.
                if(entireimage && ii>=0 && bpp==8)
                {   jjj = j-asource-1-z[ii].ypos;
                    iii = i-z[ii].xpos;               
                    if(jjj>=0 && jjj<z[ii].ysize && iii>=0 && iii<z[ii].xsize)
                        z[ii].image[z[ii].cf][jjj][iii] = saveline.p[index][xindex];
                }else
                {  
                    if(j>y1+1+asource &&( g.selected_is_square || inside_irregular_region(i,j-asource-1)))
                    setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,
                        bpp,-1,noscreen);
                }
              }
              saveline.p[index][xindex]=value;    
           }else
           {  rsum = 0;
              gsum = 0;
              bsum = 0;
              for(l=i-asource;l<=i+asource;l+=kmult)
              {   rsum += colbuf.p[0][l-x1];
                  gsum += colbuf.p[1][l-x1];
                  bsum += colbuf.p[2][l-x1];
              }    
              valuetoRGB(cpix[i-x1],rr,gg,bb,bpp);                         
              switch(filter_type)
              {   case 0:   rvalue = (int)(rsum * shnew / centerfac + rr*shold); 
                            gvalue = (int)(gsum * shnew / centerfac + gg*shold); 
                            bvalue = (int)(bsum * shnew / centerfac + bb*shold); 
                            break;
                  case 1:   rvalue = -rsum + centerfac * rr; 
                            gvalue = -gsum + centerfac * gg; 
                            bvalue = -bsum + centerfac * bb; 
                            if(sharpfac<100)
                            {  rvalue = (int)(rvalue*shnew + rr*shold); 
                               gvalue = (int)(gvalue*shnew + gg*shold); 
                               bvalue = (int)(bvalue*shnew + bb*shold); 
                            }
                            break;
                  case 2:   rvalue = (int)((rsum - centerfac * rr)*shnew + rr*shold); 
                            gvalue = (int)((gsum - centerfac * gg)*shnew + gg*shold); 
                            bvalue = (int)((bsum - centerfac * bb)*shnew + bb*shold); 
                            break;
               }   
                    //  Next set the pixel to its new value  //
               if(rvalue<0) rvalue=0;
               if(gvalue<0) gvalue=0;
               if(bvalue<0) bvalue=0;
               if(rvalue>g.maxred[bpp]  ) rvalue=g.maxred[bpp];
               if(gvalue>g.maxgreen[bpp]) gvalue=g.maxgreen[bpp];
               if(bvalue>g.maxblue[bpp] ) bvalue=g.maxblue[bpp];
               value = RGBvalue(rvalue,gvalue,bvalue,bpp);
               xindex = i - (x1 + asource);
               if(j>y1+1+asource &&(g.selected_is_square || inside_irregular_region(i,j-asource-1)))
                  setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,bpp,-1,noscreen);
               saveline.p[index][xindex] = value;
           }
         }
      }
      
      break;

           // Background subtraction or median filtering     
      case 3:
      case 5:
      { 
         bufindex = 0;                    // Collect ksize rows of pixels,
                                          // put in sequential rows of bkbuf
                                          // to avoid reading same xvalue twice
         for(l=j-asource;l<=j+asource;l+=kmult)             // Source y value
         { 
          for(k=x1; k<=x2; k++)                             // Source x value
            {   hit=1;
                bkbuf.p[bufindex][k-x1] = (int)readpixelonimage(k,l,bpp,ii); 
            }   
            bufindex++;
           }
         for(i=x1+asource; i<x2-asource; i++)               //Destination x 
         {  h = 0;    
            ii=whichimage(i,j,bpp);
            if(bpp!=obpp) 
            {   if(ii>=0 && bpp==8)
                {  obpp=bpp; 
                   switch_palette(ii);
                }
            }   
            if(ii>=0){ z[ii].hit=1; noscreen=1; } else noscreen=0;
            if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1; 
            maxbpp = max(bpp,maxbpp);
                                                            // Read pixels from buffer
            if(bpp==8 || (ii>=0 && z[ii].hitgray))
            {  for(bufindex=0;bufindex<=a;bufindex++)       // Source y value
                for(k=i-asource;k<=i+asource;k+=kmult)      // Source x value
                   sorted.p[0][h++] = bkbuf.p[bufindex][k-x1];
               if(max_bkg==0)
               {  for(ii=0;ii<=bestvalue;ii++)              // Partial sort
                    for(jj=ii;jj<count;jj++)
                      if(sorted.p[0][ii]>sorted.p[0][jj]) 
                          swap(sorted.p[0][ii],sorted.p[0][jj]);
               }else
               {  for(ii=count;ii>=bestvalue;ii--)          // Partial reverse sort
                   for(jj=ii;jj>=0;jj--)
                     if(sorted.p[0][ii]>sorted.p[0][jj]) 
                        swap(sorted.p[0][ii],sorted.p[0][jj]);
               }
               if(filter_type==3)
               {  if(max_bkg!=0) 
                  {  value = (int)readpixelonimage(i,j,bpp,ii);
                     value2 = 255-(sorted.p[0][bestvalue]-value);
                     value = (int)(shold*value + shnew*value2);
                  }   
                  else
                    value=(int)readpixelonimage(i,j,bpp,ii) - sorted.p[0][bestvalue];               
               }
               else 
               {  value = (int)readpixelonimage(i,j,bpp,ii);
                  if(abs(value-sorted.p[0][bestvalue])>range) value = sorted.p[0][bestvalue];
               }
             
                                     //  Set the pixel to its new value 
  
               if(value<0) value=0;
               else if(value>(int)g.maxvalue[bpp]) value=(int)g.maxvalue[bpp];
               xindex = i - (x1 + asource);
     
               value2 = readpixelonimage(i,j,bpp,ii);
               value = cint(shnew*value + shold*value2);            
               value = max(0, min((int)g.maxvalue[bpp],value));
               if(j>y1+1+asource && (g.selected_is_square || inside_irregular_region(i,j-asource-1)))
                  setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,bpp,-1,noscreen);
               saveline.p[index][xindex]=value;    
            }else
            {                        // Read pixels from buffer
               for(bufindex=0;bufindex<=a;bufindex++)    // Source y value
               { for(k=i-asource;k<=i+asource;k+=kmult)  // Source x value
                 {   valuetoRGB(bkbuf.p[bufindex][k-x1],rr,gg,bb,bpp);
                     sorted.p[0][h] = rr;
                     sorted.p[1][h] = gg;
                     sorted.p[2][h] = bb;
                     h++;
                 }
               }
               if(max_bkg==0)
               {  for(ii=0;ii<=bestvalue;ii++)             // Partial sort
                    for(jj=ii;jj<count;jj++)
                    { if(sorted.p[0][ii]>sorted.p[0][jj]) 
                         swap(sorted.p[0][ii],sorted.p[0][jj]);
                      if(sorted.p[1][ii]>sorted.p[1][jj]) 
                         swap(sorted.p[1][ii],sorted.p[1][jj]);
                      if(sorted.p[2][ii]>sorted.p[2][jj]) 
                         swap(sorted.p[2][ii],sorted.p[2][jj]);
                    }  
               }else
               {  for(ii=count;ii>=bestvalue;ii--)     // Partial reverse sort
                   for(jj=ii;jj>=0;jj--)
                   { if(sorted.p[0][ii]>sorted.p[0][jj])swap(sorted.p[0][ii],sorted.p[0][jj]);
                     if(sorted.p[1][ii]>sorted.p[1][jj])swap(sorted.p[1][ii],sorted.p[1][jj]);
                     if(sorted.p[2][ii]>sorted.p[2][jj])swap(sorted.p[2][ii],sorted.p[2][jj]);
                   }  
               }
               value = (int)readpixelonimage(i,j,bpp,ii);
               valuetoRGB(value,rr,gg,bb,bpp);
               orr = rr; ogg = gg; obb = bb;
               if(filter_type==3)
               {  if(max_bkg!=0) 
                  {  rr = g.maxred[bpp]   - (sorted.p[0][bestvalue] - rr);
                     gg = g.maxgreen[bpp] - (sorted.p[1][bestvalue] - gg);
                     bb = g.maxblue[bpp]  - (sorted.p[2][bestvalue] - bb);
                  }   
                  else
                  {  rr = rr - sorted.p[0][bestvalue];               
                     gg = gg - sorted.p[1][bestvalue];               
                     bb = bb - sorted.p[2][bestvalue];               
                  }
               }else
               {  
                  if(abs(rr-sorted.p[0][bestvalue])>range) rr = sorted.p[0][bestvalue];
                  if(abs(gg-sorted.p[1][bestvalue])>range) gg = sorted.p[1][bestvalue];
                  if(abs(bb-sorted.p[2][bestvalue])>range) bb = sorted.p[2][bestvalue];
               }   
             
               rvalue = cint(shnew*rr + shold*orr);            
               gvalue = cint(shnew*gg + shold*ogg);            
               bvalue = cint(shnew*bb + shold*obb);            

                                     //  Set the pixel to its new value 
               rvalue=min(max(0,rvalue),g.maxred[bpp]);
               gvalue=min(max(0,gvalue),g.maxgreen[bpp]);
               bvalue=min(max(0,bvalue),g.maxblue[bpp]);
               xindex = i - (x1 + asource);
               if(j>y1+1+asource &&(g.selected_is_square||inside_irregular_region(i,j-asource-1)))
                 setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,bpp,-1,noscreen);
               saveline.p[index][xindex]=RGBvalue(rvalue,gvalue,bvalue,bpp);    
            }
         }      
      }
      break;

    //// For background flattening, subtract dv from each pixel, where dv is
    //// dvx=fx(1-fy)(v2-v1) + fx*fy(v4-v3)     v1  o     o v2              
    //// dvy=fy(1-fx)(v3-v1) + fx*fy(v4-v2)                                 
    //// dv =fx*vx + fy*vy                                                  
    //// fx=frac. along x direction                                         
    //// fy=frac. along y direction             v3  o     o v4              
    //// For color images, this is done for each color.                     

      case 4:
      {
       fy = (double)(j-y1)/(double)(y2-y1);
       if(bpp==8 || (ii>=0 && z[ii].hitgray))
       {  fy1=v21*(1.0-fy);
          fy2=fy*v43;
       }else
       {  rfy1=(double)(rv2-rv1)*(1.0-fy);
          gfy1=(double)(gv2-gv1)*(1.0-fy);
          bfy1=(double)(bv2-bv1)*(1.0-fy);
       }
       for(k=x1; k<=x2; k++)     // Source x value
       {   
         ii=whichimage(k,j,bpp);
         if(ii>=0){ z[ii].hit=1; noscreen=1; } else noscreen=0;
         if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
         fx = (double)(k-x1)/(double)(x2-x1);
         fx1=1.0-fx;
         if(bpp==8 || (ii>=0 && z[ii].hitgray))
         {   
              dvx = fx*( fy1 + fy2 );
              dvy = fy*( fx1*v31 + fx*v42 );
              dv = fx*dvx + fy*dvy;
              value = (int)readpixelonimage(k,j,bpp,ii);
              value2 = cint(value - dv + average/8);
              value2 = max(0, min((int)g.maxvalue[bpp],value2));
              value = cint(shnew*value2 + shold*value);
              // Make sure it doesn't go to 0
              if(g.selected_is_square || inside_irregular_region(k,j))
                 setpixelonimage(k,j,value,SET,bpp,-1,noscreen);
         }else
         {    rdvx = fx*( rfy1 + fy*(double)(rv4-rv3) );
              gdvx = fx*( gfy1 + fy*(double)(gv4-gv3) );
              bdvx = fx*( bfy1 + fy*(double)(bv4-bv3) );
              rdvy = fy*( (double)(rv3-rv1)*fx1 + fx*(double)(rv4-rv2) );
              gdvy = fy*( (double)(gv3-gv1)*fx1 + fx*(double)(gv4-gv2) );
              bdvy = fy*( (double)(bv3-bv1)*fx1 + fx*(double)(bv4-bv2) );
              rdv = fx*rdvx + fy*rdvy;
              gdv = fx*gdvx + fy*gdvy;
              bdv = fx*bdvx + fy*bdvy;
              value = readpixelonimage(k,j,bpp,ii);
              valuetoRGB(value,rr,gg,bb,bpp);

              rv = rr - cint(rdv);
              gv = gg - cint(gdv);
              bv = bb - cint(bdv);

              rr = cint(shnew*rv + shold*rr);
              gg = cint(shnew*gv + shold*gg);
              bb = cint(shnew*bv + shold*bb);

              rr = max(0,min(rr, g.maxred[bpp]));
              gg = max(0,min(gg, g.maxgreen[bpp]));
              bb = max(0,min(bb, g.maxblue[bpp]));
              value = RGBvalue(rr,gg,bb,bpp);
              if(g.selected_is_square || inside_irregular_region(k,j))
                 setpixelonimage(k,j,value,SET,bpp,-1,noscreen);
         }
       }
       skip=1;
       hit=1;
      }
      break;
    
      //// Sobel edge detection           
      case 12:                                            // Sobel
      {                                                   // i,j=The Pixel
         hit=1;
         for(i=x1+asource; i<x2-asource; i++)//Dest.x pixel 
         { 
           ii=whichimage(i,j,bpp);
           if(ii>=0){ z[ii].hit=1; noscreen=1; } else noscreen=0;
           if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
           if(bpp==8 || (ii>=0 && z[ii].hitgray))         // Monochrome Sobel
           {
             s1=0;s2=0;s3=0;s4=0;                         // k,l=source pixel
             for(l=j-asource;l<=j+asource;l+=kmult)       // Source y value
             for(k=i-asource;k<=i+asource;k+=kmult)       // Source x value
             {
                kk=k-i+asource;     // kk={ 0..kernel size }
                ll=l-j+asource;     // ll={ 0..kernel size }

                v= (int)readpixelonimage(k,l,bpp,ii);     // Source pixel
                if(kk>ll) s1+=v; else if(ll>kk) s1-=v;             /* Sum \ */

                if(kk+ll> asource*2) { s2+=v; }                    // Sum /
                else if(kk+ll< asource*2){ s2-=v; }

                if(kk>asource) s3+=v; else if(kk<asource) s3-=v;   // Sum |
                if(ll>asource) s4+=v; else if(ll<asource) s4-=v;   // Sum -
             }
             v = max(max(abs(s1),abs(s2)), max(abs(s3),abs(s4)));
             value = readpixelonimage(i,j,bpp,ii);
             v = cint(shnew*v + shold*value);            
             v=max(0,min((int)g.maxvalue[bpp],v));
           }else
           {                                              // Colour Sobel
             rs1=0;rs2=0;rs3=0;rs4=0;                     // k,l=source pixel
             gs1=0;gs2=0;gs3=0;gs4=0;  
             bs1=0;bs2=0;bs3=0;bs4=0;  
             for(l=j-asource;l<=j+asource;l+=kmult)       // Source y value
             for(k=i-asource;k<=i+asource;k+=kmult)       // Source x value
             {
                kk=k-i+asource;     // kk={ 0..kernel size }
                ll=l-j+asource;     // ll={ 0..kernel size }

                v= (int)readpixelonimage(k,l,bpp,ii);     // Source pixel
                valuetoRGB(v,rr,gg,bb,bpp);
                if(kk>ll){ rs1+=rr; gs1+=gg; bs1+=bb;}             // Sum \ 
                else if(ll>kk){ rs1-=rr; gs1-=gg; bs1-=bb; }                                      

                if(kk+ll>asource*2){rs2+=rr; gs2+=gg; bs2+=bb; }   // Sum /
                else if(kk+ll<asource*2){ rs2-=rr; gs2-=gg; bs2-=bb; }

                if(kk>asource){ rs3+=rr; gs3+=gg; bs3+=bb; }       // Sum |
                else if(kk<asource){ rs3-=rr; gs3-=gg; bs3-=bb; }

                if(ll>asource){ rs4+=rr; gs4+=gg; bs4+=bb; }       // Sum -
                else if(ll<asource){ rs4-=rr; gs4-=gg; bs4-=bb; }
             }
             rv=max(max(abs(rs1),abs(rs2)),max(abs(rs3),abs(rs4)));
             gv=max(max(abs(gs1),abs(gs2)),max(abs(gs3),abs(gs4)));
             bv=max(max(abs(bs1),abs(bs2)),max(abs(bs3),abs(bs4)));
             value = readpixelonimage(i,j,bpp,ii);
             valuetoRGB(value,orr,ogg,obb,bpp);
             rv = cint(shnew*rv + shold*orr);            
             gv = cint(shnew*gv + shold*ogg);            
             bv = cint(shnew*bv + shold*obb);            
             rv=max(0,min(g.maxred[bpp],rv));
             gv=max(0,min(g.maxgreen[bpp],gv));
             bv=max(0,min(g.maxblue[bpp],bv));
             v=RGBvalue(rv,gv,bv,bpp);
           }
           xindex = i - (x1 + asource);
           if(j>y1+1+asource && (g.selected_is_square || inside_irregular_region(i,j-asource-1)))
               setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,bpp,-1,noscreen);
           saveline.p[index][xindex]=v;    
         }
      }
      break;
     
      default:                                            // Arbitrary filter 
      {  f = filter_type-6;
         for(i=x1+asource; i<x2-asource; i++)//Dest.x pixel 
         { 
           ii=whichimage(i,j,bpp);
           if(ii>=0){ z[ii].hit=1; noscreen=1; } else noscreen=0;
           if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
           if(bpp!=obpp) 
           {  if(ii>=0 && bpp==8)
              {  obpp=bpp; 
                 switch_palette(ii);
              }
           }   
           maxbpp = max(bpp,maxbpp);
           if(bpp==8 || (ii>=0 && z[ii].hitgray))
           {   value = 0;
               v = 0;
               for(l=j-asource;l<=j+asource;l+=kmult)       // Source y value
               {  l2 = l - j + asource;              
                  for(k=i-asource;k<=i+asource;k+=kmult)    // Source x value
                  {  k2 = k - i + asource;
                     hit=1;
                     value = (int)readpixelonimage(k,l,bpp,ii);
                     switch(ksize)
                     { case 3: v += three[f][l2][k2]*value;  break;
                       case 5: v += five[f][l2][k2]*value;   break;
                       case 9: v += nine[f][l2][k2]*value;   break;
                       case 15:v += fifteen[f][l2][k2]*value;break;
                       default: break; 
                     }   
                  }
               }      
               if(sharpfac<100) 
                    value = (int)(shold*readpixelonimage(i,j,bpp,ii) + shnew*v);
               else 
                    value=v;

               if(value<0) value=0;
               if(value>(int)g.maxvalue[bpp]) value=(int)g.maxvalue[bpp];
                                      //  Next set the pixel to its new value  //
               xindex = i - (x1 + asource);
               if(j>1+y1+asource && (g.selected_is_square || inside_irregular_region(i,j-asource-1)))
                 setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,
                    bpp,-1,noscreen);
               saveline.p[index][xindex]=value;    
           }else
           {   rvalue = 0;
               gvalue = 0;
               bvalue = 0;
               for(l=j-asource;l<=j+asource;l+=kmult)       // Source y value
               {  l2 = l - j + asource;              
                  for(k=i-asource;k<=i+asource;k+=kmult)    // Source x value
                  {  k2 = k - i + asource;
                     hit=1;
                     valuetoRGB((int)readpixelonimage(k,l,bpp,ii),rr,gg,bb,bpp);
                     switch(ksize)
                     {   case 3: rvalue += three[f][l2][k2] * rr;
                                 gvalue += three[f][l2][k2] * gg;
                                 bvalue += three[f][l2][k2] * bb;
                                 break;
                         case 5: rvalue += five[f][l2][k2] * rr;
                                 gvalue += five[f][l2][k2] * gg;
                                 bvalue += five[f][l2][k2] * bb;
                                 break;
                         case 9: rvalue += nine[f][l2][k2] * rr;
                                 gvalue += nine[f][l2][k2] * gg;
                                 bvalue += nine[f][l2][k2] * bb;
                                 break;
                         case 15:rvalue += fifteen[f][l2][k2] * rr;
                                 gvalue += fifteen[f][l2][k2] * gg;
                                 bvalue += fifteen[f][l2][k2] * bb;
                                 break;
                         default: break;
                      }   
                  }
               }      
               if(sharpfac<100)
               {  valuetoRGB((int)readpixelonimage(i,j,bpp,ii),rr,gg,bb,bpp);
                  rvalue = (int)(rr*shold + rvalue*shnew); 
                  gvalue = (int)(gg*shold + gvalue*shnew); 
                  bvalue = (int)(bb*shold + bvalue*shnew); 
               }
               if(rvalue<0) rvalue=0;
               if(gvalue<0) gvalue=0;
               if(bvalue<0) bvalue=0;
               if(rvalue>g.maxred[bpp]  ) rvalue=g.maxred[bpp];
               if(gvalue>g.maxgreen[bpp]) gvalue=g.maxgreen[bpp];
               if(bvalue>g.maxblue[bpp] ) bvalue=g.maxblue[bpp];
                                      //  Next set the pixel to its new value  //
               xindex = i - (x1 + asource);
               if(j>1+y1+asource && (g.selected_is_square || inside_irregular_region(i,j-asource-1)))
                 setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,
                   bpp,-1,noscreen);
               saveline.p[index][xindex]=RGBvalue(rvalue,gvalue,bvalue,bpp);    
           }
         }
      }   
    }
    if(bad) break;  
  }
 
  //------------------end of color-specific code for filtering----------//
 

               // Write out the extra pixels             
  if(!hit)
    message("Please select an image or\nlarger area to filter",ERROR);
  else if(!bad)
  {
    if(!skip)
      for(j=y2-asource; j<=y2; j++)    //destination y pixel
      {  if(j>0) index = j%(asource+1); 
         else    index = asource + (j - asource) % (asource+1);
                                                         //destination x pixel
        for(i=x1+asource; i<x2-asource; i++)
         { 
            ii=whichimage(i,j,bpp);
            if(bpp!=obpp) 
            {  if(ii>=0 && bpp==8)
               {  obpp=bpp; 
                  switch_palette(ii);
               }
            }   
            if(ii>=0){ z[ii].hit=1; noscreen=1; } else noscreen=0;
            if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
            xindex = i - (x1 + asource);
            if(j>y1+1+asource &&(g.selected_is_square||inside_irregular_region(i,j-asource-1)))
               setpixelonimage(i,j-asource-1,saveline.p[index][xindex],SET,bpp,-1,noscreen);
         }   
      }
  }

  if(want_progress_bar) progress_bar_close(www);
  if(convert24) change_image_depth(ci,8,0);

  //// Re-convert any color images touched if in 8-bpp screen mode
  for(k=0; k<g.image_count; k++) 
     if(z[k].hit){ rebuild_display(k); redraw(k); }

  redrawscreen();               // Redraw main screen
  delete[] cpix;
  if(bad) message("",BUG);
  g.want_switching = 1;
  return ii;
}


//--------------------------------------------------------------------------//
// get_average_value                                                        //
//--------------------------------------------------------------------------//
void get_average_value(int x1,int y1,int x2,int y2,int &v,int &rv,int &gv,int &bv)
{ 
    int count,bpp,ii,j,k,rr,gg,bb,value;
    v=0; count=0;
    for(j=y1;j<=y2;j++)                 // Source y value
    for(k=x1;k<=x2;k++)                 // Source x value
    {   value = readpixelonimage(k,j,bpp,ii);
        valuetoRGB(value,rr,gg,bb,bpp);
        v+=value; rv+=rr; gv+=gg; bv+=bb;
        count++;
    }
    v/=count; rv/=count; gv/=count; bv/=count;
    return;
}    


//--------------------------------------------------------------------------//
//  remove_low_frequencies                                                  //
//--------------------------------------------------------------------------//
void remove_low_frequencies(int ksize, int kmult, int ulx, int uly, int lrx, int lry)
{
   //// Duplicate image
   ulx=ulx; uly=uly; lrx=lrx; lry=lry;
   int d[10] = { 100 ,100, 100, 100, 100 ,50, 1 };
   int k, noofargs=0,bpp;
   char **arg;                // up to 20 arguments
   arg = new char*[20];
   for(k=0;k<20;k++){ arg[k] = new char[128]; arg[k][0]=0; }
   int oci = ci;
   bpp = z[ci].bpp;
   duplicate_image(ci,0,0,0,0);
   int nci = ci;
 
   //// Low pass filter new one with ksize
   for(k=1;k<=kmult;k++)
      filter_region(0, 0, z[ci].xsize-1, z[ci].ysize-1, 0, ksize, 100, 1, 1, 1, 1, 1);

   //// Multiply int by .5 in both
   
   multiplypalettecolors(d);
   switchto(oci);
   multiplypalettecolors(d);

   //// Add 127 to original
   sprintf(arg[1], "%d", 0);
   sprintf(arg[2], "%d", 0);
   sprintf(arg[3], "%d", 0);
   sprintf(arg[4], "%d", 0);
   sprintf(arg[5], "%d", 0);
   sprintf(arg[6], "%d", (int)g.maxvalue[bpp]/2);
   noofargs = 6;
   changebrightness(noofargs, arg);

   //// Subtract the duplicate image
   sprintf(arg[1], "%d", oci);
   sprintf(arg[2], "%d", nci);
   sprintf(arg[3], "%d", 3);
   sprintf(arg[4], "%d", 0);
   sprintf(arg[5], "%d", 0);
   noofargs = 5;
   execute("mask", noofargs, arg);

   //// Rescale
   switchto(oci);
   sprintf(arg[1], "%d", 5);
   noofargs = 1;
   maximize_contrast(noofargs, arg);

   eraseimage(nci,0,0,0);
   for(k=0;k<20;k++) delete[] arg[k];
   delete arg;  

}
