//--------------------------------------------------------------------------//
// xmtnimage19.cc                                                           //
// Handling of macros                                                       //
// Latest revision: 04-14-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Images     *z;
extern int         ci;
extern PrinterStruct printer;

const int MAXARGS=50;                      // max of 50 arguments

//--------------------------------------------------------------------------//

//--------------------------------------------------------------------------//
// parsemacro - returns:                                                    //
//   1st word = arg[0]                                                      //
//   remaining string in the line after 1st word = line                     //
//   additional words in the line after 1st word = arg                      //
//   total no. of words in the line = argno                                 //
//--------------------------------------------------------------------------//
void parsemacro(char *text, char *line, char *line2, int wantline, char **arg, 
   int &argno, int &opos, int &owantline, int &olineno)
{
    int j,k,kk=0,len,lineno,space1=0;
    len = (int)strlen(text);
    argno=0;
    for(k=0;k<MAXARGS;k++) arg[k][0]=0;
    line[0]=0;
    line2[0]=0;

     ////  If line no is less than before, start at beginning; otherwise continue 
     ////  starting at opos and olineno.

    if(owantline>wantline){ opos=0; olineno=0; }                                    
    lineno=olineno;

    ////  Eliminate leading spaces
    
    while(text[opos]==' ') opos++;

    ////  Find the end position kk of the string

    for(k=opos; k<=len; k++)
    {   if(text[k]=='\n' || text[k]==10 || k==len)  
        {   kk=k;
            if(lineno==wantline){ lineno++; break; }
        }
    }

    ////  Find the start of the 2nd word
    
    for(k=opos; k<=len; k++) if(text[k]==' '){ space1=k; break; }

    ////  Split the text into argno words

    for(k=opos,j=0; k<kk; k++,j++)
    {   if(text[k] == ' ')    // If 1 or more spaces, parse next argument.
        {    argno++; 
             j=0;
             arg[argno][j]=0;
             if(argno>=MAXARGS)break;
             while (text[k] == ' ') k++;
        }
        arg[argno][j]=text[k];
        arg[argno][j+1]=0;
        if(j>80) break;
    }

    ////  Fill 'line'
    
    for(k=opos,j=0;k<kk;k++,j++) line[j]=text[k]; 
    line[kk-opos]=0;

    ////  Fill 'line2'
    
    for(k=space1+1,j=0;k<kk;k++,j++) line2[j]=text[k]; 
    line2[kk-space1-1]=0;

    opos=kk+1;                    
    olineno=lineno;
    owantline=wantline;   
}

//--------------------------------------------------------------------------//
// macro                                                                    //
//--------------------------------------------------------------------------//
int macro(char *text)
{
   static char **arg;              // arguments up to 80 chars passed as strings
   static int hitarg=0;

   if(memorylessthan(16384)){  message(g.nomemory,ERROR); return(NOMEM); } 
   char line[128];
   char line2[128];
   char err[256]="Macro terminated at line ";
   int argno=0,j,k,lineno=0,status=OK,totallines=0, opos=0,owantline=0,olineno=0,
       gotformula=0,len,math_mode=0,val,owant_messages; 
                                   // math_mode==1: each line is a math formula
                                   // math_mode==0: each line is a macro
   time_t ntime,otime;
   time(&otime);                   // Don't use gettimeofday(), not avail. on Convex
  
   char tempstring[128]="";
   char tempstring2[128]="";
   owant_messages = g.want_messages;
                                   
   if(!hitarg)                     // Leave allocated forever
   {    arg = new char *[MAXARGS];
        for(k=0;k<MAXARGS;k++) arg[k] = new char[80];
        hitarg=1;
   }

   ////  Parse arguments & execute. Arg[0] is the command, arg[1] to
   ////  arg[argno] inclusive are the arguments.

   if(g.getout){ g.getout=0; return(ABORT); }
   g.getout=0;
   len = (int)strlen(text);
   for(k=0;k<len;k++) if(text[k]=='\n') totallines++;
   for(lineno=0; lineno<=totallines; lineno++)
   {
       printstatus(MACRO);
       parsemacro(text, line, line2, lineno, arg, argno, opos, owantline, olineno);

       if(opos>len+2) break;

              ////  Give other processes a chance once/second - otherwise it 
              ////  slows down the macros too much
       time(&ntime);
       if(ntime!=otime)
       {    XtAppProcessEvent(XtWidgetToApplicationContext(g.main_widget),XtIMAll);
            otime = ntime;
       }
       if(g.getout) break;
  
              ////  Unconditional goto          
       if(strcmp(arg[0],"goto")==SAME){ lineno=atoi(arg[1])-2; continue; }   

              ////  Unconditional jump to line 0 with option to abort
       if(strcmp(arg[0],"loop")==SAME)
       {   if(message("Repeat macro (y/n)?",QUESTION)!=YES) break;
           lineno=-1; 
           continue;
       }   

              ////  If line starts with 'math' it could be 1 or 0 or a formula
       gotformula=0;
       if(strcmp(arg[0],"math")==SAME)
       {   val = atoi(arg[1]);
           if(strlen(arg[1])==1 && (val==0 || val==1)) math_mode = val;
               else gotformula=1;
           if(!gotformula) continue;   // Just a control command
       }
       
       if(strcmp(arg[0],"stop")==SAME)  break;

              ////  Request parameter from user
       for(j=0;j<=argno;j++)
       {  if(arg[j][0]=='?')
          {  strcpy(tempstring,arg[j]+1);
             if(strlen(tempstring)==0){ status=BADPARAMETERS; break; }
             message(tempstring,tempstring2,PROMPT,79);
             strcpy(arg[j],tempstring2);
          }
       }  
       if(status!=OK) break; 

              ////  Execute the command
       if(g.getout) break;
       if(math_mode)              //// Math mode: entire line is a command
       {   status = math(line);
       }else if(gotformula)       //// 'math' followed by one-line formula
       {   status = math(line2);
       }else if(strlen(arg[0]))
       {                          //// A regular command
           status=execute(arg[0],argno,arg);  
       }
       if(status!=OK) break;
   }

           //---Reset parameters (end of macro)---//
   printstatus(NORMAL);
   if(!g.want_messages) status=OK;
   g.want_messages = owant_messages;
   g.want_format=-1;

   sprintf(tempstring,"%s %d\nError: ",err, lineno+1);
   switch(status)
   {  case OK:break;
      case NOMEM: strcat(tempstring,"Out of memory");break;
      case CANCEL:
      case ABORT: strcat(tempstring,"Aborted");break;
      case ERROR: strcat(tempstring,"Error in macro");break;
      case NOTFOUND: strcat(tempstring,"File not found");break;
      case NOIMAGES: strcat(tempstring,"No images");break;
      case CANTCREATEFILE: strcat(tempstring, "Can't create file");break;
      case CRITERR: strcat(tempstring,"Critical error");break;
      case ZEROLENGTH: strcat(tempstring,"File was zero length");break;
      case BADPARAMETERS: strcat(tempstring,"Bad parameters");break;
      case UNKNOWN: strcat(tempstring,"Unrecognized command");break;
      default: strcat(tempstring,"Unknown");break;
   }
   if((status!=QUIT)&&(status!=OK)) message(tempstring,ERROR);
   return(status);
}





//--------------------------------------------------------------------------//
// edit                                                                     //
// simple editor                                                            //
// count = total no. of rows in editor                                      //
//--------------------------------------------------------------------------//
void edit(char *title, char *subtitle, char *text, int rows, int columns, 
     int maxlength, int helptopic)
{
   const int HEIGHT=200;
   const int WIDTH=580;
   Widget form, editor, *w, loadbutton, savebutton, labelwidget;
   clickboxinfo c;
   XmString xms;
   int k, n, wc=0;
   Arg args[100];
   char *s;
   static char *filename=NULL;
   if(filename==NULL){ filename = new char[FILENAMELENGTH]; filename[0]=0; }
   w = new Widget[100];     // List of widgets to destroy

   //------- Create a form dialog shell widget -----------------------//
   n=0;
   XtSetArg(args[n], XmNresizable, True); n++;
   XtSetArg(args[n], XmNautoUnmanage, False); n++;
   XtSetArg(args[n], XmNtransient, False); n++;
   XtSetArg(args[n], XmNdialogTitle, xms=XmStringCreateSimple(title)); n++;
   form = w[wc++] = XmCreateFormDialog(g.main_widget, "EditForm", args, n);
   XmStringFree(xms);
 
   ////  Add label
   XtSetArg(args[n], XmNleftPosition, 2); n++;     // % of width from left
   XtSetArg(args[n], XmNrightPosition, 100); n++;  
   XtSetArg(args[n], XmNtopPosition, 0); n++;      // % of height from top
   XtSetArg(args[n], XmNtopOffset, 1); n++;     
   XtSetArg(args[n], XmNheight, 20); n++;     
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_NONE); n++;
   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
   labelwidget = XmCreateLabel(form, subtitle, args, n);
   XtManageChild(labelwidget);   

   n = 0;
   XtSetArg(args[n], XmNwidth, max(WIDTH,columns * 6)); n++;
   XtSetArg(args[n], XmNheight, HEIGHT); n++;
   XtSetArg(args[n], XmNeditable, True); n++;
   XtSetArg(args[n], XmNeditMode, XmMULTI_LINE_EDIT); n++;
   XtSetArg(args[n], XmNmaxLength, maxlength); n++;
   XtSetArg(args[n], XmNcolumns, columns); n++;
   XtSetArg(args[n], XmNrows, rows); n++;
   XtSetArg(args[n], XmNleftPosition, 1); n++;       // % of width from left
   XtSetArg(args[n], XmNmarginWidth, 1); n++;
   XtSetArg(args[n], XmNmarginHeight, 0); n++;
   XtSetArg(args[n], XmNtopPosition, 0); n++;      // % of height from top
   XtSetArg(args[n], XmNbottomPosition, 98); n++;  // % of width from top
   XtSetArg(args[n], XmNrightPosition, 99.8); n++;   // % of height from left
   XtSetArg(args[n], XmNtopOffset, 20); n++;       
   XtSetArg(args[n], XmNbottomOffset, 24); n++;       // % of height from top
   XtSetArg(args[n], XmNfractionBase, 100); n++;     // Use percentages
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
   ////  Stop Motif from trying to grab another color if none are available.
   if(g.want_colormaps)  
   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
   }
   editor = w[wc++] = XmCreateScrolledText(form, "Editor", args, n);
   XmTextSetString(editor, text);

   c.ptr = (char *)text;        // Starting text         
   c.widget[0] = form;          // Widget to put answer on
   c.widget[1] = editor;        // Widget to read string from
   c.widget[2] = labelwidget;   // Label to put filename in
   c.count = rows;              // Max. no. of rows (not enforced)
   c.maxval    = maxlength;     // Max. no. of characters
   c.title = new char[FILENAMELENGTH];
   strcpy(c.title, title);      // Title to display
   c.filename  = filename;      // Filename
   c.ask = 1;                   // Use getfilename() to get file name
   c.allowedit = 2;             // Each line is a command
   c.form      = form;          // Parent form
   c.path = NULL;
   c.maxstringsize = maxlength; // Maximum string size
   
   XtAddCallback(editor, XmNmotionVerifyCallback, (XtCBP)editcursorcb, (XtP)&c);
   XtAddCallback(editor, XmNmodifyVerifyCallback, (XtCBP)editmodifycb, (XtP)&c);
   XtManageChild(editor);
   XtManageChild(form);
   addstandardbuttons(form, &w[wc], &helptopic);
   wc += 3;    // addstandardbuttons adds 3 widgets

               // Add "load" button
   loadbutton = w[wc++] = XtVaCreateManagedWidget("Load",
         xmPushButtonWidgetClass,    form,
         XmNtopAttachment,           XmATTACH_NONE,
         XmNbottomAttachment,        XmATTACH_FORM,
         XmNleftAttachment,          XmATTACH_FORM,
         XmNrightAttachment,         XmATTACH_NONE,
         XmNwidth,                   100,
         XmNleftOffset,              335,
         XmNbottomOffset,            5,
         XmNdefaultButton,           False,
         NULL);
               // Add "save" button
   savebutton = w[wc++] = XtVaCreateManagedWidget("Save",
         xmPushButtonWidgetClass,    form,
         XmNtopAttachment,           XmATTACH_NONE,
         XmNbottomAttachment,        XmATTACH_FORM,
         XmNleftAttachment,          XmATTACH_FORM,
         XmNrightAttachment,         XmATTACH_NONE,
         XmNwidth,                   100,
         XmNleftOffset,              440,
         XmNbottomOffset,            5,
         XmNdefaultButton,           False,
         NULL);

   XtVaSetValues(loadbutton, XmNbackground, g.main_bcolor,
         XmNforeground, g.main_fcolor, NULL);
   XtVaSetValues(savebutton, XmNbackground, g.main_bcolor,
         XmNforeground, g.main_fcolor, NULL);
   XtAddCallback(savebutton, XmNactivateCallback, (XtCBP)savetextcb, (XtP)&c);
   XtAddCallback(loadbutton, XmNactivateCallback, (XtCBP)loadtextcb, (XtP)&c);
   XtAddEventHandler(XtParent((form)), FocusChangeMask, False,
        (XtEH)graph_raise_cb, (XtP)NULL);

   g.getout=0; 
   event_loop(form);
   strcpy(text, s=XmTextGetString(editor));  // This allocates memory
   XtFree(s);
   XtRemoveEventHandler(XtParent(form), StructureNotifyMask, False, 
       (XtEH)graph_raise_cb, (XtP)NULL);
   XtUnmanageChild(form);
   for(k=0;k<wc;k++) XtDestroyWidget(w[k]);
   delete[] w;
   delete[] c.title;
   return;
}


//--------------------------------------------------------------------------//
// editmodifycb  - called when text in editor changes                       //
//--------------------------------------------------------------------------//
void editmodifycb(Widget w, XtP client_data, XmACB *call_data)
{
   const int CMDLEN=FILENAMELENGTH;
   w=w;  //keep compiler quiet
   int k,len,start=0,end=0,pos=0;
   char *text=NULL;
   char command[FILENAMELENGTH];
   char newtext=0;
   char **arg;                // up to 20 arguments
   int noofargs;
   clickboxinfo *c = (clickboxinfo *)client_data;
   XmTextVerifyCallbackStruct *tvcs = (XmTextVerifyCallbackStruct *)call_data;
   if(tvcs->text->length && tvcs->text->ptr!=NULL) newtext = tvcs->text->ptr[0];

   ////  It is impossible to access the text without making a copy of it.
   ////  Getting XmNvalue also makes a copy.
   ////  Execute current line as a command. Change c.allowedit to 1 if you
   ////  don't want this.  

   ////  Local memory allocation and deallocation must be outside of Xm memory
   ////  functions for unknown reason.
   
   arg = new char*[20];
   for(k=0;k<20;k++) arg[k] = new char[CMDLEN];
   if(c->allowedit==2 && newtext == '\r') 
   {  
       text = XmTextGetString(c->widget[1]);      // This allocates memory
       if(text==NULL){ message("Error getting text string",ERROR); return; }
       if(tvcs->text->ptr) tvcs->text->length=0;  // Get motif to ignore the \r
       start = end = pos = tvcs->startPos;
       len = strlen(text);
       for(k=pos-1;k>=0;k--)                      // Find most recent \n
       {    if(text[k]=='\n'){ start=k+1; break; }
            if(k==0){ start=0; break; }
       }
       for(k=pos;k<=len;k++)                      // Find next \n
       {    if(text[k]=='\n'){ end=k; break; }
            if(k==len){ end=len; break; }
       }
       command[0]=0;
       if(start<end)
       {    strncpy(command,text+start,end-start);
            command[end-start]=0;
            for(k=strlen(command);k;k--)if(command[k]<=' ')command[k]=0; else break;
            if(g.diagnose) printf("Macro command %s\n",command);
            if(strlen(command))
            {    parsecommand(command, arg, noofargs, CMDLEN, 20);
                 XtSetKeyboardFocus(c->widget[1], g.drawing_area);
                 execute(arg[0],noofargs,arg);
                 XtSetKeyboardFocus(c->widget[1], None);
                 infocb(0,0,0);
            }
       }
       XtFree(text);
   }
   for(k=0;k<20;k++) delete[] arg[k];
   delete arg;  
}


//--------------------------------------------------------------------------//
// editcursorcb  - called when cursor moves in editor                       //
//--------------------------------------------------------------------------//
void editcursorcb(Widget w, XtP client_data, XmACB *call_data)
{
    w=w;  //keep compiler quiet
    int k,pos=0,row=1,col=1;
    Window win;
    char *s;
    char rowcol[40];
    clickboxinfo *c = (clickboxinfo *)client_data;
    XmTextVerifyCallbackStruct *tvcs = (XmTextVerifyCallbackStruct *)call_data;
    pos = tvcs->newInsert;
    s = XmTextGetString(c->widget[1]);   // This allocates memory
    for(k=0;k<pos;k++) if(s[k]=='\n'){col=1; row++;} else col++;
    sprintf(rowcol,"%d   %d   ",row,col);
    win = XtWindow(c->widget[0]);

    //  Motif doesn't care if no. of rows increases so leave this commented out.
    //    if(row>c->count)                // Truncate entry if too many rows
    //    {   s[pos]=0;
    //        XmTextSetString(c->widget[1], s);
    //    }

    XDrawImageString(g.display, win, g.gc, 200, 20, rowcol, strlen(rowcol));
    XtFree(s);
}


//--------------------------------------------------------------------//
// savetextcb - callback for saving text data from text widget        //
//--------------------------------------------------------------------//
void savetextcb(Widget w, XtP client_data, XmACB *call_data)
{
   w=w; call_data=call_data; // keep compiler quiet
   clickboxinfo *c = (clickboxinfo *)client_data;
   char *text;
   char *filename = c->filename;
   Widget editor = c->widget[1];
   text = XmTextGetString(editor);  // This allocates memory
   char tempstring[100];
   if(memorylessthan(4096)){  message(g.nomemory,ERROR); return; } 
   FILE *fp;

   message("New filename:",filename,PROMPT,FILENAMELENGTH-1);
   if(cant_overwrite_file(filename)) return;
   if ((fp=fopen(filename,"wt")) == NULL)
        error_message(filename, errno);
   else
   {    fprintf(fp,"%s\n",text);
        sprintf(tempstring,"Text saved in %s",filename);
        message(tempstring);
        fclose(fp);
   }
   XtFree(text);
}


//--------------------------------------------------------------------//
// loadtextcb - callback for reading text data into text widget       //
//--------------------------------------------------------------------//
void loadtextcb(Widget w, XtP client_data, XmACB *call_data)
{
   w=w; call_data=call_data; // keep compiler quiet
   XmString xms;
   clickboxinfo *c = (clickboxinfo *)client_data;
   char *text = (char*)c->ptr;
   char *filename = c->filename;
   char *oldfilename;
   oldfilename = new char[FILENAMELENGTH];
   strcpy(oldfilename, filename);
   
   Widget editor = c->widget[1];
   if(memorylessthan(4096)){  message(g.nomemory,ERROR); return; } 
   int size=0;
   FILE *fp;

   if(c->ask) strcpy(filename, getfilename(oldfilename));
   if(g.getout){ g.getout=0; return; }
   size = filesize(filename);            //// Check to ensure not an empty file
   if(size>=c->maxval)                   //// Check to ensure enough memory
        message("Macro file is too large!",ERROR);
   else if ((fp=fopen(filename,"rt")) == NULL)
        error_message(filename, errno);
   else                                 
   {    fread((void*)text,size,1,fp);    //// Read the entire file 
        text[size]=0;
        if(text[size-1]==10) text[size-1]=0;
        fclose(fp);
        if(c->ask && editor) XmTextSetString(editor, text);
   }
   if(strlen(filename) && c->widget[2])
   {    XtVaSetValues(c->widget[2], XmNlabelString,
           xms=XmStringCreateSimple(filename), NULL); 
        XmStringFree(xms);
   }
   delete[] oldfilename;
   return;
}


//--------------------------------------------------------------------//
// readmacrofile                                                      //
//--------------------------------------------------------------------//
int readmacrofile(char *filename, char *text)
{
    clickboxinfo c;
    c.title = new char[FILENAMELENGTH];
    c.title[0] = 0;
    c.ask = 0;
    c.allowedit = 1;
    c.ptr       = text;
    c.maxval    = MACROLENGTH;
    c.filename  = filename;
    c.widget[0] = 0;  
    c.widget[1] = 0;  
    c.widget[2] = 0;   
    loadtextcb(g.main_widget, &c, NULL);
    text = (char*) c.ptr;
    if(text[0]!=0) strcpy(text, (char *)c.ptr);
    delete[] c.title;
    return OK;
}


//--------------------------------------------------------------------//
// parsecommand                                                       //
// input: a single command with arguments  e.g. "load test 1 1 2 3".  //
// Each word k is placed in arg[k][]. arg[0][] is the command, the    //
// rest are arguments.                                                //
// The no. of arguments, not counting arg[0], is placed in `noofargs'.//
// 'maxlen' is allocated size of the string in arg[][].               //
//--------------------------------------------------------------------//
void parsecommand(char *cmd, char **arg, int &noofargs, int maxlen, int maxargs)
{   
   int count=0;
   char *buf;
   char word[FILENAMELENGTH]="";
   buf = new char[FILENAMELENGTH];
   noofargs = 0;
   do
   {    get_first_word(cmd+count, word, maxlen);
        if(word==NULL) break;
        strcpy(buf, word);
        count += 1+strlen(buf);
        count += strspn(cmd+count," ");
        strncpy(arg[noofargs++], buf, maxlen-1);
        if(count >= (int)strlen(cmd)) break;
        if(noofargs >= maxargs) break;
   }while(1);
   noofargs = max(noofargs-1,0);
   delete[] buf;
}


//--------------------------------------------------------------------//
// get_first_word - returns first word demarcated by a space in buffer//
//--------------------------------------------------------------------//
void get_first_word(char *buffer, char *word, int maxlen)
{
    int start = strspn(buffer," ");                  // Position of 1st non space
    int space = strcspn(buffer+start," ");           // Position of 1st space
    if(space==0) return;
    space = min(maxlen,space);
    strncpy(word, buffer+start, space);   
    word[space]=0;
}    

