//-------------------------------------------------------------------------//
// xmtnimage20.cc                                                          //
// Densitometry, calibration of images                                     //
// Latest revision: 05-14-1999                                             //
// Copyright (C) 1999 by Thomas J. Nelson                                  //
// All rights reserved.                                                    //
//-------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Images     *z;
extern int         ci;
extern Widget www[4];
const int SELECT_MODE = SINGLE;

//-------------------------------------------------------------------------//
// measure                                                                 //
// measure distance, area, angle, etc.                                     //
//-------------------------------------------------------------------------//
void measure(void)
{
    listinfo l;
    if(memorylessthan(16384)){  message(g.nomemory,ERROR); return; } 
    l.info = new char*[10000];
    static int measureangle=1;
    Dialog  dialog;
    Widget *widgetlist;
    double xa=0.0, ya=0.0, xb=0.0, yb=0.0, dx, dy;

    widgetlist = new Widget[100];
    double angle=0,dxa,dxb,dya,dyb,length=0,lena=0,lenb=0,xx;
    int calib_log=0,crud=0,j,k,ino,lcount=0,bpp,gotcal=0,wc=0,x1,y1,x2,y2,x3,y3,x4,y4;
    char temp[128],temp2[128],temp3[128];

    //-----Dialog box to get title & other options------------//
    dialog.needrepeat = 0;
    strcpy(dialog.title,"Measure");
    strcpy(dialog.radio[0][0],"");
    strcpy(dialog.radio[0][1],"Distance");             
    strcpy(dialog.radio[0][2],"Angle between 2 lines");
    strcpy(dialog.radio[0][3],"Angle of 1 line");
    
    dialog.radioset[0]=measureangle;

    dialog.radiono[0]=4;
    for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;

    dialog.noofradios=1;
    dialog.noofboxes=0;
    dialog.helptopic=33;  
    dialog.keep=0;
    dialog.want_changecicb = 0;
    dialog.f1 = null;
    dialog.transient = 1;
    strcpy(dialog.path,".");
    g.getout=0;

    //////////////////

    dialogbox(dialog);

    //////////////////
    if(g.getout){ g.getout=0; return; }
    measureangle = dialog.radioset[0]; 

    g.getout=0;
    g.dialogdone=0;
    g.inmenu++;
    drawselectbox(OFF);
    while(g.getout==0 && !g.dialogdone)
    {
       getline(x1,y1,x2,y2);
       if(g.getout)break;
       switch(measureangle)
       {
          case 1:       // measure distance - possibly calibrated
              ino=whichimage(x1,y1,bpp);
              gotcal=0;
              if(z[ino].cal_log[0]!=-1) gotcal=1;  
              calib_log = z[ino].cal_log[0];
              if(calib_log==-1) 
              {   length = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1));
                  gcvt(length,g.signif,temp);
              }else if(z[ino].cal_dims==0) // 1d non-directional 
              {   xa = sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1));
                  length = fabs(xa * z[ino].cal_q[0][1]);
                  gcvt(length, g.signif, temp);
              }else if(z[ino].cal_dims==1) // 1d directional 
              {  
                  calibratepixel(x2,y2,ino,0,xa,temp);
                  calibratepixel(x1,y1,ino,0,xb,temp);
                  length = xa-xb;
                  gcvt(length,g.signif,temp);
              }else                        // 2d
              { 
                  calibratepixel(x2,y2,ino,0,xa,temp);
                  calibratepixel(x2,y2,ino,1,ya,temp);
                  calibratepixel(x1,y1,ino,0,xb,temp);
                  calibratepixel(x1,y1,ino,1,yb,temp);
                  dy = yb-ya;
                  dx = xa-xb;
                  sprintf(temp,"%g   %g",dx,dy);
              }
              if(gotcal) strcpy(temp2,"Calibrated distance");
              else       strcpy(temp2,"Uncalibrated distance");
              break;
          case 2:       // measure angle between 2 lines
              line(x1,y1,x2,y2,g.maxcolor,XOR);
              getline(x3,y3,x4,y4);
              line(x1,y1,x2,y2,g.maxcolor,XOR);
              dxa = x2-x1; dxb = x4-x3;
              dya = y2-y1; dyb = y4-y3;
              lena = sqrt( dxa*dxa + dya*dya );
              lenb = sqrt( dxb*dxb + dyb*dyb );
              if(lena*lenb == 0.0)
                  angle = 0.0;
              else
              {   xx = -(dxa*dxb + dya*dyb)/(lena*lenb);
                  if(fabs(xx)<=1.00) angle = acos(xx);
              }
              angle *= 360/6.283185308;
              gcvt(angle,g.signif,temp);
              strcpy(temp2,"Angle");
              break;
          case 3:       // measure angle of 1 line
              x3=100; x4=0;
              y3=100; y4=100;
              dxa = x2-x1; dxb = x4-x3;
              dya = y2-y1; dyb = y4-y3;
              lena = sqrt( dxa*dxa + dya*dya );
              lenb = sqrt( dxb*dxb + dyb*dyb );
              if(lena*lenb == 0.0)
                  angle = 0.0;
              else
              {   xx = -(dxa*dxb + dya*dyb)/(lena*lenb);
                  if(fabs(xx)<=1.00) angle = acos(xx);
                  if(y2>y1) angle *=-1;
              }
              angle *= 360/6.283185308;
              gcvt(angle,g.signif,temp);
              strcpy(temp2,"Angle");
              break;
       }
 
       //-----Display results------//
       sprintf(temp3,"%d  %s ",lcount,temp);
       l.info[lcount] = new char[100];        
       strcpy(l.info[lcount],temp3);       
       if(lcount==0) 
       {   char listtitle[40] = "Measurement results";
           l.title = listtitle;
           l.size  = 0;
           l.itemstoshow = 2;
           l.firstitem   = 1;
           l.wantsort    = 0;
           l.wantsave    = 1;
           l.helptopic   = dialog.helptopic;
           l.allowedit   = 0;
           l.selection   = &crud;
           l.width       = 0;
           l.wantfunction = 0;
           l.autoupdate   = 0;
           l.browser = openwindow(&l, widgetlist, &wc);
       }
       lcount++;
       l.size++;
       additemtolist(l.widget, l.browser, temp3 , BOTTOM, 1);
    }
    closewindow(&l, l.browser, widgetlist, wc);
    message("Measurement mode finished");
    g.inmenu--;
    for(k=0;k<lcount;k++) delete[] l.info[k]; 
    delete l.info;
    delete[] widgetlist;
    g.getout=0;
    return;
}


//-------------------------------------------------------------------------//
// calibrate - MW as function of x or y                                    //
// If an image is selected, the image calib. is position-independent.      //
// If calibrating background, calib. is fixed.                             //
//-------------------------------------------------------------------------//
void calibrate(void)
{
    if(memorylessthan(16384)){ message(g.nomemory,ERROR); return; }
    g.inmenu++;
    g.getout=0;
    XYData data; 
    Dialog  dialog;
    double xx=0.0, yy=0.0;
    char title[64], **label, **answer, tempstring[128];  
    int count=0, j, k, ino, dims=1;
    g.getout=0;
    ino = ci;
    if(ino<0) return;

    //-----Dialog box to get title & other options------------//

    dialog.needrepeat = 0;
    strcpy(dialog.title,"Calibration");
    strcpy(dialog.radio[0][0],"Dimensions");             
    strcpy(dialog.radio[0][1],"1-D (non-directional)");             
    strcpy(dialog.radio[0][2],"1-D (directional");             
    strcpy(dialog.radio[0][3],"2-D");             
    strcpy(dialog.radio[1][0],"Calibration for 1st dim.");             
    strcpy(dialog.radio[1][1],"Linear");             
    strcpy(dialog.radio[1][2],"Logarithmic");
    strcpy(dialog.radio[1][3],"2nd-order poly.");
    strcpy(dialog.radio[1][4],"Distance from 0");
    strcpy(dialog.radio[2][0],"Calibration for 2nd dim.");             
    strcpy(dialog.radio[2][1],"Linear");             
    strcpy(dialog.radio[2][2],"Logarithmic");
    strcpy(dialog.radio[2][3],"2nd-order poly.");
    strcpy(dialog.radio[2][4],"Distance from 0");

    strcpy(dialog.boxes[0],"Data");
    strcpy(dialog.boxes[1],"Image no.");
    strcpy(dialog.boxes[2],"Units");
    strcpy(dialog.boxes[3],"1st dim. calib units");
    strcpy(dialog.boxes[4],"2nd dim. calib units");
    dialog.boxtype[0]=LABEL;
    dialog.boxtype[1]=INTCLICKBOX; 
    dialog.boxtype[2]=LABEL;
    dialog.boxtype[3]=STRING; 
    dialog.boxtype[4]=STRING; 

    sprintf(dialog.answer[1],"%d",ino);
    dialog.boxmin[1]=0; dialog.boxmax[1]=g.image_count;
    strcpy(dialog.answer[3],z[ino].cal_title[0]);
    strcpy(dialog.answer[4],z[ino].cal_title[1]);

    dialog.radioset[0] = max(1, z[ino].cal_dims + 1);
    if(z[ino].cal_log[0]<=0)dialog.radioset[1]=1;
    if(z[ino].cal_log[0]==1)dialog.radioset[1]=2;
    if(z[ino].cal_log[0]==2)dialog.radioset[1]=3;
    if(z[ino].cal_log[0]==3)dialog.radioset[1]=4;
    if(z[ino].cal_log[1]<=0)dialog.radioset[2]=1;
    if(z[ino].cal_log[1]==1)dialog.radioset[2]=2;
    if(z[ino].cal_log[1]==2)dialog.radioset[2]=3;
    if(z[ino].cal_log[1]==3)dialog.radioset[2]=4;

    dialog.radiono[0]=4;
    dialog.radiono[1]=5;
    dialog.radiono[2]=5;
    for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;

    dialog.noofradios=3;
    dialog.noofboxes=5;
    dialog.helptopic=22;  
    dialog.keep=0;
    dialog.want_changecicb = 0;
    dialog.f1 = calibcheck;
    strcpy(dialog.path,".");
    dialog.transient = 1;

    //////////////////

    dialogbox(dialog);

    //////////////////

    if(g.getout){ g.getout=0; return; }

    ino = atoi(dialog.answer[1]);
    z[ino].cal_dims = dims = dialog.radioset[0] - 1; 
    z[ino].cal_log[0] = dialog.radioset[1] -1; 
    z[ino].cal_log[1] = dialog.radioset[2] -1; 
    
    dialog.answer[3][18]=0;   // Truncate string to 18 chars
    dialog.answer[4][18]=0;   // Truncate string to 18 chars
    strcpy(z[ino].cal_title[0], dialog.answer[3]);
    strcpy(z[ino].cal_title[1], dialog.answer[4]);

    //------------//

    data.f = new double[300];
    data.g = new double[300];
    data.x = new int[300];
    data.y = new int[300];
    data.n = 0;
    data.dims      = dims;
    data.ino       = ino;
    data.nmax      = 300;
    data.width     = 0;
    data.type      = 0;
    data.duration  = TEMP;
    data.wantpause = 0;
    data.win       = 0; // Calculate automatically

    for(k=0;k<300;k++){ data.x[k]=1;  data.y[k]=1;  data.f[k]=1;  data.g[k]=1; }
    bezier_curve_start(&data);
    if(g.getout) 
        bezier_curve_end(&data);
    else
    {   
        while(g.bezier_state==CURVE)
             XtAppProcessEvent(XtWidgetToApplicationContext(g.drawing_area),XtIMAll);
        calibrate_calc(&data, 0);      // This calls bezier_curve_end
        strcpy(title,"Calibration values");
        label = new char *[data.n];
        answer = new char *[max(1,dims)*data.n];
        if(dims<=1)
        {  for(k=0; k<data.n; k++)
           {   label[k] = new char[128];
               answer[k] = new char[128];
               sprintf(label[k], "Calib. of x=%d y=%d", data.x[k], data.y[k]);
               calibratepixel(data.x[k], data.y[k], ino, 0, xx, tempstring);  // Take a guess
               sprintf(answer[k], "%d", data.x[k]);
               data.f[k] = xx;
           }
           getstrings(title, label, answer, 1, data.n, 128);
           for(k=0; k<data.n; k++) data.f[k] = atof(answer[k]);
           calibrate_calc(&data, 0);
        }else    // dims==2
        {  count = 0;
           for(k=0; k<data.n; k++)
           {   label[k] = new char[128];             
               sprintf(label[k], "x calib. of x=%d y=%d", data.x[k], data.y[k]);
               calibratepixel(data.x[k], data.y[k], ino, 0, xx, tempstring);
               answer[count] = new char[128];
               sprintf(answer[count], "%d", data.x[k]);
               data.f[k] = xx;
               count++;
           }
           for(k=0; k<data.n; k++)
           {   calibratepixel(data.x[k], data.y[k], ino, 1, yy, tempstring);
               answer[count] = new char[128];
               sprintf(answer[count], "%d", data.y[k]);
               data.g[k] = yy;
               count++;
           }
           getstrings(title, label, answer, 2, data.n, 128);
           for(k=0; k<data.n; k++)
           {   data.f[k] = atof(answer[k]);
               data.g[k] = atof(answer[k+data.n]);
           }
           calibrate_calc(&data, 0);
           calibrate_calc(&data, 1);
        }
        for(k=0;k<data.n;k++) delete[] label[k];
        for(k=0;k<count;k++)  delete[] answer[k];        
        delete[] label;
        delete[] answer;
        sprintf(tempstring,"Image %d calibrated",ino);
        message(tempstring);
    }
    delete[] data.g;
    delete[] data.f;
    delete[] data.y;
    delete[] data.x;
    g.inmenu--;
    return;
}


//--------------------------------------------------------------------------//
//  calibcheck                                                              //
//--------------------------------------------------------------------------//
void calibcheck(dialoginfo *a, int radio, int box, int boxbutton)
{
    box=box; boxbutton=boxbutton;
    if(radio==0) 
    {    if(a->radioset[0]==3)            // 2d
         {   XtSetSensitive(a->radiobox[2], True);
             XtSetSensitive(a->boxwidget[4][0], True);
         }else                            // 1d
         {   XtSetSensitive(a->radiobox[2], False);
             XtSetSensitive(a->boxwidget[4][0], False);
         }
    }
}


//-------------------------------------------------------------------------//
// calibrate_calc - MW as function of x or y                               //
// If an image is selected, the image calib. is position-independent.      //
// If calibrating background, calib. is fixed.                             //
//-------------------------------------------------------------------------//
void calibrate_calc(XYData *data, int dim)
{
    double x1,y1,m1,m2,b2;
    double *x, *y, *d, q[20], *calib;   
    int n, k, ino, calib_log;
    g.getout=0;

    q[0]=0;
    q[1]=0;
    q[2]=0;
    ino = data->ino;
    n = data->n;    
    calib_log = z[ino].cal_log[dim];
    if(dim==0) calib = data->f;
    else calib = data->g;

    if(calib_log==1)
    {  for(k=0;k<n;k++) 
       {  
          if(calib[k]>0) calib[k]=log10(calib[k]);
          else{ message("Error: negative value in logarithm",ERROR);calib[k]=0;}
       }
    }      
    x = new double[n+1];
    y = new double[n+1];
    d = new double[n+1];

    //// Correlate calib[] with distance from line CD perpendicular to AB.  
    ////                   D       slope of AB = m1                         
    ////                 A/        slope of AC = -1/m1 = m2                 
    ////                 /\        intercept of AB = b1                     
    ////               F/  \E      intercept of CD = b2                     
    ////               /\./ \      A is the origin point, obtained from     
    ////              C  P   B      linear regression of x's and y's.       

    for(k=0;k<n;k++)
    {   x[k] = data->x[k] - z[ino].xpos;  
        y[k] = data->y[k] - z[ino].ypos; 
         ////  For 2-dim calib, force axes to be almost vertical & horizontal
         ////  but not perfect to avoid slope of infinity.
        if(z[ino].cal_dims>=2)  
           if(dim==0) y[k]=0.0001+k*.0001; else x[k]=0.0001+k*.0001;
    }   

    linreg(x,y,n,1,q);       // Calculate the line AB
    m1 = q[1]; if(m1==0) m1=1e-6;
    ////  b1 = q[0];         // Slope of line segment
    x1 = data->x[0];         // Origin point x1,y1
    y1 = q[0] + q[1]*x1;     // Origin point
    m2 = -1/m1;              // Slope of perp. line thru origin point
    b2 = y1-m2*x1;           // Intercept of CD thru origin point

    for(k=0;k<n;k++)         // For each calibr. point P, calc. dist.to AB.
    {   if(data->dims==1) 
           d[k] = pointtoline(x[k],y[k],m2,b2); // Length of FP
        else if(dim==0)
           d[k] = x[k];
        else
           d[k] = y[k];
    }

    if(calib_log==2)
      linreg(d,calib,n,2,q); // Correlate data points with perp.distances
    else
      linreg(d,calib,n,1,q); // Correlate data points with perp.distances

    z[ino].cal_slope[dim] = m2;   // slope of perp.line
    z[ino].cal_int[dim]   = b2;   // intercept of perp.line
    z[ino].cal_q[dim][0] = q[0]; 
    z[ino].cal_q[dim][1] = q[1]; 
    z[ino].cal_q[dim][2] = q[2]; 
    z[ino].cal_xorigin[dim] = (int)x1;
    z[ino].cal_yorigin[dim] = (int)y1;

    delete[] d;
    delete[] y;
    delete[] x;
    return;
}


//-------------------------------------------------------------------------//
// calibratepixel - returns calibrated value from x,y of a pixel, and      //
// also puts it in the string 'tempstring'.   ux and uy are raw screen     //
// coordinates.  ino is image no, dim must be 0 or 1.                      //
//-------------------------------------------------------------------------//
void calibratepixel(int ux, int uy, int ino, int dim, double &c, char* answerstring)
{
    double cdist=0;
    double slope,intercept,q0=0,q1=0,q2=0;
    int calib_log;
    int cal_xo=0,cal_yo=0;
    if(ino<0) return;
    if(!between(dim,0,1)){ perror("Arkkkk!!!\n"); return; } 
    if(dim > z[ino].cal_dims){ c=0; strcpy(answerstring,"Uncalibrated"); return; }
    c = 0.0;
    slope = z[ino].cal_slope[dim];
    intercept = z[ino].cal_int[dim];
    calib_log = z[ino].cal_log[dim];
    if(calib_log>-1)
    {   q0 = z[ino].cal_q[dim][0];
        q1 = z[ino].cal_q[dim][1];
        q2 = z[ino].cal_q[dim][2];
        cal_xo = z[ino].cal_xorigin[dim];
        cal_yo = z[ino].cal_yorigin[dim];
        ux -= z[ino].xpos;
        uy -= z[ino].ypos;
    }
    answerstring[0]=0;

    if(z[ino].cal_dims==0)
       cdist = ux;
    else if(z[ino].cal_dims==1) 
       cdist = pointtoline((double)ux, (double)uy, slope, intercept);
    else if(dim==0)
       cdist = ux;
    else
       cdist = uy;

    switch(calib_log) 
    { 
         case 3:          // 2d scaling
            ux = ux - cal_xo;
            uy = uy - cal_yo;
            c = q0 + fabs(q1 * sqrt((double)(ux*ux + uy*uy)));
            gcvt(c,g.signif,answerstring);
            break;
         case 2:
            c = q0 + cdist*q1 + cdist*cdist*q2;
            gcvt(c,g.signif,answerstring);
            break;
         case 1:  
            c = q0 + cdist*q1;
            if(c>19)strcpy(answerstring,"Infinity");
            else if(c<-10)strcpy(answerstring,"Zero");
            else
            {   c=pow(10.0,c);
                gcvt(c,g.signif,answerstring);
            }
            break;
         case 0:
            c = q0 + cdist*q1;
            gcvt(c,g.signif,answerstring);
            break;
         case -1:
            strcpy(answerstring,"Uncalibrated");
            break;
    } 
}





//-------------------------------------------------------------------------//
// pointtoline - Calculates distance from point x,y to line y=mx+b         //
//-------------------------------------------------------------------------//
double pointtoline(double x, double y, double m, double b)
{
    double m2,b2,x2=0,y2,d;
    if(m==0) m=1e-6;
    m2 = -1/m;              // slope of perpendicular
    b2 = y-m2*x;            // y-int.of perpendicular
    if(m2!=m)
      x2 = (b-b2)/(m2-m);   // intersection x coord.
    y2 = m2*x2+b2;          // intersection y coord.
    d = sqrt((x-x2)*(x-x2)+(y-y2)*(y-y2));  // distance
    if( ((y-y2)/m)+x-x2 > 0) return(d);     // if x is right of line, d>0
    return(-d);
}



//-------------------------------------------------------------------------//
// densitometry                                                            //
// Entry point for spot densitometry                                       //
// All measurements have to be in the same image or on the main window     //
//-------------------------------------------------------------------------//
void densitometry(void)
{
  if(memorylessthan(16384)){  message(g.nomemory,ERROR); return; } 
  if(ci<0){ message("No images selected"); return; }

  const int DENS_MAX=10000;
  static int hitdefinesize=0, hitdens=0;
  static ScanParams osp;

  ScanParams sp;   // Don't make sp static in case user selects densitometry twice.
  Widget browser, *fuzzywidgetlist=0, fuzzyform=0, *widgetlist;
  
  listinfo fuzzyli;
  widgetlist = new Widget[100];
  fuzzywidgetlist = new Widget[100];
  char tempstring[1024];
  double vvv=0.0, THRESH=0.02,
      fill1=0, fill2=0, fillcolor1, fillcolor2;
  double *results;
  results = new double[FUZZY_ITEMS];
  int alloc2=0, alloc3=0, alloc4=0,  
      bbb, bpp, button=0, ino=-1, invert=0, image_moved=0,
      fuzzywc=0, iii, i, j, k, oldx, oldy,
      oino=-2, omainxpos=0, omainypos=0, oxmin=0, oxmax=0, oymin=0, oymax=0,
      oxpos=0, oypos=0, status=OK, sm, wc=0,
      x, y, x1=0, x2=0, y1=0, y2=0, xhalf=0, yhalf=0,
      xsize, ysize, xmin, xmax, ymin, ymax, xstart=0, ystart=0;

  listinfo l;       
  l.info = new char*[10000];       // Max. of 10000 densitometric measurements
  l.size = 0;
  l.title = new char[256];
  l.transient   = 0;
  l.wantfunction = 0;
  l.autoupdate   = 0;
  g.getout=0;

  g.state=DENSITOMETRY;

  uint **savebox;
  uint *savebox_1d;                 // 1d alias for Xlib

  if(!hitdens)                     // Default parameters
  {     sp.autosave           = 0; // Not used
                                   // Default background density
        sp.bdensity           = 1.0;   
        sp.fdensity           = 0;
        sp.bsignal            = 0;
        sp.fsignal            = 0;
        sp.netsignal          = 0;
        sp.netdensity         = 0;
        sp.leavemarked        = 0; // Flag if want to leave area marked
        sp.invert             = 1; // 1= max signal = od of 0
        sp.gamma_inverted     = 0; // 1=gamma table is highest for pixel=0
        sp.scanwidth          = 1; // Not used
        sp.selection_method   = 2; // Method for selecting coordinates
        sp.snapthreshold      = 0; // Not used
        sp.compensate         = 1; // Want pixel compensation
        sp.autobkg            = 0; // Want automatic background calculation
        sp.wantfixedwidth     = 0; // Not used
        sp.wantpause          = 0; // Not used
        sp.wantplot           = 0; // Not used
        sp.calfac             = 1; // Calibration factor
        sp.skip               =-2; // Image to skip (not used)
        sp.wantcolor          = 0; // Color densitometry
        sp.diameter           = 12;
        sp.scan          = g.scan; // Where to put the results
  }else
  {     sp.autosave           = osp.autosave;
        sp.bdensity           = osp.bdensity;
        sp.bsignal            = osp.bsignal;
        sp.leavemarked        = osp.leavemarked;
        sp.invert             = osp.invert;
        sp.gamma_inverted     = osp.gamma_inverted;
        sp.scanwidth          = osp.scanwidth;
        sp.selection_method   = osp.selection_method;
        sp.snapthreshold      = osp.snapthreshold;
        sp.compensate         = osp.compensate;
        sp.autobkg            = osp.autobkg;
        sp.wantfixedwidth     = osp.wantfixedwidth;
        sp.wantpause          = osp.wantpause;
        sp.wantplot           = osp.wantplot;
        sp.calfac             = osp.calfac;
        sp.skip               = osp.skip;
        sp.wantcolor          = osp.wantcolor;
        sp.diameter           = osp.diameter;
        sp.scan               = osp.scan;
  }
  sp.count       = 0;              // No. of measurements made so far
  sp.pixels      = 0;              // Total no. of pixels
  sp.fsignal     = 0.0;            // Signal density value
  sp.od = new int[256];            // Gamma table
  if(sp.od != NULL) alloc3=1;
  
  ////  This can use a lot of memory but is needed to restore the 
  ////  background in case a flood fill is bigger than expected.
  ////  If no images are in separate windows, main_xsize and
  ////  main_ysize could be used instead of the screen size.
  ////  If measurements are restricted to the current image, the image size
  ////  could be used. If only 8 bit images are present, savebox could be 
  ////  a uchar.
  ////  If pixels off the edge of the screen are to be included in the 
  ////  calculations, this array must be even bigger.
                                   
  xsize = g.xres+2; ysize=g.yres+2;
  savebox_1d = (uint*)malloc(xsize*ysize*sizeof(uint));
  savebox = make_2d_alias(savebox_1d,xsize,ysize); 
  if(savebox==NULL){ message(g.nomemory,ERROR); g.getout=1;} else alloc2=1;

                                        // Boolean array for densitometry 
  array<char> dbool(xsize+2, ysize+2);
  if(!dbool.allocated)                  // Only need to check the last malloc.
  {   message(g.nomemory,ERROR);
      g.getout=1;
  }

  if(!g.getout)
  {  
      get_densitometry_parameters(&sp,1);
      if(sp.autobkg) 
          fuzzyform = open_fuzzy_list(&fuzzyli, fuzzywidgetlist, fuzzywc);
      switch(sp.selection_method)
      {   case 1:
            message("Move to center of area\npress left mouse button\nESC or Right button to stop\n");
            break;
         case 2: message("Select area to measure"); break;
         case 3: message("Select size to measure"); break;
         case 4: 
         case 6: if(sp.autobkg)  
                 {  ////  Must be all on one line (Digital compiler can't handle '\')
                    sprintf(tempstring,"Warning: %cAuto background%c and %cirregular\nregion%c were both selected. Your area selections\nmust include similar numbers of foreground and\nbackground pixels.", '"', '"', '"', '"'); 
                    if(g.want_messages > 1) message(tempstring, WARNING);  
                 }
                 break;
      }
  }
   
  ////  Get the center point of the spot to measure, its bpp & image no.   
  ////  Use the image no. at center of selected area for all calculations. 

  while(!g.getout)
  {
         button=0;
         if(g.getout)break;
 
         ////  Recheck the parameters in case user changed something
         get_densitometry_parameters(&sp,0);
         if(sp.autobkg && fuzzyform==0) 
         {   fuzzyform = open_fuzzy_list(&fuzzyli, fuzzywidgetlist, fuzzywc);
             hitdefinesize = 0;
         }

         ////  Get boundary or starting point & ino
         sm = sp.selection_method;
         if(sm==2 || (sm==3 && !hitdefinesize))
         {    button = getbox(x1,y1,x2,y2); 
              xhalf = (x2-x1)/2;
              yhalf = (y2-y1)/2;              
              ino = whichimage(xhalf, yhalf, bpp);
              if(sp.selection_method==3) 
              {    sprintf(tempstring,
                     "Ready to begin densitometry\non %d x %d regions",x2-x1,y2-y1);
                   message(tempstring);
                   hitdefinesize = 1;
              }
         }
         if(sm==1 || sm==5 || (sm==3 && hitdefinesize))
         {    button=getpoint(x,y);
              ino = whichimage(x, y, bpp);
         }  
         if(sm==4)
         {    select_area(SELECT_MODE);
              x1 = g.ulx;
              x2 = g.lrx;
              y1 = g.uly;
              y2 = g.lry;
              xhalf = (x2+x1)/2;
              yhalf = (y2+y1)/2;              
              ino = whichimage(xhalf, yhalf, bpp);
         }
         if(sm==6)
         {    if(hitdefinesize) 
              {    button = getpoint(x,y);
                   shift_selected_area(x-xstart, y-ystart);
                   xstart = x;
                   ystart = y;
                   x1 = g.ulx;
                   x2 = g.lrx;
                   y1 = g.uly;
                   y2 = g.lry;
              }else
              {    select_area(SELECT_MODE);
                   xstart = (g.lrx + g.ulx) / 2;
                   ystart = (g.lry + g.uly) / 2;
                   hitdefinesize = 1;
                   x1 = g.ulx;
                   x2 = g.lrx;
                   y1 = g.uly;
                   y2 = g.lry;
                   x = (x2+x1)/2;
                   y = (y2+y1)/2;              
                   message("Ready to begin densitometry\non fixed shape regions");
              }
              ino = whichimage(x, y, bpp);
         }
         g.state=DENSITOMETRY;
         
         ////  Recheck the parameters in case user changed something
         get_densitometry_parameters(&sp,0);
         if(sp.autobkg && fuzzyform==0) 
         {   fuzzyform = open_fuzzy_list(&fuzzyli, fuzzywidgetlist, fuzzywc);
             hitdefinesize = 0;
         }
         if(g.getout || button==2) break;
         if(keyhit()) if(getcharacter()==27) break;

         ////  Calculate boundaries
         if(ino>=0 && z[ino].shell) 
         {    xmin = z[ino].xpos;
              ymin = z[ino].ypos;
              xmax = z[ino].xpos + z[ino].xsize;
              ymax = z[ino].ypos + z[ino].ysize;            
              xmin=min(g.xres-g.main_xpos, max(xmin,-g.main_xpos));
              xmax=min(g.xres-g.main_xpos, max(xmax,-g.main_xpos));
              ymin=min(g.yres-g.main_ypos, max(ymin,-g.main_ypos));
              ymax=min(g.yres-g.main_ypos, max(ymax,-g.main_ypos));
         }else 
         {    xmin = 0;
              ymin = 0;
              xmax = g.main_xsize;
              ymax = g.main_ysize;
         }

         ////  Save the background the 1st time, or if any windows were moved.   
         ////   'inmenu' must be 0 in order to get the true pixel value. This is 
         ////   especially important if image is not at screen bpp.            
         ////  Density calculations are done at the bpp of the image at the      
         ////    selected point.                                                 

         image_moved=0; 
         if(ino>=0){ if(z[ino].xpos!=oxpos ||  z[ino].ypos!=oypos) image_moved=1; }
         if(xmax!=oxmax || ymax!=oymax || xmin!=oxmin || ymin!=oymin) image_moved=1;
         if(g.main_xpos!=omainxpos || g.main_ypos!=omainypos) image_moved=1;

         if(ino!=oino || image_moved)
            for(j=ymin;j<=ymax;j++)
            for(k=xmin;k<=xmax;k++)
                savebox[j-ymin][k-xmin] = readpixelonimage(k,j,bbb,iii);
           
         oino=ino;
         oxmax=xmax; oxmin=xmin; oymax=ymax; oymin=ymin;
         omainxpos=g.main_xpos; omainypos=g.main_ypos;
         if(ino>=0){ oxpos=z[ino].xpos; oypos=z[ino].ypos; }
         
         ////  Check gamma table validity, determine whether it is inverted, and  
         ////  fill global array 'od[]'.                                        

         status = check_gamma_table(&sp, ino, bpp);
         if(status!=OK) break;
         invert = sp.invert;
         if(sp.gamma_inverted) invert = 1-invert;
         oldx = x; oldy = y;
         sp.pixels = 0; sp.fsignal = 0;

         ////  Start densitometry                                               
         ////  Measure total signal and number of pixels in spot          

         switch(sp.selection_method)
         {  case 1:  // Automatic

               ////  Colors to use in painting the spot                       
               ////  The two fill colors must differ by at least 2.       

              if(sp.invert)
               {  fillcolor1 = 1.0;                  // Color to fill with on 1st pass
                  fillcolor2 = 0.0;                  // Color to fill with on 2nd pass
                  fill1      = 0.98;                 // Minimum border color
                  fill2      = sp.bdensity;          // Maximum border color
               }else
               {  fillcolor1 = 0.0;                  // Color to fill with on 1st pass
                  fillcolor2 = 1.0;                  // Color to fill with on 2nd pass
                  fill1      = 0.02;                 // Minimum border color
                  fill2      = sp.bdensity;          // Maximum border color
               }

                          ////  Use fill to mark the area to measure.
                          ////  x1,y1,x2,y2 are passed by reference,
                          ////  and return the x&y min and max actually filled.

               fill(x,y,x1,y1,x2,y2,fillcolor1,fill1,fill2);

                          ////  Initialize the Boolean array for each pixel

               for(j=y1;j<=y2;j++)for(i=x1;i<=x2;i++) dbool.p[j-ymin][i-xmin]=0;

                          ////  Store 1 bit in array 'b' if the pixel is fillcolor1.

               for(i=x1;i<=x2;i++)
                  for(j=y1;j<=y2;j++)
                     if(fabs((double)readpixelonimage(i,j,bpp,ino)/
                             (double)g.maxvalue[bpp] - fillcolor1) < THRESH) 
                         dbool.p[j-ymin][i-xmin] = 1;                  

               x=oldx; y=oldy;

                          ////  fill area again with a different color guaranteed
                          ////  to overwrite the earlier fillcolor1 pixels.
                          ////  x1,y1,x2,y2 are passed by reference,
                          ////  and return the x&y min and max actually filled.

               fill(x,y,x1,y1,x2,y2,fillcolor2,fill1,fill2);
                          ////  If pixel didn't change, it is outside the area
                          ////  because it was already the fill color, so set its 
                          ////  bit to 0 if it was a 1.  There should not be any
                          ////  of these. 

               for(i=x1;i<=x2;i++)
                  for(j=y1;j<=y2;j++)
                     if(dbool.p[j-ymin][i-xmin]==1)
                        if(fabs((double)readpixelonimage(i,j,bpp,ino)/
                                (double)g.maxvalue[bpp] - fillcolor2) >= THRESH)
                           dbool.p[j-ymin][i-xmin] = 0;                      

                          //// Measure signal and restore bkgd at the same time.

               sp.pixels = 0;
               for(i=x1;i<=x2;i++)
                  for(j=y1;j<=y2;j++)
                     if(dbool.p[j-ymin][i-xmin]==1) 
                     {    sp.pixels++;
                          setpixelonimage(i,j,savebox[j-ymin][i-xmin],SET,bpp);
                          vvv = pixeldensity(&sp,i,j,invert);
                          if(sp.leavemarked) setpixelonimage(i,j,0,SET,bpp);
                          sp.fsignal += vvv;
                      } 

                          ////  Repair image in case it was remapped

               rebuild_display(ino);
               redraw(ino);
               break;
            case 2:       // Manual
               for(i=x1;i<x2;i++)
               for(j=y1;j<y2;j++)
               {    setpixel(i,j,g.maxcolor,XOR);
                    sp.pixels++;
                    sp.fsignal += pixeldensity(&sp,i,j,invert);
               }
               if(!sp.leavemarked)
               for(i=x1;i<x2;i++)
               for(j=y1;j<y2;j++)
                    setpixel(i,j,g.maxcolor,XOR);
               break;
            case 3:      // Fixed
               if(hitdefinesize)   
               {  x1 = x-xhalf;
                  x2 = x+xhalf;
                  y1 = y-yhalf;
                  y2 = y+yhalf;
                  for(i=x1;i<x2;i++)
                  for(j=y1;j<y2;j++)
                  if((i>xmin)&&(j>ymin)&&(i<xmax)&&(j<ymax))
                  {   setpixel(i,j,g.maxcolor,XOR);
                      sp.pixels++;
                      sp.fsignal += pixeldensity(&sp,i,j,invert);
                  }

                  if(!sp.leavemarked)
                  {   for(i=x1;i<x2;i++)
                      for(j=y1;j<y2;j++)
                      if((i>xmin)&&(j>ymin)&&(i<xmax)&&(j<ymax))
                          setpixel(i,j,g.maxcolor,XOR);
                  }
               }
               break;
            case 4:      //  Non-rectangular     
               for(i=x1;i<x2;i++)
               for(j=y1;j<y2;j++)
               {    if(!inside_irregular_region(i,j)) continue;
                    setpixel(i,j,g.maxcolor,XOR);
                    sp.pixels++;
                    sp.fsignal += pixeldensity(&sp,i,j,invert);
               }
#ifdef LINUX
               usleep(100000);
#endif
               if(!sp.leavemarked)
               for(i=x1;i<x2;i++)
               for(j=y1;j<y2;j++)
               {    if(!inside_irregular_region(i,j)) continue;
                    setpixel(i,j,g.maxcolor,XOR);
               }
               break;
            case 5:       // Circle
               x1 = x-sp.diameter/2;
               x2 = x+sp.diameter/2;
               y1 = y-sp.diameter/2;
               y2 = y+sp.diameter/2;
               for(i=x1;i<x2;i++)
               for(j=y1;j<y2;j++)
               if(insidecircle(i,j,x,y,sp.diameter))
               {   setpixel(i,j,g.maxcolor,XOR);
                   sp.pixels++;
                   sp.fsignal += pixeldensity(&sp,i,j,invert);
               }
               if(!sp.leavemarked)
               {   for(i=x1;i<x2;i++)
                   for(j=y1;j<y2;j++)
                      if(insidecircle(i,j,x,y,sp.diameter))
                          setpixel(i,j,g.maxcolor,XOR);
               }
               break;
            case 6:
               for(i=x1;i<x2;i++)
               for(j=y1;j<y2;j++)
               {    if(!inside_irregular_region(i,j)) continue;
                    setpixel(i,j,g.maxcolor,XOR);
                    sp.pixels++;
                    sp.fsignal += pixeldensity(&sp,i,j,invert);
               }
#ifdef LINUX
               usleep(100000);
#endif
               if(!sp.leavemarked)
               for(i=x1;i<x2;i++)
               for(j=y1;j<y2;j++)
               {    if(!inside_irregular_region(i,j)) continue;
                    setpixel(i,j,g.maxcolor,XOR);
               }
               break;
          }

         ////  Density of the background 
         ////  Calculate results         

          if(sp.pixels<2)
          {    message("Area too small");
          }else if(sp.autobkg) 
          { 
               background(&sp,x1,y1,x2,y2,invert,results,&fuzzyli);
               update_fuzzy_list(&fuzzyli);
               sp.pixels = (int)results[18];        // Total no. of pixels
               sp.bdensity = results[6];            // Bkg. density
               sp.bsignal  = sp.bdensity*sp.pixels; // Bkg. total signal
               sp.fsignal  = results[10];           // Band total signal
               sp.fdensity = results[11];           // Band density  
               sp.netdensity = results[14];         // Density spot - background
               sp.fpixels    = (int)results[22];    // Total pixels in spot
               sp.netsignal  = sp.netdensity*sp.fpixels; // Spot density * spot pixels 
          }else 
          { 
              if(sp.pixels==0)
              {    sp.fdensity = 0;
                   sp.fsignal  = 0;
                   sp.bsignal  = 0;
              }else
                   sp.fdensity = sp.fsignal/sp.pixels;
              if(bpp==8)
                   sp.bdensity = (double)(sp.od[(int(sp.bdensity*255.0))])/255.0;
              else if(invert)
                   sp.bdensity = 1.0 - sp.bdensity; 

              sp.bsignal  = sp.bdensity * sp.pixels;
              sp.netsignal = sp.fsignal - sp.bsignal;
              sp.netdensity = sp.fdensity - sp.bdensity;
              sp.fpixels  = sp.pixels;                    // Pixels in spot
          }

         //--------------------------------------------------------------------//
         // This has to be in a separate function because gcc can't handle     //
         // long functions.                                                    //
         //--------------------------------------------------------------------//
                        
         sp.count++;
         if(sp.count >= DENS_MAX-1)
         {   message("Data table is full",WARNING);
             g.getout=1;
         }
         show_densitometry_results(&sp, &l, &browser, widgetlist, &wc);
         alloc4 = 1;  // browser allocated
         if(g.getout) break;
         
  }


  osp.autosave           = sp.autosave;
  osp.bdensity           = sp.bdensity;
  osp.bsignal            = sp.bsignal;
  osp.leavemarked        = sp.leavemarked;
  osp.invert             = sp.invert;
  osp.gamma_inverted     = sp.gamma_inverted;
  osp.scanwidth          = sp.scanwidth;
  osp.selection_method   = sp.selection_method;
  osp.snapthreshold      = sp.snapthreshold;
  osp.compensate         = sp.compensate;
  osp.autobkg            = sp.autobkg;
  osp.wantfixedwidth     = sp.wantfixedwidth;
  osp.wantpause          = sp.wantpause;
  osp.wantplot           = sp.wantplot;
  osp.calfac             = sp.calfac;
  osp.skip               = sp.skip;
  osp.wantcolor          = sp.wantcolor;
  osp.diameter           = sp.diameter;
  osp.scan               = sp.scan;

  ////  Free memory in get_densitometry parameters
  get_densitometry_parameters(&sp,RESET);

  if(alloc3) delete[] sp.od;
  if(alloc2) 
  {   free(savebox_1d); 
      free(savebox);
  }
  message("Densitometry mode finished");  
  if(alloc4) closewindow(&l, browser, widgetlist, wc);
  delete[] widgetlist;
  g.state=NORMAL;
  g.getout=0;
  hitdens=1;
  hitdefinesize = 0;
   
  for(k=0;k<l.size;k++) delete[] l.info[k]; 
  delete[] l.title;
  delete l.info;
  if(fuzzyform) close_fuzzy_list(&fuzzyli, fuzzyform, fuzzywidgetlist, fuzzywc);
  delete[] fuzzywidgetlist;
  delete[] results;
  return;
}


//-------------------------------------------------------------------------//
// background - return a bkg value 0..1                                    //
//-------------------------------------------------------------------------//
void background(ScanParams *sp, int x1, int y1, int x2, int y2, int invert,
   double *results, listinfo *li)
{
    int i,j,k,count;
    double *pixels;
    count = (y2-y1)*(x2-x1);
    pixels = new double[count];
    k=0;
    for(j=y1;j<y2;j++)
    for(i=x1;i<x2;i++)
         pixels[k++] = pixeldensity(sp,i,j,invert);
    cluster(pixels, count, 2, 150, results, li);
    delete[] pixels;
    return;
}


//-------------------------------------------------------------------------//
// check_gamma_table                                                       //
//-------------------------------------------------------------------------//
int check_gamma_table(ScanParams *sp, int ino, int bpp)
{
   int highest, k, status=OK;
   int invert_table=0;
   if(bpp==8)
   {    sp->gamma_inverted = 0;
        highest=255;
        if(ino>=0)
        {   for(k=0;k<=255;k++)
                 if(z[ino].gamma[k] > z[ino].gamma[highest]) highest=k;
            if(highest!=255 && highest!=0)
            {    message("Error: Bad OD table\nTurn pixel compensation off",ERROR);
                 status = ERROR;
            }
            if(highest != 255) sp->gamma_inverted=1;
        }
        
        if(sp->gamma_inverted != sp->invert) invert_table=1;
        if(invert_table)
        {   if(sp->compensate && ino>=0)
            {   for(k=0;k<=255;k++)  sp->od[k] = z[ino].gamma[255-k];  }
            else
            {   for(k=0;k<=255;k++) sp->od[k] = 255-k; }       
        }else
        {   if(sp->compensate && ino>=0)
            {   for(k=0;k<=255;k++)  sp->od[k] = z[ino].gamma[k];  }
            else
            {   for(k=0;k<=255;k++) sp->od[k] = k; }
        }
   }
   return status;
}         



//-------------------------------------------------------------------------//
// pixeldensity - returns optical density of pixel as a                    //
//          double between 0 and 1 where 1 is the max. possible density.   //
//          Corrects for bpp of the pixel.                                 //
//          Calling function must set the od table in sp to the gammas     //
//          appropriate for the image at x,y.                              //
//-------------------------------------------------------------------------//
double pixeldensity(ScanParams *sp, int x, int y, int invert)
{ 
   //// Correct for gamma if 8 bpp. This automatically inverts value if
   //// necessary because od table should already be inverted.
   //// For pixels > 8 bpp, invert the value here.
 
   int bpp,ino;
   uint value;
   value = readpixelonimage(x,y,bpp,ino,sp->skip);
   if(bpp==8) 
       value = sp->od[value];                           
   else if(invert) 
       value = (uint)g.maxvalue[bpp] - value;   
   return ((double)value)/((double)g.maxvalue[bpp]);    // Convert to double, 0..1 
}


//-------------------------------------------------------------------------//
// colorpixeldensity - returns optical density of rgb components of pixel  //
//      as a double between 0 and 1 where 1 is the max. possible density.  //
//      Corrects for bpp of the pixel.                                     //
//-------------------------------------------------------------------------//
void colorpixeldensity(ScanParams *sp, int x, int y, double &rdensity, 
   double &gdensity, double &bdensity, int invert)
{ 
   int bpp,ino,rr,gg,bb;
   uint value;
   value = readpixelonimage(x,y,bpp,ino,sp->skip);
   valuetoRGB(value,rr,gg,bb,bpp);
   if(invert) 
   {    rr = (uint)g.maxred[bpp] - rr;   
        gg = (uint)g.maxgreen[bpp] - gg;   
        bb = (uint)g.maxblue[bpp] - bb;   
   }
   rdensity = (double)rr/(double)g.maxred[bpp];
   gdensity = (double)gg/(double)g.maxgreen[bpp];
   bdensity = (double)bb/(double)g.maxblue[bpp];
}


//-------------------------------------------------------------------------//
// show_densitometry_results                                               //
//-------------------------------------------------------------------------//
void show_densitometry_results(ScanParams *sp, listinfo *l, Widget *browser, 
   Widget *widgetlist, int *wc)
{
  char temp[220];
  char tempstring[220];
  int len=0, k;
 
  ////  If the 1st item, set up the title and 1st item in a listinfo struct,
  ////  then open a list window with no items.

  l->info[l->size] = new char[100];   // Deleted at end of densitometry()
  if(sp->count==1) 
  {     int lhelptopic=6;
        int crud=0;
        strcpy(l->info[0],  "No. Area    Total Signal  Spot Density  Total Bkgd. ");
        strcat(l->info[0],  "Bkgd.Density  Net density  Signal-Bkg. ");
        strcpy(l->title, "Densitometry results");
        l->size  = 0;
        l->itemstoshow = 10;
        l->firstitem   = 1;
        l->wantsort    = 0;
        l->wantsave    = 1;
        l->helptopic   = lhelptopic;
        l->allowedit   = 0;
        l->selection   = &crud;
        l->width       = 7 * (strlen(l->info[0]));
        *browser = openwindow(l, widgetlist, wc);
        additemtolist(l->widget, *browser, l->info[0], BOTTOM, 1);
  }
   
  ////  For other results, compose a string and append it to the browser.
  ////  Use gcvt instead of sprintf so the no. of significant
  ////  digits can be changed by the user
    
  itoa(sp->count,temp,10);
  len=strlen(temp); 

  strcat(temp,"  "); for(k=1;k<=3-len;k++) strcat(temp," ");
  ltoa(sp->pixels,tempstring,10);
  strcat(temp,tempstring);

  strcat(temp,"  ");len=strlen(temp); for(k=1;k<=13-len;k++) strcat(temp," ");
  gcvt(sp->fsignal,g.signif,tempstring);
  strcat(temp,tempstring);
  strcat(temp,"  ");len=strlen(temp); for(k=1;k<=27-len;k++) strcat(temp," ");
    
  gcvt(sp->fdensity,g.signif,tempstring);
  strcat(temp,tempstring);
  strcat(temp,"  ");len=strlen(temp); for(k=1;k<=41-len;k++) strcat(temp," ");

  gcvt(sp->bsignal,g.signif,tempstring);
  strcat(temp,tempstring);
  strcat(temp,"  ");len=strlen(temp); for(k=1;k<=53-len;k++) strcat(temp," ");

  gcvt(sp->bdensity,g.signif,tempstring);
  strcat(temp,tempstring);
  strcat(temp,"  ");len=strlen(temp); for(k=1;k<=67-len;k++) strcat(temp," ");    
   
  gcvt(sp->netdensity,g.signif,tempstring);
  strcat(temp,tempstring);
  strcat(temp,"  ");len=strlen(temp); for(k=1;k<=78-len;k++) strcat(temp," ");

  gcvt(sp->netsignal,g.signif,tempstring);
  strcat(temp,tempstring);
  strcat(temp,"  ");len=strlen(temp); for(k=1;k<=89-len;k++) strcat(temp," ");

  strcpy(l->info[l->size], temp);
  additemtolist(l->widget, *browser, temp, BOTTOM, 1);

  sp->pixels = 0;
  sp->fsignal = 0.0;
  l->size++;
}



//-------------------------------------------------------------------------//
// get_densitometry_parameters - must be called by densitometry() at least //
//    twice                                                                //
// Call with want_dialog = 1 to display dialog, then call with want_dialog //
//    = 0 to check for changes then want_dialog = RESET to close window.   //
//-------------------------------------------------------------------------//
void get_densitometry_parameters(ScanParams *sp, int want_dialog)
{
 int j,k;
 static Dialog *dialog;
 if(want_dialog==1)
 {  
      dialog = new Dialog;
      char temp[220];
      g.getout=0;
      dialog->needrepeat = 0;
      strcpy(dialog->title,"Spot densitometry");
      strcpy(dialog->radio[0][0],"Maximum signal");
      strcpy(dialog->radio[0][1],"Black");
      strcpy(dialog->radio[0][2],"White");
      strcpy(dialog->radio[1][0],"Area selection");
      strcpy(dialog->radio[1][1],"Automatic");
      strcpy(dialog->radio[1][2],"Manual (rectangular)");
      strcpy(dialog->radio[1][3],"Fixed size (rectangular)");
      strcpy(dialog->radio[1][4],"Manual (irregular)");
      strcpy(dialog->radio[1][5],"Manual (circle)");
      strcpy(dialog->radio[1][6],"Fixed shape (irregular)");
 
      if(sp->invert)     dialog->radioset[0]=1;
                    else dialog->radioset[0]=2;
      dialog->radioset[1] = sp->selection_method;
 
      dialog->radiono[0]=3;
      dialog->radiono[1]=7;
      dialog->radiono[2]=0;
 
      for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;
 
      strcpy(dialog->boxes[0],"Parameters");
      strcpy(dialog->boxes[1],"Bkgd. value");
      strcpy(dialog->boxes[2],"Calib. factor");
      strcpy(dialog->boxes[3],"Leave area marked");
      strcpy(dialog->boxes[4],"Pixel compensation");
      strcpy(dialog->boxes[5],"Auto bkgd.(fuzzy k means)");
      strcpy(dialog->boxes[6],"Diameter");

      dialog->boxtype[0]=LABEL;
      dialog->boxtype[1]=DOUBLECLICKBOX;
      dialog->boxtype[2]=STRING;
      dialog->boxtype[3]=TOGGLE;
      dialog->boxtype[4]=TOGGLE;
      dialog->boxtype[5]=TOGGLE;
      dialog->boxtype[6]=INTCLICKBOX;
 
      dialog->boxmin[1] = 0;  dialog->boxmax[1] = 1;
      dialog->boxmin[6] = 1;  dialog->boxmax[6] = 400;
 
      gcvt(sp->bdensity, g.signif, temp); strcpy(dialog->answer[1],temp);
      gcvt(sp->calfac, g.signif, temp); strcpy(dialog->answer[2],temp);
      if(sp->leavemarked==1) dialog->boxset[3]=1; else dialog->boxset[3]=0;
      if(sp->compensate==1)  dialog->boxset[4]=1; else dialog->boxset[4]=0;
      if(sp->autobkg==1)     dialog->boxset[5]=1; else dialog->boxset[5]=0;
      sprintf(dialog->answer[6], "%d", sp->diameter);
 
      dialog->noofradios=2;
      dialog->noofboxes=7;
      dialog->helptopic=6;
      dialog->transient=0;
      for(k=0;k<dialog->noofboxes;k++) dialog->wantfunc[k]=0;
      dialog->wantfunc[1]=1;
      dialog->keep=1;
      dialog->want_changecicb=0;
      dialog->f1 = null;
      dialog->transient = 0;
      strcpy(dialog->path,".");
 
      //////////////////////////////
 
      if(!g.getout) dialogbox(*dialog);
 
      //////////////////////////////

  }

  if(want_dialog != RESET)             ////  Check for new answers
  {   if(dialog->radioset[0]==1) sp->invert=1; else sp->invert=0;
      sp->selection_method = dialog->radioset[1];
      sp->bdensity = atof(dialog->answer[1]);
      sp->calfac = atof(dialog->answer[2]);
      if(dialog->boxset[3]) sp->leavemarked=1; else sp->leavemarked=0;
      if(dialog->boxset[4]) sp->compensate=1;  else sp->compensate=0;
      if(dialog->boxset[5]) sp->autobkg=1;     else sp->autobkg=0;
      sp->diameter = atoi(dialog->answer[6]);
  }else                              ////  Unmanage the dialog & free memory
  {   dialog->keep = RESET;
      dialogbox(*dialog);
      delete[] dialog;
  }
  
}
  
 
//-------------------------------------------------------------------------//
// scan_region - Entry point for strip densitometry.                       //
//-------------------------------------------------------------------------//
void scan_region(void)
{  
               /* It is dangerous to subtract 2 unsigned int's */
               /* so convert to int first                      */
   if(memorylessthan(16384)){  message(g.nomemory,ERROR); return; } 
   static ScanParams osp;
   ScanParams sp;                      
   static int hit=0;
   XYData data; 
   Dialog dialog;
   dialog.needrepeat = 0;
   int invert=0, j, k, aborted=0;
   char tempstring[128];    

   if(!hit)
   {    sprintf(sp.filename,"%s%s",g.startdir,"/1.scn");
        sp.autosave           = 0;   // Automatically save scan data
        sp.bdensity           = 0;   // Not used
        sp.count              = 0;   // No. of measurements made so far
        sp.leavemarked        = 0;   // Flag if want to leave area marked
        sp.invert             = 1;   // =1 if max value==black
        sp.gamma_inverted     = 0;   // 1=gamma table is highest for pixel=0
        sp.selection_method   = 1;   // Which method for getting coordinates
        sp.scanwidth          = 10;
        sp.snapthreshold      = 3;   // 
        sp.compensate         = 1;   // Want pixel compensation
        sp.autobkg            = 0;   // Automatically calculate background
        sp.wantfixedwidth     = 0;   // Flag if rectangle is fixed width
        sp.wantrepeat         = 0;
        sp.wantpause          = 0;   // 
        sp.wantplot           = 1;   // 1=plot scan in graph
        sp.calfac             = 1.0; // Calibration factor
        sp.skip               = -2;
        sp.wantcolor          = 0;   // 1=color scan 0=grayscale
        sp.diameter           = 12;
        sp.scan               = g.scan; // where to put the results
   }else
   {    strcpy(sp.filename, osp.filename);
        sp.autosave           = osp.autosave;
        sp.bdensity           = osp.bdensity;
        sp.count              = osp.count;
        sp.leavemarked        = osp.leavemarked;
        sp.invert             = osp.invert;
        sp.gamma_inverted     = osp.gamma_inverted;
        sp.selection_method   = osp.selection_method;
        sp.scanwidth          = osp.scanwidth;
        sp.snapthreshold      = osp.snapthreshold;
        sp.compensate         = osp.compensate;
        sp.autobkg            = osp.autobkg;
        sp.wantfixedwidth     = osp.wantfixedwidth;
        sp.wantrepeat         = osp.wantrepeat;
        sp.wantpause          = osp.wantpause;
        sp.wantplot           = osp.wantplot;
        sp.calfac             = osp.calfac;
        sp.skip               = osp.skip;
        sp.wantcolor          = osp.wantcolor;
        sp.diameter           = osp.diameter;
        sp.scan               = osp.scan;
   }

   sp.pixels      = 0;                   // Total no. of pixels
   sp.bdensity    = 0.0;                 // Background density value
   sp.fsignal     = 0.0;                 // Total signal density value
   sp.od = new int[256];                 // Gamma table
   if(sp.od==NULL){ message(g.nomemory,ERROR); return; }  

   data.x = new int[300]; if(data.x==NULL){ delete[] sp.od;
                                 message(g.nomemory,ERROR); return; }
   data.y = new int[300]; if(data.y==NULL){  delete[] sp.od; delete[] data.x; 
                                 message(g.nomemory,ERROR);  return;}  
   for(k=0;k<300;k++){ data.x[k]=0; data.y[k]=0; }
   
   dialog.needrepeat = 0;
   strcpy(dialog.title,"Strip densitometry");
   strcpy(dialog.radio[0][0],"Coordinates");             
   strcpy(dialog.radio[0][1],"Select coordinates");
   strcpy(dialog.radio[0][2],"Repeat prev. scan");
   strcpy(dialog.radio[1][0],"Maximum signal");             
   strcpy(dialog.radio[1][1],"Black");
   strcpy(dialog.radio[1][2],"White");
   strcpy(dialog.radio[2][0],"Scan type");             
   strcpy(dialog.radio[2][1],"Snap to 90 degrees");
   strcpy(dialog.radio[2][2],"Diagonal scan");
   strcpy(dialog.radio[2][3],"Fixed width");
  
   strcpy(dialog.boxes[0],"Options");
   strcpy(dialog.boxes[1],"Auto. save scan");
   strcpy(dialog.boxes[2],"Data file");
   strcpy(dialog.answer[2],sp.filename);
   strcpy(dialog.boxes[3],"Plot results");
   strcpy(dialog.boxes[4],"Pause to show region");
   strcpy(dialog.boxes[5],"Leave area marked");
   strcpy(dialog.boxes[6],"Calib. factor");
   strcpy(dialog.boxes[7],"Pixel compensation");
 
   dialog.boxtype[0]=LABEL;
   dialog.boxtype[1]=TOGGLE;
   dialog.boxtype[2]=SCROLLEDSTRING;
   dialog.boxtype[3]=TOGGLE;
   dialog.boxtype[4]=TOGGLE;
   dialog.boxtype[5]=TOGGLE;
   dialog.boxtype[6]=STRING;
   dialog.boxtype[7]=TOGGLE;

   if(g.bitsperpixel>8) sp.compensate=0;
   if(sp.wantrepeat) dialog.radioset[0]=2;
                else dialog.radioset[0]=1;
   if(sp.invert)     dialog.radioset[1]=1;
                else dialog.radioset[1]=2; 
   dialog.radioset[2] = sp.selection_method;

   //--------------------------------------------------//
   // Adjustable parameter in radio box                //
   //--------------------------------------------------//

   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   dialog.radiotype[2][3]=RADIOVALUEWIDE;     // Call slider
   dialog.radiomin[2][3]=1;
   dialog.radiomax[2][3]=200;
   itoa(sp.scanwidth, dialog.radioanswer[2][3],10);

   if(sp.autosave) dialog.boxset[1]=1; else dialog.boxset[1]=0;
   if(sp.wantplot) dialog.boxset[3]=1; else dialog.boxset[3]=0;
   if(sp.wantpause)dialog.boxset[4]=1; else dialog.boxset[4]=0;
   if(sp.leavemarked) dialog.boxset[5]=1; else dialog.boxset[5]=0;

   gcvt(sp.calfac, g.signif, tempstring);
   strcpy(dialog.answer[6], tempstring);

   if(sp.compensate) dialog.boxset[7]=1; else dialog.boxset[7]=0;
 
   dialog.radiono[0]=3;
   dialog.radiono[1]=3;
   dialog.radiono[2]=4;
   dialog.radiono[3]=0;
   dialog.noofradios=3;
   dialog.noofboxes=8;
   dialog.helptopic=8;  
   dialog.keep=0;
   dialog.want_changecicb = 0;
   dialog.f1 = null;
   strcpy(dialog.path,".");
   dialog.transient = 1;

   //////////////////
   dialogbox(dialog);
   //////////////////
   
   if(g.getout){ delete[] sp.od; delete[] data.x; delete[] data.y; return; }
   if(dialog.radioset[0]==1) sp.wantrepeat=0; else sp.wantrepeat=1;    
   if(dialog.radioset[1]==1) sp.invert=1;     else sp.invert=0;
   if(!hit) sp.wantrepeat=0;             // can't repeat if 1st time
            
   sp.selection_method = dialog.radioset[2];
   sp.scanwidth = atoi(dialog.radioanswer[2][3]);
   
   if(dialog.boxset[1]) sp.autosave=1; else sp.autosave=0;
   strcpy(sp.filename, dialog.answer[2]);
   if(dialog.boxset[3]) sp.wantplot=1; else sp.wantplot=0;
   if(dialog.boxset[4]) sp.wantpause=1; else sp.wantpause=0;
   if(dialog.boxset[5]) sp.leavemarked=1; else sp.leavemarked=0;
   sp.calfac = atof(dialog.answer[6]);
   if(dialog.boxset[7]) sp.compensate=1; else sp.compensate=0;
   if(g.bitsperpixel>8) sp.compensate=0;   // no grayscale map for color modes
    
   /*--------------------------------------------------------*/

   data.n         = 0;
   data.dims      = 1;
   data.nmax      = 300;
   data.f         = NULL;
   data.g         = NULL;
   data.width     = sp.scanwidth;
   data.type      = 4;
   data.duration  = TEMP;
   data.wantpause = 0;
   data.order     = 0;
   data.win       = 0; // calculate window automatically on drawing area
   
   g.getout = 0;
   invert = sp.invert;

   while(!g.getout)
   {   sp.invert = invert;   // sp.invert may get changed in scan_region_calc()
       scan_region_start(&data, &sp);
       if(g.getout){ g.getout=0; aborted=1; break; }
       while(g.bezier_state==CURVE)
           XtAppProcessEvent(XtWidgetToApplicationContext(g.drawing_area),XtIMAll);
       if(g.getout){ g.getout=0; aborted=1; break; }
       scan_region_calc(&data, &sp);
   }

   strcpy(osp.filename, sp.filename);
   osp.autosave           = sp.autosave;
   osp.bdensity           = sp.bdensity;
   osp.count              = sp.count;
   osp.leavemarked        = sp.leavemarked;
   osp.invert             = invert;
   osp.gamma_inverted     = sp.gamma_inverted;
   osp.selection_method   = sp.selection_method;
   osp.scanwidth          = sp.scanwidth;
   osp.snapthreshold      = sp.snapthreshold;
   osp.compensate         = sp.compensate;
   osp.autobkg            = sp.autobkg;
   osp.wantfixedwidth     = sp.wantfixedwidth;
   osp.wantrepeat         = sp.wantrepeat;
   osp.wantpause          = sp.wantpause;
   osp.wantplot           = sp.wantplot;
   osp.calfac             = sp.calfac;
   osp.skip               = sp.skip;
   osp.wantcolor          = sp.wantcolor; 
   osp.diameter           = sp.diameter;
   osp.scan               = sp.scan;

   ////  Try to clean up if user hits main Cancel button
   if(aborted)
   {   bezier_curve_end(&data);
       g.bezier_state=NORMAL;
       g.state=NORMAL;
       g.getout=0;
       message("Interrupted");
   }
   delete[] sp.od;
   delete[] data.y;
   delete[] data.x;
   hit=1;
}


//-------------------------------------------------------------------------//
// scan_region_start - Continue scanning with parameters from scan_region  //
//               or end scanning                                           //
//-------------------------------------------------------------------------//
void scan_region_start(XYData *data, ScanParams *sp)
{

   switchto(g.imageatcursor);       
   if(g.getout)
   {    g.getout=0;
        message("Densitometry mode finished"); 
   }else if(!sp->wantrepeat)
   {    sp->count = 0;
        data->n = 0;
        switch(sp->selection_method)
        { case 1: 
              data->type = 5;
              data->nmax = 4;
              bezier_curve_start(data); 
              break;
          case 2: 
              data->type = 4;
              data->nmax = 4;
              bezier_curve_start(data); 
              break;
          case 3: 
              data->type = 6;
              data->nmax = 2;
              bezier_curve_start(data); 
              break;
          default:
              g.getout=1;
              break;
        }
   }
}



//-------------------------------------------------------------------------//
// scan_region_calc - measure density of a region bounded by trapezoid     //
//-------------------------------------------------------------------------//
void scan_region_calc(XYData *data, ScanParams *sp)
{
    char tempstring[FILENAMELENGTH];    
    char extension[FILENAMELENGTH]="";
    int allnumbers, bad=0, bpp, dot, helptopic=8, i, ino, j, k, 
        nmax, number=0, oldci=0, size, status=OK, usediag=0;
    int x[6], y[6];
    double total=0.0, xx[6], yy[6];
    double *xdata;
    char **ytitle;
    char *ptr;

    size = 2*(int)sqrt(g.xres*g.xres + g.yres*g.yres);  // Size of scan[] array
    ytitle = new char*[1];
    ytitle[0] = new char[256];

    xdata = (double*)malloc(size*sizeof(double)); // Largest possible diag. distance
    if(xdata == NULL){ message(g.nomemory,ERROR); return; }

    x[1] = data->x[0];
    x[2] = data->x[1];
    x[3] = data->x[2];
    x[4] = data->x[3];
    y[1] = data->y[0];
    y[2] = data->y[1];
    y[3] = data->y[2];
    y[4] = data->y[3];
    x[5] = (data->x[0] + data->x[1] + data->x[2] + data->x[3])/4;
    y[5] = (data->y[0] + data->y[1] + data->y[2] + data->y[3])/4;

    if(sgn(x[2]-x[1])!=sgn(x[4]-x[3])){ swap(x[3],x[4]); swap(y[3],y[4]); }
    if(sgn(y[2]-y[1])!=sgn(y[4]-y[3])){ swap(x[3],x[4]); swap(y[3],y[4]); }
 
    xx[1] = (double)x[1];
    xx[2] = (double)x[2];
    xx[3] = (double)x[3];
    xx[4] = (double)x[4];
    xx[5] = (double)x[5];
    yy[1] = (double)y[1];
    yy[2] = (double)y[2];
    yy[3] = (double)y[3];
    yy[4] = (double)y[4];
    yy[5] = (double)y[5];


    //----------------------------------------------------------------------//
    // Validate gamma tables & test if need to invert gamma.                //
    // Do center point last. Center determines whether inverting is needed. //
    //----------------------------------------------------------------------//

    nmax = data->nmax;
    ino = whichimage(x[nmax-1],y[nmax-1],bpp);
    status = check_gamma_table(sp, ino, bpp);
    if(status!=OK)
    {  
            g.getout=0;
            free(xdata);
            message("Densitometry mode finished"); 
            return;
    }
    if(sp->gamma_inverted) sp->invert = 1 - sp->invert;

    //----------------------------------------------------------------------//
    // Measure signal in selected region                                    //
    //----------------------------------------------------------------------//

    for(i=0;i<size;i++){ sp->scan[i]=0; xdata[i]=(double)i; }
 
    switch(sp->selection_method)
    {   case 1: usediag=0; break;
        case 2:          // If diagonal scan was selected but points were 
                         // vertical or horizontal
           if((abs(x[2]-x[1]) < sp->snapthreshold)||
              (abs(y[2]-y[1]) < sp->snapthreshold))
              usediag=0;
           else 
              usediag=1;
           break;
        case 3: usediag=2; break;
    }        
 
    switch(usediag)
    {   case 0: total = scan_normal_polygon(sp,x,y,g.scancount); break;
        case 1: total = scan_arbitrary_polygon(sp,xx,yy,g.scancount); break;
        case 2: total = scan_fixed_width_area(sp,xx,yy,g.scancount); break;
    }
  
    //----------------------------------------------------------------------//
    // Plot scan results                                                    //
    //----------------------------------------------------------------------//
 
    sprintf(tempstring,"Total signal scanned %g",total);
    message(tempstring);
    strcpy(ytitle[0],"Signal");
 
    if(sp->wantplot) 
       status=plot(sp->filename, "Data Point", ytitle, xdata, sp->scan, 
          g.scancount, 1, MEASURE, 0, 0, null, null, null, helptopic);
    g.getout = 0;          // In case user clicks 'cancel' button like a schmuck
   if((sp->autosave)&&(!bad))
    {  save_scan(sp->filename, sp->scan, g.scancount, 1);
          

                         // Find last dot and slash in filename
       ptr = 1 + strrchr(sp->filename,'/');
       strcpy(tempstring, ptr);
       dot = (int)strlen(tempstring);
       for(j=0;j<(int)strlen(tempstring);j++) if(tempstring[j]=='.') dot=j;
                         
                         // If text to left of dot is all digits, increment
                         // the filename.
       allnumbers=1;
       for(k=0;k<dot;k++) if(!isdigit(tempstring[k])) allnumbers=0;
       if(allnumbers)         
       {    memset(extension,0,strlen(extension));
            strncpy(extension,tempstring+dot+1,4);
            number = atoi(tempstring);
            tempstring[dot]=0;
            number++;
            itoa(number,tempstring,10);
            sprintf(sp->filename,"%s/%s.%s",g.startdir,tempstring,extension);
       }

    }
    if(status==GOTNEW)      // If plot created an image, put it in background
    {   switchto(oldci);      
        status=OK;
    }
    delete[] ytitle[0];
    delete[] ytitle;
    free(xdata);
}


//-------------------------------------------------------------------------//
// scan_normal_polygon  (called by scan_region_calc)                       //
// 4-sided polygon, vertical/horizontal starting edge, all points are at   //
// the center of a pixel.                                                  //
//-------------------------------------------------------------------------//
double scan_normal_polygon(ScanParams *sp, int x[], int y[], int &scancount) 
{
     int i,k,count,direction;
     double total=0.0, fvalue;
     int mode[3];
     mode[0]=XOR;
     mode[1]=XOR;
     if(( sp->compensate)&&(!sp->invert)) mode[2]=SUM;
     if((!sp->compensate)&&(!sp->invert)) mode[2]=SUM2;
     if(( sp->compensate)&&( sp->invert)) mode[2]=NOT;
     if((!sp->compensate)&&( sp->invert)) mode[2]=NOT2;

           //-------------Scan vertically or horizontally-------------//     

     if(abs(y[1]-y[2]) < sp->snapthreshold) direction=VERTICAL;
     else direction=HORIZONTAL;

     if(direction==VERTICAL)
     { if((y[3]!=y[1])&&(y[4]!=y[2]))
       {
           /*-You have to have (int) before both terms here or it-----*/
           /*-doesn't divide correctly for some unfathomable reason---*/
           //-----------------Scan top to bottom----------------------//
             
          scancount = abs((int)y[3]-(int)y[1]);
          for(k=0;k<=2;k++)
          {  count=0;
             for(i=0;count<abs(y[3]-y[1]);i+=sgn(y[3]-y[1]))      
             {   if(k==2)
                 {  fvalue = scanline(sp, x[1]+(int)(i*(x[3]-x[1]))/(int)(y[3]-y[1]), 
                             y[1]+i, x[2]+(int)(i*(x[4]-x[2]))/(int)(y[4]-y[2]), 
                             y[2]+i);
                    sp->scan[count] = fvalue; 
                    total+=fvalue;
                 }else
                 {  g.inmenu++;
                    line(x[1]+(int)(i*(x[3]-x[1]))/(int)(y[3]-y[1]), y[1]+i,
                             x[2]+(int)(i*(x[4]-x[2]))/(int)(y[4]-y[2]), y[2]+i,
                             g.maxcolor,mode[k]);                  
                    g.inmenu--;
                 }
                 count++;      
             } 
          }                    
       }
     }     
     else 
     { if((x[3]!=x[1])&&(x[4]!=x[2]))   
       {
           //-----------------Scan left to right----------------------//   
          scancount = abs(x[3]-x[1]); 
          for(k=0;k<=2;k++)
          {  count=0;
             for(i=0;count<abs(x[3]-x[1]);i+=sgn(x[3]-x[1]))
             {  if(k==2)
                {   fvalue = scanline(sp, x[1]+i,
                             y[1]+(int)(i*(y[3]-y[1]))/(int)(x[3]-x[1]),
                             x[2]+i, y[2]+(int)(i*(y[4]-y[2]))/(int)(x[4]-x[2]));
                    sp->scan[count] = fvalue;       
                    total += fvalue;
                }else
                    line(x[1]+i,y[1]+(int)(i*(y[3]-y[1]))/(int)(x[3]-x[1]),
                             x[2]+i,y[2]+(int)(i*(y[4]-y[2]))/(int)(x[4]-x[2]),
                             g.maxcolor,mode[k]); 
                count++;
             }                                               
          }    
       }   
     }  
     return total;
}  

//-------------------------------------------------------------------------//
// scan_arbitrary_polygon   (called by scan_region_calc)                   //
// xx and yy range from 1 to 4, not 0 to 3                                 //
//-------------------------------------------------------------------------//
double scan_arbitrary_polygon(ScanParams *sp, double xx[],double yy[], int &scancount) 
{
                           // Arbitrary 4-sided polygon. Points do not
                           // necessarily coincide with centers of pixels.
     int j,k,x,y;
     int parallelogram;
                                              // dx = x increment
                                              // dy = y increment
                                              // m  = slope
                                              // b  = y intercept
                                              // numbers = line segments
     double m,m12,m13,m14,m23,m24,m34;
     double b,b12,b13,b14,b23,b24,b34;
     double dx,dx12,dx31,dx41,dx32,dx42,dx43;
     double denom,total=0.0;
  
                        //-------Reorder the segments from left to right----   
                        //----After finding the starting edge. This makes---
                        //----calculations easier.--------------------------

     dx = (double)(xx[2]-xx[1]); if(dx==0) dx=1e-6;
     m = (yy[2] - yy[1])/dx; if(m==0) m=1e-6;     //--Slope of starting edge
     b = yy[1] - xx[1]*m;                         //--Intercept of starting edge
     scancount = 0;                               // No. points in scan perp.to 1--2
     if((m!=0)&&(m+1/m !=0)) denom = 1/(m + 1/m); else denom=1;          


     for(j=1;j<=4;j++)
     for(k=1;k<=4;k++)
        if(xx[k] > xx[j]){ fswap(xx[k],xx[j]); fswap(yy[k],yy[j]); }

     //-------Try to fix any bad parameters--------------
     //-------Prevent divide by 0 errors-----------------
     //---Slope & y-intercept based on original start
     //---and end points of segment 1--2 

     dx12 = (double)(xx[2]-xx[1]); if(dx12==0) dx12=1e-6;
     dx31 = (double)(xx[3]-xx[1]); if(dx31==0) dx31=1e-6;
     dx41 = (double)(xx[4]-xx[1]); if(dx41==0) dx41=1e-6;
     dx32 = (double)(xx[3]-xx[2]); if(dx32==0) dx32=1e-6;
     dx42 = (double)(xx[4]-xx[2]); if(dx42==0) dx42=1e-6;
     dx43 = (double)(xx[4]-xx[3]); if(dx43==0) dx43=1e-6;

     m12 = (yy[2] - yy[1])/dx12; if(m12==0) m12=1e-6;
     m13 = (yy[3] - yy[1])/dx31; if(m13==0) m12=1e-6;
     m14 = (yy[4] - yy[1])/dx41; if(m14==0) m12=1e-6;
     m23 = (yy[3] - yy[2])/dx32; if(m23==0) m12=1e-6;
     m24 = (yy[4] - yy[2])/dx42; if(m24==0) m12=1e-6;
     m34 = (yy[4] - yy[3])/dx43; if(m34==0) m12=1e-6;

     b12 = yy[1] - xx[1]*m12;
     b13 = yy[1] - xx[1]*m13;
     b14 = yy[1] - xx[1]*m14;
     b23 = yy[2] - xx[2]*m23;
     b24 = yy[2] - xx[2]*m24;
     b34 = yy[3] - xx[3]*m34;
                                    
     // The global 'scancount' is incremented in diagonalscan()      
     // which contains common code. For trapezoids, the area is      
     // split into 3 sections and 3 of these loops are executed.    
     // For a rectangle, only loop 'e' is executed.		     
     // If the edges are exactly vertical or horizontal, this method 
     // doesn't work. This case is handled by the rectilinear scan   
     // method above, which is faster.  			     
     // 							     
     // In the section below, the pixels are always scanned in the   
     // same order (y's changing fastest). The pixel values are      
     // automatically placed in the correct element of the array.    
     // The 6 cases are for the 6 possible shapes.		     

     //------Show the region being scanned---------------
     //------then scan diagonally------------------------
     //---If 2--3 is parallel to 1--4  
     // (difference in slope is less than 
     //  the diff. in slope between 1--2 and
     //  3--4 and 1--3 and 2--4)
    
     parallelogram=0;
     if( (fabs(m23-m14)<fabs(m12-m34)) && (fabs(m23-m14)<fabs(m13-m24)) )
        parallelogram=1;
   
     if(parallelogram)   
     {  
        for(k=1;k<=3;k++)           // Loop 'a' - change k loop to 1..3 if you
        {                           // want the area temporarily inverted.
           if(fabs(xx[1]-xx[2])>1.0)
           {    for(x=(int)xx[1]; x<(int)xx[2]; x++)
                for(y=(int)(m12*x+b12); y!=(int)(m14*x+b14); y+=sgn(m14-m12))
                {   if(k<3) setpixel(x,y,g.maxcolor,XOR);
                    else total+=diagonalscan(sp,x,y,m,b,denom);
                    if((y<=0)||(y>g.main_ysize-1))break;
                }   
           }
           if(fabs(xx[2]-xx[3])>2.0) // Loop 'b'
           {    for(x=(int)xx[2]; x<(int)xx[3]; x++)
                for(y=(int)(m23*x+b23); y!=(int)(m14*x+b14); y+=sgn(m14-m12))
                {   if(k<3) setpixel(x,y,g.maxcolor,XOR);
                    else total+=diagonalscan(sp,x,y,m,b,denom);
                    if((y<=0)||(y>g.main_ysize-1))break;
                }   
           }
           if(fabs(xx[3]-xx[4])>2.0) // Loop 'c'
           {   for(x=(int)xx[3]; x<(int)xx[4]; x+=sgn(xx[4]-xx[3]))
               for(y=(int)(m34*x+b34); y!=(int)(m14*x+b14); y+=sgn(m14-m12))
               {   if(k<3) setpixel(x,y,g.maxcolor,XOR);
                   else total+=diagonalscan(sp,x,y,m,b,denom);
                   if((y<=0)||(y>g.main_ysize-1))break;
               }   
           }  
        }
     }
     else                           //---If 1--2 is parallel to 3--4
     {                              // Change k loop to 1..3 if you
                                    // want the area temporarily inverted.
        for(k=1;k<=3;k++)              
        {  if(fabs(xx[1]-xx[2])>1.0)  // Loop 'd'
           { 
             for(x=(int)xx[1]; x<(int)xx[2]; x++)
             for(y=(int)(m13*x+b13); y!=(int)(m12*x+b12); y+=sgn(m12-m13))
             {     if(k<3) setpixel(x,y,g.maxcolor,XOR);
                   else total+= diagonalscan(sp,x,y,m,b,denom);
                   if((y<=0)||(y>g.main_ysize-1))break;
             }
           }     
           if(fabs(xx[2]-xx[3])>1.0)  // Loop 'e'
           {   for(x=(int)xx[2]; x<(int)xx[3]; x++)
               {   for(y=(int)(m13*x+b13); y!=(int)(m24*x+b24); y+=sgn(m12-m13))
                   {  if(k<3) setpixel(x,y,g.maxcolor,XOR);
                      else total+=diagonalscan(sp,x,y,m,b,denom);
                      if((y<=0)||(y>g.main_ysize-1))break;
                   }   
               }
           }
           if(fabs(xx[3]-xx[4])>1.0) // Loop 'f'
           { 
             for(x=(int)xx[3]; x<(int)xx[4]; x+=sgn(xx[4]-xx[3]))
             for(y=(int)(m34*x+b34); y!=(int)(m24*x+b24); y+=sgn(m12-m13))
             {     if(k<3) setpixel(x,y,g.maxcolor,XOR);
                   else total+=diagonalscan(sp,x,y,m,b,denom);
                   if((y<=0)||(y>g.main_ysize-1))break;
             }   
           }  
        }   
     }
     return total;
}
   

//-------------------------------------------------------------------------//
// scan_fixed_width_area    (called by scan_region_calc)                   //
// xx and yy range from 1 to 4, not 0 to 3                                 //
//-------------------------------------------------------------------------//
double scan_fixed_width_area(ScanParams *sp, double xx[], double yy[], 
  int &scancount) 
{
                                              // dx = x increment
                                              // dy = y increment
                                              // m  = slope
                                              // b  = y intercept
                                              // numbers = line segments
  double A,B,C,D,a,m12,dx12,fvalue,total=0.0,fr,fg,fb;
  double d=0.0,dp,dx,dy,ddx,ddy,fx,fy,mp,rx,ry,w,xs,ys;
  int h,hmax,j,k,ix,iy,bpp,ino,ia,ra,ga,ba,ib,rb,gb,ic,rc,gc,bc,
      id,rd,gd,bd,vv,rr,gg,bb;
  int obpp=-1;

    //------------------------------------------------------------------//
    // Fixed width scan. Analyze by brute force.                        //
    // The starting and ending coordinates are the only points guaran-  //
    // teed to coincide with a pixel. Calculate all other points at     //
    // multiples of 1 pixel from each other, read the surrounding 4     //
    // pixels, and calculate the value for that point based on a        //
    // value proportional to how far away the pixel is.                 //
    //------------------------------------------------------------------//

    //---Calculate line between the starting and ending points---//

  dx12 = (double)(xx[2]-xx[1]); if(dx12==0) dx12=1e-6;
  m12 = (yy[2] - yy[1])/dx12; if(m12==0) m12=1e-6;

    //---Length of line between x1,y1 and x2,y2------------------//
  a = (xx[2]-xx[1]) * (xx[2]-xx[1]) + (yy[2]-yy[1]) * (yy[2]-yy[1]);
  d = sqrt(a);
  scancount = (int)d;

    //---X & Y incremental distances along line parallel to 1--2 //
  ddx = sqrt(1.0/(1.0+m12*m12));
  if(xx[1]>xx[2]) ddx *= -1.0;
  ddy = ddx*m12;

    //---Calculate line perpendicular to 1--2, through x1,y1 ----//
  if(fabs(m12)>1e-10)
    mp = -1.0/m12;
  else
    mp = -100000000.0;

  //---Initialize everything-------------------------------------//
  //   For k=-width to +width, calculate the starting and        //
  //   ending coordinates on the perpendicular line.             //
  //   Read the values of the 4 closest pixels to this point.    //
  //   Take a fraction of these pixel values proportional to     //
  //   how close they are.                                       //
  //                                                             //
  //   Pixel boundary is at upper left of pixel A.               //
  //   X,Y coordinate is a non-integer remainder into            //
  //   pixel A = rx,ry. This ensures that the pixels             //
  //   to use are always to the right and below pixel A.         //
  //                                                             //
  //    Contributing pixel   Use                                 //
  //        A                A * (1-rx) * (1-ry)                 //
  //        B               +B * rx * (1-ry)                     // 
  //        C               +C * (1-rx) * ry                     //
  //        D               +D * rx * ry                         //
  //                   Total should be 1.                        //
  //-------------------------------------------------------------//

  double fracx,fracy;
  for(k=0;k<scancount;k++) sp->scan[k]=0;
  total = 0;
  if(sp->leavemarked) hmax=1; else hmax=0;
  for(h=0;h<=hmax;h++)
  {
     for(k=0;k<=sp->scanwidth;k++)
     {  w = (double)sp->scanwidth;
        dp = (double)k - floor(w/2.0-0.5);
                                         // 1 2 3 4 5 -> -2 -1 0 -1 2, etc.
                                         // 1 2 3 4   -> -1  0 1  2
        dx = sqrt((dp*dp)/(1.0+mp*mp));  // Incremental distance along perp.
        dx *= sgn(dp);
        dy = mp * dx;       
        xs = xx[1] + dx;                 // Starting coordinates
        ys = yy[1] + dy;
       
       //---Calculate the coords. of each point on this line.-------//
        fx = xs; fy=ys;                  // floating-point real coordinates
  
        for(j=0;j<scancount;j++)
        {   
           rx = modf(fx, &fracx);
           ry = modf(fy, &fracy);
           ix = int(fracx);
           iy = int(fracy);

           if(h==1)
                setpixel(ix,iy,g.maxcolor,SET);
           else switch(sp->wantcolor){
           case 0:          // grayscale mode 0..1 for scanning
                A = pixeldensity(sp, ix, iy   , sp->invert);
                B = pixeldensity(sp, ix+1,iy  , sp->invert);
                C = pixeldensity(sp, ix  ,iy+1, sp->invert);
                D = pixeldensity(sp, ix+1,iy+1, sp->invert);

                fvalue = A * (1.0-rx) * (1.0-ry)
                       + B * rx * (1.0-ry)
                       + C * (1.0-rx) * ry
                       + D * rx * ry;
     
                total += fvalue;
                sp->scan[j] += fvalue; 
                break;
           case 1:          // color mode for rotating images
                            // w/antialiasing
                ia = readpixelonimage(ix,  iy,  bpp, ino);
                valuetoRGB(ia, ra, ga, ba, bpp);    
                ib = readpixelonimage(ix+1,iy,  bpp, ino);
                valuetoRGB(ib, rb, gb, bb, bpp);    
                ic = readpixelonimage(ix,  iy+1,bpp, ino);
                valuetoRGB(ic, rc, gc, bc, bpp);    
                id = readpixelonimage(ix+1,iy+1,bpp, ino);
                valuetoRGB(id, rd, gd, bd, bpp);    
                fr = (double)(
                       ra * (1.0-rx) * (1.0-ry)
                     + rb * rx * (1.0-ry)
                     + rc * (1.0-rx) * ry
                     + rd * rx * ry );
                fg = (double)(
                       ga * (1.0-rx) * (1.0-ry)
                     + gb * rx * (1.0-ry)
                     + gc * (1.0-rx) * ry
                     + gd * rx * ry );
                fb = (double)(
                       ba * (1.0-rx) * (1.0-ry)
                     + bb * rx * (1.0-ry)
                     + bc * (1.0-rx) * ry
                     + bd * rx * ry );
              
                ////  Source pixel is on a different image, switch to new colormap 
                if(bpp!=obpp && ino>=0 && z[ino].colortype!=GRAY)
                {    memcpy(g.palette, z[ino].palette, 768);
                     obpp = bpp;
                }
                vv = (int)RGBvalue((int)fr,(int)fg,(int)fb,bpp);

                ////  Source pixel is grayscale, convert to color 
                if(ino>=0 && z[ino].colortype==GRAY)
                {    vv=convertpixel(vv, bpp, sp->bpp, 1);
                     vv=graypixel(vv,sp->bpp);
                     bpp=sp->bpp;
                }
                sp->scan[j] = convertpixel(vv, bpp, sp->bpp, 1);
                break;
           case 2:          // color mode for rotating images w/o 
                            // antialiasing
                sp->scan[j] = (double)readpixelonimage(ix,iy,bpp,ino);
                break;
           case 3:          // grayscale mode 0..maxcolor for rotating
                A = (double)readpixelonimage(ix,  iy,  bpp, ino);
                B = (double)readpixelonimage(ix+1,iy,  bpp, ino);
                C = (double)readpixelonimage(ix,  iy+1,bpp, ino);
                D = (double)readpixelonimage(ix+1,iy+1,bpp, ino);
                fvalue = A * (1.0-rx) * (1.0-ry)
                       + B * rx * (1.0-ry)
                       + C * (1.0-rx) * ry
                       + D * rx * ry;
                vv = (int)fvalue; 
                   
                ////  Source pixel is color or different bpp, convert to gray 
                if(bpp!=sp->bpp)
                {     fvalue /= g.maxgray[sp->bpp];
                      rr = (int)fvalue*g.maxred[sp->bpp];
                      gg = (int)fvalue*g.maxgreen[sp->bpp];
                      bb = (int)fvalue*g.maxblue[sp->bpp];
                     
                      ////  Source pixel is on a different image, switch to new colormap 
                      if(bpp!=obpp && ino>=0)
                      {    memcpy(g.palette, z[ino].palette, 768);
                           obpp = bpp;
                      }
                      vv = RGBvalue(rr,gg,bb,sp->bpp);
                }
                sp->scan[j] = vv;
                break;
           case 4:          // color mode for rotating 1 image w/o 
                            // antialiasing
               ino = ci;
               bpp = z[ino].bpp;
               if(between(ix,0,z[ino].xsize-1) && between(iy,0,z[ino].ysize-1))
                   sp->scan[j]=pixelat(z[ino].image[0][iy]+ix*g.off[bpp], bpp);
               break;
           }
           fx+=ddx;
           fy+=ddy; 
        } 
    }
  }
  return total;
}



//-------------------------------------------------------------------------//
// diagonalscan - calculate distance from x,y to line given by slope m     //
//      and y intercept b, read the pixel at x,y, and put its value into   //
//      array 'scan[]', using anti-aliasing to partition portions of the   //
//      pixel value into the appropriate element depending how close it    //
//      is to the calculated position.                                     //
//      This does not take into account pixels at edge of the figure.      //
//      'denom' is a pre-calculated factor 1/(m+1/m) for speed.            //
//-------------------------------------------------------------------------//
double diagonalscan(ScanParams *sp, double x, double y, double m, double b, 
                    double denom)
{ 
        double bperp, d, dx, dy, fval, r, xint, yint;
        int n;
        
        if(m==0) 
           bperp = y + x*1e6;         // Y-intercept of line perpendicular 
        else                          // to y=mx+b, through x,y.
           bperp = y + x/m;
        xint = (bperp - b)*denom;     // Intersection coordinates
        yint = m*xint + b;
        dx = x-xint;                  // X distance point to line
        dy = y-yint;                  // Y distance point to line
        d = sqrt(dx*dx + dy*dy);      // Distance from point to line
        n = int(d);                   // Integer distance point to line
        r = d - (double)n;             // Remainder (determines partitioning)
        g.scancount = max(g.scancount,n);
        fval = pixeldensity(sp, (int)x, (int)y, sp->invert);
                                      // Partition the pixel value
        if(r >= 0.5)
        {   sp->scan[n+1] += (r - 0.5)*fval;
            sp->scan[n]   += (1.5 - r)*fval; 
        }else
        {   sp->scan[n]   += (r + 0.5)*fval;
            if(n>0) sp->scan[n-1] += (0.5 - r)*fval;
        }        
        return(fval);
}        



//-------------------------------------------------------------------------//
// trace curve                                                             //
//-------------------------------------------------------------------------//
int trace_curve(void)
{
  static int wantsave=0,wantplot=1,wantslow=0;
  static char filename[FILENAMELENGTH] = "1.trace";
  static int track_color=0;
  if(memorylessthan(16384)){ message(g.nomemory,ERROR); return(NOMEM); }
  Dialog dialog;
  char tempstring[256];
  int helptopic=32;
  int *trace;
  int d,oldy,x,y1,y2,xstart,ystart,xend=0;
  int test1=0,test2=0,tmin=0,bpp,ino;
  int ymax=0;
  int done1=0,done2=0,done3=0;
  double *xdata;
  xdata = (double*)malloc(2048*sizeof(double));                          
  char **ytitle;

  ytitle = new char*[1];
  ytitle[0] = new char[256];
  if(ci<0) { message("Please select an image",ERROR); return(NOIMAGES); }
  trace = new int[z[ci].xsize];
  if(trace==NULL){ message(g.nomemory,ERROR); return(NOMEM); }

  g.getout=0;
  while(!g.getout)
  {
    dialog.needrepeat = 0;
    strcpy(dialog.title,"Trace Curve");
    strcpy(dialog.boxes[0],"Options");
    strcpy(dialog.boxes[1],"Save trace to disk");
    strcpy(dialog.boxes[2],"Plot trace");
    strcpy(dialog.boxes[3],"Debug");
    strcpy(dialog.boxes[4],"Color to track");
    dialog.boxtype[0]=LABEL;
    dialog.boxtype[1]=TOGGLE;
    dialog.boxtype[2]=TOGGLE;
    dialog.boxtype[3]=TOGGLE;
    dialog.boxtype[4]=INTCLICKBOX;

    if(wantsave==1) dialog.boxset[1]=1; else dialog.boxset[1]=0;
    if(wantplot==1) dialog.boxset[2]=1; else dialog.boxset[2]=0;
    if(wantslow==1) dialog.boxset[3]=1; else dialog.boxset[3]=0;
    itoa(track_color, dialog.answer[4],10);
    dialog.boxmin[4]=0;  dialog.boxmax[4]=(int)g.maxvalue[z[ci].bpp];

    dialog.noofradios=0;
    dialog.noofboxes=5;
    dialog.helptopic=32;  

    dialog.radiono[0]=1;
    dialog.keep=0;
    dialog.want_changecicb = 0;
    dialog.f1 = null;
    dialog.transient = 1;
    strcpy(dialog.path,".");

    g.getout=0;
 
    //////////////////

    dialogbox(dialog);

    //////////////////

    if(g.getout)break;
    track_color = atoi(dialog.answer[4]);
    if(dialog.boxset[1]) wantsave=1; else wantsave=0;
    if(dialog.boxset[2]) wantplot=1; else wantplot=0;
    if(dialog.boxset[3]) wantslow=1; else wantslow=0;

    message("Move to starting point and click \nthe mouse to start");
    getpoint(xstart,ystart);
    done3=0;
    oldy=ystart;

    for(x=xstart;x<z[ci].xpos+z[ci].xsize;x++)
    { 
        xend=x;
        done1=0;done2=0;
        g.inmenu++;
        if(x>xstart)
        {  line(x-1,oldy-15,x-1,oldy+15,g.maxcolor,XOR);  // Cross hairs
           line(x-15-1,oldy,x+15-1,oldy,g.maxcolor,XOR);
        }
        g.inmenu--;
    
        if(g.getout) done3=0;
        ymax=oldy;
        tmin = (int)g.maxvalue[z[ci].bpp];
        for(y1=oldy,y2=oldy;;y1--,y2++)
        {  if(y1>z[ci].ypos) 
           {   test1 = readpixelonimage(x,y1,bpp,ino); 
               d = abs(test1 - track_color);
               if(d<tmin) { tmin=d; ymax=y1; }
           }else done1=1;
           if(y2<z[ci].ypos+z[ci].ysize) 
           {   test2 = readpixelonimage(x,y2,bpp,ino);
               d = abs(test2 - track_color);
               if(d<tmin) { tmin=d; ymax=y2; }
           }else done2=1;
           if((done1==1)&&(done2==1))break;
        }  

        // Calculate calibrated x value

        ino = whichimage(x,ymax,bpp);
        if(ino>=0 && z[ino].cal_log[0] >= 0)
            calibratepixel(x,ymax,ino,0,xdata[x-xstart],tempstring); 
        else
            xdata[x-xstart] = (double)(x-xstart);
  
        if(wantslow)
        { 
           printf("x %d y %d value %d Max %d\n",x,ymax,tmin,
              z[ci].ysize-ymax+z[ci].ypos);
#ifdef LINUX
           usleep(100000);
#endif
           XtAppProcessEvent(XtWidgetToApplicationContext(g.drawing_area),XtIMAll);
           if(g.getout)done3=1;
        }
        oldy=ymax;
        g.inmenu++;
        line(x,oldy-15,x,oldy+15,g.maxcolor,XOR); // Cross hairs
        line(x-15,oldy,x+15,oldy,g.maxcolor,XOR);
        g.inmenu--;

        printcoordinates(x,ymax,0);

        trace[x-xstart]=z[ci].ysize-ymax+z[ci].ypos;
        if(done3==1)break;
    }

    g.inmenu++;
    line(x-1,oldy-15,x-1,oldy+15,g.maxcolor,XOR);  // Cross hairs
    line(x-15-1,oldy,x+15-1,oldy,g.maxcolor,XOR);
    g.inmenu--;
    strcpy(ytitle[0],"Optical Density");      
    if(wantplot) plot(filename, "X", ytitle, xdata, trace, xend-xstart, 1,
         MEASURE, 0, 0,  null, null, null, helptopic);
    if(wantsave) save_scan(filename, trace, xend-xstart, 0);
  }
  delete[] ytitle[0];
  delete[] ytitle;
  delete[] trace;
  free(xdata);
  return(OK);
}


//-------------------------------------------------------------------//
// scanline - use Bresenham algorithm to measure the pixel value     //
// along the length of a line.                                       //
//-------------------------------------------------------------------//
double scanline(ScanParams *sp, int xs, int ys, int xe, int ye)
{
     int decision, dx, dy, x_sign, y_sign, x, y;
     double total=0.0;
     dx = abs(xe - xs);
     dy = abs(ye - ys);
     if(dx > dy)
     {  if(xs > xe){ swap(xs,xe); swap(ys,ye); }
        if(ye - ys < 0) y_sign = -1; else y_sign = 1;
        for (x=xs,y=ys,decision=0; x<=xe; x++,decision+=dy)
        {   if(decision>=dx){ decision -= dx; y+= y_sign; }
            total += pixeldensity(sp, x, y, sp->invert);
        } 
     }else
     {  if(ys > ye){ swap(ys,ye); swap(xs,xe); }
        if(xe - xs < 0) x_sign = -1; else x_sign = 1;
        for (x=xs,y=ys,decision=0; y<=ye; y++,decision+=dx)
        {   if(decision>=dy){ decision -= dy; x+=x_sign;  }
            total += pixeldensity(sp, x, y, sp->invert);
        }
     }
     return total;
}


//-------------------------------------------------------------------//
// insidecircle                                                      //
//-------------------------------------------------------------------//
int insidecircle(int i, int j, int x, int y, int d)
{
   double r = (double)(d)/2;
   double r2 = r*r;
   double dist2;
   dist2 = (i-x)*(i-x) + (j-y)*(j-y);
   if(dist2<=r2) return 1;
   return 0;
}


//------------------------------------------------------------------//
// fill_polygon (4 sides only)                                      //
// xx and yy range from 1 to 4, not 0 to 3                          //
//------------------------------------------------------------------//
void fill_polygon(int *xx, int *yy, int color) 
{
   fill_triangle(xx[1], yy[1], xx[4], yy[4], xx[3], yy[3], color);
   fill_triangle(xx[1], yy[1], xx[4], yy[4], xx[2], yy[2], color);
}
