//--------------------------------------------------------------------------//
// xmtnimage21.cc                                                           //
// Evaluation of string expressions                                         //
// Latest revision: 02-15-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h" 

extern Globals     g;
extern Images     *z;
extern int         ci;
char *text=NULL;

//-------------------------------------------------------------------------//
// formula_error                                                           //
//-------------------------------------------------------------------------//
void formula_error(char *error_message)
{
    message(error_message);
    g.getout=1; 
}

 
//-------------------------------------------------------------------------//
// eval - entry point into evaluating a string                             //
//-------------------------------------------------------------------------//
double eval(char* s)
{   
   if(strlen(s)==0 || strlen(s)>1023) return 0;
   if(g.getout) return 0;
   text = (char*)malloc(strlen(s)+8);
   strcpy(text,s);
   strcat(text,"\n\255");
   yyparse();
   free(text);
   return read_variable("temp");
}    


//-------------------------------------------------------------------------//
// evaluate  - check a variable                                            //
//-------------------------------------------------------------------------//
void evaluate(int noofargs, char **arg)
{   
   int k;
   char *tempstring = new char[FILENAMELENGTH];
   if(noofargs<1) return;
   add_all_variables(ci);
   strcpy(tempstring, arg[1]);
   for(k=2;k<=noofargs;k++)
   {   strcat(tempstring," "); 
       strcat(tempstring,arg[k]); 
   }
   double dd = eval(tempstring);
   sprintf(tempstring, "value is %g",dd);
   message(tempstring);
   delete[] tempstring;
}    


//--------------------------------------------------------------------------//
//  erf                                                                     //
//--------------------------------------------------------------------------//
double erf(double x)
{
   double s1,s2,s3,s4,s5,y;
   s1 = fabs(x);
   s2 = 1 / (1 + .5 * s1);
   s3 = -.82215223 + s2 * .17087277;
   s3 = 1.48851587 + s2 * s3;
   s3 = -1.13520398 + s2 * s3;
   s3 = .27886807 + s2 * s3;            // Chebyshev fitting for
   s3 = -.18628806 + s2 * s3;           // estimation of
   s3 = s2 * (.09678418 + s2 * s3);     // complementary error function
   s3 = s2 * (1.00002368 + s2 * (.37049196 + s3));
   s5 = (-s1 * s1 - 1.26551223 + s3);
   if(s5 < -77) s5 = -77;
   s4 = s2 * exp(s5);
   if(x < 0) s4 = 2 - s4;
   y = s4;
   return y;
}


//-------------------------------------------------------------------------//
// mask                                                                    //
//-------------------------------------------------------------------------//
void mask(int noofargs, char **arg)
{
  static int ino2=ci;
  enum{ MASK, IMASK, ADDMASK, SUBMASK, MULMASK };
  static int mode=MASK, xoffset=0, yoffset=0;
  int bpp, ino1=ci, f,i,i2,i3,i4,j,k,rr1,gg1,bb1,rr2,gg2,bb2,value1,value2;
  if(ci<0 && noofargs==0){ message("Select an image first"); return; }
  Dialog dialog;
  dialog.needrepeat=0;

  if(noofargs==0)
  {
     strcpy(dialog.title,"Mask  (Combine two images)");
     strcpy(dialog.radio[0][0],"Operation"); 
     strcpy(dialog.radio[0][1],"Mask (1&=2)");             
     strcpy(dialog.radio[0][2],"Inverse Mask (1&=~2)");             
     strcpy(dialog.radio[0][3],"Add (1+=2)");
     strcpy(dialog.radio[0][4],"Subtract (1-=2)");
     strcpy(dialog.radio[0][5],"Multiply (1*=2)");
     dialog.radiono[0]=6;
     dialog.radioset[0]=mode+1;

     strcpy(dialog.boxes[0],"Data");
     strcpy(dialog.boxes[1],"Image to change");
     strcpy(dialog.boxes[2],"Image for mask");
     strcpy(dialog.boxes[3],"X offset");
     strcpy(dialog.boxes[4],"Y offset");

     sprintf(dialog.answer[1],"%d",ino1);
     sprintf(dialog.answer[2],"%d",ino2);

     dialog.boxtype[0]=LABEL;
     dialog.boxtype[1]=INTCLICKBOX;
     dialog.boxtype[2]=INTCLICKBOX;
     dialog.boxtype[3]=INTCLICKBOX;
     dialog.boxtype[4]=INTCLICKBOX;

     dialog.boxmin[1]=1; dialog.boxmax[1]=g.image_count-1;
     dialog.boxmin[2]=1; dialog.boxmax[2]=g.image_count-1;
     dialog.boxmin[3]=-1000; dialog.boxmax[3]=1000;
     dialog.boxmin[4]=-1000; dialog.boxmax[4]=1000;
      
     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
     itoa(ino1,dialog.answer[1],10);
     itoa(ino2,dialog.answer[2],10);
     itoa(xoffset,dialog.answer[3],10);
     itoa(yoffset,dialog.answer[4],10);   

     dialog.noofradios=1;
     dialog.noofboxes=5;
     dialog.helptopic=0;  
     dialog.keep=0;
     dialog.want_changecicb = 0;
     dialog.f1 = null;
     dialog.transient = 0;
     strcpy(dialog.path,".");
  
     //---------------//

     dialogbox(dialog);

     //---------------//
   
     if(g.getout) return;
     mode = dialog.radioset[0] - 1;
     ino1 = atoi(dialog.answer[1]);
     ino2 = atoi(dialog.answer[2]);
     xoffset = atoi(dialog.answer[3]);
     yoffset = atoi(dialog.answer[4]);

  }else
  {  if(noofargs>=1) ino1 = image_number(arg[1]);
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=2) ino2 = image_number(arg[2]);
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=3) mode = atoi(arg[3]); 
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=4) xoffset = atoi(arg[4]); 
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=5) yoffset = atoi(arg[5]); 
  }

  if(z[ino1].bpp != z[ino2].bpp){message("Both images must be\nsame bits/pixel"); return; } 
  switch_palette(ino1);
  drawselectbox(OFF);
  bpp = z[ino1].bpp;  
  f = z[ino1].cf;  

  if(z[ino1].colortype==GRAY)
  {    for(j=0; j<z[ino1].ysize; j++)
       for(i=0,i2=0,i3=xoffset,i4=xoffset*g.off[bpp]; i<z[ino1].xsize; 
           i++, i2+=g.off[bpp], i3++, i4+=g.off[bpp])
       {   if(i>=z[ino2].xsize || j>=z[ino2].ysize) continue;
           if(i3>= z[ino1].xsize || j+yoffset >= z[ino1].ysize) continue;
           value1 = pixelat(z[ino1].image[f][j+yoffset]+i4, bpp);
           value2 = pixelat(z[ino2].image[f][j]+i2, bpp);
           switch(mode)
           {   case MASK: if(!value2) value1=0; break;
               case IMASK: if(value2) value1=0; break;
               case ADDMASK: value1 += value2; break;
               case SUBMASK: value1 -= value2; break;
               case MULMASK: value1 *= value2; break;
           }
           value1 = max(0, min(value1, (int)g.maxvalue[bpp]));
           putpixelbytes(z[ino1].image[f][j+yoffset]+i4,value1,1,bpp);
       }
       z[ino1].hitgray = 0;   // Force remapping of gray levels
  }else
  {
       for(j=0; j<z[ino1].ysize; j++)
       for(i=0,i2=0,i3=xoffset,i4=xoffset*g.off[bpp]; i<z[ino1].xsize; 
           i++, i2+=g.off[bpp], i3++, i4+=g.off[bpp])
       {   if(i>=z[ino2].xsize || j>=z[ino2].ysize) continue;
           if(i3 >= z[ino1].xsize || j+yoffset >= z[ino1].ysize) continue;
           value1 = pixelat(z[ino1].image[f][j+yoffset]+i4, bpp);
           value2 = pixelat(z[ino2].image[f][j]+i2, bpp);
           valuetoRGB(value1,rr1,gg1,bb1,bpp);
           valuetoRGB(value2,rr2,gg2,bb2,bpp);

           switch(mode)
           {   case MASK: if(!rr2) rr1=0; if(!gg2) gg1=0; if(!bb2) bb1=0; break;
               case IMASK: if(rr2) rr1=0; if(gg2) gg1=0; if(bb2) bb1=0; break;
               case ADDMASK: rr1+=rr2; gg1+=gg2; bb1+=bb2; break;
               case SUBMASK: rr1-=rr2; gg1-=gg2; bb1-=bb2; break;
               case MULMASK: rr1*=rr2; gg1*=gg2; bb1*=bb2; break;
           }
           rr1 = max(0, min(rr1, g.maxred[bpp]));
           gg1 = max(0, min(gg1, g.maxgreen[bpp]));
           bb1 = max(0, min(bb1, g.maxblue[bpp]));
           value1 = RGBvalue(rr1,gg1,bb1,bpp);
           putpixelbytes(z[ino1].image[f][j+yoffset]+i4,value1,1,bpp);
       }
  }
  z[ino1].touched=1;
  repair(ino1);
  return;
}
