//--------------------------------------------------------------------------//
// xmtnimage25.cc                                                           //
// IPC                                                                      //
// Latest revision: 03-12-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Images     *z;
extern int         ci;

int s_pipe(int fd[2]);

//--------------------------------------------------------------------------//
//  select_plugin                                                           //
//  mode = 0 (test plugin) or 1 (execute plugin). In testing mode, the      //
//  plugin writes to stdout to allow user to debug it.                      //
//  In a stream pipe, each channel is bidirectional. The child reads and    //
//  writes to fd[1] and the parent reads and writes to fd[0].               //
//--------------------------------------------------------------------------//
int select_plugin(int nargs, char **arg)
{
  const int CMDLEN=128;
  char *info[1024];            // Maximum of 1024 plugins
  FILE *fp;
  int k, mode, count=0, selection = 0, noofargs;
  listinfo l;
  if(ci<0) { message("Please select an image",ERROR); return(NOIMAGES); }
  char tempstring[FILENAMELENGTH];
  char command[FILENAMELENGTH];
  char pluginfile[FILENAMELENGTH];
  if(nargs==0)
  {   mode=1;
      sprintf(pluginfile,"%s/%s", g.helpdir, "plugins");
      if((fp=fopen(pluginfile,"rt")) == NULL)
      {   sprintf(pluginfile,"%s/%s",g.homedir,"plugins");
          if((fp=fopen(pluginfile,"rt")) == NULL)
          {   error_message(pluginfile, errno);
              sprintf(tempstring, "Can't find list of plugins\nshould be %s",pluginfile);
              message(tempstring,ERROR);
              return ERROR;
          }
      }
      while(!feof(fp))             // Count no. of plugins
      {    fgets(tempstring,FILENAMELENGTH-1,fp); 
           if(strlen(tempstring)==0) break;
           if(tempstring[0]==10) break;
           tempstring[100]=0;
           info[count] = new char[100]; 
           for(k=0;k<(int)strlen(tempstring);k++) 
               if(tempstring[k]==10) tempstring[k]=0;
           strcpy(info[count],tempstring);
           if(feof(fp))break;
           count++;
      }
      fclose(fp);
 
      sprintf(pluginfile,"%s%s",g.homedir,"/plugins");
      if((fp=fopen(pluginfile,"rt")) != NULL)
      {    while(!feof(fp))             // Count no. of plugins
           {    fgets(tempstring,FILENAMELENGTH-1,fp); 
                if(strlen(tempstring)==0) break;
                if(tempstring[0]==10) break;
                tempstring[100]=0;
                info[count] = new char[100]; 
                for(k=0;k<(int)strlen(tempstring);k++) 
                    if(tempstring[k]==10) tempstring[k]=0;
                strcpy(info[count],tempstring);
                if(feof(fp))break;
                count++;
           }
           fclose(fp);
      }


      l.title = new char[100];
      strcpy(l.title,"Available plugins");
      l.info = info;
      l.size = count;
      l.itemstoshow = 8;
      l.firstitem   = 0;
      l.wantsort    = 0;
      l.wantsave    = 0;
      l.helptopic   = 0;
      l.selection   = &selection;
      l.allowedit   = 1;
      l.edittitle  = "Enter arguments for plugin";
      l.editrow    = 0;
      l.editcol    = 0;
      l.maxstringsize = 99;
      l.width       = 0;
      l.transient   = 1;    
      l.wantfunction = 0;
      l.autoupdate   = 0;
      list(&l);
      selection = *l.selection;
      strcpy(command, info[selection]);
      if(g.getout){ g.getout=0; return ABORT; }
  }else
  {   strcpy(command, arg[1]);
      mode = atoi(arg[2]);
  }

  parsecommand(command, arg, noofargs, CMDLEN, 20);
  execute_plugin(arg[0], arg, noofargs, mode);
  return OK;
}


//--------------------------------------------------------------------------//
//  execute_plugin                                                          //
//--------------------------------------------------------------------------//
int execute_plugin(char *filename, char **arg, int noofargs, int mode)
{
   enum{ TEST, EXECUTE };
   int count, k, n=0, fd[2], have_message=0, size,x1,y1,x2,y2;
   pid_t pid;
   char tempstring[1024];
   char zero[10] = {0,0,0,0,0,0,0,0,0,0};
   int changed[MAXIMAGES];
   int oimagecount = g.image_count;
   if(ci>=0) {  x1=0;     y1=0;     x2=z[ci].xsize; y2=z[ci].ysize;  }
   else      {  x1=g.ulx; y1=g.uly; x2=g.lrx;       y2=g.lry;  }

   
   if(signal(SIGPIPE, sig_pipe) == SIG_ERR) fprintf(stderr, "signal error\n");
   if(signal(SIGCHLD, sig_child) == SIG_ERR) fprintf(stderr, "sigchld\n");
   if(s_pipe(fd)<0) fprintf(stderr, "pipe error\n");
   if((pid=fork())<0) fprintf(stderr, "the fork is not with you\n");
   else if(pid>0)   // Parent
   {   
        close(fd[1]);

        ////  Up to 20 command line arguments length up to 128
        ////  separated by spaces
        for(k=0;k<=noofargs;k++)
        {  n+=write(fd[0], arg[k], strlen(arg[k]));
           n+=write(fd[0], " ", 1);
        }

        ////  Send currently-selected area as command line arguments
        sprintf(tempstring,"%d ",x1);        
        n+=write(fd[0], tempstring, strlen(tempstring));
        sprintf(tempstring,"%d ",y1); 
        n+=write(fd[0], tempstring, strlen(tempstring));
        sprintf(tempstring,"%d ",x2); 
        n+=write(fd[0], tempstring, strlen(tempstring));
        sprintf(tempstring,"%d ",y2); 
        n+=write(fd[0], tempstring, strlen(tempstring));

        ////  Mark the end of command-line arguments with a null
        n+=write(fd[0], zero, 1);

        ////  Current image no.
        write(fd[0], &ci, sizeof(int));
        
        ////  'mode' can be TEST or EXECUTE
        write(fd[0], &mode, sizeof(mode));

        ////  No. of images present
        write(fd[0], &g.image_count,  sizeof(g.image_count));

        ////  Current screen depth
        write(fd[0], &g.bitsperpixel, sizeof(g.bitsperpixel));

        ////  Send image params.
        ////  Send all the images so plugin can do multi-image calculations.
        for(k=1;k<g.image_count;k++)
        {
            changed[k]=0;
            write(fd[0], &z[k], sizeof(z[k]));
            size = z[k].frames * z[k].xsize * z[k].ysize * g.off[z[k].bpp];
            write_fd(z[k].image_1d, fd[0], size);
            if(z[k].floatexists)
            {     size = z[k].fftxsize * z[k].fftysize * sizeof(complex);
                  write_fd((uchar *)z[k].fft_1d, fd[0], size);
            }
            write(fd[0], z[k].palette, 768);
            write(fd[0], z[k].opalette, 768);
            write(fd[0], z[k].spalette, 768);
            write(fd[0], z[k].name, FILENAMELENGTH);
            write(fd[0], z[k].cal_title[0], FILENAMELENGTH);
            write(fd[0], z[k].cal_title[1], FILENAMELENGTH);
            write(fd[0], z[k].gamma, 256*sizeof(short int));
        }

        ///////////////////////////////////////////////////////////////////
        ////  Finished writing data - block here until input comes in  ////
        ///////////////////////////////////////////////////////////////////

        ////  Read back and display error message from plugin

        n=read(fd[0], &have_message, sizeof(int));
        n=read(fd[0], tempstring, 1024);

        ////  These lines should match `write_data()' in plugin.
        n=read(fd[0], &ci, sizeof(int));
        n=read(fd[0], &g.ulx, sizeof(int));
        n=read(fd[0], &g.uly, sizeof(int));
        n=read(fd[0], &g.lrx, sizeof(int));
        n=read(fd[0], &g.lry, sizeof(int));
        n=read(fd[0], &g.image_count, sizeof(int));
        count = g.image_count;

        ////  Can't read z[k] because it contains pointers.
        for(k=1; k<count; k++)
        {  
             n=read(fd[0], &changed[k], sizeof(int));
             n=read(fd[0], &z[k].xpos, sizeof(z[k].xpos));
             n=read(fd[0], &z[k].ypos, sizeof(z[k].ypos));
             n=read(fd[0], &z[k].xsize, sizeof(z[k].xsize));
             n=read(fd[0], &z[k].ysize, sizeof(z[k].ysize));
             n=read(fd[0], &z[k].bpp, sizeof(z[k].bpp));
             n=read(fd[0], &z[k].frames, sizeof(z[k].frames));
             n=read(fd[0], &z[k].colortype, sizeof(z[k].colortype));

             ////  If they created a new image
             if(k>=oimagecount)
             {   ////  This changes g.image_count and ci
                 g.image_count--;
                 newimage(0,0,z[k].xsize,z[k].ysize,z[k].bpp,z[k].colortype,
                       z[k].frames,g.want_shell,PERM,1,g.window_border,0);
             }
             size = z[k].frames * z[k].xsize * z[k].ysize * g.off[z[k].bpp];

             n=read_fd(z[k].image_1d, fd[0], size);
             if(z[k].floatexists)
             {     size = z[k].fftxsize * z[k].fftysize * sizeof(complex);
                   read_fd((uchar *)z[k].fft_1d, fd[0], size);
             }
             n=read(fd[0], &z[k].floatexists, sizeof(z[k].floatexists));
             n=read(fd[0], &z[k].backedup, sizeof(int));
             n=read(fd[0], &z[k].fftdisplay, sizeof(int));
             n=read(fd[0], &z[k].fmin, sizeof(double));
             n=read(fd[0], &z[k].fmax, sizeof(double));
             n=read(fd[0], z[k].gray, 4*sizeof(int));
             n=read(fd[0], &z[k].fftstate, sizeof(int));
             n=read(fd[0], z[k].palette, 768);
             n=read(fd[0], z[k].opalette, 768);
             n=read(fd[0], z[k].spalette, 768);
             n=read(fd[0], z[k].name,  FILENAMELENGTH);
             n=read(fd[0], z[k].cal_title[0], FILENAMELENGTH);
             n=read(fd[0], z[k].cal_title[1], FILENAMELENGTH);
             n=read(fd[0], z[k].gamma, 256*sizeof(short int));
             if(k>=oimagecount)
             {    setpalette(z[k].palette);
                  rebuild_display(ci);
             }
        }
        sleep(1);   
        wait(0);    // Wait for child process to terminate
        close(fd[0]);
        if(n==0) message("Child closed pipe",ERROR);
        if(ferror(stdin)) message("fgets error on stdin",ERROR);

   }else            // Child
   { 
        close(fd[0]);
        if(fd[1] != STDIN_FILENO)
        {    if(dup2(fd[1], STDIN_FILENO) != STDIN_FILENO)
                 fprintf(stderr, "dup2 error to stdin\n");
        }
        if(fd[1] != STDOUT_FILENO && mode==EXECUTE)
        {    if(dup2(fd[1], STDOUT_FILENO) != STDOUT_FILENO)
             fprintf(stderr, "dup2 error to stdout\n");
        }
        if(execl(filename, filename, NULL) <0) fprintf(stderr, "execl error\n");
        close(fd[1]);
        _exit(0);     // Child must exit
   }


    if(have_message) message(tempstring);
    drawselectbox(OFF);
    for(k=1; k<g.image_count; k++) if(changed[k]) repair(k);
    return OK;
}    

        
//--------------------------------------------------------------------------//
//  sig_pipe                                                                //
//--------------------------------------------------------------------------//
void sig_pipe(int signo)
{
   fprintf(stderr, "Caught SIGPIPE (%d), crashing\n",signo);
   exit(1);
}


//--------------------------------------------------------------------------//
//  sig_child                                                               //
//--------------------------------------------------------------------------//
void sig_child(int signo)
{
   signo=signo;
   char tempstring[1024];
   int statloc;
   wait(&statloc);
   sprintf(tempstring, "Process returned %d\n", statloc);
   if(WIFEXITED(statloc)) strcat(tempstring,"(Successful)");
   else if(WIFSIGNALED(statloc)) strcat(tempstring,"(Abnormal termination)"); 
   else if(WIFSTOPPED(statloc)) strcat(tempstring,"(Stopped)");  
   else if(statloc) strcat(tempstring,"(Unsuccessful)");  
   message(tempstring);
}


//--------------------------------------------------------------------------//
//  s_pipe - open bidirectional BSD style stream pipe                       //
//--------------------------------------------------------------------------//
int s_pipe(int fd[2])
{
    return(socketpair(AF_UNIX, SOCK_STREAM, 0, fd));
}
 


//--------------------------------------------------------------------------//
//  write_fd - write to a file descriptor. Can only send 62000 bytes at a   //
//  time so break it up into <=1k chunks.  Returns bytes written.           //
//--------------------------------------------------------------------------//
int write_fd(uchar *source, int fd, int size)
{
   int n;
   int total=0;        // bytes sent
   int rem = size;     // bytes remaining to send
   while(total<size)   
   {     n = write(fd, source + total, min(rem,1024)); 
         if(n<=0){ message("Error writing to plugin"); break; }
         total += n;
         rem -= n;
   }
   return total;
}


//--------------------------------------------------------------------------//
//  read_fd - read from a file descriptor. Can only read 62000 bytes at a   //
//  time so break it up into <=1k chunks.  Returns bytes read.              //
//--------------------------------------------------------------------------//
int read_fd(uchar *dest, int fd, int size)
{
   int n;
   int total=0;        // bytes read
   int rem = size;     // bytes remaining to read
   while(total<size)   
   {    n = read(fd, dest + total, min(rem,1024));
        if(n<=0){ message("Error reading from plugin"); break; }
        total += n;
        rem -= n;
   }
   return total;
}

