//--------------------------------------------------------------------------//
// xmtnimage28.cc                                                           //
// dialog box                                                               //
// Latest revision: 04-14-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Images     *z;
extern int         ci;
extern PrinterStruct printer;

//--------------------------------------------------------------------//
// dialogbox (Motif)                                                  //
//  radiotype      method                                             //
//  --------      -----------                                         //
//  RADIO         simple toggle radio button.                         //
//  RADIOTEXT     get text and toggle radio button. Returns           //
//                    answer as string in a.radioanswer[][].          //
//  RADIOVALUE    get value from slider and toggle radio button.      //
//                    Returns answer as string in a.radioanswer[][].  //
//                    If a.radio[][] contains an '=', the string is   //
//                    also substituted for whatever was after the '='.//
//  RADIOVALUEWIDE  same as RADIOVALUE except wider.                  //
//  boxtype       method                                              //
//  -------       ---------------                                     //
//  TOGGLE        toggle button with label                            //
//  TOGGLESTRING  toggle button with label and a string.              //
//  INTCLICKBOX   integer clickbox() using sliders                    //
//  FILENAME      getfilename()                                       //
//  MULTICLICKBOX multiclickbox()(Assumes you want a rgb)             //
//  TOGGLEMULTICLICKBOX  toggle button with label and a multiclickbox.//
//  NON_EDIT_LIST non-editable list                                   //
//  STRING        getstring                                           //
//  LIST          editable list                                       //
//  LABEL         heading only                                        //
//  NO_LABEL_LIST editable list except no label on button             //
//  DOUBLECLICKBOX float clickbox() using sliders, uses clickboxstep[]//
//                                                                    //
// NOTE: It is the calling routine's responsibility to restore the    //
//     original parameters if "Cancel" is selected, using the global  //
//     variable 'getout', and then to restore 'getout' to 0.          //
// All widgets created here must be destroyed before leaving.         //
//     This is done by setting w[wc] to each widget as it is created, //
//     then deleting all the widgets at once.                         //
// All answers are placed into dialoginfo struct by callbacks. This   //
//     allows the calling routine to keep the dialog on the screen by //
//     setting 'a.keep', and then change the parameters asynchronous- //
//     ly. If a.keep is set, calling routine must call dialogbox      //
//     again, setting a.keep to RESET, to close the dialog.           //
// Boxes and radios can set up groups or gray out unwanted options    //
//     using a.f1 function pointer.                                   //
// All elements in a.boxwidget[][] must be set to 0 !                 //
// The callback f1() can use XtSetSensitive to gray out unwanted      //
//     options based on the radio and box settings. It must check     //
//     to make sure the widget is not 0 first, because unused widgets //
//     are initialized to 0.                                          //
//--------------------------------------------------------------------//
void dialogbox(dialoginfo &a)
{
 static char dirmask[FILENAMELENGTH]="*";
 char tempstring[FILENAMELENGTH];
 clickboxinfo *clickboxdata=NULL;
 Widget *w=NULL;
 Widget form=0;
 Widget radiopush[10];
 int n,wc=0,i,j,k,value,boxcount=0,radiocount=0,count,result,startpos=0,
     multiclickboxcount=0,nonmulticlickboxcount=0;
 int boxlabelwidth=0;
 int color[4];
 XmString xms;
 double spacing, bspacing, rspacing, y;
 char junk[2];
 junk[0]=32;
 junk[1]=0;
 if(a.keep!=RESET)
 {  
      Arg args[100];
      clickboxdata = new clickboxinfo[DCOUNT];  // Up to DCOUNT sliders
      w = new Widget[200];
      char  *ptr, temp[256], *boxstring, radiostring[256];
      Widget wid[4], radiobox[DCOUNT], radiobutton[10][DCOUNT], boxbutton[DCOUNT];
      Widget radioslider[10][DCOUNT],widget;

      ////  Percentages of dialog box width.
      ////  To get uniform positioning of widgets, make sure each widget
      ////  is attached to a space proportionately equal to its expected width.
      ////  Motif expands the dialog box and all other widgets to give the
      ////  smallest widget enough pixels and simultaneously keep the 
      ////  percentages to the set values.

      g.getout=0;
      g.dialogdone=0;

      //----Calculate how large it has to be-----------------------------//

      double TOPMARGIN = 2;
      double RADIOSTART= 1;
      double RADIOEND  = 37;
      double BOXSTART  = 37.2;
      double CHECKWIDTH= 8;
      double LABELEND  = 99;
      double BOXEND, LABELSTART;
 
      for(j=0;j<a.noofboxes;j++) 
      {    boxcount++;
           if(a.boxtype[j]!=TOGGLE && a.boxtype[j]!=LABEL)
               boxlabelwidth = max(boxlabelwidth, (int)strlen(a.boxes[j]));
           if(a.boxtype[j]==MULTICLICKBOX) multiclickboxcount++;
           else if(a.boxtype[j]!=LABEL) nonmulticlickboxcount++;
      }
      BOXEND = LABELSTART = min(80,100-(int)(1.4*boxlabelwidth));

      for(i=0;i<10;i++) radiopush[i]=0;
      for(i=0;i<a.noofradios;i++)
      {   radiopush[i]=0;
          for(j=0;j<a.radiono[i];j++)
          {   radiocount++; 
              if(a.radiotype[i][j]==RADIOVALUE || a.radiotype[i][j]==RADIOVALUEWIDE) 
              {    RADIOEND=29;BOXSTART=48;BOXEND=82;LABELSTART=82; }
              if(a.radiotype[i][j]==RADIOTEXT) 
              {    RADIOEND=31;BOXSTART=40;BOXEND=72;LABELSTART=72; }
          }
      }
      if(a.noofboxes==0) RADIOEND=BOXSTART=BOXEND=LABELSTART=LABELEND=98;
      if(a.noofradios==0)
      {     RADIOSTART=RADIOEND=BOXSTART=2; 
            TOPMARGIN=4; 
            BOXEND=LABELSTART=55; 
            CHECKWIDTH=8;
      }
      if(nonmulticlickboxcount==0 && multiclickboxcount>0)
      {     RADIOEND=39;
            BOXSTART=40; 
            BOXEND=LABELSTART=63; 
      }
 
      count = max(boxcount,radiocount)+3;
      spacing = (100.0/(double)count);

      //------- Create a form dialog shell widget -----------------------//

      n=0;
      if(a.transient){  XtSetArg(args[n], XmNtransient, True); n++; }
      else           {  XtSetArg(args[n], XmNtransient, False); n++; }
      XtSetArg(args[n], XmNresizable, True); n++;
      XtSetArg(args[n], XmNmarginHeight, 0); n++;
      XtSetArg(args[n], XmNmarginWidth, 0); n++;
      XtSetArg(args[n], XmNheight, count*28); n++;
      XtSetArg(args[n], XmNautoUnmanage, False); n++;

      ////  Stop Motif from trying to grab another color if none are available.
      if(g.want_colormaps)  
      {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
           XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
      }
      XtSetArg(args[n], XmNkeyboardFocusPolicy, XmEXPLICIT); n++;
      XtSetArg(args[n], XmNnavigationType, XmEXCLUSIVE_TAB_GROUP); n++;
      XtSetArg(args[n], XmNdialogTitle, xms = XmStringCreateSimple(a.title));n++;
      form = w[wc++] = XmCreateFormDialog(g.main_widget, "DialogForm", args, n);
      XmStringFree(xms);

      w[wc++] = addlabel(form, a.title, CENTER, 20, 3, 80, 3+spacing); 
      y=TOPMARGIN+spacing;
      rspacing = 1.03*spacing;
      for(i=0;i<a.noofradios;i++)
      {
            ////  'radiobox' is a container widget so leave bottom position
            ////  unattached so it can accommodate new pushbuttons. Otherwise,
            ////  Motif makes them all different sizes.
            ////  'Horizontal' orientation means buttons are stacked vertically.

            w[wc++] = addlabel(form,a.radio[i][0],LEFT,RADIOSTART,y,RADIOEND,y+rspacing); 
            y+=rspacing;

            n=0;
            XtSetArg(args[n], XmNspacing, 0); n++;
            XtSetArg(args[n], XmNradioBehavior, True); n++;
            XtSetArg(args[n], XmNradioAlwaysOne, True); n++;
            XtSetArg(args[n], XmNleftPosition, RADIOSTART); n++;  
            XtSetArg(args[n], XmNtopPosition, y); n++;    
            XtSetArg(args[n], XmNrightPosition, RADIOEND); n++;  
            XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
            XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
            XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
            XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
            XtSetArg(args[n], XmNbottomPosition, y+rspacing*.91*(-1+(double)a.radiono[i])); n++; 
            XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
            XtSetArg(args[n], XmNpacking,XmPACK_COLUMN); n++;
            XtSetArg(args[n], XmNnumColumns,50); n++;
            XtSetArg(args[n], XmNorientation,XmHORIZONTAL); n++;
            XtSetArg(args[n], XmNresizable, False); n++;
            XtSetArg(args[n], XmNadjustLast, False); n++;
            ////  Stop Motif from trying to grab another color if none are available.
            if(g.want_colormaps)  
            {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                 XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
            }
            radiobox[i] = w[wc++] = XmCreateRadioBox(form, "radiobox", args, n);
            XtManageChild(radiobox[i]);
            a.radiobox[i]=radiobox[i];
            for(j=1;j<a.radiono[i];j++)
            {  
                ////  Create main radio buttons - set shadowThickness to 1 to create
                ////  a border.

                n=0;
                XtSetArg(args[n], XmNresizable, False); n++;
                XtSetArg(args[n], XmNspacing, 0); n++;
                XtSetArg(args[n], XmNmarginWidth, 4); n++;
                XtSetArg(args[n], XmNleftPosition, RADIOSTART); n++; 
                XtSetArg(args[n], XmNtopPosition, y); n++;  
                XtSetArg(args[n], XmNbottomPosition, y+rspacing); n++; 
                XtSetArg(args[n], XmNrightPosition, RADIOEND); n++;  
                XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                XtSetArg(args[n], XmNpacking,XmPACK_COLUMN); n++;
                XtSetArg(args[n], XmNnumColumns,50); n++;
                XtSetArg(args[n], XmNorientation,XmHORIZONTAL); n++;
                strcpy(radiostring, a.radio[i][j]);
                for(k=strlen(radiostring);k<20;k++)strcat(radiostring,junk);
                XtSetArg(args[n], XmNindicatorOn, True); n++;
                XtSetArg(args[n], XmNindicatorType, XmONE_OF_MANY); n++;
                XtSetArg(args[n], XmNfillOnSelect, True); n++;
                XtSetArg(args[n], XmNspacing, 5); n++;
                ////  Stop Motif from trying to grab another color if none are available.
                if(g.want_colormaps)  
                {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                     XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                }
                radiobutton[i][j] = w[wc++] =  
                     XmCreateToggleButton(radiobox[i],radiostring,args,n);
                a.radiowidget[i][j] = radiobutton[i][j];
                XtManageChild(radiobutton[i][j]);
                if(a.radioset[i]==j)
                    XmToggleButtonSetState(radiobutton[i][j], True, True); 
                else
                    XmToggleButtonSetState(radiobutton[i][j], False, True);                

                ////  Set up a callback for each radio button. They must both be
                ////  disarmcallbacks (called when they unclick on it).
                ////  If armcallbacks - it will read wrong button state before calling f1.
                ////  If valuechanged - it will call f1 twice.
                ////  If different - it will only call one of them.

#ifndef LESSTIF_VERSION
                XtAddCallback(radiobutton[i][j], XmNdisarmCallback, (XtCBP)togglecb, (XtP)&a);
                XtAddCallback(radiobutton[i][j], XmNdisarmCallback, (XtCBP)armboxcb, (XtP)&a);
#endif
                ////  However, in lesstif the above doesn't work for some reason.
		////  Apparently, no disarmcb.
		////  So add yet another hack specific for lesstif.
#ifdef LESSTIF_VERSION
                XtAddCallback(radiobutton[i][j], XmNarmCallback, (XtCBP)togglecb, (XtP)&a);
                XtAddCallback(radiobutton[i][j], XmNarmCallback, (XtCBP)armboxcb, (XtP)&a);
#endif

                XtManageChild(radiobutton[i][j]);
                
                ////  If pushbutton needed too in radio button.
                ////  Put on form Widget so it doesn't mess up size and spacing 
                ////  of the radio buttons.

                if(a.radiotype[i][j]==RADIOPUSHBUTTON) 
                {   n = 0;
                    XtSetArg(args[n], XmNleftPosition, RADIOSTART+3); n++; 
                    XtSetArg(args[n], XmNtopPosition, y+0.85*rspacing); n++;  
                    XtSetArg(args[n], XmNbottomPosition, y+1.85*rspacing); n++; 
                    XtSetArg(args[n], XmNrightPosition, RADIOEND-5); n++;  
                    XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                    XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNresizable, False); n++;  
#ifdef MOTIF2
                    XtSetArg(args[n], XmNindicatorOn, XmINDICATOR_NONE); n++;
#else 
                    XtSetArg(args[n], XmNindicatorOn, False); n++;
#endif
                    XtSetArg(args[n], XmNtraversalOn, True); n++;
                    XtSetArg(args[n], XmNlabelString,
                      xms = XmStringCreateSimple(a.radiopushbuttonlabel[i]));n++; 
                    ////  Stop Motif from trying to grab another color if none are available.
                    if(g.want_colormaps)  
                    {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                         XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                    }
                    radiopush[i] = w[wc++] = XmCreatePushButton(form,//radiobox[i], 
                         a.radiopushbuttonlabel[i], args,n);
                    XtAddCallback(radiopush[i], XmNactivateCallback, 
                        (XtCBP)a.radiocb[i], (XtP)(a.radiopushbuttonptr[i]));
                    XmStringFree(xms);
                }
                 
                ////  Text needed too in radio button
                if(a.radiotype[i][j]==RADIOTEXT) 
                {   n = 0;
                    XtSetArg(args[n], XmNresizable, False); n++;
                    XtSetArg(args[n], XmNspacing, 0); n++;
                    XtSetArg(args[n], XmNtopPosition, y-rspacing); n++;    
                    XtSetArg(args[n], XmNbottomPosition, y+0*rspacing); n++; 
                    XtSetArg(args[n], XmNmarginWidth, 2); n++;
                    XtSetArg(args[n], XmNleftPosition, RADIOEND); n++;
                    XtSetArg(args[n], XmNrightPosition, BOXSTART); n++;  
                    XtSetArg(args[n], XmNfractionBase, 100); n++;   
                    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNeditMode, XmSINGLE_LINE_EDIT); n++;
                    XtSetArg(args[n], XmNcolumns, 2); n++;
                    XtSetArg(args[n], XmNrows, 1); n++;
                    XtSetArg(args[n], XmNwordWrap, False); n++;
                    XtSetArg(args[n], XmNeditable, True); n++;
                    ////  Stop Motif from trying to grab another color if none are available.
                    if(g.want_colormaps)  
                    {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                         XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                    }
                    radioslider[i][j] = w[wc++] = XmCreateText(form, "text", args, n);
                    a.radioslider[i][j] = radioslider[i][j];
                    XtManageChild(radioslider[i][j]);
                    XmTextSetString(radioslider[i][j], a.radioanswer[i][j]);
	        }

                //// Slider needed too in radio button
                if(a.radiotype[i][j]==RADIOVALUE || a.radiotype[i][j]==RADIOVALUEWIDE) 
                {   value = atoi(a.radioanswer[i][j]);
                    n = 0;
                    XtSetArg(args[n], XmNresizable, False); n++;
                    XtSetArg(args[n], XmNspacing, 0); n++;
                    XtSetArg(args[n], XmNleftPosition, RADIOEND); n++;
                    XtSetArg(args[n], XmNtopPosition, y-.2*rspacing); n++;    
                    XtSetArg(args[n], XmNrightPosition, BOXSTART); n++;  
                    XtSetArg(args[n], XmNfractionBase, 100); n++;   
                    XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                    XtSetArg(args[n], XmNminimum, a.radiomin[i][j]); n++;
                    XtSetArg(args[n], XmNmaximum, a.radiomax[i][j]); n++;
                    XtSetArg(args[n], XmNprocessingDirection,XmMAX_ON_RIGHT); n++;
                    XtSetArg(args[n], XmNhighlightOnEnter, True); n++;      
                    XtSetArg(args[n], XmNincrement, 1); n++;
                    XtSetArg(args[n], XmNvalue, value); n++;
                    XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
#ifdef MOTIF2
                    XtSetArg(args[n], XmNshowValue, XmNEAR_BORDER); n++;
                    XtSetArg(args[n], XmNshowArrows, XmEACH_SIDE); n++;      
#else
                    XtSetArg(args[n], XmNshowArrows, True); n++;      
#endif
                    XtSetArg(args[n], XmNdecimalPoints, 0); n++;      
                    ////  Stop Motif from trying to grab another color if none are available.
                    if(g.want_colormaps)  
                    {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                         XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                    }

                    ////  This will create Motif warnings of "Cannot allocate 
                    ////  colormap entry" if want_colormaps is 1, when Motif 
                    ////  tries to grab colors for the descendants Scrollbar 
                    ////  and Title. The warnings are suppressed by defining
                    ////  a warning handler.

                    radioslider[i][j] = w[wc++] = XmCreateScale(form, "slider", args, n);
                    a.radioslider[i][j] = radioslider[i][j];
                    XtManageChild(radioslider[i][j]);
                }
                y+=rspacing;
            }
       }

       //------Answer boxes (input boxes)-------------------------------//

       y=TOPMARGIN+spacing;
       bspacing = spacing;
       for(j=0;j<a.noofboxes;j++)
       {     
           boxstring = a.boxes[j];          // boxstring is a temporary string 
           boxstring[40] = 0;               // Make sure it fits on button
           switch(a.boxtype[j])
           {
              case TOGGLE:
                   n = 0;
                   XtSetArg(args[n], XmNmarginLeft, 15); n++;
                   XtSetArg(args[n], XmNresizable, True); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;     // % of width from left
                   XtSetArg(args[n], XmNtopPosition, y); n++;     // % of height from top
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  // % of width from top
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;   // % of height from left
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    #ifdef MOTIF2
                   XtSetArg(args[n], XmNindicatorOn, XmINDICATOR_CROSS_BOX); n++;
    #endif
                   XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
                   if(a.boxset[j])
                       XtSetArg(args[n], XmNset, True); 
                   else
                       XtSetArg(args[n], XmNset, False);
                   n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   boxbutton[j] = w[wc++] = XmCreateToggleButton(form,boxstring,args,n);
                   a.boxwidget[j][0] = boxbutton[j];
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   XtManageChild(boxbutton[j]);

                   ////  Set up callbacks for each box button
                   XtAddCallback(boxbutton[j], XmNvalueChangedCallback,
                        (XtCBP)toggleboxcb, (XtP)&a);
                   XtAddCallback(boxbutton[j], XmNvalueChangedCallback,
                        (XtCBP)armboxcb, (XtP)&a);
                   break;
              case STRING:
                   n = 0;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;     // % of width from left
                   XtSetArg(args[n], XmNtopPosition, y); n++;     // % of height from top
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  // % of width 
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;   // % of height from left
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrows, 1); n++;
                   XtSetArg(args[n], XmNmarginHeight, 1); n++;
                   XtSetArg(args[n], XmNmarginWidth, 4); n++;
                   XtSetArg(args[n], XmNmaxLength, 256); n++;
                   XtSetArg(args[n], XmNvalue, a.answer[j]);n++; 
                   XtSetArg(args[n], XmNautoShowCursorPosition, True); n++;
                   XtSetArg(args[n], XmNcursorPositionVisible, True); n++;
                   ////  Motif ignores this for some reason
                   XtSetArg(args[n], XmNcursorPosition,(int)strlen(a.answer[j])); n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   boxbutton[j] = w[wc++] = XmCreateText(form, boxstring, args, n);
                   a.boxwidget[j][0] = boxbutton[j];
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   XtManageChild(boxbutton[j]);
                   XtAddCallback(boxbutton[j], XmNvalueChangedCallback, 
                        (XtCBP)dialogstringcb, (XtP)&a);
                   XtAddCallback(boxbutton[j], XmNfocusCallback, 
                        (XtCBP)dialogcursorcb, (XtP)&a);
                   break;
              case SCROLLEDSTRING:
                   n = 0;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;     // % of width from left
                   XtSetArg(args[n], XmNtopPosition, y); n++;     // % of height from top
                   XtSetArg(args[n], XmNbottomPosition, y+1.5*bspacing); n++;  // % of width 
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;   // % of height from left
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrows, 1); n++;
                   XtSetArg(args[n], XmNmarginHeight, 1); n++;
                   XtSetArg(args[n], XmNmarginWidth, 4); n++;
                   XtSetArg(args[n], XmNmaxLength, 256); n++;
                   XtSetArg(args[n], XmNvalue, a.answer[j]);n++; 
                   XtSetArg(args[n], XmNautoShowCursorPosition, True); n++;
                   XtSetArg(args[n], XmNcursorPositionVisible, True); n++;
                   ////  Motif ignores this for some reason
                   XtSetArg(args[n], XmNcursorPosition,(int)strlen(a.answer[j])); n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   boxbutton[j] = w[wc++] = XmCreateScrolledText(form, boxstring, args, n);
                   a.boxwidget[j][0] = boxbutton[j];
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   XtManageChild(boxbutton[j]);
                   XtAddCallback(boxbutton[j], XmNvalueChangedCallback, 
                        (XtCBP)dialogstringcb, (XtP)&a);
                   XtAddCallback(boxbutton[j], XmNfocusCallback, 
                        (XtCBP)dialogcursorcb, (XtP)&a);
                   y+=0.4*bspacing;
                   break;
              case TOGGLESTRING:
                   //// Toggle button portion - The callback f1 must unset other
                   //// buttons, gray out unwanted options, etc. - it is not done
                   //// automatically.

                   n = 0;
                   XtSetArg(args[n], XmNmarginLeft, 15); n++;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;   
                   XtSetArg(args[n], XmNtopPosition, y); n++;    
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  
                   XtSetArg(args[n], XmNrightPosition, BOXSTART+CHECKWIDTH); n++; 
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    #ifdef MOTIF2
                   XtSetArg(args[n], XmNindicatorOn, XmINDICATOR_CROSS_BOX); n++;
    #endif
                   XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
                   if(a.boxset[j]){  XtSetArg(args[n], XmNset, True); n++; }
                   else           {  XtSetArg(args[n], XmNset, False); n++; }
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   w[wc++] = boxbutton[j] =XmCreateToggleButton(form," ",args,n);
                   a.boxwidget[j][0] = boxbutton[j];
                   XtManageChild(boxbutton[j]);
                   ////  Set up callbacks for each box button

                   XtAddCallback(a.boxwidget[j][0], XmNarmCallback,(XtCBP)armboxcb, 
                        (XtP)&a);
                   XtAddCallback(a.boxwidget[j][0], XmNvalueChangedCallback,
                        (XtCBP)toggleboxcb, (XtP)&a);
                   if(a.boxset[j])
                        XmToggleButtonSetState(a.boxwidget[j][0], True, False); 

                   ////  String portion
                   n = 0;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNresizeWidth, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART+CHECKWIDTH); n++;   
                   XtSetArg(args[n], XmNtopPosition, y); n++;    
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;  
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrows, 1); n++;
                   XtSetArg(args[n], XmNcolumns, 10); n++;
                   XtSetArg(args[n], XmNmarginHeight, 1); n++;
                   XtSetArg(args[n], XmNmarginWidth, 1); n++;
                   XtSetArg(args[n], XmNmaxLength, 256); n++;
                   XtSetArg(args[n], XmNvalue, a.answer[j]);n++; 
                   XtSetArg(args[n], XmNcursorPosition,(int)strlen(a.answer[j])); n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   widget = w[wc++] = XmCreateText(form, boxstring, args, n);
                   w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   a.boxwidget[j][1] = widget;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   XtManageChild(widget);
                   XtAddCallback(widget, XmNvalueChangedCallback, 
                      (XtCBP)dialogstringcb, (XtP)&a);
                   break;
              case INTCLICKBOX:       // Pushbutton calling a clickbox (Slider)
                   n = 0;
                   XtSetArg(args[n], XmNmarginLeft, 4); n++;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;     
                   XtSetArg(args[n], XmNtopPosition, y); n++;     
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++; 
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;  
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNresizable, False); n++; 
                   XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++; 
                   XtSetArg(args[n], XmNlabelString,
                      xms = XmStringCreateSimple(a.answer[j]));n++; 
                   XtSetArg(args[n], XmNmarginWidth, 1); n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   boxbutton[j] = w[wc++] = XmCreatePushButton(form,boxstring,args,n);
                   a.boxwidget[j][0] = boxbutton[j];
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   XtManageChild(boxbutton[j]);
                   XmStringFree(xms);
                   w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   clickboxdata[j].title = boxstring;
                   clickboxdata[j].answer = atoi(a.answer[j]);
                   clickboxdata[j].startval = atoi(a.answer[j]);
                   clickboxdata[j].minval = a.boxmin[j];
                   clickboxdata[j].maxval = a.boxmax[j];
                   clickboxdata[j].type = INTEGER;
                   clickboxdata[j].wantpreview = 0;
                   clickboxdata[j].widget[0] = boxbutton[j];
                   clickboxdata[j].dirmask = 0;
                   clickboxdata[j].f1 = null;
                   clickboxdata[j].f2 = null;
                   clickboxdata[j].f3 = null;
                   a.c[j] = &clickboxdata[j];
                   XtAddCallback(boxbutton[j], XmNactivateCallback, 
                      (XtCBP)dialogclickboxcb, (XtP)&a);
                   XtAddCallback(boxbutton[j], XmNarmCallback, (XtCBP)armboxcb, 
                      (XtP)&a);
                   break;
              case FILENAME:          // Get filename
              case MULTIFILENAME:     // Get list of filenames. dialogclickboxcb()
	                              //  allocates a list & puts it in a->boxlist[][].
                                      // Calling routine must free this list.
                   if((int)a.answer[j][0]==0) strcpy(a.answer[j],"none     ");
                   else strcpy(tempstring, basefilename(a.answer[j]));
                   n = 0;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;     // % of width from left
                   XtSetArg(args[n], XmNtopPosition, y); n++;     // % of height from top
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  // % of width from top
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;   // % of height from left
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNresizable, False); n++; 
                   XtSetArg(args[n], XmNalignment, XmALIGNMENT_END); n++; 
                   XtSetArg(args[n], XmNlabelString, 
                      xms=XmStringCreateSimple(tempstring)); n++;  
                   XtSetArg(args[n], XmNcursorPosition,(int)strlen(tempstring)); n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   boxbutton[j] = w[wc++] = XmCreatePushButton(form,boxstring,args,n);
                   a.boxwidget[j][0] = boxbutton[j];
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   clickboxdata[j].title = a.answer[j];
                   clickboxdata[j].widget[1] = boxbutton[j];
                   clickboxdata[j].f1 = null;
                   clickboxdata[j].f2 = null;
                   clickboxdata[j].f3 = null;
                   clickboxdata[j].wc = 0;
                   clickboxdata[j].w  = new Widget[DCOUNT];
                   clickboxdata[j].form = form;
                   clickboxdata[j].list = NULL;
                   clickboxdata[j].dirmask = new char[FILENAMELENGTH];
                   clickboxdata[j].path = a.path;
                   strcpy(clickboxdata[j].dirmask, dirmask);
                   if(a.boxtype[j] == MULTIFILENAME)
                        clickboxdata[j].type = MULTIPLE;
                   else
                        clickboxdata[j].type = SINGLE;
                   clickboxdata[j].count = 0;
                   a.c[j] = &clickboxdata[j];
                   XtAddCallback(boxbutton[j], XmNactivateCallback, 
                      (XtCBP)dialogfilenamecb, (XtP)&a);
                   XtAddCallback(boxbutton[j], XmNarmCallback, (XtCBP)armboxcb, (XtP)&a);
                   XtManageChild(boxbutton[j]);
                   XmStringFree(xms);
                   startpos = strlen(tempstring);
                   startpos = max(0,startpos-36);
                   break;
              case MULTICLICKBOX:      // RGB multi clickbox - Display 3 boxes that all 
                   n = 0;              // change the same data.
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;     // % of width from left
                   XtSetArg(args[n], XmNtopPosition, y); n++;     // % of height from top
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  // % of width from top
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;   // % of height from left
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   value = atoi(a.answer[j]);
                   valuetoRGB(value,color[1],color[2],color[3],a.bpp);
                   clickboxdata[j].title = boxstring;
                   clickboxdata[j].answer = atoi(a.answer[j]);
                   clickboxdata[j].startval = atoi(a.answer[j]);
                   clickboxdata[j].minval = a.boxmin[j];
                   clickboxdata[j].maxval = a.boxmax[j];
                   clickboxdata[j].type = INTEGER;
                   clickboxdata[j].wantpreview = 0;
                   clickboxdata[j].wantdragcb = 1;
                   clickboxdata[j].f1 = null;
                   clickboxdata[j].f2 = null;
                   clickboxdata[j].f3 = null;
                   a.c[j] = &clickboxdata[j];
                   a.boxwidget[j][0] = 0;
                   for(k=1;k<4;k++)
                   {   XtSetArg(args[n], XmNleftPosition, BOXSTART+(k-1)*7); n++;
                       XtSetArg(args[n], XmNtopPosition, y); n++;    
                       XtSetArg(args[n], XmNrightPosition, BOXSTART+k*7-1); n++;   
                       XtSetArg(args[n], XmNresizable, False); n++; 
                       XtSetArg(args[n], XmNspacing, 0); n++;
                       XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++; 
                       XtSetArg(args[n], XmNresizable, False); n++;
                       XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                       XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                       XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                       XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                       XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                       itoa(color[k],temp,10);
                       XtSetArg(args[n], XmNlabelString,
                          xms = XmStringCreateSimple(temp));n++; 
                       wid[k] = w[wc++] = XmCreatePushButton(form,boxstring,args,n);
                       a.boxwidget[j][k] = wid[k];
                       XtManageChild(wid[k]);
                       XmStringFree(xms);
                       XtAddCallback(wid[k], XmNactivateCallback, 
                          (XtCBP)dialogmulticlickboxcb, (XtP)&a);
                       XtAddCallback(wid[k], XmNarmCallback,(XtCBP)armboxcb, (XtP)&a);
                       w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   }
                   for(i=0;i<4;i++) clickboxdata[j].widget[i] = wid[i];
                   clickboxdata[j].dirmask = 0;
                   break;
              case TOGGLEMULTICLICKBOX:
                   //// Toggle button portion - The callback f1 must unset other
                   //// buttons, gray out unwanted options, etc. - it is not done
                   //// automatically.

                   n = 0;
                   XtSetArg(args[n], XmNmarginLeft, 1); n++;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;   
                   XtSetArg(args[n], XmNtopPosition, y); n++;    
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  
                   XtSetArg(args[n], XmNrightPosition, BOXSTART+CHECKWIDTH); n++; 
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    #ifdef MOTIF2
                   XtSetArg(args[n], XmNindicatorOn, XmINDICATOR_CROSS_BOX); n++;
    #endif
                   XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
                   if(a.boxset[j]){  XtSetArg(args[n], XmNset, True); n++; }
                   else           {  XtSetArg(args[n], XmNset, False); n++; }
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   w[wc++] = boxbutton[j] =XmCreateToggleButton(form," ",args,n);
                   a.boxwidget[j][0] = boxbutton[j];
                   XtManageChild(boxbutton[j]);
                   ////  Set up callbacks for each box button
                   XtAddCallback(boxbutton[j], XmNarmCallback,(XtCBP)armboxcb, (XtP)&a);

                   ////  Multiclickbox portion

                   n = 0;             
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART+CHECKWIDTH); n++;   
                   XtSetArg(args[n], XmNtopPosition, y); n++;    
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;  
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   value = atoi(a.answer[j]);
                   valuetoRGB(value,color[1],color[2],color[3],a.bpp);
                   clickboxdata[j].title = boxstring;
                   clickboxdata[j].answer = atoi(a.answer[j]);
                   clickboxdata[j].startval = atoi(a.answer[j]);
                   clickboxdata[j].minval = a.boxmin[j];
                   clickboxdata[j].maxval = a.boxmax[j];
                   clickboxdata[j].type = INTEGER;
                   clickboxdata[j].wantpreview = 0;
                   clickboxdata[j].wantdragcb = 1;
                   clickboxdata[j].f1 = null;
                   clickboxdata[j].f2 = null;
                   clickboxdata[j].f3 = null;
                   a.c[j] = &clickboxdata[j];
                   for(k=1;k<4;k++)
                   {   XtSetArg(args[n], XmNleftPosition, BOXSTART+k*7); n++;
                       XtSetArg(args[n], XmNtopPosition, y); n++;    
                       XtSetArg(args[n], XmNrightPosition, BOXSTART+(k+1)*7-1); n++;   
                       XtSetArg(args[n], XmNresizable, False); n++; 
                       XtSetArg(args[n], XmNspacing, 0); n++;
                       XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++; 
                       XtSetArg(args[n], XmNresizable, False); n++;
                       XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
                       XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                       XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                       XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                       XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                       itoa(color[k],temp,10);
                       XtSetArg(args[n], XmNlabelString,
                          xms = XmStringCreateSimple(temp));n++; 
                       wid[k] = w[wc++] = XmCreatePushButton(form,boxstring,args,n);
                       a.boxwidget[j][k] = wid[k];
                       XtManageChild(wid[k]);
                       XmStringFree(xms);
                       XtAddCallback(wid[k], XmNactivateCallback, 
                          (XtCBP)dialogmulticlickboxcb, (XtP) &a);
                       XtAddCallback(wid[k], XmNarmCallback,(XtCBP)armboxcb, (XtP)&a);
                       w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   }
                   for(i=0;i<4;i++) clickboxdata[j].widget[i] = wid[i];
                   clickboxdata[j].dirmask = 0;
                   break;
              case LIST:
              case NON_EDIT_LIST:
              case NO_LABEL_LIST:
                   n = 0;
                   XtSetArg(args[n], XmNmarginLeft, 4); n++;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;
                   XtSetArg(args[n], XmNtopPosition, y); n++;    
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++;  
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++;   
                   XtSetArg(args[n], XmNfractionBase, 100); n++;   
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNresizable, False); n++; 
                   XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++; 
                   k = a.boxmenuselection[j];
                   XtSetArg(args[n], XmNlabelString,
                      xms = XmStringCreateSimple(a.boxmenu[j][k]));n++; 
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   boxbutton[j] = w[wc++] = XmCreatePushButton(form,boxstring,args,n);
                   a.boxwidget[j][0] = boxbutton[j];
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   XtManageChild(boxbutton[j]);
                   XmStringFree(xms);
                   w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   clickboxdata[j].title = a.boxmenutitle[j];
                   clickboxdata[j].answer = a.boxmenuselection[j];
                   clickboxdata[j].widget[0] = boxbutton[j];
                   clickboxdata[j].dirmask = 0;
                   clickboxdata[j].list = a.boxmenu[j];
                   clickboxdata[j].helptopic = a.helptopic;
                   clickboxdata[j].listsize = a.boxmax[j];
                   clickboxdata[j].f1 = null;
                   clickboxdata[j].f2 = null;
                   clickboxdata[j].f3 = null;
                   if(a.boxtype[j]==NON_EDIT_LIST) 
                        clickboxdata[j].allowedit = 0;
                   else 
                        clickboxdata[j].allowedit = 1;
                   clickboxdata[j].editcol = a.startcol[j];
                   clickboxdata[j].maxstringsize = 50;
                   a.c[j] = &clickboxdata[j];
                   XtAddCallback(boxbutton[j], XmNactivateCallback, 
                      (XtCBP)dialoglistcb, (XtP)&a);
                   break;
              case LABEL:      // Just a heading, don't do anything
                   w[wc++] = addlabel(form,boxstring,LEFT,BOXSTART,y,BOXEND,y+bspacing);
                   a.boxwidget[j][0] = 0;
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   break;
              case DOUBLECLICKBOX:
                   n = 0;
                   XtSetArg(args[n], XmNmarginLeft, 4); n++;
                   XtSetArg(args[n], XmNresizable, False); n++;
                   XtSetArg(args[n], XmNspacing, 0); n++;
                   XtSetArg(args[n], XmNleftPosition, BOXSTART); n++;
                   XtSetArg(args[n], XmNtopPosition, y); n++;     
                   XtSetArg(args[n], XmNbottomPosition, y+bspacing); n++; 
                   XtSetArg(args[n], XmNrightPosition, BOXEND); n++; 
                   XtSetArg(args[n], XmNfractionBase, 100); n++; 
                   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
                   XtSetArg(args[n], XmNresizable, False); n++; 
                   XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++; 
                   XtSetArg(args[n], XmNlabelString,
                      xms = XmStringCreateSimple(a.answer[j]));n++; 
                   ////  Stop Motif from trying to grab another color if none are available.
                   if(g.want_colormaps)  
                   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
                        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
                   }
                   boxbutton[j] = w[wc++] = XmCreatePushButton(form,boxstring,args,n);
                   a.boxwidget[j][0] = boxbutton[j];
                   a.boxwidget[j][1] = 0;
                   a.boxwidget[j][2] = 0;
                   a.boxwidget[j][3] = 0;
                   XtManageChild(boxbutton[j]);
                   XmStringFree(xms);
                   w[wc++] = addlabel(form,boxstring,LEFT,LABELSTART,y,LABELEND,y+bspacing);
                   clickboxdata[j].title = boxstring;
                   clickboxdata[j].fanswer = atof(a.answer[j]);
                   clickboxdata[j].fstartval = atof(a.answer[j]);
                   clickboxdata[j].fminval = (double)a.boxmin[j];
                   clickboxdata[j].fmaxval = (double)a.boxmax[j];
                   clickboxdata[j].type = FLOAT;
                   clickboxdata[j].wantpreview = 0;
                   clickboxdata[j].widget[0] = boxbutton[j];
                   clickboxdata[j].dirmask = 0;
                   clickboxdata[j].f1 = null;
                   clickboxdata[j].f2 = null;
                   clickboxdata[j].f3 = null;
                   a.c[j] = &clickboxdata[j];
                   XtAddCallback(boxbutton[j], XmNactivateCallback, 
                      (XtCBP)dialogdoubleclickboxcb, (XtP)&a);
                   XtAddCallback(boxbutton[j], XmNarmCallback,(XtCBP)armboxcb, (XtP)&a);
                   break;
              default: fprintf(stderr,"error in dialogboxes\n");
           }
           y+=bspacing;
       }
       addstandardbuttons(form, &w[wc], &a.helptopic);
       wc += 3;

      // This makes the Ok button the default, but also trashes the dialog box.
      // So we are forced to add an event handler to every button.
      // XtVaSetValues(form, XmNdefaultButton, okbutton, NULL); 
      // Make Return key activate Ok button:

      for(k=0;k<wc;k++)
           XtAddEventHandler(w[k], KeyPressMask, FALSE, (XtEH)entercb, (XtP)form);
    
      if(a.want_changecicb)
           XtAddEventHandler(g.drawing_area, SubstructureNotifyMask, FALSE,
                (XtEH)dialogchangecicb, (XtP)&a);

      ////  Initialize any grayed-out items in the callback if any

      armboxcb(0,&a,0);

      //-------------Event loop-------------------------------------------//

      XtManageChild(form);
      for(i=0;i<10;i++) if(radiopush[i]) XtManageChild(radiopush[i]);

      ////  Show any error messages.
      ////  The calling routine checks the returned answers, sets a.needrepeat,
      ////  a.message1, and a.messagemode, and calls dialogbox() again if it finds 
      ////  an inconsistency.  This allows the message to appear on top of the 
      ////  dialog box (simply raising the window in message() doesn't work).

      if(a.needrepeat) message(a.message1, a.messagemode);
      a.needrepeat = 0;
      event_loop(form);

      //// Put result back in radioboxes.  Only needed for widgets that   
      ////   don't have callbacks.                                        
      //// Also, if radiotype[radio][item] is RADIOTEXT get a string or if
      ////   it is RADIOVALUE get an integer (=2). (Return as a string in 
      ////   a.radioanswer[i][j]).   Should be replaced entirely.         

      for(i=0;i<a.noofradios;i++)
      { 
         for(j=1;j<a.radiono[i];j++)
         {   switch(a.radiotype[i][j])
             {    case RADIOTEXT:       // Get input string
                    ptr = XmTextGetString(radioslider[i][j]);
                    if(ptr!=NULL)
                    {  strcpy(a.radioanswer[i][j],ptr); 
                       XtFree(ptr);
                    }
                    break;
                  case RADIOVALUE:      // Small slider
                  case RADIOVALUEWIDE:  // Wide slider
                    result = atoi(a.radioanswer[i][j]);
                    XmScaleGetValue(radioslider[i][j], &result);
                    itoa(result,a.radioanswer[i][j],10);
                    for(k=strlen(a.radio[i][j]);k;k--)
                    {   if(a.radio[i][j][k]=='=')
                        {  a.radio[i][j][k+1]=0;    
                           strcat(a.radio[i][j],a.radioanswer[i][j]);
                           break;
                        }
                    }
                    break;
             }
         }
      }

      ////  Put result in answer boxes.   Only needed for widgets that   
      ////  don't have callbacks. Should be replaced entirely.           

      for(j=0;j<a.noofboxes;j++)
      {   switch(a.boxtype[j])
          {       case FILENAME:                                 // Filename 
                     strcpy(a.answer[j],clickboxdata[j].title);  // Single 
                     break;
                  case MULTIFILENAME:                            // Multiple

                     ////  Caller must XtFree clickboxdata[j].list[k]'s 
                     ////  and delete[] clickboxdata[j].list

                     a.boxcount[j] = clickboxdata[j].count;

                     ////  User did not select any files so no list was created
                     ////  in fsokcb().  Create a list and use the old filename.
                        
                     if(a.boxcount[j]==0)
                     {   clickboxdata[j].list = new char*[1];          
                         clickboxdata[j].list[0] = XtMalloc(FILENAMELENGTH); 
                         strcpy(clickboxdata[j].list[0], clickboxdata[j].title); 
                         a.boxcount[j]=1;
                     }
                     a.boxlist[j] = clickboxdata[j].list;
                     strcpy(a.answer[j],a.boxlist[j][0]);
                     break;
               }
       }
     
       ////  Save all the information needed to unmanage the dialog.
       ////  This may happen right away or later, depending on `a.keep' .
     
       a.widget = w;                            // array of 200 Widgets
       a.form = form;                           // DialogForm Widget
       a.clickboxdata = clickboxdata;           // Array of clickboxinfo structs
       a.widgetcount = wc;                      // No. of widgets on form
 }
 
 ////  If it is a normal dialog, delete the current widgets & data.
 ////  If dialog was kept on screen from before, use the data that 
 ////  was saved previously in its dialoginfo struct.

 if(a.keep!=ON)
 {  
      for(j=0;j<a.noofboxes;j++)
      {    if(a.boxtype[j]==FILENAME || a.boxtype[j]==MULTIFILENAME)
           {     
                 if(a.clickboxdata[j].dirmask) 
                 {    strcpy(dirmask,a.clickboxdata[j].dirmask);
                      delete[] a.clickboxdata[j].dirmask;
                 }
                 ////  The widgets were already destroyed in getfilenames().
                 delete[] clickboxdata[j].w;
            }
       }
       if(a.want_changecicb)
           XtRemoveEventHandler(g.drawing_area, SubstructureNotifyMask, 
               FALSE,(XtEH)dialogchangecicb, (XtP)&a);
       XtUnmanageChild(a.form);
       if(a.clickboxdata!=NULL) delete[] a.clickboxdata;
       for(k=0;k<a.widgetcount;k++) 
             if(a.widget[k]!=NULL)
	     {    XtDestroyWidget(a.widget[k]);
                  a.widget[k]=0;
	     }
       if(a.widget!=NULL) delete[] a.widget;
       a.widget=0;
 }
 return;
}  


//--------------------------------------------------------------------//
// dialogclickboxcb - callback for getting a value in dialog box.     //
// Calls clickbox(), using 'answer' in a clickboxinfo struct as the   //
// initial value, converts the integer returned by clickbox() to a    //
// string and stores it in the  'c->widget's XmNlabelString.          //
//--------------------------------------------------------------------//
void dialogclickboxcb(Widget w, XtP client_data, XmACB *call_data)
{
  w=w; call_data=call_data;  // Keep compiler quiet
  int box=0,j,k,answer;
  Arg args[100];
  XmString xms;
  char tempstring[256];
  int n=0;
  dialoginfo *a = (dialoginfo*)client_data;

  ////  Figure out which box
  for(k=0;k<4;k++)
  for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0) box = j;

  clickboxinfo *c = a->c[box];

  answer = clickbox(c->title, c->startval, c->minval, c->maxval, null, 10);
  if(g.getout){ g.getout=0; return; }
  c->answer = answer;
  c->startval = answer;

  ////  Redraw the label in the pushbutton widget that called the clickbox.
  itoa(answer,tempstring,10);
  XtSetArg(args[n], XmNlabelString, xms=XmStringCreateSimple(tempstring)); n++;
  XtSetValues(c->widget[0], args, n);
  XmStringFree(xms);

  ////  Put answer back in dialoginfo in case dialog remains on screen
  itoa(answer, a->answer[box], 10);
  return;
}


//--------------------------------------------------------------------//
// dialogdoubleclickboxcb - callback for getting a value in dialog box.//
//--------------------------------------------------------------------//
void dialogdoubleclickboxcb(Widget w, XtP client_data, XmACB *call_data)
{
  w=w; call_data=call_data;  // Keep compiler quiet
  double fanswer;
  Arg args[100];
  XmString xms;
  char tempstring[256];
  int box=0,j,k,n;

  dialoginfo *a = (dialoginfo*)client_data;

  ////  Figure out which box
  for(k=0;k<4;k++)
  for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0) box = j;

  clickboxinfo *c = a->c[box];

  fanswer = (double)clickbox(c->title, c->fstartval, c->fminval, c->fmaxval, null, 10);
  if(g.getout){ g.getout=0; return; }
  c->fanswer = fanswer;
  c->fstartval = fanswer;

  ////  Redraw the label in the pushbutton widget that called the clickbox.
  gcvt(c->fanswer, g.signif, tempstring);
  n = 0;
  XtSetArg(args[n], XmNlabelString, xms=XmStringCreateSimple(tempstring)); n++;
  XtSetValues(c->widget[0], args, n);
  XmStringFree(xms);

  gcvt(fanswer, g.signif, a->answer[box]); 
  return;

}


//--------------------------------------------------------------------//
// dialogmulticlickboxcb - callback for getting a value in dialog box //
// using multiclickbox().  Use widgets 1,2,and 3.                     //
//--------------------------------------------------------------------//
void dialogmulticlickboxcb(Widget w, XtP client_data, XmACB *call_data)
{
  w=w; call_data=call_data;  // Keep compiler quiet
  int col[4];
  uint color;
  Arg args[100];
  XmString xms;
  char tempstring[256];
  int box=0, cmyk, j, k, n=0;

  dialoginfo *a = (dialoginfo*)client_data;
  ////  Figure out which box
  for(j=0;j<a->noofboxes;j++)
     for(k=0;k<4;k++)
        if(w==a->boxwidget[j][k] && w!=0) box = j;

  clickboxinfo *c = a->c[box];
  //if(printer.palette>1) cmyk=1; else 
  cmyk=0;

  color = getcolor("Color adjustment (128=normal)",c->answer,a->bpp,cmyk);
  valuetoRGB(color,col[1],col[2],col[3],a->bpp);
  
  // Redraw the labels in the pushbutton widgets that called the clickbox.
  
  for(k=1;k<4;k++)
  {    n = 0;
       itoa(col[k],tempstring,10);
       XtSetArg(args[n], XmNlabelString, xms=XmStringCreateSimple(tempstring)); n++;
       XtSetValues(c->widget[k], args, n);
       XmStringFree(xms);
  }
  itoa(color, a->answer[box], 10);
  c->answer = color;
  c->startval = color;
} 



//--------------------------------------------------------------------//
// dialogstringcb - callback for getting a string in dialog box.      //
//--------------------------------------------------------------------//
void dialogstringcb(Widget w, XtP client_data, XmACB *call_data)
{
  w=w; call_data=call_data;  // Keep compiler quiet
  int box=0,boxbutton=0,j,k;
  char *ptr;
  dialoginfo *a = (dialoginfo*)client_data;

  ////  Figure out which box
  for(k=0;k<4;k++)
  for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0){ box = j; boxbutton=k; }

  ////  Put answer back in dialoginfo in case dialog remains on screen

  ptr = XmTextGetString(w);
  if(ptr != NULL){  strcpy(a->answer[box],ptr); XtFree(ptr); }
  if(!strlen(a->lastanswer[box])) strcpy(a->lastanswer[box],"1");
  a->f1(a,0,box,boxbutton);
  if(strlen(a->answer[box])) strncpy(a->lastanswer[box], a->answer[box],99);
  return;
}


//--------------------------------------------------------------------//
// dialogcursorcb - set cursor. Should not be necessary but Motif     //
// ignores cursor settings at initialization.                         //
//--------------------------------------------------------------------//
void dialogcursorcb(Widget w, XtP client_data, XmACB *call_data)
{
  call_data=call_data;
  int box=0,j,k;
  ////  Figure out which box
  dialoginfo *a = (dialoginfo*)client_data;
  for(k=0;k<4;k++)
  for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0){ box = j; break; }
  int len =(int)strlen(a->answer[box]);
  ////  Set string and position at same time so it takes it.
  if(len)
     XtVaSetValues(w, XmNvalue, a->answer[box], XmNcursorPosition, len, NULL); 
}


//--------------------------------------------------------------------//
// dialoglistcb - callback for getting item from list in dialog box.  //
// Calls list(), stores the selected string in the  'c->widget's      //
// XmNlabelString. The list data must be already allocated.           //
//--------------------------------------------------------------------//
void dialoglistcb(Widget w, XtP client_data, XmACB *call_data)
{
  w=w; call_data=call_data;  // Keep compiler quiet
  Arg args[100];
  XmString xms;
  int box=0,boxbutton=0,j,k,n,choice;
  dialoginfo *a = (dialoginfo*)client_data;
  for(k=0;k<4;k++)
  for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0){ box = j; boxbutton=k; }
  clickboxinfo *c = a->c[box];
  char edittitle[20] = "Enter value";
  listinfo l;
  l.title = c->title;
  l.info  = c->list;
  l.size  = c->listsize;
  l.itemstoshow = 11;
  l.firstitem   = 0;
  l.wantsort    = 0;
  l.wantsave    = 0;
  l.allowedit   = c->allowedit;
  l.helptopic   = c->helptopic;
  l.selection   = &c->answer;
  l.edittitle   = edittitle;
  l.editcol     = c->editcol;
  l.editrow     = 0;
  l.maxstringsize = c->maxstringsize;
  l.width       = 0;
  l.transient   = 1;
  l.wantfunction = 0;
  l.autoupdate   = 0;
  list(&l);
  
  c->answer = *l.selection; 

  ////  Redraw the label in the pushbutton widget that called the clickbox.
  n=0;
  xms = XmStringCreateSimple(c->list[c->answer]);
  XtSetArg(args[n], XmNlabelString, xms); n++;
  XtSetValues(c->widget[0], args, n);
  XmStringFree(xms);

  choice = c->answer;
  strcpy(a->answer[box],a->boxmenu[box][choice]);
  a->boxmenuselection[box] = choice;
  a->f1(a,0,box,boxbutton);
  g.getout=0;
}


//--------------------------------------------------------------------//
// dialogchangecicb                                                   //
//--------------------------------------------------------------------//
void dialogchangecicb(Widget w, XtP client_data, XmACB *call_data)
{
    w=w;call_data=call_data;      // keep compiler quiet
    dialoginfo *a = (dialoginfo *)client_data;
    armboxcb(0,a,0);
}


//--------------------------------------------------------------------//
// armboxcb - callback for dialog-specific functions when button is   //
// clicked in dialog. Togglecb must be called before armboxcb in      //
// radioboxes.                                                        //
//--------------------------------------------------------------------//
void armboxcb(Widget w, XtP client_data, XmACB *call_data)
{
    call_data=call_data;      // keep compiler quiet
    int i,j,k,radio=-1,box=-1,boxbutton=-1;
    dialoginfo *a = (dialoginfo *)client_data;

    for(i=0;i<a->noofradios;i++)
    for(j=0;j<a->radiono[i];j++)
       if(w==a->radiowidget[i][j]) radio=i;

    ////  Set the cursor position on any string boxes. This should be done
    ////  at initialization but Motif ignores it.
    
    for(j=0;j<a->noofboxes;j++)
       if(a->boxtype[j]==STRING)
          dialogcursorcb(a->boxwidget[j][0],client_data,call_data);

    for(k=0;k<4;k++)
    for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0)
       {   boxbutton=k;  box=j; }        
    if((int)w==0){ box=0; radio=0;}
    a->f1(a,radio,box,boxbutton);
}


//--------------------------------------------------------------------//
// toggleboxcb - callback to toggle a box button in dialog            //
//--------------------------------------------------------------------//
void toggleboxcb(Widget w, XtP client_data, XmACB *call_data)
{
    call_data=call_data;      // keep compiler quiet
    int j,k,state;
    state = XmToggleButtonGetState(w);
    dialoginfo *a = (dialoginfo *)client_data;

    for(k=0;k<=2;k++)
    for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0)
          if(state==True) a->boxset[j]=1; else a->boxset[j]=0;
    if(w) XmToggleButtonSetState(w,state,False);

}


//--------------------------------------------------------------------//
// togglecb - callback to toggle a list button in dialog              //
//--------------------------------------------------------------------//
void togglecb(Widget w, XtP client_data, XmACB *call_data)
{
    call_data=call_data;      // keep compiler quiet
    int i,j,state;
    state = XmToggleButtonGetState(w);
    dialoginfo *a = (dialoginfo *)client_data;

    ////  Lesstif hack
#ifdef LESSTIF_VERSION
    for(i=0;i<a->noofradios;i++)
    for(j=0;j<a->radiono[i];j++)
       if(w==a->radiowidget[i][j]) 
       {   if(state==True)
           {   state=False;
               XmToggleButtonSetState(w, False, True);
           }else
           {   XmToggleButtonSetState(w, True, True);
	       state=True;
	   }
       }
#endif

    for(i=0;i<a->noofradios;i++)
    for(j=0;j<a->radiono[i];j++)
       if(state==True && w==a->radiowidget[i][j]) a->radioset[i]=j;
}



//--------------------------------------------------------------------//
// getfilename                                                        //
//--------------------------------------------------------------------//
char *getfilename(char *oldtitle)
{ 
   static char dirmask[FILENAMELENGTH]="*";
   int k;
   clickboxinfo c;
   c.title = oldtitle;
   c.widget[1] = NULL;
   c.helptopic = 0;
   c.done      = 0;
   c.wc        = 0;
   c.w         = new Widget[20];
   c.f1 = null;
   c.f2 = null;
   c.f3 = null;
   c.f4 = NULL;
   c.wc = 0;
   c.form = NULL;
   c.path = NULL;
   c.type = SINGLE;
   c.list = (char**)NULL;
   c.dirmask = new char[FILENAMELENGTH];
   strcpy(c.dirmask, dirmask);
   ////  This changes c.wc
   filenamecb(g.main_widget, &c, (XmACB*)NULL);  
   while(!c.done)
        XtAppProcessEvent (XtWidgetToApplicationContext(g.main_widget), XtIMAll);
   for(k=0;k<c.wc;k++) XtDestroyWidget(c.w[k]);
   strcpy(dirmask, c.dirmask);
   delete[] c.w;
   delete[] c.dirmask;
   c.wc = 0;
   if(strlen(c.title)>0) return c.title;
   else return oldtitle;
}



//--------------------------------------------------------------------//
// dialogfilenamecb - callback for getting filename or list of        //
// filenames in dialog box.                                           //
//--------------------------------------------------------------------//
void dialogfilenamecb(Widget w, XtP client_data, XmACB *call_data)
{
  w=w; call_data=call_data;  // Keep compiler quiet
  int box=0,j,k;
  dialoginfo *a = (dialoginfo*)client_data;
  for(k=0;k<4;k++)
  for(j=0;j<a->noofboxes;j++)
       if(w==a->boxwidget[j][k] && w!=0) box = j;
  clickboxinfo *c = a->c[box];

  //// If c->type is SINGLE, one filename is returned in c->title.
  //// Otherwise, it allocates a list of filenames and puts it in array c->list[].

  if(c->type == SINGLE)
      strcpy(c->title, getfilename(c->title));
  else
      getfilenames(c);

  if(strlen(c->title)) strcpy(a->answer[box], c->title);
  if(strlen(c->path)) strcpy(a->path, c->path);
  return;
}


//--------------------------------------------------------------------//
// getfilenames                                                       //
//--------------------------------------------------------------------//
void getfilenames(clickboxinfo *c)
{  
   static int ingfn=0;
   int k;
   c->done      = 0;
   c->count     = 0;
   if(ingfn){ c->done=1; return; }
   c->wc        = 0;
   ingfn=1;
   static char dirmask[FILENAMELENGTH]="*";
   c->type      = MULTIPLE;
   c->dirmask   = new char[FILENAMELENGTH];

   strcpy(c->dirmask, dirmask);
   ////  This changes c->wc, allocates and fills c->list, and sets c->count.
   filenamecb(g.main_widget, c, (XmACB*)NULL);   
   while(!c->done)
      XtAppProcessEvent (XtWidgetToApplicationContext(g.main_widget), XtIMAll);
   strcpy(dirmask, c->dirmask);
   
   //// Destroy the 4 fileselector widgets here in case usr clicks twice.
   //// Don't deallocate array here, wait until dialogbox is deactivated.
   for(k=0;k<c->wc;k++) XtDestroyWidget(c->w[k]);
   ingfn=0;
}


