//--------------------------------------------------------------------------//
// xmtnimage34.cc                                                           //
// Latest revision: 04-23-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
// warp, split, combine, flip, create image                                 //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Images     *z;
extern int         ci;
extern PrinterStruct printer;

//-------------------------------------------------------------------------//
// setfcolor                                                               //
//-------------------------------------------------------------------------//
void setfcolor(void)
{ 
   int color=0;
   color = getcolor("Foreground color",g.fcolor);
   if(g.getout){ g.getout=0; return; }
   g.fcolor = color;
   g.line.color = g.fcolor;
   if(g.bitsperpixel==8) 
   {   g.fc.red = g.palette[g.fcolor].red; 
       g.fc.green = g.palette[g.fcolor].green; 
       g.fc.blue = g.palette[g.fcolor].blue; 
   }
}


//-------------------------------------------------------------------------//
// setbcolor                                                               //
//-------------------------------------------------------------------------//
void setbcolor(void)
{ 
   int color=0;
   color = getcolor("Background color",g.bcolor);
   if(g.getout){ g.getout=0; return; }
   g.bcolor = color;
   if(g.bitsperpixel==8) 
   {   g.bc.red = g.palette[g.bcolor].red; 
       g.bc.green = g.palette[g.bcolor].green; 
       g.bc.blue = g.palette[g.bcolor].blue; 
   }
}



//-------------------------------------------------------------------------//
// warp                                                                    //
// Change shape of image                                                   //
//-------------------------------------------------------------------------//
void warp(void)
{
   if(ci<0){  message("Please select an image first\n");  return; }
   if(memorylessthan(16384)){  message(g.nomemory,ERROR); return; } 
   const int NMAX = 300;
   int bpp, ino, d, i, i2, j, k, xstart=0, xend, x, y, yalign;
   for(k=0; k<g.image_count; k++) z[k].hit=0;
   uint value;  
  
   XYData data;
   data.x = new int[NMAX]; if(data.x==NULL){ message(g.nomemory,ERROR); return; }
   data.y = new int[NMAX]; 
          if(data.y==NULL){ delete[] data.x; message(g.nomemory,ERROR); return; }
   data.f = NULL;
   data.g = NULL;
   data.n = 0;
   data.dims = 1;
   data.nmax = NMAX;
   data.duration = TEMP;
   data.wantpause = 0;
   data.type = 0;
   data.win  = 0; // calculate window automatically on drawing area
 
   for(k=0;k<g.xres;k++) g.highest[k]=0;        // Initialize baseline buf.

   ////  bezier_curve_start() sets bezier_state to CURVE.                    
   ////  When user presses a key to stop the Bezier curve, bezier_curve_end()
   ////  sets bezier_state to 'NORMAL'.                                             
   ////  bezier_curve_end() also deallocates arrays x & y.                        

   bezier_curve_start(&data);
   while(g.bezier_state==CURVE)
        XtAppProcessEvent(XtWidgetToApplicationContext(g.drawing_area),XtIMAll);
   
   xstart = 0;               // Left x coord of user's Bezier
   xend = g.xres;            // Right x coord of user's Bezier

   for(k=0;k<g.xres;k++) if(g.highest[k]!=0) { xstart=k; break; }
   for(k=g.xres;k>=xstart;k--) if(g.highest[k]!=0) { xend=k; break; }

   message("Click at y value to align to");
   getpoint(x,y);
   yalign = y;

   ////  This section warps all the pixels in the selected image. Pixels outside
   ////  the image are  not affected.
   ////  highest[] is the Bezier curve for each x relative to the left of the 
   ////  screen. Subtract main_xpos to get the coordinate for drawing pixels
   ////  (which may be negative).

   for(i=xstart;i<=xend;i++)
   {   
       d = yalign - g.highest[i];    // d is distance to move down
       i2 = i-g.main_xpos;
       if(d>0)                     // Pixels to be moved down 
       {   for(j=z[ci].ypos+z[ci].ysize-1;j>=z[ci].ypos;j--)
           {   if(j-d >= z[ci].ypos)
                 value = readpixelonimage(i2,j-d,bpp,ino);
               else
                 value = g.bcolor;
               if(ino>=0)
               {  if((g.bitsperpixel==8)&&(bpp!=8)&&(z[ino].hitgray==0))
                      z[ino].hit=1;
                  if(!g.wantr || !g.wantg || !g.wantb) z[ino].hit=1;
                  if(bpp==8 || z[ino].colortype==GRAY) z[ino].hit=1;
               }
               setpixelonimage(i2,j,value,g.imode,bpp);
           }
       }else                       // Pixels to be moved up
       {   for(j=z[ci].ypos;j<z[ci].ysize+z[ci].ypos;j++)
           {   if(j-d < z[ci].ysize+z[ci].ypos)
                  value = readpixelonimage(i2,j-d,bpp,ino);
               else
                  value = g.bcolor;
               if(ino>=0)
               {  if((g.bitsperpixel==8)&&(bpp!=8)&&(z[ino].hitgray==0))
                      z[ino].hit=1;
                  if(!g.wantr || !g.wantg || !g.wantb) z[ino].hit=1;
                  if(bpp==8 || z[ino].colortype==GRAY) z[ino].hit=1;
               }
               setpixelonimage(i2,j,value,g.imode,bpp);
           }
       }
       if(g.getout){ g.getout=0; break;}
   }    

   g.state=NORMAL;
   g.bezier_state=NORMAL;
   delete[] data.x;
   delete[] data.y;
   for(k=0; k<g.image_count; k++) 
   {  if(z[k].hit)
      {    rebuild_display(k); 
           repairimg(k,0,0,z[k].xsize,z[k].ysize); 
           redraw(k); 
      }
   }
  switch_palette(ci);
  rebuild_display(ci); 
  redraw(ci); 
}


//-------------------------------------------------------------------------//
// flip_horiz                                                              //
//-------------------------------------------------------------------------//
void flip_horiz(void)
{
    int h,i2,i3,j,k,bpp,ii;
    uint value;
    uint *buffer;
    for(k=0; k<g.image_count; k++) z[k].hit=0;

    buffer=new uint[6*(g.selected_lrx-g.selected_ulx+2)];
    if(buffer==NULL){ message(g.nomemory,ERROR); return; }

    drawselectbox(OFF);
    if(g.selectedimage>-1 && ci >=0) // Use faster method if region is all same image
    {   bpp = z[ci].bpp;
        for(j=0;j<z[ci].ysize;j++)
        {   for(h=0,i2=0; h<z[ci].xsize; h++,i2+=g.off[bpp])
               buffer[h] = pixelat(z[ci].image[z[ci].cf][j]+i2,bpp);
            i3 = (z[ci].xsize-1)*g.off[bpp];
            for(h=0,i2=0; h<z[ci].xsize; h++,i2+=g.off[bpp])
            {  value=buffer[h];
               putpixelbytes(z[ci].image[z[ci].cf][j]+i3-i2,value,1,bpp);
            }      
        }
        z[ci].hit=1;
    }else
    {   for(j=g.selected_uly;j<g.selected_lry;j++)
        {   for(h=g.selected_ulx,i2=0;h<g.selected_lrx;h++,i2+=4)
            {  value = readpixelonimage(h,j,bpp,ii);
               if(ii>=0){ if((g.bitsperpixel==8)&&(bpp!=8)) z[ii].hit=1;}
               if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
               buffer[h-g.selected_ulx]=value;               
            }
            for(h=g.selected_ulx,i2=0;h<=g.selected_lrx;h++,i2+=4)
            {  if(!g.selected_is_square && !inside_irregular_region(h,j)) continue;      
               bpp = whatbpp(h,j);
               value = buffer[h-g.selected_ulx];
               setpixelonimage(g.selected_lrx-(h-g.selected_ulx),j,value,SET,bpp);  
            }      
        }       
    }
    delete[] buffer;
    for(k=0; k<g.image_count; k++) 
       if(z[k].hit){ rebuild_display(k); redraw(k);  }
    return;
}




//-------------------------------------------------------------------------//
// flip_vert                                                               //
//-------------------------------------------------------------------------//
void flip_vert(void)
{
    int i,i2,j3,j,k,bpp,ii;
    uint value;
    uint *buffer;
    drawselectbox(OFF);
    for(k=0; k<g.image_count; k++) z[k].hit=0;
    buffer=new uint[6*(g.selected_lry-g.selected_uly+2)];
    if(buffer==NULL){ message(g.nomemory,ERROR); return; }
    if(g.selectedimage>-1 && ci >=0) // Use faster method if region is all same image
    {    bpp = z[ci].bpp;
         j3 = z[ci].ysize-1;
         for(i=0,i2=0;i<z[ci].xsize;i++,i2+=g.off[bpp])
         {   for(j=0;j<z[ci].ysize;j++)
                 buffer[j] = pixelat(z[ci].image[z[ci].cf][j]+i2,bpp);
             for(j=0;j<z[ci].ysize;j++)
             {  value = buffer[j];
                putpixelbytes(z[ci].image[z[ci].cf][j3-j]+i2,value,1,bpp);
             }      
         }
         z[ci].hit=1;
    }else
    {    for(i=g.selected_ulx;i<g.selected_lrx;i++)
         {   for(j=g.selected_uly;j<g.selected_lry;j++)
             {  value = readpixelonimage(i,j,bpp,ii);
                if(ii>=0){ if((g.bitsperpixel==8)&&(bpp!=8)) z[ii].hit=1;}
                if(ii>=0 && (!g.wantr || !g.wantg || !g.wantb)) z[ii].hit=1;
                buffer[j-g.selected_uly]=value;
             }
             for(j=g.selected_uly;j<g.selected_lry;j++)
             {  if(!g.selected_is_square && !inside_irregular_region(i,j)) continue;      
                bpp = whatbpp(i,j);
                value = buffer[j-g.selected_uly];
                setpixelonimage(i,g.selected_lry-(j-g.selected_uly),value,SET,bpp);  
             }       
         }
    }
    delete[] buffer;
    for(k=0; k<g.image_count; k++) 
        if(z[k].hit){ rebuild_display(k); redraw(k); }
}



//-------------------------------------------------------------------------//
// image_border - draw a border around the image                           //
//-------------------------------------------------------------------------//
void image_border(void)
{  
  if(ci<0) return;
  box(z[ci].xpos+1, z[ci].ypos, z[ci].xpos+z[ci].xsize-1, z[ci].ypos+z[ci].ysize-1, 
      g.fcolor, g.imode, &g.line);
  rebuild_display(ci);
  z[ci].touched=1;
  switchto(ci);             
  redraw(ci);
}
 
 

//-------------------------------------------------------------------------//
// paint region                                                            //
//-------------------------------------------------------------------------//
void paint(int color)
{   
    int i, j, ok=1, xx1=g.get_x1, xx2=g.get_x2, yy1=g.get_y1, yy2=g.get_y2;
    message("Select area(s) to paint\nclick cancel button when finished");
    g.getout=0;
    int square = g.selected_is_square;
    while(!g.getout)
    { 
        getbox(xx1, yy1, xx2, yy2);
        for(j=yy1;j<=yy2;j++)
        for(i=xx1;i<=xx2;i++)
        {   
            ok = 0;
            if(square) ok=1;
            else if(inside_irregular_region(i,j)) ok=1;
            if(ok) setpixelonimage(i,j,color,SET,0,-1,0,0,1);
        }
    }
}        


//-------------------------------------------------------------------------//
// show od table                                                           //
//-------------------------------------------------------------------------//
void showodtable(void)
{  
    int k,helptopic;
    int *tt;               // Can't plot grayscale directly because it is short int
    double *xdata;
    char **ytitle;

    char title[FILENAMELENGTH] = "Gamma_table";
    ytitle = new char*[1];
    ytitle[0] = new char[256];
    strcpy(ytitle[0],"Optical Density");      

    xdata = new double[256];
    tt = new int[256];     
    if(tt==NULL){  message(g.nomemory,ERROR); return; } 
    for(k=0;k<256;k++) 
    {  if(ci>=0) tt[k] = z[ci].gamma[k];
       else      tt[k] = k;   
       xdata[k] = (double)k;
    }
    helptopic=17;
    plot(title,"Pixel Value",ytitle,xdata,tt,256,1,MEASURE,0,256,null,null,null,helptopic);
    if(ci>=0) for(k=0;k<256;k++) z[ci].gamma[k]=(short int)tt[k];
 
    delete[] ytitle[0];
    delete[] ytitle;
    delete[] tt;
    delete[] xdata;
}    
 

//-------------------------------------------------------------------------//
// erase background                                                        //
//-------------------------------------------------------------------------//
void erasebackground(void)
{  int j;
   for(j=0;j<z[0].ysize;j++) 
      putpixelbytes(z[0].image[0][j], g.bcolor, z[0].xsize, g.bitsperpixel, 1);
   copybackground(1,1,z[0].xsize-1,z[0].ysize-1,-1);
   repair(0);
   g.background_touched=0;
}



//-------------------------------------------------------------------------//
// createimage - allocate space for a new image at current bits/pixel.     //
//-------------------------------------------------------------------------//
int createimage(int noofargs, char **arg)
{ 
   static int method=1;   
   static int border=0;   // 0=no 1=yes
   static int xsize=128,ysize=128,xpos=0,ypos=0,frames=1;
   static int cimage=-1;  // image no. to copy
   static int shell=g.want_shell, auto_edge=0;
   if(memorylessthan(16384)){  message(g.nomemory,ERROR); return(NOMEM); } 
   Dialog dialog;
   int j,k,status=OK;
   char tempname[FILENAMELENGTH]="Untitled";
   if(ci>=0) cimage=ci;
   dialog.needrepeat=0;
   if(noofargs==0) 
   {
      strcpy(dialog.title,"Create/Resize Image");
      strcpy(dialog.radio[0][0],"Method");             
      strcpy(dialog.radio[0][1],"Use mouse");   
      strcpy(dialog.radio[0][2],"Fixed size");  
      strcpy(dialog.radio[0][3],"Duplicate image");   
      strcpy(dialog.radio[0][4],"Panel->multiframe");   
      strcpy(dialog.radio[0][5],"Multiframe->panel");   
      strcpy(dialog.radio[0][6],"Combine images->multiframe");   
      strcpy(dialog.radio[0][7],"Split multiframe->images");   
      strcpy(dialog.radio[0][8],"Add frame to image");   
      strcpy(dialog.radio[0][9],"Grab from screen");   
      strcpy(dialog.radio[0][10],"Composite 2 or more images");   
      strcpy(dialog.radio[0][11],"Resize image window only");   

      strcpy(dialog.radio[1][0],"Add border");             
      strcpy(dialog.radio[1][1],"No");   
      strcpy(dialog.radio[1][2],"Yes");  
      
      dialog.radioset[0]=method;
      dialog.radioset[1]=border+1;
    
      strcpy(dialog.boxes[0],"Fixed size parameters");
      strcpy(dialog.boxes[1],"X size");
      strcpy(dialog.boxes[2],"Y size");
      strcpy(dialog.boxes[3],"X position");
      strcpy(dialog.boxes[4],"Y position");
      strcpy(dialog.boxes[5],"New title");
      strcpy(dialog.boxes[6],"Image# to copy");
      strcpy(dialog.boxes[7],"No. of frames");
      strcpy(dialog.boxes[8],"Put in separate window");
      strcpy(dialog.boxes[9],"Automatically find frame edge");
   
      dialog.boxtype[0]=LABEL;
      dialog.boxtype[1]=STRING;
      dialog.boxtype[2]=STRING;
      dialog.boxtype[3]=STRING;
      dialog.boxtype[4]=STRING;
      dialog.boxtype[5]=STRING;
      dialog.boxtype[6]=INTCLICKBOX;
      dialog.boxtype[7]=INTCLICKBOX;
      dialog.boxtype[8]=TOGGLE;
      dialog.boxtype[9]=TOGGLE;
    
      dialog.boxmin[6]=0; dialog.boxmax[6]=g.image_count-1;
      dialog.boxmin[7]=1; dialog.boxmax[7]=1000;
  
      sprintf(dialog.answer[1], "%d",xsize);
      sprintf(dialog.answer[2], "%d",ysize);
      sprintf(dialog.answer[3], "%d",xpos);
      sprintf(dialog.answer[4], "%d",ypos);
      strcpy(dialog.answer[5],tempname);
      sprintf(dialog.answer[6], "%d",cimage);
      sprintf(dialog.answer[7], "%d",frames);
      if(shell) dialog.boxset[8]=1; else dialog.boxset[8]=0;
      if(auto_edge) dialog.boxset[9]=1; else dialog.boxset[9]=0;
 
      dialog.radiono[0]=12;
      dialog.radiono[1]=3;
      dialog.radiono[2]=0;
      for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
      dialog.noofradios=2;
      dialog.noofboxes=10; 
      dialog.helptopic=31;  
      dialog.keep=0;
      dialog.want_changecicb = 0;
      dialog.f1 = null;
      dialog.transient = 1;
      strcpy(dialog.path,".");
    
      ///////---///////
      dialogbox(dialog);
      ///////---///////
      if(g.getout)return(ABORT);

      method = dialog.radioset[0];
      if(dialog.radioset[1]==2) border=1; else border=0;
      xsize = atoi(dialog.answer[1]);
      ysize = atoi(dialog.answer[2]);
      xpos  = atoi(dialog.answer[3]);
      ypos  = atoi(dialog.answer[4]);
      strcpy(tempname,dialog.answer[5]);
      cimage= atoi(dialog.answer[6]);
      frames= atoi(dialog.answer[7]);
      if(dialog.boxset[8]) shell=1; else shell=0;
      if(dialog.boxset[9]) auto_edge=1; else auto_edge=0;
   }else
   {  if(noofargs>=1) method=atoi(arg[1]);
      if(noofargs>=2) border=atoi(arg[2]);
      if(noofargs>=3) xsize =atoi(arg[3]);
      if(noofargs>=4) ysize =atoi(arg[4]);
      if(noofargs>=5) xpos  =atoi(arg[5]);
      if(noofargs>=6) ypos  =atoi(arg[6]);
      if(noofargs>=7) cimage=atoi(arg[7]);
      if(noofargs>=8) frames=atoi(arg[8]);
      if(noofargs>=9) shell =atoi(arg[9]);
      if(noofargs>=10) auto_edge =atoi(arg[10]);
   }   
   
   xsize = max(1,xsize);
   ysize = max(1,ysize);
   frames = max(1,frames);
   drawselectbox(OFF);
   ////  Prevent switching to another image when selecting new area
   if(method!=4) g.want_switching=0;
   switch(method)
   {  case 1: status = create_image_with_mouse(xpos,ypos,frames,shell,border); break;
      case 2: status = create_fixed_size_image(xsize,ysize,xpos,ypos,frames,
                          shell,border); break;
      case 3: status = duplicate_image(cimage,xpos,ypos,shell,border); break;
      case 4: status = create_multiframe_image(cimage,xsize,ysize,xpos,ypos,
                          frames,shell,border,auto_edge);break;
      case 5: status = create_panel(cimage,xpos,ypos,10,10,shell,border); break;
      case 6: status = combine_frames(xpos,ypos,shell,border); break;
      case 7: status = split_frames(cimage,shell,border); break;
      case 8: status = add_frame(cimage); break;
      case 9: grab_window(); break;
      case 10: status = composite_images(xsize,ysize,xpos,ypos,shell,border); break;
      case 11: resize_image(cimage,xsize,ysize); status=NOIMAGES; break;
      default: message("Are you, like, sure?"); 
   }
   g.want_switching=1;

   ////  Can't just set xpos and ypos or it won't redraw
   if(ci>=0 && z[ci].shell) moveimage(ci,0,0);
   
   if(status==OK)
   {  message("New image created");
      z[ci].touched = 1;
      memcpy(z[ci].palette, g.palette, 768);
      memcpy(z[ci].opalette, g.palette, 768);
      memcpy(z[ci].spalette, g.palette, 768);
      rebuild_display(ci);
      redraw(ci);
      switchto(ci);
      if(border) image_border();
      if(g.autoundo) backupimage(ci,0);
      setimagetitle(ci,tempname);
      return(OK);
   }
   if(status==NOIMAGES) return OK;
   return(NOMEM);
}


//-------------------------------------------------------------------------//
// create_image_with_mouse                                                 //
//-------------------------------------------------------------------------//
int create_image_with_mouse(int xpos, int ypos, int frames, int shell, int border)
{
   int f,i,i2,j,j2,x1,x2,y1,y2,ct=g.colortype,ino,oino=-1,bpp,ibpp;
   uint value;
   message("Select region for new image");
   getbox(x1,y1,x2,y2);

   ino = whichimage((x1+x2)/2, (y1+y2)/2, bpp);
   int needconvert=0;
   bpp=g.bitsperpixel;
   if(bpp==8){ bpp=24; needconvert=1; }
   if(ino>=0) ct = z[ino].colortype;
   if(newimage(xpos,ypos,x2-x1,y2-y1,bpp,ct,frames,shell,PERM,0,border,0)==OK)
   {   for(f=0; f<frames; f++)
       for(j=y1,j2=0;j<y2;j++,j2++)
       for(i=x1,i2=0;i<x2;i++,i2+=g.off[bpp])
       { 
           ////  Read pixel, skipping pixels on image ci. 
           ////  Make all frames the same.
           value = readpixelonimage(i,j,ibpp,ino,ci);  
           if(ino!=oino){ switch_palette(ino); oino=ino; }
           value = convertpixel(value, ibpp, bpp,1);
           putpixelbytes(z[ci].image[f][j2]+i2,value,1,bpp,1);
       }
   }else return NOMEM;
   if(needconvert) change_image_depth(ci,8,1);
   return OK;
}


//-------------------------------------------------------------------------//
//  create_fixed_size_image                                                //
//-------------------------------------------------------------------------//
int create_fixed_size_image(int xsize, int ysize, int xpos, int ypos,  
    int frames, int shell, int border)
{
   //// Copy from preselected region, possibly off screen
   int f,i,i2,j,j2,x1,x2,y1,y2,ct=g.colortype,ino,oino=-1,bpp,ibpp;
   uint value;
   message("Click at upper left \nof area for new image");
   getpoint(x1,y1);
   x2 = x1+xsize;
   y2 = y1+ysize;

   ino = whichimage((x1+x2)/2, (y1+y2)/2, bpp);
   int needconvert=0;
   bpp=g.bitsperpixel;
   if(bpp==8){ bpp=24; needconvert=1; }
   if(ino>=0) ct = z[ino].colortype;
   if(newimage(xpos,ypos,x2-x1,y2-y1,bpp,ct,frames,shell,PERM,0,border,0)==OK)
   {   for(f=0; f<frames; f++)
       for(j=y1,j2=0;j<y2;j++,j2++)
       for(i=x1,i2=0;i<x2;i++,i2+=g.off[bpp])
       { 
           ////  Read pixel, skipping pixels on image ci. 
           ////  Make all frames the same.
           value = readpixelonimage(i,j,ibpp,ino,ci);  
           if(ino!=oino){ switch_palette(ino); oino=ino; }
           value = convertpixel(value, ibpp, bpp,1);
           putpixelbytes(z[ci].image[f][j2]+i2,value,1,bpp,1);
       }
   }else return NOMEM;
   if(needconvert) change_image_depth(ci,8,1);
   return OK;
}


//-------------------------------------------------------------------------//
// duplicate                                                               //
//-------------------------------------------------------------------------//
int duplicate(int noofargs, char **arg)
{
   int status,ino=ci;   // If no argument specified, make something up
   if(noofargs>=1) ino = image_number(arg[1]); else ino=ci;
   if(g.getout){ g.getout=0; return ERROR; }
   status = duplicate_image(ino, z[ino].xpos+10, z[ino].ypos+10, z[ino].shell,
            z[ino].window_border);
   return status;
}


//-------------------------------------------------------------------------//
// duplicate_image                                                         //
//-------------------------------------------------------------------------//
int duplicate_image(int cimage, int xpos, int ypos, int shell, int border)
{
   int f,i,j,frames,xsize,ysize,xbytes=0,ct,bpp,xsize3,ysize3;
   if(cimage>=g.image_count || cimage<0)
   {   message("Non-existent image",ERROR);
       return(BADPARAMETERS);
   }   
   if(cimage==0)
   {   message("Cannot duplicate the background",ERROR);
       return(BADPARAMETERS);
   }   
   xsize = z[cimage].xsize;
   ysize = z[cimage].ysize;
   ct    = z[cimage].colortype;
   bpp   = z[cimage].bpp;
   frames= z[cimage].frames;
   if(newimage(xpos,ypos,xsize,ysize,bpp,ct,frames,shell,PERM,0,border,0)==OK)
   {   
       for(f=0;f<frames;f++)
       {   xbytes=g.off[bpp] * xsize;
           for(j=0;j<ysize;j++)
               memcpy(z[ci].image[f][j],z[cimage].image[f][j],xbytes);
       }
       if(z[cimage].floatexists)
       {   setupfft(ci, xsize3, ysize3);
           for(j=0;j<ysize;j++)
           for(i=0;i<xsize;i++)
           {    z[ci].fft[j][i].real() = z[cimage].fft[j][i].real();
                z[ci].fft[j][i].imag() = z[cimage].fft[j][i].imag();
           }
           z[ci].floatexists = 1;
           z[ci].fftstate = z[cimage].fftstate;
       }
       if(z[cimage].waveletexists) 
       {   allocate_image_wavelet(ci, xsize, ysize);
           for(j=0;j<ysize;j++)
           for(i=0;i<xsize;i++)
                z[ci].wavelet[j][i] = z[cimage].wavelet[j][i];
           z[ci].waveletexists = 1;
           z[ci].fftstate = z[cimage].fftstate;
       }
   }else return NOMEM;
   copyimageparameters(ci, cimage);
   repair(ci);
   switchto(ci);
   moveimage(ci, xpos, ypos);
   redraw(ci);
   return OK;
}


//-------------------------------------------------------------------------//
// create_multiframe_image                                                 //
//-------------------------------------------------------------------------//
int create_multiframe_image(int cimage, int xsize, int ysize, int xpos, int ypos,  
    int frames, int shell, int border, int auto_edge)
{
   int cf,f,i,i2,j,j2,x,y,xx,yy,ct,bpp,ibpp,ino;
   uint value;
   uchar *add;
   switchto(cimage);
   if(cimage>=g.image_count || cimage<0)
   {   message("Non-existent image",ERROR);
       return(BADPARAMETERS);
   }   
   message("Click near upper left\nof region for each frame");
   cf = z[cimage].cf;
   ct = z[cimage].colortype;
   bpp = z[cimage].bpp;
   if(newimage(-xsize,-ysize,xsize,ysize,bpp,ct,frames,shell,PERM,0,border,0)!=OK)
       return NOMEM;
   ino = ci;
   switchto(cimage);
   for(f=0;f<frames;f++)
   {   
       getpoint(x,y);
       if(auto_edge) findedge(cimage,x,y);
       xx = x - z[cimage].xpos;
       yy = y - z[cimage].ypos;
       ////  The xor's can't cross a getpoint() because image 
       ////  might move.
       xor_box(x,y,x+xsize,y+ysize);
#ifdef LINUX
       usleep(200000);   
#else
       sleep(1);
#endif                      
       cf = z[cimage].cf;
       ibpp = z[cimage].bpp;
       for(j=yy,j2=0; j<yy+ysize; j++,j2++)
       for(i=xx,i2=0; i<xx+xsize; i++,i2+=g.off[bpp])
       {   if(!between(i,0,z[cimage].xsize-1) ||
              !between(j,0,z[cimage].ysize-1))  value=0;
           else
           {   
               add = z[cimage].image[cf][j] + g.off[bpp]*i;
               value = pixelat(add,ibpp);  
               value = convertpixel(value, ibpp, bpp,1);
           }
           putpixelbytes(z[ino].image[f][j2]+i2,value,1,bpp,1);
       }
       xor_box(x,y,x+xsize,y+ysize);
   }
   switchto(ino);
   moveimage(ino, xpos, ypos);
   return OK;
}


//-------------------------------------------------------------------------//
//  create_panel                                                           //
//-------------------------------------------------------------------------//
int create_panel(int cimage, int xpos, int ypos, int xoff, int yoff, int shell, 
    int border)
{
   static int cols = 4;
   int f,frames,frameheight,framewidth,i,i2,j,ct,bpp,rows,x,y,xsize,ysize,
      xstart;
   uint value;
   uchar *add;
   switchto(cimage);
   if(cimage>=g.image_count || cimage<0)
   {   message("Non-existent image",ERROR);
       return(BADPARAMETERS);
   }   
   frames = z[cimage].frames;
   if(frames<2)
   {   message("Cannot create panel from \nsingle-frame image",ERROR);
       return(ERROR);
   }   
   ct = z[cimage].colortype;
   bpp = z[cimage].bpp;
   framewidth = z[cimage].xsize;
   frameheight = z[cimage].ysize;

   cols = clickbox("No. of columns", cols, 1, frames, null, 0);
   xsize = 1+cols*(framewidth+xoff)+xoff;
   rows = int(0.9999 + (double)frames / (double)cols);
   ysize = 1+rows*(frameheight+yoff)+yoff;
   if(newimage(xpos,ypos,xsize,ysize,bpp,ct,1,shell,PERM,0,border,0)!=OK) return NOMEM;
   xstart = cint(g.off[bpp]*(0.5+(double)xoff/2.0));
   x = xstart;
   y = cint(g.off[bpp]*(0.5+(double)yoff/2.0));
   for(j=0;j<ysize;j++) memset(z[ci].image[0][j], 0, z[ci].xsize*g.off[bpp]);
   for(f=0;f<frames;f++)
   {  for(j=0; j<frameheight; j++)
      for(i=0,i2=0; i<framewidth; i++,i2+=g.off[bpp])
      {   add = z[cimage].image[f][j] + i2;
          value = pixelat(add,bpp);  
          if(j+y<ysize) putpixelbytes(z[ci].image[0][j+y]+i2+x,value,1,bpp,1);
      }
      x += g.off[bpp]*(xoff+framewidth);
      if(x+framewidth>=xsize){ x= xstart; y+=yoff+frameheight;}
   }
   return OK;
}


//-------------------------------------------------------------------------//
//  parse_digits  convert "1 2 3 4-6 8 23-33 44" into int array            //
//-------------------------------------------------------------------------//
int parse_digits(char *inputstring, int *number, int maxcount)
{
   char c, tempstring[16];
   int j=0,k,num1,num2,len,count=0,pos=0;
   len = strlen(inputstring);
   while(j<len && count<maxcount)
   {   c = inputstring[j++];
       pos = 0;
       while(isdigit(c)){ tempstring[pos++] = c; c = inputstring[j++]; }
       num1 = number[count++] = atoi(tempstring);
       if(j>=len || count>=maxcount) break;
       if(c=='-')
       {   pos = 0;
           c = inputstring[j++];
           while(isdigit(c)){ tempstring[pos++] = c; c = inputstring[j++]; }
           num2 = atoi(tempstring);
           if(num1>num2) swap(num1,num2);
           for(k=num1+1;k<=min(maxcount,num2);k++) number[count++]=k;
        }
   }
   return count;
}


//-------------------------------------------------------------------------//
//  split_frames                                                           //
//-------------------------------------------------------------------------//
int split_frames(int cimage, int shell, int border)
{
   int ct,bpp,f,i,i2,j,xsize,ysize,xpos,ypos;
   uchar *add;
   uint value;
   if(!between(cimage,1,g.image_count-1)){ message("Bad image number"); return ERROR; }
   ct = z[cimage].colortype;
   bpp = z[cimage].bpp;
   xsize = z[cimage].xsize;
   ysize = z[cimage].ysize;
   xpos = z[cimage].xpos+10;
   ypos = z[cimage].ypos+10;
   for(f=0;f<z[cimage].frames;f++)
   {   if(newimage(xpos+10*f,ypos+10*f,xsize,ysize,bpp,ct,1,shell,PERM,0,border,0)!=OK) return NOMEM;
       for(j=0; j<ysize; j++)
       for(i=0,i2=0; i<xsize; i++,i2+=g.off[bpp])
       {   add = z[cimage].image[f][j] + i2;
           value = pixelat(add,bpp);  
           putpixelbytes(z[ci].image[0][j]+i2,value,1,bpp,1);
       }
       rebuild_display(ci);
       redraw(ci);
   }
   return OK;
}
 

//-------------------------------------------------------------------------//
// findedge                                                                //
// given starting point x,y on image ino, find upper left corner of item   //
// to make into a frame.                                                   //
//-------------------------------------------------------------------------//
void findedge(int cimage, int &x, int &y)
{
   const int BORDER=1;  // Each frame must have a border at least this size
   int count=0, i,j, ino, bpp, start, value, rr, gg, bb, rr1, gg1, bb1, ulx, uly;
   float d;
   ulx = x;
   uly = y;
   if(!between(cimage,1,g.image_count-1)){ message("Bad image number"); return; }
   start = readpixelonimage(x,y,bpp,ino);  
   valuetoRGB(start,rr1,gg1,bb1,bpp);
   for(i=x;i>=z[cimage].xpos;i--)
   {   value = readpixelonimage(i,y,bpp,ino);
       valuetoRGB(value,rr,gg,bb,bpp);
       d = (float)(abs(rr-rr1) + abs(gg-gg1) + abs(bb-bb1))/(float)g.maxgreen[bpp]; 
       if(d>0.5) count++;
       ulx = i;
       if(count>BORDER){ ulx = i+BORDER+1; break;}
   }
   count=0;
   for(j=y;j>=z[cimage].ypos;j--)
   {   value = readpixelonimage(x,j,bpp,ino);
       valuetoRGB(value,rr,gg,bb,bpp);
       d = (float)(abs(rr-rr1) + abs(gg-gg1) + abs(bb-bb1))/(float)g.maxgreen[bpp]; 
       if(d>0.5) count++;
       uly = j;
       if(count>BORDER){ uly = j+BORDER+1; break;}
   }
   x = ulx;
   y = uly;
}

//-------------------------------------------------------------------------//
// separate_colors                                                         //
//-------------------------------------------------------------------------//
int separate_colors(int ino, int noofargs, char **arg)
{
   noofargs=noofargs; arg=arg;
   int f,i,j,ct,x,y,w,h,rr,gg,bb,bpp,obpp,value,frames,wantpalette,
        shell,rino,gino,bino;
   if(ino<0) { message("Please select an image",WARNING); return NOIMAGES; }
   if(z[ino].colortype!=COLOR) { message("Recommend converting image\n\
to 24 bit color before proceeding",WARNING); }
   x = z[ino].xpos;
   y = z[ino].ypos;
   w = z[ino].xsize;
   h = z[ino].ysize;
   frames = z[ino].frames;
   ct = GRAY;
   bpp = 8;
   obpp = z[ino].bpp;
   shell = z[ino].shell;
   wantpalette = 1;

   ////  Create 3 8-bit images for r,g, and b
      
   if(newimage(x+10,y+10,w,h,bpp,ct,frames,shell,PERM,wantpalette,g.window_border,0)!=OK)
   {  return NOMEM; }
   rino = ci;
   if(newimage(x+20,y+20,w,h,bpp,ct,frames,shell,PERM,wantpalette,g.window_border,0)!=OK)
   {  eraseimage(rino,0,0,1); return NOMEM; }
   gino = ci;
   if(newimage(x+30,y+30,w,h,bpp,ct,frames,shell,PERM,wantpalette,g.window_border,0)!=OK)
   {   eraseimage(rino,0,0,1);  eraseimage(gino,0,0,1); return NOMEM; }
   bino = ci;

      
   for(f=0;f<frames;f++)
   for(j=0;j<h;j++)
   for(i=0;i<w;i++)
   {    value = pixelat(z[ino].image[f][j] + i*g.off[obpp], obpp);  
        valuetoRGB(value,rr,gg,bb,obpp);
        z[rino].image[f][j][i] = rr;
        z[gino].image[f][j][i] = gg;
        z[bino].image[f][j][i] = bb;
   }
   setimagetitle(rino,"Red");
   setimagetitle(gino,"Green");
   setimagetitle(bino,"Blue");
   z[rino].touched = 1;
   z[gino].touched = 1;
   z[bino].touched = 1;
   rebuild_display(rino);
   rebuild_display(gino);
   rebuild_display(bino);
   redraw(rino);
   redraw(gino);
   redraw(bino);
   switchto(rino);
   switchto(gino);
   switchto(bino);
   if(g.autoundo){ backupimage(rino,0); backupimage(gino,0); backupimage(bino,0); }
   return OK;
}


//-------------------------------------------------------------------------//
// combine_colors                                                          //
//-------------------------------------------------------------------------//
int combine_colors(int noofargs, char **arg)
{
  noofargs=noofargs; arg=arg;
  int f,h,i,j,ct,x,y,w,rr,gg,bb,value,frames,wantpalette,shell;
  static int rino=ci, gino=ci, bino=ci;
  char temp[100];
  int helptopic = 0;
   
  if(noofargs==0)
  {
        clickboxinfo *item;
        item = new clickboxinfo[3];
        item[0].wantpreview=0;
        item[0].title = new char[128];
        strcpy(item[0].title,"Red");
        item[0].startval = rino;
        item[0].minval = 0;
        item[0].maxval = g.image_count;
        item[0].type = 2;
        item[0].wantdragcb = 1;
        item[0].answers = new int[10]; // Must be 10 for multiclickbox cb
        item[0].form = NULL;
        item[0].path = NULL;

        item[1].title = new char[128];
        strcpy(item[1].title,"Green");
        item[1].startval = gino;
        item[1].minval = 0;
        item[1].maxval = g.image_count;
        item[1].type = 2;
        item[1].wantdragcb = 1;
        item[1].form = NULL;
        item[1].path = NULL;

        item[2].title = new char[128];
        strcpy(item[2].title,"Blue");
        item[2].startval = bino;
        item[2].minval = 0;
        item[2].maxval = g.image_count;
        item[2].type = 2;
        item[2].wantdragcb = 1;
        item[2].form = NULL;
        item[2].path = NULL;
        strcpy(temp,"Source images for red, green, and blue");
        multiclickbox(temp, 3, item, null, helptopic);
        rino = item[0].answer;
        gino = item[1].answer;
        bino = item[2].answer;
        delete[] item[0].answers;
        delete[] item[0].title;
        delete[] item[1].title;
        delete[] item[2].title;
        delete[] item;                                   
   }else
   {
        if(noofargs>1) rino = image_number(arg[1]);
        if(g.getout){ g.getout=0; return ERROR; }
        if(noofargs>2) gino = image_number(arg[2]);
        if(g.getout){ g.getout=0; return ERROR; }
        if(noofargs>3) bino = image_number(arg[3]); 
        if(g.getout){ g.getout=0; return ERROR; }
   }

   if(z[rino].bpp!=8 || z[gino].bpp!=8 || z[bino].bpp!=8)
   {   message("Source images must be 8 bits/pixel",ERROR); return BADPARAMETERS; }
   if(z[rino].xsize!= z[gino].xsize || 
      z[rino].xsize!= z[bino].xsize || 
      z[gino].xsize!= z[bino].xsize || 
      z[rino].ysize!= z[gino].ysize || 
      z[rino].ysize!= z[bino].ysize || 
      z[gino].ysize!= z[bino].ysize )
   {   message("Source images must be the same size",ERROR); return BADPARAMETERS; }


   ////  Create 24-bit image for combination


   x = z[rino].xpos;
   y = z[rino].ypos;
   w = z[rino].xsize;
   h = z[rino].ysize;
   frames = min(min(z[rino].frames,z[gino].frames),z[bino].frames);
   ct = COLOR;
   shell = z[rino].shell;
   wantpalette = 0;
    
   if(newimage(x+10,y+10,w,h,24,ct,frames,shell,PERM,wantpalette,g.window_border,0)!=OK)
   {  return NOMEM; }

   for(f=0;f<frames;f++)
   for(j=0;j<h;j++)
   for(i=0;i<w;i++)
   {    rr = z[rino].image[f][j][i];
        gg = z[gino].image[f][j][i];
        bb = z[bino].image[f][j][i];
        value = RGBvalue(rr,gg,bb,24);
        putpixelbytes(z[ci].image[f][j]+i*g.off[24], value, 1, 24);
   }
   setimagetitle(ci,"Composite");
   z[ci].touched = 1;
   rebuild_display(ci);
   redraw(ci);
   switchto(ci);
   if(g.autoundo) backupimage(ci,0); 
   return OK;
}


//-------------------------------------------------------------------------//
// shift_frame                                                             //
//-------------------------------------------------------------------------//
int shift_frame(int noofargs, char **arg)
{
  static int vshift=0, hshift=0;
  char temp[100];
  int bpp,f,helptopic = 0,i,j,ir,iw,jr,jw,bytesperline,value,xstart,ystart,xend,yend;
  uchar *radd, *wadd;
  if(ci==0){ message("Can't shift background"); return ERROR; }
  if(ci<0 || ci>=g.image_count) { message("Bad image number"); return ERROR; }
   
  if(noofargs==0)
  {    clickboxinfo *item;
       item = new clickboxinfo[2];
       item[0].wantpreview=0;
       item[0].title = new char[128];        
       strcpy(item[0].title,"Vertical");
       item[0].startval = vshift;
       item[0].minval = -100;
       item[0].maxval = 100;
       item[0].type = VALSLIDER;
       item[0].wantdragcb = 0;
       item[0].answers = new int[10]; // Must be 10 for multiclickbox cb

       item[1].title = new char[128];        
       strcpy(item[1].title,"Horizontal");
       item[1].startval = hshift;
       item[1].minval = -100;
       item[1].maxval = 100;
       item[1].type = VALSLIDER;
       item[1].wantdragcb = 0;
       strcpy(temp,"Shift frame for current image");
       multiclickbox(temp, 2, item, null, helptopic);
       vshift = item[0].answer;
       hshift = item[1].answer;
       delete[] item[0].title;
       delete[] item[0].answers;
       delete[] item;                                   
  }else
  {    if(noofargs>1) vshift = atoi(arg[1]);
       if(noofargs>2) hshift = atoi(arg[2]);
  }

  f = z[ci].cf;
  bpp = z[ci].bpp;
  xstart = -hshift * g.off[bpp];
  ystart = -vshift;
  bytesperline = z[ci].xsize * g.off[bpp];
  xend   = bytesperline - g.off[bpp];
  yend   = z[ci].ysize - 1;

  if(hshift<0)
  {    for(j=0; j<z[ci].ysize;j++)
       for(iw=0,ir=xstart; iw<bytesperline; iw+=g.off[bpp],ir+=g.off[bpp])
       {    if(ir>=0 && ir<bytesperline)
            {   radd = z[ci].image[f][j] + ir;
                value = pixelat(radd,bpp);
            }else value=0;
            wadd = z[ci].image[f][j] + iw;
            putpixelbytes(wadd, value, 1, bpp);
       }
  }
  if(hshift>0)
  {    for(j=0; j<z[ci].ysize;j++)
       for(iw=xend, ir=xstart+xend; iw>=0; iw-=g.off[bpp],ir-=g.off[bpp])
       {    if(ir>=0 && ir<bytesperline)
            {   radd = z[ci].image[f][j] + ir;
                value = pixelat(radd,bpp);
            }else value=0;
            wadd = z[ci].image[f][j] + iw;
            putpixelbytes(wadd, value, 1, bpp);
       }
  }
  
  if(vshift<0)
  {    for(jw=0,jr=ystart; jw<z[ci].ysize; jw++,jr++)
       for(i=0; i<bytesperline; i+=g.off[bpp])
       {    if(jr>=0 && jr<z[ci].ysize)
            {   radd = z[ci].image[f][jr] + i;
                value = pixelat(radd,bpp);
            }else value=0;
            wadd = z[ci].image[f][jw] + i;
            putpixelbytes(wadd, value, 1, bpp);
       }
  }
  if(vshift>0)
  {    for(jw=yend, jr=ystart+yend; jw>=0; jw--, jr--)
       for(i=0; i<bytesperline; i+=g.off[bpp])
       {    if(jr>=0 && jr<z[ci].ysize)
            {   radd = z[ci].image[f][jr] + i;
                value = pixelat(radd,bpp);
            }else value=0;
            wadd = z[ci].image[f][jw] + i;
            putpixelbytes(wadd, value, 1, bpp);
       }
  }
  rebuild_display(ci); 
  redraw(ci);
  return OK;
}



//-------------------------------------------------------------------------//
// grab_window                                                             //
//-------------------------------------------------------------------------//
void grab_window(void)
{
  int rx,ry,wx,wy,button=0,ct=g.colortype; 
  uint uw,uh,ubw,bpp;   
  int xroot,yroot,xstart=0,ystart=0;
  uint keys;
  Window rwin, cwin;
  message("Click on desired window to grab");
  while(!button)
  {     XQueryPointer(g.display,g.root_window,&rwin,&cwin,&rx,&ry,&wx,&wy,&keys);
        button = keys & Button1Mask;
  }
  XGetGeometry(g.display, cwin, &rwin, &xroot, &yroot, &uw, &uh, &ubw, &bpp);
  // ignore the sometimes incorrect bpp information from XGetGeometry
  bpp = g.bitsperpixel;  
  // Make sure grabbed window is not off screen
  uw = min(g.xres - xroot, (int)uw);
  uh = min(g.yres - yroot, (int)uh);
  if(xroot < 0){ xstart = -xroot; uw -= xstart; } 
  if(yroot < 0){ ystart = -yroot; uh -= ystart; } 
  if(newimage(0,0,uw,uh,bpp,ct,1,g.want_shell,PERM,0,g.window_border,0)==OK)
     XGetSubImage(g.display, cwin, xstart, ystart, uw, uh, XAllPlanes(), 
        ZPixmap, z[ci].image_ximage, 0, 0); 
  else{ message(g.nomemory); return; }
}


//-------------------------------------------------------------------------//
//  button_clicked                                                         //
//-------------------------------------------------------------------------//
int button_clicked(void)
{
  int rx,ry,wx,wy,button=0;   
  uint keys;
  Window rwin, cwin;
  XQueryPointer(g.display,g.root_window,&rwin,&cwin,&rx,&ry,&wx,&wy,&keys);
  button = keys & Button1Mask;
  return button;
}


//-------------------------------------------------------------------------//
//  get_image_list                                                         //
//-------------------------------------------------------------------------//
int get_image_list(int *source, int &count)
{
   int k, bpp;
   char inputstring[1024];
   memset(inputstring,0,1024);
   char tempstring[1024] = "Enter list of image numbers,\ne.g. 1 2 3 6 8 11-23 34\nNote: all images must have the same bits/pixel";
   message(tempstring, inputstring, PROMPT, 1023);
   count = parse_digits(inputstring, source, MAXIMAGES);
   for(k=0;k<count;k++) if(source[k]<=0 || source[k]>=g.image_count)
   {   sprintf(tempstring,"Bad image number(%d)",source[k]); 
       message(tempstring,ERROR);
       return ERROR;
   }
   bpp = z[source[0]].bpp;
   for(k=0;k<count;k++) if(z[source[k]].bpp !=bpp)
   {   sprintf(tempstring,"Image number(%d) has wrong bits/pixel (%d)\n\
Convert image to %d bits/pixel first",source[k],z[source[k]].bpp,bpp); 
       message(tempstring,ERROR);
       return ERROR;
   }
   if(count<1) return BADPARAMETERS;
   return OK;
}


//-------------------------------------------------------------------------//
//  combine_frames                                                         //
//-------------------------------------------------------------------------//
int combine_frames(int xpos, int ypos, int shell, int border)
{
   int ct,bpp,f,i,i2,ino,j,count=0,xsize,ysize;
   int source[MAXIMAGES];
   uchar *add;
   uint value;
   if(get_image_list(source, count)!=OK) return ERROR;
   ct = z[source[0]].colortype;
   bpp = z[source[0]].bpp;
   xsize = z[source[0]].xsize;
   ysize = z[source[0]].ysize;
   if(xsize<1 || ysize<1){ xsize=z[source[0]].xsize; ysize=z[source[0]].ysize; }
   if(newimage(xpos,ypos,xsize,ysize,bpp,ct,count,shell,PERM,0,border,0)!=OK) return NOMEM;
   for(f=0; f<count; f++)
   {  ino = source[f];      
      for(j=0; j<ysize; j++)
      for(i=0,i2=0; i<xsize; i++,i2+=g.off[bpp])
      {   add = z[ino].image[0][j] + i2;
          value = pixelat(add,bpp);  
          putpixelbytes(z[ci].image[f][j]+i2,value,1,bpp,1);
      }
   }
   return OK;
}


//-------------------------------------------------------------------------//
// composite_images                                                        //
//-------------------------------------------------------------------------//
int composite_images(int xsize, int ysize, int xpos, int ypos, int shell, int border)
{
   int ct,bpp,bpp2,i,i2,i3,ino,j,k,count=0,subimage,xsubsize,ysubsize,xsub,ysub,oci,
       needconvert=0;
   int source[MAXIMAGES];
   char title[64], **label, **answer, tempstring[1024];  
   uchar *add;
   int value;

   if(get_image_list(source, count)!=OK) return ERROR;
   label = new char *[count];
   answer = new char *[2*count];

   strcpy(title, "Positions for subimages in composite image");
   ct = z[source[0]].colortype;
   for(k=0; k<count; k++)
   {   label[k] = new char[128];        
       ino = source[k];     
       strcpy(tempstring, basefilename(z[ino].name));
       tempstring[20]=0;  // truncate name
       sprintf(label[k], "Position for image %c%d (%s)", '#', ino, tempstring);
       answer[k] = new char[128];
       sprintf(answer[k], "%d", 0);
       answer[k+count] = new char[128];
       sprintf(answer[k+count], "%d", 0);
       if(z[ino].colortype!=GRAY) ct = z[ino].colortype;
   }
   getstrings(title, label, answer, 2, count, 128);

   bpp = bpp2 = z[source[0]].bpp;
   if(ct!=GRAY){ bpp2 = 24; ct=COLOR; needconvert=1; }
   if(xsize<1 || ysize<1){ xsize=z[source[0]].xsize; ysize=z[source[0]].ysize; }
   if(newimage(xpos,ypos,xsize,ysize,bpp2,ct,1,shell,PERM,0,border,0)!=OK) return NOMEM;
   oci = ci;
   for(j=0; j<z[ci].ysize; j++) memset(z[ci].image[0][j],255,z[ci].xsize*g.off[bpp2]);
   bpp = z[source[0]].bpp;
   for(subimage=0; subimage<count; subimage++)
   {   ino = source[subimage];
       switchto(ino); 
       memcpy(g.palette, z[ino].palette, 768);
       xsub = atoi(answer[subimage]);        // x position for subimage in composite
       ysub = atoi(answer[subimage+count]);  // y position for subimage in composite
       xsubsize = z[ino].xsize;              // x size of subimage
       ysubsize = z[ino].ysize;              // y size of subimage
       for(j=0; j<ysubsize; j++)
       for(i=0,i2=0,i3=xsub*g.off[bpp2]; i<xsubsize; i++,i2+=g.off[bpp],i3+=g.off[bpp2])
       {   if(j+ysub >= z[oci].ysize || i+xsub >= z[oci].xsize) continue;
           if(j+ysub < 0 || i3 < 0) continue;
           add = z[ino].image[0][j] + i2;
           value = pixelat(add, bpp);  
           if(ct==GRAY) value = convertpixel(value, bpp, bpp2, 0);
           else         value = convertpixel(value, bpp, bpp2, 1);
           putpixelbytes(z[oci].image[0][j+ysub]+i3,value,1,bpp2,1);
       }
   }
   memcpy(z[ci].palette, z[source[0]].palette, 768);
   memcpy(z[ci].opalette, z[source[0]].opalette, 768);
   memcpy(z[ci].spalette, z[source[0]].spalette, 768);
   memcpy(g.palette, z[source[0]].palette, 768);
   switchto(oci);
   if(needconvert) change_image_depth(oci, bpp, 1);
   for(k=0;k<count;k++) delete[] label[k];
   for(k=0;k<2*count;k++) delete[] answer[k];
   delete[] answer;
   return OK;
}
