//--------------------------------------------------------------------------//
// xmtnimage41.cc                                                           //
// Graphs                                                                   //
// Latest revision: 05-22-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Images     *z;
extern int         ci;
const int AUTOSCALE=0;
inline double cosine(double x);
inline double sine(double x);

#define SLIDERS 15     //// Change this if adding slider to 3d dialog
#define BUTTONS3D 5    //// Change this if adding button to 3d dialog

//--------------------------------------------------------------------------//
// Plot      -     Motif version                                            //
// Simple bar graph - returns GOTNEW if user converted it to a new image    //
//  calx = 1 if x values are to be calculated using calibratepixel(), 0 if  //
//          some other kind of graph                                        //
//  f = y calibration factor                                                //
//  For 2D data - 1st array index is dimension, 2nd is element.             //
//  'mode' can be MEASURE, CHANGE, SUBTRACT                                 //
//  'title' should be char[FILENAMELENGTH] since it can be changed by user. //
//  'ytitle' is a list of 1 title for each panel on y axis.                 //
//  'draw_callback' is called once every time part of graph is redrawn.     //
//  'point_callback' is called once every time a point is changed.          //
//--------------------------------------------------------------------------//
int plot(char *title, char *xtitle, char **ytitle, double *xdata, int *pdata, 
   int n, int dims, int mode, int ymin, int ymax,
   void (*f1cb)(pcb_struct v), 
   void (*f2cb)(void *pd, int n1, int n2),
   void (*f3cb)(void *pd, int n1, int n2),
   int helptopic)
{  int k,status=OK;
   dims=1;    
   array<double> ddd(n,1);
   for(k=0;k<n;k++) ddd.p[0][k] = (double)pdata[k];
   status = plot(title,xtitle,ytitle,xdata,ddd.p,n,dims,mode,ymin,ymax,f1cb,f2cb,f3cb,helptopic);
   for(k=0;k<n;k++) pdata[k]=(int)ddd.p[0][k];
   return(status);
}
int plot(char *title, char *xtitle, char **ytitle, double *xdata, double *pdata, 
   int n, int dims, int mode, int ymin, int ymax, 
   void (*f1cb)(pcb_struct v),
   void (*f2cb)(void *pd, int n1, int n2),
   void (*f3cb)(void *pd, int n1, int n2),
   int helptopic)
{  int k,status=OK;
   dims=1;    
   array<double> ddd(n,1);
   for(k=0;k<n;k++) ddd.p[0][k] = pdata[k];
   status = plot(title,xtitle,ytitle,xdata,ddd.p,n,dims,mode,ymin,ymax,f1cb,f2cb,f3cb,helptopic);
   for(k=0;k<n;k++) pdata[k]=ddd.p[0][k];
   return(status);
}
int plot(char *title, char *xtitle, char **ytitle, double *xdata, int **pdata, 
   int n, int dims, int mode, int ymin, int ymax,
   void (*f1cb)(pcb_struct v),
   void (*f2cb)(void *pd, int n1, int n2),
   void (*f3cb)(void *pd, int n1, int n2),
   int helptopic)
{  int j,k,status=OK;
   array<double> ddd(n+10,dims+2);
   for(j=0;j<dims;j++) for(k=0;k<n;k++) ddd.p[j][k]=(double)pdata[j][k];
   status = plot(title,xtitle,ytitle,xdata,ddd.p,n,dims,mode,ymin,ymax,f1cb,f2cb,f3cb,helptopic);
   for(j=0;j<dims;j++) for(k=0;k<n;k++) pdata[j][k]=(int)ddd.p[j][k];
   return(status);
}
int plot(char *title, char *xtitle, char **ytitle, double *xdata, double **pdata, 
   int npoints, int dims, int mode, int ymin, int ymax,
   void (*f1cb)(pcb_struct v),
   void (*f2cb)(void *pd, int n1, int n2),
   void (*f3cb)(void *pd, int n1, int n2),
   int helptopic)
{

   //////////////////////////////////////////////////////////////////////
   // To add a new button:                                             //
   // 1. New button must be second last, otherwise clickboxdata[] and  //
   //    buttonlabels must be changed in several places.               //
   // 2. Change NOOFBUTTONS                                            //
   // 3. Add label to buttonlabel[] (last button must be 'finished').  //
   // 4. Add new CASE statement in graphbuttoncb()                     //
   // 5. Set the button type to PUSHBUTTON or TOGGLE.                  //
   //////////////////////////////////////////////////////////////////////

  if(memorylessthan(16384)){ message(g.nomemory,ERROR); return(ABORT); }
#ifdef DIGITAL
  #define NOOFBUTTONS 18;  // Change this if adding new button
  #define BOXSIZE 21;      // Vertical size of buttons in pixels
  #define BOXWIDTH 100;    // Horizontal size of buttons in pixels
  #define LABELWIDTH 30;   // Horizontal size of y axis label in pixels
  #define LABELHEIGHT 30;  // Vertical size of y axis label in pixels
  #define TOPPOS 1;        // Top of graph window in % of form size
  #define BOTPOS 92;       // Bottom of graph window in % of form size
  #define LEFTPOS 4;       // Left of graph window in % of form size
  #define PUSHBUTTON 0;
  #define TOGGLEBUTTON 1;
#else
  const int NOOFBUTTONS=18;  // Change this if adding new button
  const int BOXSIZE=21;      // Vertical size of buttons in pixels
  const int BOXWIDTH=100;    // Horizontal size of buttons in pixels
  const int LABELWIDTH=30;   // Horizontal size of y axis label in pixels
  const int LABELHEIGHT=30;  // Vertical size of y axis label in pixels
  const int TOPPOS=1;        // Top of graph window in % of form size
  const int BOTPOS=92;       // Bottom of graph window in % of form size
  const int LEFTPOS=4;       // Left of graph window in % of form size
  const int PUSHBUTTON=0;
  const int TOGGLEBUTTON=1;
#endif
  int buttontype[NOOFBUTTONS];
  int button_initial_state[NOOFBUTTONS];
  char **buttonlabel;
  XmString xms;
  Arg args[100];
  int gsize, top, xsize, ysize = 500, j, k, n, wc=0, x,y,rr,gg,bb;
  xsize = max(350, BOXWIDTH + LABELWIDTH + npoints + 10);
  int *markmin, *markmax, *markstart, *markend;
  double *scale, *area;
  Window *windows, *ylabelwindow, xlabelwindow;
  if(helptopic==-1) helptopic=13;

  buttonlabel = new char*[NOOFBUTTONS];
  buttonlabel[0] = new char[20]; strcpy(buttonlabel[0], "Save to file ");
  buttonlabel[1] = new char[20]; strcpy(buttonlabel[1], "Read file    ");
  buttonlabel[2] = new char[20]; strcpy(buttonlabel[2], "Smooth       ");
  buttonlabel[3] = new char[20]; strcpy(buttonlabel[3], "Auto.Baseline");
  buttonlabel[4] = new char[20]; strcpy(buttonlabel[4], "Manl.Baseline");
  buttonlabel[5] = new char[20]; strcpy(buttonlabel[5], "Capture image");
  buttonlabel[6] = new char[20]; strcpy(buttonlabel[6], "Peak areas   ");
  buttonlabel[7] = new char[20]; strcpy(buttonlabel[7], "BL Smoothness");
  buttonlabel[8] = new char[20]; strcpy(buttonlabel[8], "Scale        ");
  buttonlabel[9] = new char[20]; strcpy(buttonlabel[9], "Multiply     ");
  buttonlabel[10]= new char[20];strcpy(buttonlabel[10], "Add value    ");
  buttonlabel[11]= new char[20];strcpy(buttonlabel[11], "Change       ");
  buttonlabel[12]= new char[20];strcpy(buttonlabel[12], "Bezier change");
  buttonlabel[13]= new char[20];strcpy(buttonlabel[13], "Maximize     ");
  buttonlabel[14]= new char[20];strcpy(buttonlabel[14], "Invert       ");
  buttonlabel[15]= new char[20];strcpy(buttonlabel[15], "Reset        ");
  buttonlabel[16]= new char[20];strcpy(buttonlabel[16], "Rescale      ");
  buttonlabel[17]= new char[20];strcpy(buttonlabel[17], "Finished     ");//<-Must be last!

  ////  Button type and initial state (for toggle buttons only)
  for(k=0;k<NOOFBUTTONS;k++)
  {   buttontype[k]=PUSHBUTTON; 
      button_initial_state[k]=False; 
  }
  buttontype[11]=TOGGLEBUTTON;
  button_initial_state[11] = (mode==CHANGE);
    
  clickboxinfo *clickboxdata;
  clickboxdata = new clickboxinfo[NOOFBUTTONS];
  for(k=0;k<NOOFBUTTONS;k++)
  {    clickboxdata[k].title = new char[128]; 
       clickboxdata[k].title[0] = 0; 
       clickboxdata[k].form = NULL; 
       clickboxdata[k].path = NULL; 
  }
#ifdef DIGITAL
  Widget graph[3], ylabelwidget[3], xlabelwidget, form, button[NOOFBUTTONS], *w; 
#else
  Widget graph[dims], ylabelwidget[dims], xlabelwidget, form, button[NOOFBUTTONS], *w; 
#endif
  PlotData pd;                         // Data 

  markmin   = new int[dims];           // small enough not to check
  markmax   = new int[dims];
  markstart = new int[dims];
  markend   = new int[dims];
  windows   = new Window[dims];
  ylabelwindow = new Window[dims];
  scale     = new double[dims];
  area      = new double[dims];
  w         = new Widget[100];
  g.inmenu++;

 //-------------------graph form--------------------------------------//
  n=0;
  XtSetArg(args[n], XmNwidth, xsize); n++;
  XtSetArg(args[n], XmNheight, ysize); n++;
  XtSetArg(args[n], XmNmarginWidth, 1); n++;
  XtSetArg(args[n], XmNtransient, False); n++;
  XtSetArg(args[n], XmNresizable, True); n++;
  XtSetArg(args[n], XmNautoUnmanage, False); n++;
  XtSetArg(args[n], XmNdialogTitle, xms=XmStringCreateSimple(title)); n++;
  ////  Stop Motif from trying to grab another color if none are available.
  if(g.want_colormaps)  
  {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
       XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
  }
  form = w[wc++] = XmCreateFormDialog(g.main_widget, "GraphForm", args, n);
  XmStringFree(xms);
  XtAddEventHandler(XtParent(form), FocusChangeMask, False,
        (XtEH)graph_raise_cb, (XtP)NULL);


 //-------------------graph area (at right)---------------------------//
  top = TOPPOS;
  gsize = (BOTPOS-TOPPOS)/dims;
  for(k=0;k<dims;k++)
  {    n=0;
       XtSetArg(args[n], XmNtitle, "Graph"); n++;
       XtSetArg(args[n], XmNwidth, npoints); n++;
       XtSetArg(args[n], XmNmarginHeight, 0); n++;
       XtSetArg(args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
       XtSetArg(args[n], XmNmappedWhenManaged, True); n++;
       XtSetArg(args[n], XmNleftPosition, 0); n++;            // % of width fromleft
       XtSetArg(args[n], XmNleftOffset, LEFTPOS+BOXWIDTH+LABELWIDTH); n++; // pixels from left
       XtSetArg(args[n], XmNtopPosition, top); n++;           // % of height from top
       XtSetArg(args[n], XmNbottomPosition, top+gsize); n++;  // % of width from top
       XtSetArg(args[n], XmNfractionBase, 100); n++;          // Use percentages
       XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNrightAttachment, XmATTACH_NONE); n++;
       XtSetArg(args[n], XmNbackground, g.bcolor); n++;
       ////  Stop Motif from trying to grab another color if none are available.
       if(g.want_colormaps)  
       {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
            XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
       }
       graph[k] = w[wc++] = XmCreateDrawingArea(form, "graph", args, n);
       XtAddCallback(graph[k],XmNexposeCallback, (XtCBP)graphexposecb,
           (XtP)&pd);
       XtAddCallback(graph[k], XmNinputCallback, (XtCBP)graphcb, (XtP)&pd);

       // Need event handler to catch those wacky mouse movement events.
       XtAddEventHandler(graph[k],g.mouse_mask,False,(XtEH)graphmousecb,(XtP)&pd);
       XtManageChild(graph[k]);
 
  //--------------y axis labels---------------------------------------//

       n=0;
       XtSetArg(args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
       XtSetArg(args[n], XmNmappedWhenManaged, True); n++;
       XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
       XtSetArg(args[n], XmNfractionBase, 100); n++;          // Use percentages
       XtSetArg(args[n], XmNtopPosition, top); n++;           
       XtSetArg(args[n], XmNbottomPosition, top+gsize); n++;  
       XtSetArg(args[n], XmNleftPosition, 0); n++;  
       XtSetArg(args[n], XmNleftOffset, LEFTPOS+BOXWIDTH); n++;  // pixels from left
       XtSetArg(args[n], XmNrightWidget, graph[k]); n++;   
       ////  Stop Motif from trying to grab another color if none are available.
       if(g.want_colormaps)  
       {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
            XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
       }
       ylabelwidget[k] = w[wc++] = XmCreateDrawingArea(form, "YLabelWidget", args, n);
       XtAddCallback(ylabelwidget[k],XmNexposeCallback,
           (XtCBP)graphylabelexposecb, (XtP)&pd);
       XtManageChild(ylabelwidget[k]);
       top += gsize;
  }

  //--------------x axis label----------------------------------------//

  n=0;
  XtSetArg(args[n], XmNmarginWidth, 0); n++;
  XtSetArg(args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  XtSetArg(args[n], XmNmappedWhenManaged, True); n++;
  XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(args[n], XmNfractionBase, 100); n++;          // Use percentages
  XtSetArg(args[n], XmNtopPosition, top); n++;           
  XtSetArg(args[n], XmNbottomPosition, 100); n++;  
  XtSetArg(args[n], XmNbottomWidget, form); n++;  
  XtSetArg(args[n], XmNbottomOffset, LABELHEIGHT); n++;  
  XtSetArg(args[n], XmNleftOffset, LEFTPOS+BOXWIDTH+LABELWIDTH); n++; // pixels from left
  XtSetArg(args[n], XmNleftPosition, 0); n++;  
  XtSetArg(args[n], XmNrightPosition, 100); n++;           
  ////  Stop Motif from trying to grab another color if none are available.
  if(g.want_colormaps)  
  {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
       XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
  }
  xlabelwidget = w[wc++] = XmCreateDrawingArea(form, "XLabelWidget", args, n);
  XtAddCallback(xlabelwidget,XmNexposeCallback,
       (XtCBP)graphxlabelexposecb, (XtP)&pd);
  XtManageChild(xlabelwidget);

    
  //-----Initialize data----------------------------------------------//

  rr = g.maxred[g.bitsperpixel]/2;
  gg = g.maxgreen[g.bitsperpixel]/2;
  bb = g.maxblue[g.bitsperpixel]/2;
  pd.gcolor = RGBvalue(rr,gg,bb, g.bitsperpixel);
  pd.win     = windows;           // Xlib windows for each dimension - must wait until 
                                  //   form widget is managed before filling array.
  pd.ylabelwindow = ylabelwindow; // Xlib windows for y axis label
  pd.title   = title;             // Title of graph
  pd.xtitle  = xtitle;            // X axis label
  pd.ytitle  = ytitle;            // Y axis label
  pd.data    = pdata;             // Arrays of data to plot
  pd.xdata   = xdata;             // Calibrated x values
  pd.dims    = dims;              // No. of graphs in the plot 
  pd.xsize   = xsize;             // Width of entire plot
                                  // Height of plot
  pd.ysize   = ysize*(BOTPOS-TOPPOS)/100/dims;     
  pd.n       = npoints;           // No. of data points
  pd.focus   = 0;                 // Which graph has mouse focus
  pd.xcalib  = 1.0;               // User-specified x axis calibration factor
  pd.ycalib  = 1.0;               // User-specified y axis calibration factor
  pd.scale   = scale;             // Automatic scaling factor
  pd.xscale  = 1.0;               // Not used
  pd.yscale  = 1.0;               // Button-selected y axis scaling factor
  pd.markmin = markmin;           // Min value of selected area
  pd.markmax = markmax;           // Max value of selected area
  pd.markstart = markstart;       // Actual start of mouse dragging
  pd.markend   = markend;         // Actual end of mouse dragging
  pd.helptopic = helptopic;       // Help topic for the graph
  pd.type      = mode;            // 0=get areas, 1=change values when dragging
  pd.f         = f1cb;            // Callback: what to do when mouse is dragged
  pd.f2        = f2cb;            // Callback: what to do when graph is redrawn
  pd.f3        = f3cb;            // Callback: what to do when click on maximize button
  pd.area      = area;            // Most recently-selected area
  pd.button    = 0;               // Mouse button currently pressed
  pd.ymin      = (double)ymin;    // Minimum observed value 
  pd.ymax      = (double)ymax;    // Maximum observed value 
  pd.llimit    = (double)ymin;    // Minimum allowable value (0=don't care)
  pd.ulimit    = (double)ymax;    // Maximum allowable value (0=don't care)
  pd.reason    = NORMAL;          // Why draw_graph was called
  // pd.formwin is assigned below  
  
  for(k=0;k<dims;k++)
  {    markmin[k]=0;  
       markmax[k]=0; 
       markstart[k]=0; 
       markend[k]=0; 
       scale[k] = 0.0;
       area[k] = 0.0;
  }

  //------Add buttons on left-----------------------------------------//
  //  The last button ('finished') is bigger than the others and is   //
  //  left unmanaged.                                                 //
  //------------------------------------------------------------------//
  
  x = 5;
  y = 5;
  for(j=0;j<NOOFBUTTONS;j++)
  {     n = 0;
        XtSetArg(args[n], XmNx, x); n++;
        XtSetArg(args[n], XmNy, y); n++;
        XtSetArg(args[n], XmNwidth, BOXWIDTH); n++;  
        XtSetArg(args[n], XmNresizable, False); n++;  
#ifdef MOTIF2
        XtSetArg(args[n], XmNindicatorOn, XmINDICATOR_NONE); n++;
#else
        XtSetArg(args[n], XmNindicatorOn, False); n++;
#endif
        XtSetArg(args[n], XmNtraversalOn, True); n++;
        if(j==NOOFBUTTONS-1)
        {   XtSetArg(args[n], XmNheight, BOXSIZE*2); n++; }
        else
        {   XtSetArg(args[n], XmNheight, BOXSIZE); n++;  }
        XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
        ////  Stop Motif from trying to grab another color if none are available.
        if(g.want_colormaps)  
        {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
             XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
        }
        clickboxdata[j].button = j;
        clickboxdata[j].helptopic = helptopic;
        clickboxdata[j].ptr = &pd;
        clickboxdata[j].f1 = null;
        clickboxdata[j].f2 = null;
        clickboxdata[j].f3 = null;     
        clickboxdata[j].f3 = NULL;     
        clickboxdata[j].maxval = NOOFBUTTONS-1;
        clickboxdata[j].list = buttonlabel;
        if(buttontype[j]==PUSHBUTTON)
        {  button[j] = w[wc++] = XmCreatePushButton(form,buttonlabel[j],args,n);
           XtManageChild(button[j]);
           XtAddCallback(button[j], XmNactivateCallback, 
                (XtCBP)graphbuttoncb, (XtP)(&clickboxdata[j]));
        }else if(buttontype[j]==TOGGLEBUTTON)
        {  button[j] = w[wc++] = XmCreateToggleButton(form,buttonlabel[j],args,n);
           XtManageChild(button[j]);
           XtAddCallback(button[j], XmNvalueChangedCallback, 
                (XtCBP)graphbuttoncb, (XtP)(&clickboxdata[j]));
           XmToggleButtonSetState(button[j],button_initial_state[j],False);
        }
        y += BOXSIZE;
  }

  ////  Let manual baseline button map & unmap the last button
  for(j=0;j<NOOFBUTTONS;j++)
        clickboxdata[j].widget[0] = button[NOOFBUTTONS-1];


  //--------Ok, Cancel, Help buttons & their callbacks----------------//

  addstandardbuttons(form, &w[wc], &pd.helptopic);
  wc += 3;    // addstandardbuttons adds 3 widgets

  XtManageChild(form);
  for(k=0;k<dims;k++)  
  {    windows[k] = XtWindow(graph[k]);        // Window to draw in
       ylabelwindow[k] = XtWindow(ylabelwidget[k]);  // Window for y axis labels
  }
  xlabelwindow = XtWindow(xlabelwidget);       // Window for y axis labels
  pd.xlabelwindow = xlabelwindow; // Xlib windows for x axis label
  pd.formwin = XtWindow(form);    //  Window for labels
  XtUnmapWidget(button[NOOFBUTTONS-1]);

  //--------Wait around for something to happen-----------------------//
 
  event_loop(form);

  //--------Get rid of event handlers before leaving------------------//
  for(k=0;k<dims;k++)
       XtRemoveEventHandler(graph[k], g.mouse_mask, False, 
            (XtEH)graphmousecb, (XtP)&pd);

  XtRemoveEventHandler(XtParent(form), StructureNotifyMask, False, 
       (XtEH)graph_raise_cb, (XtP)NULL);

  XtUnmanageChild(form);
  for(k=0;k<NOOFBUTTONS;k++)
  {     delete[] buttonlabel[k];
        delete[] clickboxdata[k].title; 
  }
  delete[] clickboxdata;
  delete[] area;
  delete[] scale;
  delete[] markstart;
  delete[] markend;
  delete[] markmax;
  delete[] markmin;
  delete[] windows;
  delete[] ylabelwindow;
  delete[] buttonlabel;
  for(k=0;k<wc;k++) XtDestroyWidget(w[k]);
  delete[] w;
  g.inmenu--;
  return OK;
}


//--------------------------------------------------------------------------//
//  draw_graph                                                              //
//  Draw segment of graph from data point n1 to data point n2               //
//  Also calculates area of selected region of graph                        //
//--------------------------------------------------------------------------//
double draw_graph(PlotData *pd, int n1, int n2)
{ 
    int focus,k,mark,mmstart,mmend,n,y,ye,win;
    double answer=0.0, scale=0.0, largest;
    n      = pd->n;
    y      = pd->ysize;
    focus  = pd->focus;
    win    = pd->win[focus];
    mmstart= pd->markstart[focus];
    mmend  = pd->markend[focus];
    if(mmstart>mmend) swap(mmstart,mmend);
    if(n1>n2) swap(n1,n2);
    n1     = max(0,min(n,n1     ));
    n2     = max(0,min(n,n2     ));
    mmstart= max(0,min(n,mmstart));
    mmend  = max(0,min(n,mmend  ));
    scale  = pd->scale[focus];
    ////  Automatically rescale graph
    if(!scale || AUTOSCALE)
    {
        largest= findbiggest(pd->data[focus], n);
        if(largest!=0) scale = (double)(pd->ysize)/largest;
        if(scale==0.0) scale = 1e6; 
    }
    pd->scale[focus] = scale;
    
    pd->gcolor = WhitePixel(g.display, g.screen);
    blackbox(n1,0,n2-1,pd->ysize,pd->gcolor,win);
    XSetForeground(g.display, g.image_gc, BlackPixel(g.display, g.screen));
    XSetBackground(g.display, g.image_gc, WhitePixel(g.display, g.screen));
      
    if(focus < pd->dims) XDrawLine(g.display, win, g.image_gc, n1, y-1, n2, y-1);
    mark=0;
    for(k=n1;k<n2;k++) 
    {   if(k>=mmstart && pd->type==MEASURE) mark=1;
        if(k>=mmend) mark=0;
        ye = pd->ysize - (int)(pd->data[focus][k]*scale);
        if(mark)
        {    answer += pd->data[focus][k];
             XDrawLine(g.display, win, g.image_gc, k, 0, k, ye);
        }else
        {    XDrawLine(g.display, win, g.image_gc, k, y, k, ye);
        }	
    }

    if(pd->type != MEASURE) answer = 0.0;
    pd->area[focus] = answer;        
    ////  Don't call callback if just redrawing graph after expose event
    if(pd->reason==NORMAL) pd->f2(pd,n1,n2); 
    return answer;
}


//--------------------------------------------------------------------------//
//  redraw_graph                                                            //
//--------------------------------------------------------------------------//
void redraw_graph(PlotData *pd)
{
   pd->reason=EXPOSE;
   draw_graph(pd, 0, pd->n);
   pd->reason=NORMAL;
}


//--------------------------------------------------------------------------//
//  graph_raise_cb                                                          //
//--------------------------------------------------------------------------//
void graph_raise_cb(Widget w, XtP client_data, XEvent *event)
{
  client_data=client_data; event=event;
  if(g.state != MESSAGE) XRaiseWindow(g.display, XtWindow(w));
}


//--------------------------------------------------------------------//
// graphcb - callback for graph input                                 //
//--------------------------------------------------------------------//
void graphcb(Widget w, XtP client_data, XmACB *call_data)
{  
  static int ofocus=-1,ox=-1;
  XmDrawingAreaCallbackStruct *ptr;
  Window win;
  XEvent event;  

  int focus,k,x,y;
  ptr = (XmDrawingAreaCallbackStruct*) call_data;
  if(ptr==NULL) return;
  PlotData *pd = (PlotData *)client_data;
  event = *(ptr->event);
  focus = pd->focus;
  x = (int)event.xbutton.x;
  y = (int)event.xbutton.y;

   // All X events except MotionNotify in the graph window are handled here.

  switch(ptr->event->type)
  {    case ButtonPress: 
            pd->button = (int)event.xbutton.button;
            win = XtWindow(w);
            for(k=0;k<pd->dims;k++) if(win==pd->win[k]){ focus = k; break; }
            pd->focus = focus;
            print_graph_coordinates(pd,x,y);

            pd->markstart[focus] = x;
            pd->markend[focus]   = x;
                 // Unmark old marked area
            draw_graph(pd, pd->markmin[focus], pd->markmax[focus]); 

            pd->markmin[focus] = x;
            pd->markmax[focus] = x;
            ox=x;
            break;
       case ButtonRelease: 
            pd->button = 0;
            pd->markmin[focus] = pd->n;
            pd->markmax[focus] = 0;
            ////  If only changing focus, don't call callback
            if(focus==ofocus || x!=ox) pd->f2(pd,0,pd->n);   
            ofocus = focus;
            break;
   }  
   return;
}  


//--------------------------------------------------------------------//
// graphexposecb - callback for graph expose                          //
//--------------------------------------------------------------------//
void graphexposecb(Widget w, XtP client_data, XmACB *call_data)
{  
  call_data=call_data;  // keep compiler quiet
  int k,ofocus;
  Window win;
  PlotData *pd = (PlotData *)client_data;
  ofocus = pd->focus;
  win = XtWindow(w);
  for(k=0;k<pd->dims;k++) if(win==pd->win[k]){ pd->focus = k; break; }
  redraw_graph(pd);
  pd->focus = ofocus;
}  


//--------------------------------------------------------------------//
// graphxlabelexposecb - callback for graph expose x axis label       //
//--------------------------------------------------------------------//
void graphxlabelexposecb(Widget w, XtP client_data, XEvent *event)
{
  w=w;client_data=client_data, event=event;
  PlotData *pd = (PlotData *)client_data;
  print(pd->xtitle, 15, 10, g.main_fcolor, g.main_bcolor, &g.gc,
      pd->xlabelwindow, 0, 0);

}


//--------------------------------------------------------------------//
// graphylabelexposecb - callback for graph expose y axis labels      //
// This must be in a callback because it is impossible to draw in a   //
// window until it receives its first expose event.                   //
//--------------------------------------------------------------------//
void graphylabelexposecb(Widget w, XtP client_data, XEvent *event)
{
  client_data=client_data, event=event;
  PlotData *pd = (PlotData *)client_data;
  int k, panel = 0;
  for(k=0;k<pd->dims;k++) if(XtWindow(w) == pd->ylabelwindow[k]) panel=k;
  print(pd->ytitle[panel], 15, pd->ysize-16, g.main_fcolor, 255, &g.gc,
      pd->ylabelwindow[panel], 0, 1);

}


//--------------------------------------------------------------------//
// graphmousecb - callback for graph mouse movements                  //
//--------------------------------------------------------------------//
void graphmousecb(Widget w, XtP client_data, XEvent *event)
{  
  w=w;  // keep compiler quiet
  double yvalue;
  int focus, mode, x, y; 
  PlotData *pd = (PlotData *)client_data;

  focus = pd->focus;
  mode = pd->type;
  x = event->xmotion.x;
  y = event->xmotion.y;
  if(x<0 || x>=pd->n) return;
  pcb_struct pcb;

  pd->reason=EXPOSE;  // Stop draw_graph from calling f2
  XSetForeground(g.display, g.gc, g.main_fcolor);
  XSetBackground(g.display, g.gc, g.main_bcolor);

  if(pd->button)
  {    if(mode==MEASURE)
       {    pd->markend[focus] = x;
            pd->markmin[focus] = min(pd->markmin[focus],x);
            pd->markmax[focus] = max(pd->markmax[focus],x);
            draw_graph(pd, pd->markmin[focus], pd->markmax[focus]); 
       }
       if(mode==CHANGE && y>=0)
       {    if(pd->scale[focus] > 0.0)
                 yvalue = (double)(pd->ysize - y) / pd->scale[focus];
            else yvalue = (double)(pd->ysize - y); 
            yvalue = min(yvalue, (double)MAXINT/2);
            pd->data[focus][x] = yvalue;
            pd->data[focus][max(0,x-1)] = yvalue; // make mouse drag 3 pixels wide
            pd->data[focus][min(pd->n-1,x+1)] = yvalue;
            draw_graph(pd, x-1, x+1); 
            pcb.x = x;
            pcb.y = (int)yvalue;
            pcb.focus = focus;
            pd->f(pcb);  // Callback for single point
       }
  }
  print_graph_coordinates(pd,x,y);
  return;              
} 

//--------------------------------------------------------------------//
//  print_graph_coordinates                                           //
//--------------------------------------------------------------------//
void print_graph_coordinates(PlotData *pd, int x, int y)
{
  y=y;
  const int SPACING=14;
  char tempstring2[100];
  char tempstring[256];
  int ystart = 380;
  int focus = pd->focus;
  Window win = pd->formwin;
  
                  // Print x value
  gcvt(pd->xcalib * (double)x, g.signif, tempstring2);
  sprintf(tempstring,"x=%s          ",tempstring2);
  XDrawImageString(g.display, win, g.gc, 5, ystart ,tempstring,13);
  ystart+=SPACING;
        
                 // Print y value
  gcvt(pd->ycalib * (double)pd->data[focus][x], g.signif, tempstring2);
  sprintf(tempstring,"y=%s             ",tempstring2);
  XDrawImageString(g.display, win, g.gc, 5, ystart, tempstring,16);
  ystart+=SPACING;

                 // Print area
  gcvt(pd->area[focus], g.signif, tempstring2);
  sprintf(tempstring,"Area=%s                           ",tempstring2);
  XDrawImageString(g.display, win, g.gc, 5, ystart, tempstring,26);
  ystart+=SPACING;

                 // Print scale
  gcvt(pd->scale[focus], g.signif, tempstring2);
  sprintf(tempstring,"Scale=%s            ",tempstring2);
  XDrawImageString(g.display, win, g.gc, 5, ystart, tempstring,17);
  ystart+=SPACING;

                // Focus is also printed in graphcb
  sprintf(tempstring,"Graph=%d           ",pd->focus+1);
  XDrawImageString(g.display, win, g.gc, 5, ystart, tempstring,17);
  ystart+=SPACING;
  return;              
} 


//--------------------------------------------------------------------//
// graphbuttoncb - callback for buttons at left of graph              //
//--------------------------------------------------------------------//
void graphbuttoncb(Widget w, XtP client_data, XmACB *call_data)
{  
  static double fvalue=1.0, addvalue=0;
  static int blcount = 200;
  static XYData data;    // Data for manual baseline
  call_data=call_data;   // keep compiler quiet
  int status=OK, height, usef2;
  Window win = XtWindow(w);
  clickboxinfo *c = (clickboxinfo *)client_data;
  PlotData *pd = (PlotData *)c->ptr;
  pcb_struct v;
  int button = c->button;
  int nn = pd->n;
  int focus = pd->focus;
  int ct, i, j, k, otype=pd->type;

  double biggest=0, datamax, scale=1.0;
  v.focus=focus;
  pd->reason=NORMAL;
  for(k=0;k<nn;k++) if(pd->data[focus][k] > biggest) biggest=pd->data[focus][k];
  if(pd->f2 != (void (*)(void *, int, int)) null) usef2=1; else usef2=0;

  switch(button)
  {    case 0:  save_scan(pd->title,pd->data[focus],nn,0);  // Save to disk
                break;
       case 1:  read_data(pd->title,pd->data[focus],nn);    // Read from disk
                draw_graph(pd,0,nn);
                break;
       case 2:  smooth(pd->data[focus],nn,15);    // Smooth
                draw_graph(pd,0,nn);
                if(usef2) pd->f2(pd,0,nn);   
                else for(k=0;k<nn;k++)
                {  v.x=k;
                   v.y=(int)pd->data[focus][k];
                   pd->f(v);
                }
                break;
       case 3:                                    // Auto baseline
                baseline(pd, blcount);
                draw_graph(pd,0,nn);
                if(usef2) pd->f2(pd,0,nn);   
                else for(k=0;k<nn;k++)
                {  v.x=k;
                   v.y=(int)pd->data[focus][k];
                   pd->f(v);
                }
                break;
       case 4:                                    // Manual baseline
                XtMapWidget(c->widget[0]);
                pd->type = SUBTRACT;
                manual_baseline(pd, &data);
                draw_graph(pd,0,nn);
                if(usef2) pd->f2(pd,0,nn);   
                else for(k=0;k<nn;k++)
                {  v.x=k;
                   v.y=(int)pd->data[focus][k];
                   pd->f(v);
                }
                pd->type = otype;
                break;
       case 5:  setSVGApalette(1);                // Capture image
                if(g.bitsperpixel<=8) ct=INDEXED; else ct=COLOR;
                if(newimage(0, 0, nn, pd->ysize, g.bitsperpixel, ct, 1,
                     g.want_shell, PERM, 1, g.window_border,0)==OK) 
                {    status=GOTNEW;
                     win = pd->win[focus];
                     height = pd->ysize - (pd->dims-1);
                     copyimage(0,0,nn-1,height-1,0,0,GET,z[ci].image_ximage,win);
                     setimagetitle(ci,"graph");    
                     for(j=0;j<z[ci].ysize;j++)
                     for(i=0;i<z[ci].xsize;i++)
                         if(z[ci].image[z[ci].cf][j][i]) 
                             z[ci].image[z[ci].cf][j][i]= g.maxcolor;
                     rebuild_display(ci);
                     switchto(ci);  
                     z[ci].touched=1;
                }
                break;
       case 6:  findpeaks(pd);
                break;
       case 7:  blcount = clickbox("Baseline smoothness",blcount,1,1000,null,0);
                break;
       case 8:  pd->yscale = clickbox("Y Scale",pd->yscale,0.0,10.0,null,0);
                draw_graph(pd,0,nn);
                break;
       case 9:  fvalue = clickbox("Factor to multiply by",fvalue,0.0,10.0,null,0);
                if(g.getout){ g.getout=0; break; }
                for(k=0;k<nn;k++) pd->data[focus][k] *= fvalue;
                draw_graph(pd,0,nn);
                if(usef2) pd->f2(pd,0,nn);   
                else for(k=0;k<nn;k++)
                {  v.x=k;
                   v.y=(int)pd->data[focus][k];
                   pd->f(v);
                }
                break;                                     
       case 10:                                   // Add
                addvalue = clickbox("Value to add",addvalue,-biggest,biggest,null,0);
                if(g.getout){ g.getout=0; break; }
                for(k=0;k<nn;k++) pd->data[focus][k] += addvalue;
                draw_graph(pd,0,nn);
                if(usef2) pd->f2(pd,0,nn);   
                else for(k=0;k<nn;k++)
                {  v.x=k;
                   v.y=(int)pd->data[focus][k];
                   pd->f(v);
                }   
                break;                                     
       case 11: if(pd->type == MEASURE) pd->type = CHANGE; 
                else pd->type = MEASURE;
                break;
       case 12:                                   // Bezier change
                XtMapWidget(c->widget[0]);
                pd->type = CHANGE;
                manual_baseline(pd, &data);
                draw_graph(pd,0,nn);
                if(usef2) pd->f2(pd,0,nn);   
                else for(k=0;k<nn;k++)
                {  v.x=k;
                   v.y=(int)pd->data[focus][k];
                   pd->f(v);
                }   
                pd->type = otype;
                break;
       case 13:                                   // Maximize
                pd->f3(pd,0,nn);
                if(pd->ymax != pd->ymin)
                   scale = (pd->ulimit - pd->llimit) / (pd->ymax - pd->ymin);
                else scale = 1.0;
                if(usef2) pd->f2(pd,0,nn);   
                else for(k=0;k<nn;k++)
                {  v.x = k;
                   pd->data[focus][k] = scale * (pd->data[focus][k] - pd->ymin);
                   v.y = (int)(pd->data[focus][k]);
                   pd->f(v);
                }
                draw_graph(pd,0,nn);
                break;
       case 14:                                   // Invert
                pd->f3(pd,0,nn);
                datamax=0.0;
                if(pd->llimit == pd->ulimit) datamax = biggest;
                else datamax = pd->ulimit;
                for(k=0;k<nn;k++)
                {  v.x=k;
                   pd->data[focus][k] = max(pd->llimit,min(pd->ulimit,(datamax - pd->data[focus][k])));
                   v.y = (int)(pd->data[focus][k]);
                   if(!usef2) pd->f(v);
                }   
                draw_graph(pd,0,nn);   
                break;
       case 15:                                   // Reset
                for(k=0;k<nn;k++) pd->data[pd->focus][k] =k;
                redraw_graph(pd);                 // Reset graph
                break;
       case 16:                                   // Rescale
                pd->scale[focus]=0;
                draw_graph(pd,0,nn);              // Reset graph
                break;
       default: if(g.bezier_state != NORMAL)      // Last button ("Finished")
                      bezier_curve_end(&data);    // Finished with manual baseline       
                XtUnmapWidget(w);                 // Get rid of the finish button
                break;
                 
  }
  pd->status = status;
  g.getout=0;
  return;
}  



//--------------------------------------------------------------------------//
// findbiggest                                                              //
// Return the largest element in a 1-dim array of length n                  //
//--------------------------------------------------------------------------//
double findbiggest(double* pdata, int n)
{   int k;
    double biggest=0;
    for(k=0;k<n;k++) if(pdata[k] > biggest) biggest=pdata[k]; 
    if(biggest<=0.0) biggest=1.0;
    return(biggest);
}


//--------------------------------------------------------------------------//
//  baseline                                                                //
//  Automatically subtracts baseline from data set of count points          //
//--------------------------------------------------------------------------//
void baseline(PlotData *pd, int iterations)
{  
   int j, k, count;
   count       = pd->n;
   int ysize   = pd->ysize;
   int focus   = pd->focus;
   Window win  = pd->win[focus];
   double scale = pd->scale[focus];

   printstatus(CALCULATING);
   double *tdata;
   double *pdata = pd->data[pd->focus];
   tdata = new double[count];
   for(k=0;k<count;k++) tdata[k] = pdata[k];
   for(j=0;j<iterations;j++)
   {  smooth(tdata,count,21);
      for(k=0;k<count;k++)
      {   if(tdata[k]>pdata[k]) tdata[k]=pdata[k];
          if(tdata[k]<0) tdata[k]=0;
      }    
   }
   smooth(tdata,count,15);
   printstatus(NORMAL);
   for(k=0;k<count;k++)
   {   setpixel(k, (int)(  ysize-tdata[k]*scale), g.maxcolor,SET,win);
       setpixel(k, (int)(1+ysize-tdata[k]*scale), g.maxcolor,SET,win);
       setpixel(k, (int)(2+ysize-tdata[k]*scale), g.maxcolor,SET,win);    
   }
   sleep(1);
   for(k=1;k<count;k++)
   {   pdata[k] -= tdata[k];
       if(pdata[k]<0) pdata[k]=0;
   }
   pdata[0]=pdata[1];
   delete[] tdata;
}


//--------------------------------------------------------------------------//
//  manual_baseline                                                         //
//--------------------------------------------------------------------------//
void manual_baseline(PlotData *pd, XYData *data)
{  
   const int POINTS = 100;   // No. of control points user can select
   int k;    
   double dy=0.0;   
   Widget widget;
   int focus = pd->focus;
   data->x = new int[POINTS];
   data->y = new int[POINTS];
   data->f = NULL;
   data->g = NULL;
   data->n = 0;
   data->nmax      = POINTS;
   data->width     = 0;
   data->type      = 0;
   data->duration  = TEMP;
   data->wantpause = 0;
   data->order     = 0;
   data->win       = pd->win[focus]; 
   uint bezier_mask = 
                   KeyPressMask      |
                   KeyReleaseMask    |
                   ButtonPressMask   |
                   ButtonReleaseMask |
                   Button1MotionMask |      // Click & Drag
                   PointerMotionMask;       // Mouse movement  

  //// Replace graphcb and graphmousecb callbacks for graph windows 
  //// with beziercb

  for(k=1;k<pd->n;k++) g.highest[k+g.main_xpos] = pd->ysize;

  for(k=0;k<pd->dims;k++)
  {     widget = XtWindowToWidget(g.display, pd->win[k]);
        XtRemoveEventHandler(widget, g.mouse_mask,False,(XtEH)graphmousecb,(XtP)pd);
        XtAddEventHandler(widget,bezier_mask,False,(XtEH)beziercb,(XtP)data); 
        XtRemoveCallback(widget, XmNinputCallback,(XtCBP)graphcb, (XtP)pd);
        XtAddCallback(widget, XmNinputCallback,(XtCBP)beziercb, (XtP)data);
  }

  //// Wait around for something to happen
  
  XSetForeground(g.display, g.image_gc, g.maxcolor);
  g.bezier_state = CURVE;
  while(g.bezier_state==CURVE)
        XtAppProcessEvent(XtWidgetToApplicationContext(g.drawing_area),XtIMAll);

  //// Restore original callbacks

  for(k=0;k<pd->dims;k++)
  {     widget = XtWindowToWidget(g.display, pd->win[k]);
        XtRemoveEventHandler(widget,bezier_mask,False,(XtEH)beziercb,(XtP)data);
        XtAddEventHandler(widget, g.mouse_mask,False,(XtEH)graphmousecb,(XtP)pd);
        XtRemoveCallback(widget, XmNinputCallback, (XtCBP)beziercb, (XtP)data);
        XtAddCallback(widget, XmNinputCallback, (XtCBP)graphcb, (XtP)pd);
  }
  g.bezier_state = NORMAL;

  //// Subtract the baseline

  for(k=1;k<pd->n;k++)
  {   if(pd->scale[focus] != 0.0) 
          dy = (double)(pd->ysize - g.highest[k+g.main_xpos]) / pd->scale[focus];
      if(pd->type==SUBTRACT)
          pd->data[focus][k] -= dy;
      else if(pd->type==CHANGE)
          pd->data[focus][k] = dy;
      if(pd->data[focus][k] < 0.0) pd->data[focus][k] = 0.0;
  }
  
  draw_graph(pd,0,data->n);
  delete[] data->y;
  delete[] data->x;
}



//--------------------------------------------------------------------------//
// findpeaks                                                                //
// Calculate & display peaks                                                //
//--------------------------------------------------------------------------//
void findpeaks(PlotData *pd)
{
   if(memorylessthan(16384)){ message(g.nomemory,ERROR); return; }

   int count, end, focus, j, k, osign=0, peak=0, sign=0, size;
   double *tdata, *dy, *d2y, *area, *yy;
   double dx=0.0;
   int *xafter, *xx;
   char tempstring[100]; 

   focus = pd->focus;
   count = pd->n;   
   tdata = new double[count+10];              // Temporary smoothed data
   if(tdata==NULL){ message(g.nomemory,ERROR); return; }
   dy  = new double[count+10];                // 1st deriv               
   d2y = new double[count+10];                // 2nd deriv
   xx  = new int[count+10];
   yy  = new double[count+10];

   xx[0]=0;
   yy[0]=pd->data[focus][0];
   
   for(k=0;k<count;k++) tdata[k] = pd->data[focus][k];
   smooth(tdata, count, 3);

   ////  Count the number of peaks.
                                             // Take deriv.
   for(k=1;k<count;k++) dy[k] = tdata[k] - tdata[k-1];
   for(k=1;k<count;k++) d2y[k] = dy[k] - dy[k-1];
   smooth(dy, count, 5);
   smooth(d2y, count, 5);
   for(k=2;k<count-2;k++)
   {                                         // If a maximum
       if(sgn(dy[k+1])== 1) sign=1;  
       if(sgn(dy[k+1])==-1) sign=-1;  
       if( sign!=osign && sgn(d2y[k])==-1 && k>3 )
       {   peak++;
           if(peak>=200){ message("Too many peaks",ERROR); break; }
           xx[peak] = k;
           yy[peak] = pd->data[focus][k];    // Use original data for y at max.
           k++;
       }    
       osign = sign;
   }
               
   ////  Find start, end, maximum, and area of each peak
 
   area = new double[peak+10];               // Peak areas
   xafter  = new int[peak+10];               // x position of end of peak

   xafter[0] = 0;
   for(k=1;k<=peak;k++)
   {   area[k]=0;
       osign=1;

       ////  Measure left half of peak, stop at end of previous peak

       for(j=xx[k];j>xx[k-1];j--)
       {   sign = sgn(dy[j]);  
           dx = fabs(pd->xdata[j+1] - pd->xdata[j]);
           area[k] += dx * tdata[j];
           if(xx[k]==0)break;    
           if(sign!=osign && j<xx[k]-3)break;
       }                                     // At this point start of peak=j    
       osign=-1;
       if(k<peak) end=xx[k+1]; else end=count;

       ////  Measure right half of peak, stop at start of next peak

       for(j=xx[k]+1;j<end;j++)
       {   sign = sgn(dy[j]); 
           dx = fabs(pd->xdata[j+1] - pd->xdata[j]);
           area[k] += dx * tdata[j];
           xafter[k] = j;
           if(xx[k]==0) break;    
           if(sign!=osign && j>xx[k]+3)break;
       }                                    // At this point end of peak=xafter[k]
   }   

   ////  Put results in list box

   char listtitle[100]="Peaks";                  
   char **info;                                    
   info = new char*[peak+10];
   info[0] = new char[100];                       
   strcpy(info[0],"Peak#  x      y");
   for(k=0;k<13;k++) strcat(info[0]," ");
   strcat(info[0], "Area");
   if(peak)
   { for(k=1;k<=peak;k++)                                
     {   info[k] = new char[100];                       
         itoa(k,tempstring,10);
         strcpy(info[k],tempstring);
         for(j=strlen(info[k]);j<6;j++)strcat(info[k]," ");
 
         itoa(xx[k],tempstring,10);
         strcat(info[k],tempstring);
         for(j=strlen(info[k]);j<14;j++)strcat(info[k]," ");
        
         gcvt(yy[k], g.signif, tempstring);
         strcat(info[k],tempstring);       
         for(j=strlen(info[k]); j<22+g.signif; j++) strcat(info[k]," ");

         gcvt(area[k], g.signif, tempstring);
         strcat(info[k], tempstring);
     }                                                  
   }
   if(peak>0)                                   // Don't sort list
   {  size=max(5,min(peak+1,30)); 
      char edittitle[50] = "Edit item";
      
      listinfo l;
      int crud=0;
      l.title      = listtitle;
      l.info       = info;
      l.size       = peak+1;
      l.itemstoshow= size;
      l.firstitem  = 0;
      l.wantsort   = 0;
      l.wantsave   = 1;
      l.helptopic  = 0;
      l.selection  = &crud;
      l.allowedit  = 1;
      l.edittitle  = edittitle;
      l.editrow    = 0;
      l.editcol    = 0;
      l.maxstringsize = 99;
      l.width      = 0;
      l.transient   = 0;
      l.maxstringsize = 99;
      l.wantfunction = 0;
      l.autoupdate   = 0;
      list(&l);
   }
   else
     message("No peaks");
   
   //----------------------------------------------------------------//

   for(k=0;k<=peak;k++) delete[] info[k];              
   delete[] info;
   delete[] area; 
   delete[] xafter;
   delete[] tdata;
   delete[] dy;
   delete[] d2y;
   delete[] yy;
   delete[] xx;
   g.getout=0;
}


//--------------------------------------------------------------------------//
//  smooth                                                                  //
//  Smooth dataset with 'width'-point smoothing                             //
//--------------------------------------------------------------------------//
void smooth(int* pdata,int count,int width)
{  int k;
   double* ddd;
   ddd = new double[count+10]; 
   if(ddd==NULL){ message(g.nomemory,ERROR); return; }
   for(k=0;k<count;k++) ddd[k] = (double)pdata[k];
   smooth(ddd,count,width);
   for(k=0;k<count;k++) pdata[k] = (int)ddd[k];
   delete[] ddd;
}
void smooth(double* pdata,int count,int width)
{
    int j,k,tot,w;
    static int gau3[ 4] = {  1, 2,  1                     };
    static int gau5[ 6] = {  1, 4,  6,  4,   1            };
    static int gau7[ 8] = {  1, 6, 15, 20,  15,   6,   1  };
    static int gau9[10] = {  1, 8, 28, 56,  70,  56,  28,   8,   1         };
    static int gau11[12]= {  1,10, 45,120, 210, 252, 210, 120,  45,  10, 1 };
    static int gau15[16]= {  1,14, 91,364,1001,2002,3003,3432,3003,2002,1001,
                           364,91, 14,  1 }; 
    static int gau19[20]= {  1,18,153,816,3060,8568,18564,31824,43758,48620,
                           43758,31824,18564,8568,3060,816,153,18,1 };
    static int gau21[22]= {  1,20,190, 1140, 4845, 15504,  38760, 77520, 
                            125970, 167960, 184756, 167960, 125970, 77520,
                             38760,  15504,   4845,   1140, 190, 20, 1 }; 
    double testval;
    double *tdata;
    tdata = new double[count+10];
    if(tdata==NULL){ message(g.nomemory,ERROR); return; }
    for(k=0;k<count;k++) tdata[k]=pdata[k];
    switch(width)
    {    case 3: { w=1; tot=4;  break;  }
         case 5: { w=2; tot=16; break;  }
         case 7: { w=3; tot=64; break;  }
         case 9:  { w=4; tot=256;     break; }
         case 11: { w=5; tot=1024;    break; }
         case 15: { w=7; tot=16384;   break; }
         case 19: { w=9; tot=262144;  break; }
         case 21: { w=10;tot=1048576; break; }
         default: { message("Error",ERROR); delete[] tdata; return; }
    }
    for(k=0;k<count;k++)
    {  tdata[k]=0;
       for(j=-w;j<=w;j++)
       {  if(k+j<0) testval = pdata[0];
          else if(k+j>=count) testval = pdata[count-1];
          else testval = pdata[k+j];
          switch(width)
          {    case 3: { tdata[k] += testval*gau3[w+j]; break; }
               case 5: { tdata[k] += testval*gau5[w+j]; break; }
               case 7: { tdata[k] += testval*gau7[w+j]; break; }
               case 9: { tdata[k] += testval*gau9[w+j]; break; }
               case 11:{ tdata[k] += testval*gau11[w+j]; break; }
               case 15:{ tdata[k] += testval*gau15[w+j]; break; }
               case 19:{ tdata[k] += testval*gau19[w+j]; break; }
               case 21:{ tdata[k] += testval*gau21[w+j]; break; }
          }
       }
       if(tot!=0) tdata[k] /= tot;
    }      
    for(k=0;k<count;k++) pdata[k]=tdata[k];
    delete[] tdata;
    return;
}



//--------------------------------------------------------------------------//
//  plot3d                                                                  //
//--------------------------------------------------------------------------//
void plot3d(int ino) 
{
   const int LLABEL=12;
   const int RLABEL=87;
   Arg args[100];
   int k, n, wc=0, rw=200;
   if(ino<=0) return;
   char tempstring[1024];
   XmString xms;
   Widget *slider, *w, form, *labelwidget, *buttonwidget;
   char title[100];
   int helptopic=0;
   double a,b,c, xscale,yscale,zscale,xcenter,ycenter,zcenter,xtrans,ytrans,ztrans;
   int colormode = 0;
   int granularity = 4;
   a=330;
   b=10;
   c=20;
   xscale=yscale=zscale=1;
   xcenter=ycenter=zcenter=0;
   xtrans=ytrans=ztrans=0;

   void **ptr;
   int *answer, *buttonstate;
   answer = new int[20];
   buttonstate = new int[20];
   if(z[ino].colortype==GRAPH)
   {  
       if(z[ino].gparam[0]<=0){ message("Original image is no longer present\nCan't redraw graph!"); return; }
       switchto(ino);           // Set ci to The Graph ****ino and ci change here***
       ino = z[ci].gparam[0];   // Set ino to original image being graphed
                                // Restore graph params 
       for(k=0;k<20;k++) answer[k] = z[ci].gparam[k];
       for(k=0;k<20;k++) buttonstate[k] = z[ci].gbutton[k];
   }else 
   {   if(newimage(200, 200, cint(1.5*z[ino].xsize), cint(1.5*z[ino].ysize), g.bitsperpixel, 
           GRAPH, 1, 1, PERM, 0, 1, 0)!=OK) 
       {   message(g.nomemory); 
           delete[] answer;
           delete[] buttonstate;
           return;
       }
       answer[0] = ino;        // Original image to plot
       answer[1] = cint(a);    // Angle around x axis
       answer[2] = cint(b);    // Angle around y axis
       answer[3] = cint(c);    // Angle around z axis
       answer[4] = cint(100*xscale);
       answer[5] = cint(100*yscale);
       answer[6] = cint(100*zscale);
       answer[7] = cint(xcenter);
       answer[8] = cint(ycenter);
       answer[9] = cint(zcenter);
       answer[10] = cint(xtrans);
       answer[11] = cint(ytrans);
       answer[12] = cint(ztrans);
       answer[13] = granularity;
       answer[14] = colormode;
       answer[15] = ci;        // Image no. of the graph
       answer[16] = 0;
       answer[17] = 0;
       answer[18] = 0;
       answer[19] = 0;
   }

   w = new Widget[100];
   ptr = new void*[5];
   slider = new Widget[SLIDERS];
   labelwidget = new Widget[SLIDERS];
   buttonwidget = new Widget[BUTTONS3D];
   ptr[0] = answer;
   ptr[1] = (void*)slider;
   ptr[2] = (void*)labelwidget;
   ptr[3] = (void*)buttonwidget;
   ptr[4] = (void*)buttonstate;

   sprintf(title, "Image # %d 3D Graph controls",ino);
   
   //------- Create a form dialog shell widget -----------------------//

   n=0;
   XtSetArg(args[n], XmNwidth, 400); n++;
   XtSetArg(args[n], XmNheight, 400); n++;
   XtSetArg(args[n], XmNtransient, False); n++;
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNautoUnmanage, False); n++;
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
   XtSetArg(args[n], XmNdialogTitle, xms = XmStringCreateSimple(title));n++;
   ////  Stop Motif from trying to grab another color if none are available.
   if(g.want_colormaps)  
   {    XtSetArg(args[n], XmNbackground, g.main_bcolor); n++;
        XtSetArg(args[n], XmNforeground, g.main_fcolor); n++;
   }
   form = w[wc++] = XmCreateFormDialog(g.main_widget, "Plot3DForm", args, n);
   XmStringFree(xms);
       
   //--------Title for dialog box-------------------------------------//

   w[wc++] = addlabel(form, title, CENTER, 20, 2, 80, 10); 

   //--------Sliders--------------------------------------------------//

   slider[0] = w[wc++] = add_slider(form, "Img#", rw, LLABEL, RLABEL, 10, 15, 1, max(1,g.image_count-1), answer[0], 1);  
   sprintf(tempstring, "Img#");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 10, 99, 15); 
   sprintf(tempstring, "%d", answer[0]);
   w[wc++] = labelwidget[0] = addlabel(form, tempstring, LEFT, 1, 10, LLABEL, 15); 
                                                                                                       
   slider[1] = w[wc++] = add_slider(form, "Xrot", rw, LLABEL, RLABEL, 15, 20, 0, 360, answer[1], 1);  
   sprintf(tempstring, "x%c",XK_degree);
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 15, 99, 20); 
   sprintf(tempstring, "%d", answer[1]);
   w[wc++] = labelwidget[1] = addlabel(form, tempstring, LEFT, 1, 15, LLABEL, 20); 
                                                                                
   slider[2] = w[wc++] = add_slider(form, "Yrot", rw, LLABEL, RLABEL, 20, 25, 0, 360, answer[2], 1);  
   sprintf(tempstring, "y%c",XK_degree);
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 20, 99, 25); 
   sprintf(tempstring, "%d", answer[2]/100);
   w[wc++] = labelwidget[2] = addlabel(form, tempstring, LEFT, 1, 20, LLABEL, 25); 
                                                                                
   slider[3] = w[wc++] = add_slider(form, "Zrot", rw, LLABEL, RLABEL, 25, 30, 0, 360, answer[3], 1);  
   sprintf(tempstring, "z%c",XK_degree);
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 25, 99, 30); 
   sprintf(tempstring, "%d", answer[3]/100);
   w[wc++] = labelwidget[3] = addlabel(form, tempstring, LEFT, 1, 25, LLABEL, 30); 
                                                                                
   slider[4] = w[wc++] = add_slider(form, "X scale", rw, LLABEL, RLABEL, 30, 35,  1, 1000, answer[4], 1);  
   sprintf(tempstring, "x scale");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 30, 99, 35); 
   sprintf(tempstring, "%3.2g", (double)answer[4]/100);
   w[wc++] = labelwidget[4] = addlabel(form, tempstring, LEFT, 1, 30, LLABEL, 35); 
                                                                                
   slider[5] = w[wc++] = add_slider(form, "Y scale", rw, LLABEL, RLABEL, 35, 40,  1, 1000, answer[5], 1);  
   sprintf(tempstring, "y scale");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 35, 99, 40); 
   sprintf(tempstring, "%3.2g", (double)answer[5]/100);
   w[wc++] = labelwidget[5] = addlabel(form, tempstring, LEFT, 1, 35, LLABEL, 40); 
                                                                                
   slider[6] = w[wc++] = add_slider(form, "Z scale", rw, LLABEL, RLABEL, 40, 45,  1, 1000, answer[6], 1);  
   sprintf(tempstring, "z scale");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 40, 99, 45); 
   sprintf(tempstring, "%3.2g", (double)answer[6]/100);
   w[wc++] = labelwidget[6] = addlabel(form, tempstring, LEFT, 1, 40, LLABEL, 45); 
                                                                                
   slider[7] = w[wc++] = add_slider(form, "X ctr", rw, LLABEL, RLABEL, 45, 50,  -1000, 1000, answer[7], 1);  
   sprintf(tempstring, "x ctr");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 45, 99, 50); 
   sprintf(tempstring, "%d", answer[7]/100);
   w[wc++] = labelwidget[7] = addlabel(form, tempstring, LEFT, 1, 45, LLABEL, 50); 
                                                                                
   slider[8] = w[wc++] = add_slider(form, "Y ctr", rw, LLABEL, RLABEL, 50, 55,  -1000, 1000, answer[8], 1);  
   sprintf(tempstring, "y ctr");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 50, 99, 55); 
   sprintf(tempstring, "%d", answer[8]/100);
   w[wc++] = labelwidget[8] = addlabel(form, tempstring, LEFT, 1, 50, LLABEL, 55); 
                                                                                
   slider[9] = w[wc++] = add_slider(form, "Z ctr", rw, LLABEL, RLABEL, 55, 60,  -1000, 1000, answer[9], 1);  
   sprintf(tempstring, "z ctr");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 55, 99, 60); 
   sprintf(tempstring, "%d", answer[9]/100);
   w[wc++] = labelwidget[9] = addlabel(form, tempstring, LEFT, 1, 55, LLABEL, 60); 
                                                                                
   slider[10] = w[wc++] = add_slider(form, "X trans", rw, LLABEL, RLABEL, 60, 65,  -1000, 1000, answer[10], 1);  
   sprintf(tempstring, "x trans");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 60, 99, 65); 
   sprintf(tempstring, "%d", answer[10]/100);
   w[wc++] = labelwidget[10] = addlabel(form, tempstring, LEFT, 1, 60, LLABEL, 65); 
                                                                                
   slider[11] = w[wc++] = add_slider(form, "Y trans", rw, LLABEL, RLABEL, 65, 70,  -1000, 1000, answer[11], 1);  
   sprintf(tempstring, "y trans");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 65, 99, 70); 
   sprintf(tempstring, "%d", answer[11]/100);
   w[wc++] = labelwidget[11] = addlabel(form, tempstring, LEFT, 1, 65, LLABEL, 70); 
                                                                                
   slider[12] = w[wc++] = add_slider(form, "Z trans", rw, LLABEL, RLABEL, 70, 75, -1000, 1000, answer[12], 1);  
   sprintf(tempstring, "z trans");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 70, 99, 75); 
   sprintf(tempstring, "%d", answer[12]/100);
   w[wc++] = labelwidget[12] = addlabel(form, tempstring, LEFT, 1, 70, LLABEL, 75); 
                                                                                
   slider[13] = w[wc++] = add_slider(form, "Gran", rw, LLABEL, RLABEL, 75, 80, 1, 100, answer[13], 1);  
   sprintf(tempstring, "gran");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 75, 99, 80); 
   sprintf(tempstring, "%d", answer[13]);
   w[wc++] = labelwidget[13] = addlabel(form, tempstring, LEFT, 1, 75, LLABEL, 80); 
                                                                                
   slider[14] = w[wc++] = add_slider(form, "Color", rw, LLABEL, RLABEL, 80, 85, 0, 2, answer[14], 1);  
   sprintf(tempstring, "color");
   w[wc++] = addlabel(form, tempstring, LEFT, RLABEL, 80, 99, 85); 
   sprintf(tempstring, "%d", answer[14]);
   w[wc++] = labelwidget[14] = addlabel(form, tempstring, LEFT, 1, 80, LLABEL, 85); 
   //// Change SLIDERS at top if adding a new scale
                                                                                
   //// Toggle buttons

   w[wc++] = buttonwidget[0] = add_togglebutton(form, "Invert Z", 12, 86, plot3dbuttoncb, ptr);
   w[wc++] = buttonwidget[1] = add_togglebutton(form, "Rotate X", 27, 86, plot3dbuttoncb, ptr);
   w[wc++] = buttonwidget[2] = add_togglebutton(form, "Rotate Y", 42, 86, plot3dbuttoncb, ptr);
   w[wc++] = buttonwidget[3] = add_togglebutton(form, "Rotate Z", 57, 86, plot3dbuttoncb, ptr);
   w[wc++] = buttonwidget[4] = add_togglebutton(form, "Surface ", 72, 86, plot3dbuttoncb, ptr);

   //// Change BUTTONS3D at top if adding a new button

   for(k=0;k<BUTTONS3D;k++)
   {   buttonstate[k] = z[ci].gbutton[k];
       XmToggleButtonSetState(buttonwidget[k], buttonstate[k], False); 
   }

   //--------Ok, Cancel, Help buttons & their callbacks---------------//

   addstandardbuttons(form, &w[wc], &helptopic);
   wc += 3;    // addstandardbuttons adds 3 widgets

   ////  Callbacks for x,y,z rotation sliders
   for(k=0;k<SLIDERS;k++)
   {    XtAddCallback(slider[k], XmNvalueChangedCallback, (XtCBP)plot3dslidercb, ptr);
        //// Comment this out entirely for slow computers
        //// Remove the 'if' for really fast computers.
        if(z[ci].xsize*z[ci].ysize<160000)
            XtAddCallback(slider[k], XmNdragCallback, (XtCBP)plot3dslidercb, ptr);
        //// Not this
        XtManageChild(slider[k]);        
   }
   XtManageChild(form);
   for(k=0;k<20;k++) z[ci].gparam[k] = answer[k];
   for(k=0;k<20;k++) z[ci].gbutton[k] = buttonstate[k];
   plot3dslidercb(slider[0], (XtP)ptr, 0);  // initial graph

   event_loop(form);

   //// Params & button states are contained in graph not original image.
   for(k=0;k<20;k++) z[ci].gparam[k] = answer[k];
   for(k=0;k<20;k++) z[ci].gbutton[k] = buttonstate[k];
   XtUnmanageChild(form);
   if(g.autoundo) backupimage(ci,0);
   for(k=0;k<wc;k++) XtDestroyWidget(w[k]);
   delete[] w;
   delete[] answer;
   delete[] slider;
   delete[] buttonstate;
   delete[] buttonwidget;
   delete[] labelwidget;
   delete[] ptr;
   return;
}


//--------------------------------------------------------------------------//
// plot3dslidercb                                                           //
//--------------------------------------------------------------------------//
void plot3dslidercb(Widget w, XtP client_data, XmACB *call_data)
{
   call_data=call_data;  // Keep compiler quiet
   XmString xms;
   char tempstring[64];
   int k, value = 0, row=0;
   double showvalue;
   void **ptr          = (void**)client_data;
   int *parameter      = (int*)ptr[0];
   Widget *slider      = (Widget*)ptr[1];
   Widget *labelwidget = (Widget*)ptr[2];
   for(k=0;k<SLIDERS;k++) if(w == slider[k]){ row = k;  break;}
   XmScaleGetValue(w, &value);
   parameter[row] = value;
   showvalue = (double)value;
   if(between(row,4,6)) showvalue/=100;
   sprintf(tempstring, "%g", showvalue);
   XtVaSetValues(labelwidget[row], XmNlabelString, xms=XmStringCreateSimple(tempstring), NULL);
   XmStringFree(xms);
   XFlush(g.display);
   plot3dgraph(ptr);
}


//--------------------------------------------------------------------------//
//  setslider                                                               //
//--------------------------------------------------------------------------//
void setslider(void **ptr, int sliderno, int value)
{
   XmString xms;
   char tempstring[64];
   Widget *slider      = (Widget*)ptr[1];
   Widget *labelwidget = (Widget*)ptr[2];
   XmScaleSetValue(slider[sliderno], value); 
   sprintf(tempstring, "%d", value);
   XtVaSetValues(labelwidget[sliderno], XmNlabelString, xms=XmStringCreateSimple(tempstring), NULL);
   XmStringFree(xms);
}


//--------------------------------------------------------------------------//
// plot3buttoncb                                                            //
//--------------------------------------------------------------------------//
void plot3dbuttoncb(Widget w, XtP client_data, XmACB *call_data)
{
   call_data=call_data;  // Keep compiler quiet
   if(z[ci].colortype != GRAPH) return;
   int ino, k, but=0, state;
   void **ptr       = (void**)client_data;
   int *parameter   = (int*)ptr[0];
   Widget *button   = (Widget*)ptr[3];
   int *buttonstate = (int*)ptr[4];
   for(k=0; k<BUTTONS3D; k++) { if(w == button[k]) { but = k; break; }}
   state = XmToggleButtonGetState(w); 
   buttonstate[but] = state;
   ino = ci;
   do{
      check_event();
      if(buttonstate[1]){ parameter[1] += 2; setslider(ptr, 1, parameter[1]);}
      if(buttonstate[2]){ parameter[2] += 2; setslider(ptr, 2, parameter[2]);}
      if(buttonstate[3]){ parameter[3] += 2; setslider(ptr, 3, parameter[3]);}
      if(parameter[1]>=360) parameter[1]-=360;
      if(parameter[2]>=360) parameter[2]-=360;
      if(parameter[3]>=360) parameter[3]-=360;
      plot3dgraph(ptr); 
      //// In case they unload the graph or image while it's rotating
      if(z[ino].gparam[0] == 0)
      {   message("Glrrk!!!!\nHey! Where did that image go?",ERROR); 
          for(k=0;k<20;k++) z[ino].gbutton[k]=0;
          z[ino].gparam[0] = -1;
      }
      if(g.getout || z[ino].gparam[0] <= 0 || ino>g.image_count-1)
      {  buttonstate[1] = buttonstate[2] = buttonstate[3] = 0; g.getout=0; break; }
   }while (buttonstate[1] || buttonstate[2] || buttonstate[3]);

   XFlush(g.display);
}


//--------------------------------------------------------------------------//
//   plot3dgraph                                                            //
//--------------------------------------------------------------------------//
void plot3dgraph(void **ptr)
{
   const int ZGRAY=1;
   const int ZORIG=2;
   int colormode = NONE, ino, imno, hit=0, i, i2, i3, i4, imax, xmax, 
       j, j2, k, value=0, ii, jj, granularity, ovalue, oxx=0, oyy=0, v, 
       xtemp=0, ytemp=0, wantcolor, want_surface=0, xcorner[5], ycorner[5];
   int jj1, jj2, jj3, ogimode;
   int color = (int)g.maxcolor; 

   int *parameter      = (int*)ptr[0];
   int *buttonstate    = (int*)ptr[4];

   ino = parameter[15];    // graph image no
   imno  = parameter[0];   // image no of image being plotted
   if(z[ino].colortype != GRAPH || ino>g.image_count-1) return;
   if(z[ino].gparam[0] == 0)
   {   message("Glrrk!!!!\nHey! Where did that image go?",ERROR); 
       for(k=0;k<20;k++) z[ino].gbutton[k]=0;
       z[ino].gparam[0] = -1;
   }
   if(z[ino].gparam[0] == -1) return;

   int bpp = z[ino].bpp;
   int ibpp = z[imno].bpp;
   imax = max(2*z[imno].xsize, 2*z[imno].ysize);
   int xlastline[imax+1];
   int ylastline[imax+1];
   int prevx[imax+1];
   int prevy[imax+1];

   int xcenter, ycenter, zcenter, xtrans, ytrans, ztrans, xhalf, yhalf;
   int xgraphcenter = z[ino].xpos + z[ino].xsize/2;
   int ygraphcenter = z[ino].ypos + z[ino].ysize/2;
   int granoff,total,count,rtot,gtot,btot,rr,gg,bb;
   double mat[3][3];
   double a,b,c;                  // x,y,x angle
   double xx,yy,zz;               // coordinates
   double xx1,yy1;                // new coordinates
   double xscale, yscale, zscale =  0.5;
   double zhalf = 127;            // All z values are converted to 8 bits
   double yymat01, yymat11;

   lineinfo li;

   li.ino=ino;    // constrain drawing to ino
   li.type=0;
   li.color=0;
   li.width=1;
   li.wantgrad=0;
   li.grad=0;
   li.skew=0;
   li.perp_ends=0;
   li.wantarrow=0;
   li.arrow_width=0;
   li.arrow_inner_length=0;
   li.arrow_outer_length=0;
   li.window=z[ino].win;
   li.wantprint=0;
  
   a = (double)parameter[1];
   b = (double)parameter[2];
   c = (double)parameter[3];
   xscale  = (double)parameter[4] / 100;
   yscale  = (double)parameter[5] / 100;
   zscale  = (double)parameter[6] / 100;
   xcenter = parameter[7];
   ycenter = parameter[8];
   zcenter = parameter[9];
   xtrans  = parameter[10];
   ytrans  = parameter[11];
   ztrans  = parameter[12];
   granularity = parameter[13];
   colormode   = parameter[14];

   xhalf = z[ino].xsize/2 + xcenter;
   yhalf = z[ino].ysize/2 + ycenter;
   xmax = z[imno].xsize * g.off[ibpp];
   wantcolor = (z[imno].colortype != GRAY);

   hit = 0;
   while(a<0.0) a+=360.0;
   while(b<0.0) b+=360.0;
   while(c<0.0) c+=360.0;
   while(a>=360) a-=360.0;
   while(b>=360) b-=360.0;
   while(c>=360) c-=360.0;

   for(k=0;k<z[ino].ysize;k++) memset(z[ino].image[0][k], 0, z[ino].xsize*g.off[z[ino].bpp]);

   granoff = granularity * g.off[ibpp];
   mat[0][0] = cosine(b)*cosine(c);
   mat[0][1] = sine(a)*sine(b)*cosine(c) - cosine(a)*sine(c);
   mat[0][2] = cosine(a)*sine(b)*cosine(c) + sine(a)*sine(c);

   mat[1][0] = cosine(b)*sine(c);
   mat[1][1] = sine(a)*sine(b)*sine(c) + cosine(a)*cosine(c);
   mat[1][2] = cosine(a)*sine(b)*sine(c) - sine(a)*cosine(c);             

   mat[2][0] = -sine(a);
   mat[2][1] = sine(a)*cosine(b);
   mat[2][2] = cosine(a)*cosine(b);              

   jj1 = 0;
   jj2 = z[imno].ysize;
   jj3 = granularity;
   ogimode = g.imode;
   g.imode = BUFFER;  //// This makes surface plotting much faster

   for(j=jj1; j<jj2; j+=jj3)
   { 
     jj = j - yhalf;
     if(yscale !=1.0) yy = (double) jj * yscale; else yy = (double)jj;
     yymat11 = yy*mat[1][1]; 
     yymat01 = yy*mat[0][1];
     for(i=0,i2=0; i<z[imno].xsize; i+=granularity, i2+=granoff)
     {  
       count = total = rtot = gtot = btot = 0;
       for(j2=j; j2<j+granularity; j2++)
       for(i4=i,i3=i2; i4<i+granularity; i4++,i3+=g.off[ibpp])
       {   
            if(j2 >= z[imno].ysize) continue;
            if(i3 >= xmax) continue;
            v = pixel_equivalent(imno, i4, i3, j);
            if(!wantcolor)
                total += v;
            else 
            {   valuetoRGB(v,rr,gg,bb,ibpp);
                rtot += rr;
                gtot += gg;
                btot += bb;
            }
            count++;
       }
       if(count) 
       {   if(!wantcolor)
                value = total / count;
           else
           {    rr = rtot / count;
                gg = gtot / count;
                bb = btot / count;
                value = RGBvalue(rr,gg,bb,ibpp);
           }
       }else value = 0;
       ovalue = value;

       //// Make it 8 bits for scaling purposes
       if(bpp!=8) value = convertpixel(value, ibpp, 8, wantcolor);

       ii = i - xhalf;
       if(xscale !=1.0) xx = (double) ii * xscale; else xx = (double)ii;
       if(buttonstate[0])  // invert z
           zz = (double) (value+ztrans) * zscale - zhalf + zcenter;
       else
           zz = (double) (-value+ztrans) * zscale + zhalf + zcenter; 
       if(buttonstate[4])  // surface
           want_surface=1;

       xx1 = xx*mat[0][0] + yymat01 + zz*mat[0][2];
       yy1 = xx*mat[1][0] + yymat11 + zz*mat[1][2];
       //// zz1 = xx*mat[2][0] + yy*mat[2][1] + zz*mat[2][2];
       xx = xx1;
       yy = yy1;
       ////  zz = zz1 + ztrans;

       switch(colormode)
       {   case ZGRAY:
                color = cint(max(30, min(255, zhalf - zz)));
                color = RGBvalue(color, color, color, bpp);
                break;
           case ZORIG:
                color = convertpixel(ovalue, ibpp, bpp, 1); 
                break;
           case NONE:
                break;
       }

       xtemp = cint(xx) + xgraphcenter + xtrans;
       ytemp = cint(yy) + ygraphcenter + ytrans;

       if(want_surface)
       {  if(i>0 && j>0)
          {   xcorner[1] = prevx[i-granularity];
	      ycorner[1] = prevy[i-granularity];
	      xcorner[2] = prevx[i];
	      ycorner[2] = prevy[i];
	      xcorner[3] = oxx;
	      ycorner[3] = oyy;
	      xcorner[4] = xtemp;
	      ycorner[4] = ytemp;
              fill_polygon(xcorner, ycorner, color);
          }
       }else
       {  if(j>0 && hit && between(i,0,imax)) 
              line(xlastline[i], ylastline[i], xtemp, ytemp, color, BUFFER, &li);
          if(i>0 && hit) 
              line(oxx, oyy, xtemp, ytemp, color, BUFFER, &li);
       }

       xlastline[i] = xtemp;
       ylastline[i] = ytemp;
       oxx = xtemp;
       oyy = ytemp;
       hit = 1;
     }
     if(want_surface)
     {  for(i=0; i<z[imno].xsize; i+=granularity)
        {  prevx[i] = xlastline[i];
           prevy[i] = ylastline[i];
        } 
     }
   }
   g.imode = ogimode;
   rebuild_display(ino);  
   redraw(ino);
}


//--------------------------------------------------------------------------//
// sine  - x is in degrees                                                  //
//--------------------------------------------------------------------------//
inline double sine(double x)
{ 
    static double s[360], d[360];
    static int hit=0;
    int k;
    int intx;
    double answer,rem;
    if(!hit)
    {  hit=1;
       for(k=0;k<360;k++) s[k] = sin(k*0.01745329252);
       d[0] = 0;
       for(k=1;k<360;k++) d[k] = s[k] - s[k-1];
    }
    intx = int(x);
    rem = x - (double)intx;
    answer = s[intx] + rem*d[intx];
    return(answer);
}


//--------------------------------------------------------------------------//
// cosine  - x is in degrees                                                //
//--------------------------------------------------------------------------//
inline double cosine(double x)
{ 
    static double c[360], d[360];
    static int hit=0;
    int k;
    int intx;
    double answer,rem;
    if(!hit)
    {  hit=1;
       for(k=0;k<360;k++) c[k] = cos(k*0.01745329252);
       d[0] = 0;
       for(k=1;k<360;k++) d[k] = c[k] - c[k-1];
    }
    intx = int(x);
    rem = x - (double)intx;
    answer = c[intx] + rem*d[intx];
    return(answer);
}
