//--------------------------------------------------------------------------//
// xmtnimage43.cc                                                           //
// Latest revision: 03-12-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
// grain counting                                                           //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Images     *z;
extern int         ci;
Pattern opattern, npattern;
inline int pixel_below_threshold(int v, int vpat, int pix_bpp, int pat_bpp, 
   int pattern_type, int is_color, int threshold);

//-------------------------------------------------------------------------//
//  pattern_recognition                                                    //
//-------------------------------------------------------------------------//
void pattern_recognition(void)
{
  static Dialog *dialog;
  static double thresh1 = 0.5;
  static double pweight=1.0, bweight=-0.05;
  static char filename[FILENAMELENGTH] = "Pattern";
  static int in_pattern_recognition=0;
  static int want_print=0;
  int j,k;
  if(in_pattern_recognition) return;
  in_pattern_recognition = 1;
  dialog = new Dialog;
  if(dialog==NULL){ message(g.nomemory); in_pattern_recognition = 0; return; }
  dialog->needrepeat=0;
  strcpy(dialog->title,"Grain/Pattern Counting");

  //-----Dialog box to get title & other options------------//

  strcpy(dialog->radio[0][0],"Operation");
  strcpy(dialog->radio[0][1],"Select grain/pattern");
  strcpy(dialog->radio[0][2],"Read pattern from disk");
  strcpy(dialog->radio[0][3],"Save pattern on disk");
  strcpy(dialog->radio[0][4],"Count grains");             
  strcpy(dialog->radio[0][5],"Count std. size grains");   
  strcpy(dialog->radio[0][6],"Count patterns");             
  strcpy(dialog->radio[0][7],"Erode (experimental)");   
  strcpy(dialog->radio[0][8],"Enhance grains");   
  dialog->radiono[0]=9;
  dialog->radioset[0] = -1;
  
  strcpy(dialog->boxes[0],"Parameters");
  strcpy(dialog->boxes[1],"Filename");
  strcpy(dialog->boxes[2],"Threshold (0-1)");
  strcpy(dialog->boxes[3],"Weight (-1 to 1)");
  strcpy(dialog->boxes[4],"Match");
  strcpy(dialog->boxes[5],"Mismatch");
  strcpy(dialog->boxes[6],"Print results to stdout");
  dialog->boxtype[0]=LABEL;
  dialog->boxtype[1]=FILENAME;
  dialog->boxtype[2]=STRING;
  dialog->boxtype[3]=LABEL;
  dialog->boxtype[4]=TOGGLESTRING;
  dialog->boxtype[5]=TOGGLESTRING;
  dialog->boxtype[6]=TOGGLE;
  dialog->boxset[4]=1;
  dialog->boxset[5]=1;
  strcpy(dialog->answer[1], filename);
  sprintf(dialog->answer[2],"%g", thresh1); 
  sprintf(dialog->answer[4], "%g", pweight);
  sprintf(dialog->answer[5], "%g", bweight);
  if(want_print) dialog->boxset[6]=1; else dialog->boxset[6]=0;

  for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;  

  dialog->keep=1;
  dialog->want_changecicb = 0;
  dialog->f1 = patterncheck;
  dialog->noofradios=1;
  dialog->noofboxes=7;
  dialog->helptopic=50;
  dialog->transient = 1;
  strcpy(dialog->path,".");

  /////////////////////////////

  dialogbox(*dialog);

  /////////////////////////////

  thresh1 = max(0, min(1, atof(dialog->answer[2])));
  if(dialog->boxset[4]) pweight = atof(dialog->answer[4]);
  if(dialog->boxset[5]) bweight = atof(dialog->answer[5]);
  dialog->keep = RESET;
  dialogbox(*dialog);
  delete[] dialog;
  if(opattern.xsize) free_pattern(&opattern);
  if(npattern.xsize) free_pattern(&npattern);
  in_pattern_recognition = 0;
  return;
}


//--------------------------------------------------------------------------//
//  patterncheck                                                            //
//--------------------------------------------------------------------------//
void patterncheck(dialoginfo *a, int radio, int box, int boxbutton)
{
   static int inpatterncheck=0;
   double thresh = max(0, min(1, atof(a->answer[2])));
   double pweight=1.0, bweight=0.0;
   int want_print;
   if(a->boxset[4]) pweight = atof(a->answer[4]);
   if(a->boxset[5]) bweight = atof(a->answer[5]);
   want_print = a->boxset[6];
   if(inpatterncheck) return;  //// XtSetValues->dialogstringcb->fftcheck
   inpatterncheck=1;
   box=box; boxbutton=boxbutton;
   if(radio==0) switch(a->radioset[0])
   {
       case 1: set_user_pattern(&opattern);
               break;
       case 2: break;
       case 3: break;
       case 4:                //// Grains
               if(opattern.xsize==0) set_user_pattern(&opattern);
               set_pattern_threshold(&opattern, &npattern, thresh);
               patterns(npattern, GRAIN, thresh, pweight, bweight, want_print);
               break;
       case 5:                //// Standard size grains
               set_grain_pattern(&npattern);
               patterns(npattern, GRAIN, thresh, pweight, bweight, want_print); 
               break;
       case 6:                //// Patterns
               if(opattern.xsize==0) set_user_pattern(&opattern);
               patterns(opattern, PATTERN, thresh, pweight, bweight, want_print);
               break;
       case 7: erode(thresh); break;
       case 8: enhance(); break;
   }
   a->radioset[0] = -1;
   inpatterncheck=0;
}


//-------------------------------------------------------------------------//
// erode                                                                   //
//-------------------------------------------------------------------------//
void erode(double thresh)
{
#ifdef DIGITAL
   #define N 25
#else
   const int N=25;
#endif
   double dev=0.0, thresh2=2*thresh;
   int v[N];
   int b,center=0,f,i,i2,j,k,ino=ci,bpp,total,vout=0,count=0;
   int x1,y1,x2,y2;

   f = z[ino].cf;
   bpp = z[ino].bpp;
   b = g.off[bpp];
   x1 = g.ulx + 2;
   x2 = g.lrx - 2;
   y1 = g.uly + 2;
   y2 = g.lry - 2;
   for(j=y1; j<y2; j++)
   for(i=x1,i2=x1*g.off[bpp]; i<x2; i++,i2+=g.off[bpp])
   {    
        v[0] = pixelat(z[ino].image[f][j-1]+i2-b, bpp);
        v[1] = pixelat(z[ino].image[f][j-1]+i2  , bpp);
        v[2] = pixelat(z[ino].image[f][j-1]+i2+b, bpp);
        v[3] = pixelat(z[ino].image[f][j  ]+i2-b, bpp);
        v[4] = pixelat(z[ino].image[f][j  ]+i2  , bpp);
        v[5] = pixelat(z[ino].image[f][j  ]+i2+b, bpp);
        v[6] = pixelat(z[ino].image[f][j+1]+i2-b, bpp);
        v[7] = pixelat(z[ino].image[f][j+1]+i2  , bpp);
        v[8] = pixelat(z[ino].image[f][j+1]+i2+b, bpp);

        v[9]  = pixelat(z[ino].image[f][j-2]+i2-2*b, bpp);
        v[10] = pixelat(z[ino].image[f][j-2]+i2-1*b, bpp);
        v[11] = pixelat(z[ino].image[f][j-2]+i2-0*b, bpp);
        v[12] = pixelat(z[ino].image[f][j-2]+i2+1*b, bpp);
        v[13] = pixelat(z[ino].image[f][j-2]+i2+2*b, bpp);

        v[14] = pixelat(z[ino].image[f][j-1]+i2-2*b, bpp);
        v[15] = pixelat(z[ino].image[f][j-1]+i2+2*b, bpp);
        v[16] = pixelat(z[ino].image[f][j  ]+i2-2*b, bpp);
        v[17] = pixelat(z[ino].image[f][j  ]+i2+2*b, bpp);
        v[18] = pixelat(z[ino].image[f][j+1]+i2-2*b, bpp);
        v[19] = pixelat(z[ino].image[f][j+1]+i2+2*b, bpp);

        v[20] = pixelat(z[ino].image[f][j+2]+i2-2*b, bpp);
        v[21] = pixelat(z[ino].image[f][j+2]+i2-1*b, bpp);
        v[22] = pixelat(z[ino].image[f][j+2]+i2-0*b, bpp);
        v[23] = pixelat(z[ino].image[f][j+2]+i2+1*b, bpp);
        v[24] = pixelat(z[ino].image[f][j+2]+i2+2*b, bpp);

        total = 0;
        for(k=0;k<N;k++) total += v[k];
        //// ave = total / N;
        center = v[4];
        dev = 0.0;
        for(k=0;k<N;k++) dev += abs(v[k] - center);
        dev /= N;
        vout = 0;
        if(between(dev,thresh,thresh2)) vout = 0; else vout=g.maxcolor;
        if(vout<0) vout=0;
        putpixelbytes(z[ino].img[j]+i*g.off[g.bitsperpixel], vout, 1, g.bitsperpixel, 1);
        count++;
   }
   redraw(ino);
}


//-------------------------------------------------------------------------//
//  enhance                                                                //
//-------------------------------------------------------------------------//
void enhance(void)
{
   int k,noofargs;
   char **arg;                // up to 20 arguments
   arg = new char*[20];
   for(k=0;k<20;k++){ arg[k] = new char[128]; arg[k][0]=0; }

   noofargs = 0;
   strcpy(arg[0],"filter");
   strcpy(arg[1],"12"); noofargs++;
   strcpy(arg[2],"2"); noofargs++;
   strcpy(arg[3],"4"); noofargs++;
   filter(noofargs, arg);
   invertimage();
   for(k=0;k<20;k++) delete[] arg[k];
   delete arg;  
 }


//-------------------------------------------------------------------------//
// patterns                                                                //
//-------------------------------------------------------------------------//
void patterns(Pattern p, int pattern_type, double thresh, double pweight, 
     double bweight, int want_print)
{
   int count,ino,
       bpp,k,x1,y1,x2,y2,x11,y11,x22,y22,total=0,size,ototal,iterations=0,
       threshold, xmax=0, ymax=0, start_ino;
   float min_signal, vmax = 0.0; 
   char tempstring[1024];
   ino = p.ino;
   size = p.xsize * p.ysize;

   if(ci != ino) switchto(ino);
   start_ino = ino;
   bpp = z[ino].bpp;

   int is_color = 1;
   if(z[ino].colortype==GRAY) is_color=0;
   if(z[ino].colortype==INDEXED && unmodified_grayscale(z[ino].palette)) is_color=0;

   threshold = cint((1.0 - thresh) * g.maxgray[bpp]);

   array<float> layer(z[ino].xsize+1, z[ino].ysize+1); 
   if(!layer.allocated) { message(g.nomemory); return; }

   x1 = g.ulx;
   x2 = g.lrx;
   y1 = g.uly;
   y2 = g.lry;

   ////  Find all neurons over threshold. 
   ////  Neurons fire in order of highest to lowest signal. Each
   ////  firing causes the cells in the vicinity to be recalculated
   ////  and the cycle is repeated.

   total = 0;
   if(pattern_type==GRAIN) 
   {   min_signal = 0.5 * thresh * sqrt(size);
       strcpy(tempstring, "grains"); 
   }else 
   {   min_signal = thresh * (float)size;
       strcpy(tempstring, "patterns");
   }
   if(want_print)
      printf("\nSearching image %d for %s between %d,%d and %d,%d\n\
Pattern = %d x %d\n\
Min signal = %g\nThreshold = %g\nMatch weight = %g\nMismatch weight = %g\n", 
ino, tempstring, x1, y1, x2, y2, 
p.xsize, p.ysize,
min_signal, thresh, pweight, bweight);

   printstatus(CALCULATING);
   do
   {  ototal = total;
      check_event();
      if(ci != start_ino) { message("Unexpected image change",ERROR); g.getout=1; }
      if(g.getout) { g.getout=0; break; }
      iterations++;
      if(neuron_count_mismatch(ino, x1, y1, x2, y2)) break;
      count = find_pattern(p, pattern_type, layer.p, ino, threshold, pweight, bweight, 
              min_signal, is_color, x1, y1, x2, y2, 1);
      //// vmax is no. of pixels below threshold for each pattern  
      //// xmax, ymax is location of best fitting point   
      if(want_print) printf("Iteration %d  count %d\n",iterations, count);

      for(k=0;k<count;k++)
      {   
          check_event();
          if(ci != start_ino) { message("Unexpected image change",ERROR); g.getout=1; }
          if(g.getout) break; 
          if(neuron_count_mismatch(ino, x1, y1, x2, y2)) break;
          find_highest_neuron(layer.p, x1, y1, x2, y2, xmax, ymax, vmax);           
          if(g.getout) break; 
          if(vmax >= min_signal)
          {   subtract_pattern(p, pattern_type, layer.p, ino, is_color, threshold, xmax, ymax);
              total++;
              sprintf(tempstring, "Tot.%d Sig.%g                 ", total,vmax);
              if(want_print) printf("  Total %d\t at %4d %4d signal=%g\n",total,xmax,ymax,vmax);
              printstatus(MESSAGE, tempstring);
          }else break;
          x11 = xmax - p.xsize; 
          y11 = ymax - p.ysize;
          x22 = xmax + p.xsize; 
          y22 = ymax + p.ysize;
          find_pattern(p, pattern_type, layer.p, ino, threshold, pweight, bweight, 
               min_signal, is_color, x11, y11, x22, y22, 0);
      }
      if(g.getout) { g.getout=0; break; }
   }while(total>ototal);
   if(pattern_type == GRAIN) sprintf(tempstring,"%d grains found",total);
   else sprintf(tempstring,"%d patterns found",total);
   message(tempstring);
}



//-------------------------------------------------------------------------//
// find_pattern                                                            //
//-------------------------------------------------------------------------//
int find_pattern(Pattern p, int pattern_type, float **layer, int ino, int threshold, 
   double pweight, double bweight, float min_signal, int is_color,
   int x1, int y1, int x2, int y2, int want_progress_bar)
{
   Widget www, scrollbar;
   double layertemp;
   int f,i,ii,j,jj,k,k1,k2,l,l1,bpp,v=0,vpat,total=0,xsize,ysize;
   f = z[ino].cf;
   uchar *pp;
   uchar **add;
   add = z[ino].image[f];
   bpp = z[ino].bpp;
   xsize = z[ino].xsize;
   ysize = z[ino].ysize;
   if(want_progress_bar) progress_bar_open(www, scrollbar);

   for(j=y1; j<y2; j++)
   {
     check_event();
     if(want_progress_bar) progress_bar_update(scrollbar, (j-y1)*100/(y2-y1));
     if(g.getout) return 0;
     for(i=x1; i<x2; i++)
     {  
        if(!between(j,0,z[ino].ysize-1)) continue;
        if(!between(i,0,z[ino].xsize-1)) continue;
        layer[j][i] = 0.0;
        layertemp = 0.0;
        jj = j - p.ysize/2 - z[ino].ypos;
        ii = i - p.xsize/2 - z[ino].xpos;
        for(l=0; l<p.ysize; l++)
        for(k=0; k<p.xsize; k++)
        {    
             vpat = pixelat(p.pat[l]+k*g.off[p.bpp], p.bpp); 
             l1 = l + jj;
             k1 = k + ii;
             if(l1<0 || l1 >= ysize) continue;
             if(k1<0 || k1 >= xsize) continue;
             k2 = k1*g.off[bpp]; 
             pp = add[l1]+k2;


             //// This abomination is pixelat(), i.e., 
             ////    v = pixelat(pp, bpp);.  Copying it here makes
             //// find_pattern() 22% faster. Making pixelat() inline has 
             //// no effect with gcc.

             if(g.ximage_byte_order == LSBFirst)
             {    switch(bpp)
                  {   case 8:   v = ( *pp ); break;
                      case 15:  v = ( 0x7fff & (*(pp+1)<<8) + *pp ); break;
                      case 16:  v = ( (*(pp+1)<<8) + *pp ); break;
                      case 24:
                      case 32:  v = ( (*(pp+2)<<16) + (*(pp+1)<<8) + *pp ); break;
                      case 48:  v = ( (*(pp+4)<<16) + (*(pp+2)<<8) + *pp ); break;
                  }
             }else
             {    switch(bpp)
                  {   case 8:   v = ( *pp ); break;
                      case 15:  v = ( 0x7fff & (*(pp+1)<<8) + *pp ); break;
                      case 16:  v = ( (*(pp)<<8) + *(pp+1) ); break;
                      case 24:  v = ( (*(pp+2)<<16) + (*(pp+1)<<8) + *(pp+0) ); break;
                      case 32:  v = ( (*(pp+1)<<16) + (*(pp+2)<<8) + *(pp+3) ); break;
                      case 48:  v = ( (*(pp+1)<<16) + (*(pp+3)<<8) + *(pp+5) ); break;
                  }
             } 


             if(pixel_below_threshold(v, vpat, bpp, p.bpp, pattern_type, is_color, threshold))
                 layertemp += pweight; 
             else
                 layertemp += bweight; 

/*
             switch(pixel_below_threshold(v, vpat, bpp, p.bpp, pattern_type, is_color, threshold))
             {   case 1:  layertemp += pweight; break;
                 case 0:  break;
                 case -1: layertemp += bweight; 
             }

*/
        }
        layer[j][i] = layertemp;  ////  Signal for each neuron
        if(layer[j][i] >= min_signal) total++;
     }
   }
   if(want_progress_bar) progress_bar_close(www);
   return total;
}


//-------------------------------------------------------------------------//
// find_highest_neuron                                                     //
//-------------------------------------------------------------------------//
void find_highest_neuron(float **layer, int x1, int y1, int x2, int y2, 
     int &xmax, int &ymax, float &vmax)
{
   int i,j;
   vmax = 0.0;
   for(j=y1; j<y2; j++)
   for(i=x1; i<x2; i++)
      if(layer[j][i] > vmax)
      {    vmax = layer[j][i];
           xmax = i;
           ymax = j;
      }
}

//-------------------------------------------------------------------------//
// neuron_count_mismatch                                                   //
//-------------------------------------------------------------------------//
int neuron_count_mismatch(int ino, int x1, int y1, int x2, int y2) 
{
   if(!between(x1,0,z[ino].xsize-1) || !between(x2,0,z[ino].xsize-1) ||
      !between(y1,0,z[ino].ysize-1) || !between(y2,0,z[ino].ysize-1))
      { message("Neuron count mismatch");  return 1; }
   return 0;
}


//-------------------------------------------------------------------------//
// pixel_below_threshold                                                   //
//-------------------------------------------------------------------------//
inline int pixel_below_threshold(int v, int vpat, int pix_bpp, int pat_bpp, 
   int pattern_type, int is_color, int threshold)
{
   int diff, rr, gg, bb, pat_rr, pat_gg, pat_bb;
   if(pattern_type==GRAIN)
   {   if(is_color)
       {    valuetoRGB(v, rr, gg, bb, pix_bpp);
            valuetoRGB(vpat, pat_rr, pat_gg, pat_bb, pat_bpp);
            if((rr + gg + bb)/3 <= threshold &&
                (pat_rr + pat_gg + pat_bb)/3 <= threshold)
                 return 1;
            if((rr + gg + bb)/3 > threshold &&
                (pat_rr + pat_gg + pat_bb)/3 > threshold)
                 return 0;
//          else 
//               return -1;
       }else
       {    if(v <= threshold && vpat <= threshold)  // match and grain
                return 1;
            if(v > threshold && vpat > threshold)    // match not grain
                return 0;
//          return -1;                               // mismatch
       }
   }else      // PATTERN
   {   if(is_color)
       {    valuetoRGB(v, rr, gg, bb, pix_bpp);
            valuetoRGB(vpat, pat_rr, pat_gg, pat_bb, pat_bpp);
            diff = (abs(rr-pat_rr) + abs(gg-pat_gg) + abs(bb-pat_bb))/3;
            if(diff < threshold && v != int(g.maxvalue[pix_bpp]))
                return 1;
            else 
                return 0;
       }else
       {    if(abs(v - vpat) <= threshold)
                return 1;
            else 
                return 0;
       }
   }
   return 0;
}


//-------------------------------------------------------------------------//
// subtract_pattern                                                        //
//-------------------------------------------------------------------------//
void subtract_pattern(Pattern p, int pattern_type, float **layer, int ino, int is_color, 
     int threshold, int xmax, int ymax)
{
   int f, k, k1, k2, k3, l, l1, l2, bpp, v, vpat;
   f = z[ino].cf;
   bpp = z[ino].bpp;

   for(l=0; l<p.ysize; l++)
   for(k=0; k<p.xsize; k++)
   {    
        l1 = l + ymax - p.ysize/2;
        k1 = k + xmax - p.xsize/2;
        l2 = l1 - z[ino].ypos;
        k2 = k1 - z[ino].xpos;
        if(k1<0 || l1<0 || l1>=z[ino].ysize || k1>=z[ino].xsize) continue;
        k3 = k2*g.off[bpp]; 
        v = pixelat(z[ino].image[f][l2]+k3, bpp);
        vpat = pixelat(p.pat[l]+k*g.off[p.bpp], p.bpp); 
        if(pixel_below_threshold(v, vpat, bpp, p.bpp, pattern_type, is_color, threshold)==1)
        {    setpixelonimage(k1, l1, int(g.maxvalue[bpp]), SET,8,-1,0);
             layer[l1][k1] = 0.0;
        } 
   }
}


//-------------------------------------------------------------------------//
// set_user_pattern                                                        //
//-------------------------------------------------------------------------//
void set_user_pattern(Pattern *p)
{
   char tempstring[1024];
   int bytesperpixel, bpp, ino, i, j, xx1, yy1, xx2, yy2, ulx, uly, lrx, lry, xsize, ysize, v;
   if(p->xsize) free_pattern(p);
   message("Use mouse to select\na representative grain or pattern\n(draw box around pattern)");
   ulx = g.ulx; uly = g.uly; lrx = g.lrx; lry = g.lry;
   getbox(xx1, yy1, xx2, yy2);
   g.ulx = ulx; g.uly = uly; g.lrx = lrx; g.lry = lry;
   ino = ci;
   bytesperpixel = (7+z[ino].bpp)/8;
   p->ino = ino;
   p->bpp = z[ino].bpp;
   xsize = yy2 - yy1 + 1;
   ysize = xx2 - xx1 + 1;
   if(!(xsize & 1)) xsize++;
   if(!(ysize & 1)) ysize++;
   p->ysize = xsize;
   p->xsize = ysize;

   p->colortype = z[ino].colortype; 
   if(z[ino].colortype==INDEXED && unmodified_grayscale(z[ino].palette)) p->colortype=GRAY;
   p->pat_1d = (uchar*)malloc(p->ysize * p->xsize * sizeof(uchar)*bytesperpixel);
   if(p->pat_1d==NULL) { message(g.nomemory); return; }
   p->pat    = make_2d_alias(p->pat_1d, p->xsize*bytesperpixel, p->ysize);
   for(j=0; j<p->ysize; j++)
   for(i=0; i<p->xsize; i++)
   {   v = readpixelonimage(xx1+i, yy1+j, bpp, ino, -1);  
       putpixelbytes(p->pat[j]+i*g.off[p->bpp], v, 1, p->bpp);
   }
   sprintf(tempstring, "%d x %d pattern stored", p->xsize, p->ysize);
   message(tempstring);
}


//-------------------------------------------------------------------------//
//  set_pattern_threshold                                                  //
//-------------------------------------------------------------------------//
void set_pattern_threshold(Pattern *p1, Pattern *p2, double thresh)
{
   int bytesperpixel, i, j, rr, gg, bb, size, vpat;
   int threshold = cint(thresh * 255);
   if(p2->xsize) free_pattern(p2);

   ////  Copy p1 into p2
   p2->ino       = p1->ino;
   p2->bpp       = p1->bpp;
   p2->ysize     = p1->ysize;
   p2->xsize     = p1->xsize;
   p2->colortype = p1->colortype;
   bytesperpixel = (7+p1->bpp)/8;
   p2->pat_1d    = (uchar*)malloc(p1->ysize*p1->xsize*sizeof(uchar)*bytesperpixel);
   if(p2->pat_1d==NULL) { message(g.nomemory); return; }
   p2->pat       = make_2d_alias(p1->pat_1d, p1->xsize*bytesperpixel, p1->ysize);
   size = p2->xsize * p2->ysize * bytesperpixel;
   memcpy(p2->pat_1d, p1->pat_1d, size);

   //// Threshold pattern p2
   for(j=0; j<p2->ysize; j++)
   for(i=0; i<p2->xsize; i++)
   {  
       vpat = pixelat(p2->pat[j] + i*g.off[p2->bpp], p2->bpp); 
       if(p2->colortype==GRAY)
       {    
           if(vpat < threshold) p2->pat[j][i]=0; else  p2->pat[j][i]=255;
       }else
       {   valuetoRGB(vpat, rr, gg, bb, p2->bpp);
           if((rr + gg + bb)/3 < threshold) 
                rr = gg = bb = 0; 
           else
           {    rr = g.maxred[p2->bpp];
                gg = g.maxgreen[p2->bpp];
                bb = g.maxblue[p2->bpp];
           }
           vpat = RGBvalue(rr,gg,bb,p2->bpp);
           putpixelbytes(p2->pat[j]+i*g.off[p2->bpp], vpat, 1, p2->bpp);
      }
   }
}


//-------------------------------------------------------------------------//
// set_grain_pattern                                                       //
//-------------------------------------------------------------------------//
void set_grain_pattern(Pattern *p)
{
   double dist;
   int ino = ci, i, j;
   int bytesperpixel = (7+z[ino].bpp)/8;
   if(p->xsize) free_pattern(p);
   p->ino = ino;
   p->bpp = z[ino].bpp;
   p->ysize = 11;
   p->xsize = 11;
   p->colortype = z[ino].colortype;
   p->pat_1d = (uchar*)malloc(p->ysize*p->xsize*sizeof(uchar)*bytesperpixel);
   if(p->pat_1d==NULL) { message(g.nomemory); return; }
   p->pat    = make_2d_alias(p->pat_1d, p->xsize*bytesperpixel, p->ysize);
   for(j=0; j<11; j++)
   for(i=0; i<11; i++)
   {   
      dist = sqrt(abs(i-5)*abs(i-5) + abs(j-5)*abs(j-5));
      if(dist<2.8) p->pat[j][i] = 0; else p->pat[j][i] = (int)g.maxcolor;
   }
}

//-------------------------------------------------------------------------//
// free_pattern                                                            //
//-------------------------------------------------------------------------//
void free_pattern(Pattern *p)
{
   free(p->pat);
   free(p->pat_1d);
   p->ysize = 0;
   p->xsize = 0;
}
