//--------------------------------------------------------------------------//
// xmtnimage44.cc                                                           //
// Latest revision: 05-18-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
// Wavelets                                                                 //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"
#include "xmtnimagec.h"

extern Globals     g;
extern Images     *z;
extern int         ci;

//--------------------------------------------------------------------------//
// wavelet                                                                  //
//--------------------------------------------------------------------------//
void wavelet(int noofargs, char **arg)
{
   static int  transform_type=1, want_grid=0, want_gray_offset, gray_offset, maxlevels=0,
     levels = 0;
   static char wavefilename[1024] = "none";
   static char wpath[1024] = "x";
   static char index_range[80] = "all";
   static char coeff_range[80] = "all";
   static char ignore_range[80] = "none";
   int j, k, ino=ci;
   if(maxlevels==0) 
   {   for(k=2; k<min(z[ino].xsize, z[ino].ysize); k*=2) maxlevels++;
       levels = maxlevels;
       for(k=maxlevels; k>=z[ino].wavelet_xminres; k/=2) levels--;
   }
   arg=arg; noofargs=noofargs; 

   ////------------------------////

   if(!strcmp(wpath,"x")) strcpy(wpath, g.waveletpath);

   Dialog dialog;
   dialog.needrepeat = 0;  
   strcpy(dialog.radio[0][0],"Execute"); 
   strcpy(dialog.radio[0][1],"Decompose");             
   strcpy(dialog.radio[0][2],"Reconstitute");
   dialog.radiono[0]=3;
   dialog.radioset[0] = -1; 

   strcpy(dialog.radio[1][0],"Algorithm"); 
   strcpy(dialog.radio[1][1],"Pyramidal");             
   strcpy(dialog.radio[1][2],"Mallat");
   strcpy(dialog.radio[1][3],"Multiresolution");
   dialog.radiono[1]=4;
   dialog.radioset[1] = transform_type;

   strcpy(dialog.radio[2][0],"Other functions"); 
   strcpy(dialog.radio[2][1],"Save coordinates");             
   strcpy(dialog.radio[2][2],"Read coordinates");
   strcpy(dialog.radio[2][3],"Restore original");
   dialog.radiono[2]=4;
   dialog.radioset[2] = -1;

   strcpy(dialog.title,"Wavelet transform");
   strcpy(dialog.boxes[0],"Data");
   dialog.boxtype[0]=LABEL;

   strcpy(dialog.boxes[1],"Image no. ");
   dialog.boxtype[1]=INTCLICKBOX;
   sprintf(dialog.answer[1],"%d",ino);
   dialog.boxmin[1]=0; 
   dialog.boxmax[1]=g.image_count-1;

   strcpy(dialog.boxes[2],"Wavelet");
   strcpy(dialog.answer[2], wavefilename);
   dialog.boxtype[2]=FILENAME;

   strcpy(dialog.boxes[3],"Reconstitution coefficients");
   dialog.boxtype[3]=LABEL;

   strcpy(dialog.boxes[4],"Index range to use");
   dialog.boxtype[4]=STRING;
   strcpy(dialog.answer[4], index_range);

   strcpy(dialog.boxes[5],"Value range to use");
   dialog.boxtype[5]=STRING;
   strcpy(dialog.answer[5], coeff_range);

   strcpy(dialog.boxes[6],"Value range to ignore");
   dialog.boxtype[6]=STRING;
   strcpy(dialog.answer[6], ignore_range);

   strcpy(dialog.boxes[7],"Options");
   dialog.boxtype[7]=LABEL;

   strcpy(dialog.boxes[8],"Show grid");
   dialog.boxtype[8]=TOGGLE;
   dialog.boxset[8] = want_grid;

   strcpy(dialog.boxes[9],"Gray value offset");
   dialog.boxtype[9]=TOGGLESTRING;
   sprintf(dialog.answer[9], "%d", gray_offset);
   dialog.boxset[9] = want_gray_offset;

   strcpy(dialog.boxes[10],"Levels of detail");
   dialog.boxtype[10]=INTCLICKBOX;
   sprintf(dialog.answer[10],"%d",levels);
   dialog.boxmin[10]=1; 
   dialog.boxmax[10] = maxlevels;

 
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
   for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
   for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
   dialog.noofradios=3;
   dialog.noofboxes=11;
   dialog.helptopic=52;
   dialog.keep=0;
   dialog.needrepeat=0;
   dialog.want_changecicb = 0;
   dialog.f1 = wavelet_check;
   dialog.transient = 1;
   strcpy(dialog.path, wpath);

   dialogbox(dialog);

   ino = atoi(dialog.answer[1]);
   strcpy(wavefilename, dialog.answer[2]);
   strcpy(wpath, dialog.c[2]->path);
   strcpy(index_range, dialog.answer[4]);
   strcpy(coeff_range, dialog.answer[5]);
   strcpy(ignore_range, dialog.answer[6]);
   want_grid = dialog.boxset[8];
   want_gray_offset = dialog.boxset[9];
   gray_offset = atoi(dialog.answer[9]);
   levels = atoi(dialog.answer[10]);
}


//--------------------------------------------------------------------------//
// read_wavelet                                                             //
// File Format:                                                             //
//   Lines starting with # or blank = comments                              //
//    [Range low] [Range high]                                              //
//    [Decomposition LP filter coefficients]                                //
//    . . .                                                                 // 
//    [Range low] [Range high]                                              //
//    . . .                                                                 // 
//    [Reconstitution LP filter coefficients]                               //
//    . . . (If absent, use decomposition coefficients)                     // 
//    Any characters after column 80 may be ignored.                        //
//--------------------------------------------------------------------------//
void read_wavelet(Wavelet &w)
{
  FILE *fp;
  int  k, count=0, low=0, high=0, sign = -1;
  char buffer[1024];
  if((fp=fopen(w.filename,"rt")) == NULL)
      error_message(w.filename, errno);
  else
  {   
      if(read_next_non_comment(fp, buffer)) goto readb;
      sscanf(buffer, "%d %d", &low, &high);
      w.ioffset = low;
      w.size = high - low + 1; 
      count = 0;
      ////  Decomposition filter
      for(k=low; k<=high; k++)
      {   if(read_next_non_comment(fp, buffer)) goto readb;
          if(count<1024) w.filter[0][count] = atof(buffer);
          w.filter[1][count] = w.filter[0][count];
          count++;
      }
      if(read_next_non_comment(fp, buffer)) goto readb;
      sscanf(buffer, "%d %d", &low, &high);
      w.joffset = low;
      count = 0;
      ////  Reconstitution filter
      for(k=low; k<=high; k++)
      {   if(read_next_non_comment(fp, buffer)) goto readb; 
          if(count<1024) w.filter[1][count] = atof(buffer);
          count++;
      }
      ////  High-pass filters - reversed with alternating sign changes
readb:
      for(k=0; k<w.size; k++)
      {   w.hifilter[0][w.size-k-1] = sign * w.filter[0][k];
          w.hifilter[1][w.size-k-1] = sign * w.filter[1][k];
          sign = -sign;
      }
      if(w.ioffset == 0 && w.joffset == 0) w.ioffset = w.joffset = -w.size/2; 
      fclose(fp); 
  }

//printf("wavelet %d %d %d\n",w.size,w.ioffset,w.joffset);
//for(k=0;k<w.size;k++) printf(" %g\t %g\t %g\t %g\n",w.filter[0][k],w.filter[1][k],
//w.hifilter[0][k],w.hifilter[1][k]);

  return;
}


//--------------------------------------------------------------------------//
// read_next_non_comment                                                    //
//--------------------------------------------------------------------------//
int read_next_non_comment(FILE *fp, char *buffer)
{
   int eof = 0;
   do
   {   fgets(buffer, 1024, fp);
       if(feof(fp)){ eof=1; break; }
       remove_trailing_space(buffer);
   } while( (buffer[0] == '#') || (!strlen(buffer)) );
   return eof;
}


//--------------------------------------------------------------------------//
//  remove_trailing_space - including \f \n \r \t \v and space              //
//--------------------------------------------------------------------------//
void remove_trailing_space(char *s)
{
   int k;
   ////  isspace doesn't match \n as it is supposed to 
   for(k=strlen(s)-1; k>=0; k--){ if(isspace(s[k]) || s[k]==10) s[k] = 0; else break; }
}


//--------------------------------------------------------------------------//
//  wavelet_check                                                           //
//--------------------------------------------------------------------------//
void wavelet_check(dialoginfo *a, int radio, int box, int boxbutton)
{
   static Wavelet_params wp;
   static int incheck=0, hit=0;
   static Wavelet wave;    
   int transform_type=1;
   if(incheck) return;  //// XtSetValues->dialogstringcb->fftcheck
   int k, noofargs, xss=0, yss=0, xlen, ylen, want_grid, status=OK;
   char **buf;
   XmString xms1, xms2, xms3;
   char tempstring[128];
 
   wp.ino = atoi(a->answer[1]);
   wp.levels = atoi(a->answer[10]);
   wp.minres = min(z[wp.ino].xsize, z[wp.ino].ysize);
   for(k=1;k<wp.levels;k++) wp.minres/=2;
   wp.minres = max(4, wp.minres);
   
   ////  Bail out if bad ino unless reading one from disk
   if(radio==0 && box==-1 && boxbutton==-1)
   {    if(wp.ino<=0 || wp.ino>=g.image_count) return;
        ////  Make sure image size is a power of 2
        xlen = z[wp.ino].xsize;
        ylen = z[wp.ino].ysize;
        for(xss=1; xss<=xlen*2; xss*=2) if(xss>=xlen) break; 
        for(yss=1; yss<=ylen*2; yss*=2) if(yss>=ylen) break; 
        if(z[wp.ino].xsize!= xss || z[wp.ino].ysize!= yss) 
        {     message("Enlarging image to nearest\npower of 2");
              if(g.getout){ g.getout=0; return; }
              resize_image(wp.ino, xss, yss);        
        }
        if(z[wp.ino].colortype != GRAY) 
        {     message("Warning: converting image\nto grayscale for wavelet transform");
              if(g.getout){ g.getout=0; return; }
              converttograyscale(wp.ino);
        }
   }
 
   buf = new char*[4];
   for(k=0;k<4;k++){ buf[k] = new char[32]; buf[k][0] = 0; }

   ////  Permanently allocate space for vavelet
   if(!hit)
   {   wave.filename    = new char[FILENAMELENGTH];
       wave.filter      = new double*[3];
       wave.filter[0]   = new double[1024];  // decomposition filter
       wave.filter[1]   = new double[1024];  // reconstruction filter
       wave.hifilter    = new double*[3];
       wave.hifilter[0] = new double[1024];  // reverse decomposition filter
       wave.hifilter[1] = new double[1024];  // reverse reconstruction filter
       strcpy(wave.filename, "none");
       wp.format        = new char[128];
       strcpy(wp.format, "pyramidal");
       wp.levels        = 0;                 // not used yet
   }

   incheck=1;
   box=box; boxbutton=boxbutton;

   if(a->radioset[0] == 1) wp.direction=1; else wp.direction = 0;
   transform_type = a->radioset[1];
   strcpy(wave.filename, a->answer[2]);

   z[wp.ino].wmin = smallest(wp.ino);
   z[wp.ino].wmax = largest(wp.ino);
                                 // index range to use
   strip_extraneous_dashes(a->answer[4]);
   parse_string(a->answer[4], buf, noofargs, 0, 32);
   if(!strcmp(buf[0],"all"))
   {   wp.imin=0; wp.imax=xss*yss;}
   else if(!strcmp(buf[0],"none"))
   {   wp.imin=0; wp.imax=0; }
   else if(noofargs>=2)
   {   wp.imin = atoi(buf[0]);
       if(noofargs==2) wp.imax = atoi(buf[1]);
       else wp.imax = atoi(buf[2]);
   }else{ wp.imin=0; wp.imax=xss*yss;}

                              // values to use
   strip_extraneous_dashes(a->answer[5]);
   parse_string(a->answer[5], buf, noofargs, 0, 32);
   if(!strcmp(buf[0],"all"))
   {   wp.cmin = z[wp.ino].wmin; wp.cmax = z[wp.ino].wmax; }
   else if(!strcmp(buf[0],"none"))
   {   wp.cmin=0.0; wp.cmax=0.0; }
   else if(noofargs>=2)
   {   wp.cmin = atof(buf[0]);
       if(noofargs==2) wp.cmax = atof(buf[1]);
       else wp.cmax = atof(buf[2]);
   }else{ wp.cmin = z[wp.ino].wmin; wp.cmax = z[wp.ino].wmax; }

                              // values to ignore
   strip_extraneous_dashes(a->answer[6]);
   parse_string(a->answer[6], buf, noofargs, 0, 32);
   if(!strcmp(buf[0],"all"))
   {   wp.ignmin = z[wp.ino].wmin; wp.ignmax = z[wp.ino].wmax; }
   else if(!strcmp(buf[0],"none"))
   {   wp.ignmin = 0.0; wp.ignmax = 0.0; }
   else if(noofargs>=2)
   {   wp.ignmin = atof(buf[0]);
       if(noofargs==2) wp.ignmax = atof(buf[1]);
       else wp.ignmax = atof(buf[2]);
   }else{ wp.ignmin = 0.0; wp.ignmax = 0.0; }

   want_grid = a->boxset[8];
   wp.want_gray_offset = a->boxset[9];
   wp.gray_offset = atoi(a->answer[9]);

//if(wp.direction==0) printf("using %d %d   %g %g   %g %g\n",wp.imin,wp.imax,wp.cmin,wp.cmax,
//wp.ignmin,wp.ignmax);
   
   //// Radio buttons
   if(radio==0 && box==-1 && boxbutton==-1)
   {    if(!strcmp(wave.filename,"none")) 
             message("Select a wavelet first");
        else
        {    read_wavelet(wave);
             ////  Allocate wavelet array for image
             if(!z[wp.ino].waveletexists) status = allocate_image_wavelet(wp.ino, xss, yss);
             if(status==OK)
             {  switch(transform_type)
                {  case 1: 
                     wavelet_pyramidal_calc(wave, wp);
                     repair(wp.ino);
                     refreshspreadsheet(wp.ino);
                     if(want_grid && z[wp.ino].waveletstate) 
                         show_wavelet_grid(wp.ino, transform_type);
                     break;
                 default: message("Not implemented\n",ERROR); break;
                }     
             }
        }
        for(k=0;k<g.openlistcount;k++) update_list(k);
   }
   if(radio==2 &&box==-1 && boxbutton==-1) 
   {    switch(a->radioset[2])
        {    case 1: write_image_wavelet(wave, wp); break;
             case 2: read_image_wavelet(0, NULL, wave, wp); 

                     a->boxmax[1] = g.image_count-1;
                     a->c[1]->maxval = g.image_count-1;
                     sprintf(tempstring, "%d", ci);
                     a->c[1]->answer = ci;
                     a->c[1]->startval = ci;
                     strcpy(a->answer[1], tempstring);
                     XtVaSetValues(a->boxwidget[1][0], XmNlabelString, 
                              xms1=XmStringCreateSimple(tempstring), 
                              XmNcursorPosition, strlen(tempstring),NULL); 

                     strcpy(a->answer[2], wave.filename);
                     strcpy(a->c[2]->title, wave.filename);
                     XtVaSetValues(a->boxwidget[2][0], XmNlabelString, 
                              xms2=XmStringCreateSimple(wave.filename), 
                              XmNcursorPosition, strlen(wave.filename),NULL); 

                     sprintf(tempstring, "%d", wp.levels);
                     a->c[10]->answer = wp.levels;
                     a->c[10]->startval = wp.levels;
                     strcpy(a->answer[10], tempstring);
                     XtVaSetValues(a->boxwidget[10][0], XmNlabelString, 
                              xms3=XmStringCreateSimple(tempstring), 
                              XmNcursorPosition, strlen(tempstring),NULL); 

                     XmStringFree(xms3);
                     XmStringFree(xms2);
                     XmStringFree(xms1);
                     break;
             case 3: restoreimage(); break;
        }
   }
 
   incheck=0;
   for(k=0;k<4;k++) delete[] buf[k];
   delete[] buf;
   hit = 1;
}


//--------------------------------------------------------------------------//
//  allocate_image_wavelet                                                  //
//--------------------------------------------------------------------------//
int allocate_image_wavelet(int ino, int xsize, int ysize)
{
    if(ino>=g.image_count || z[ino].image==NULL){ message("Non-existent image",ERROR); return ERROR; } 
    z[ino].fftxsize = xsize;
    z[ino].fftysize = ysize;
    z[ino].wavelet_1d = (double*)malloc(ysize*xsize*sizeof(double));                        
    if(z[ino].wavelet_1d==NULL){ message("Error allocating wavelet array",ERROR); return NOMEM;} 
    z[ino].wavelet = make_2d_alias(z[ino].wavelet_1d, xsize, ysize);
    if(z[ino].wavelet==NULL){ message(g.nomemory,ERROR); return NOMEM; } 
    z[ino].waveletexists = 1;
    z[ino].waveletstate = 0;
    z[ino].fmin = 0;
    z[ino].fmax = 0;
    z[ino].wmin = 0;
    z[ino].wmax = 0;
    return OK;
}


//--------------------------------------------------------------------------//
// wavelet_pyramidal_calc                                                   //
// 2-dimensional vavelet transform                                          //
//--------------------------------------------------------------------------//
void wavelet_pyramidal_calc(Wavelet &wave, Wavelet_params &wp)
{
   int bpp, ino, bytesperpixel,f,h,i,j,k,n,nt,value;
   double *ww;
   double coef=0, frac;
   ino = wp.ino;
   bpp = z[ino].bpp;
   bytesperpixel = (7+bpp)/8;
   int xsize = z[ino].xsize;
   int ysize = z[ino].ysize;
   f = z[ino].cf;

   z[ino].fmax =-(int)g.maxvalue[bpp];
   z[ino].fmin = (int)g.maxvalue[bpp];
   z[ino].wmax =-(int)g.maxvalue[bpp];
   z[ino].wmin = (int)g.maxvalue[bpp];
   z[ino].wavelet_xminres = wp.minres;
   z[ino].wavelet_yminres = wp.minres * ysize / xsize;
   ww = new double[max(xsize,ysize)];
   frac = (double)wp.gray_offset / g.maxvalue[bpp];
 
   for(h=0; h<ysize+xsize; h++)
   {      if(wp.direction==1)    // 1=forward   data >> transform y >> transform x
          { 
              k = 0;
              if(h<ysize)
              {  n = xsize;
                 for(i=0; i<n; i++) ww[k++] = (double)pixelat(z[ino].image[f][h]+
                                        i*bytesperpixel, bpp);
              }else
              {  n = ysize;
                 for(j=0; j<ysize; j++) ww[k++] = z[ino].wavelet[j][h-ysize];
              }
              for(nt=n; nt>wp.minres; nt/=2) wavelet_transform(wave, ww, nt, 1);

              k = 0;
              if(h<ysize)
              {  for(i=0; i<xsize; i++) z[ino].wavelet[h][i] = ww[k++]; 
                 
              }else
              {  for(j=0; j<ysize; j++)
                 {   z[ino].wavelet[j][h-ysize] = ww[k]; 
                     if(j>ysize/2 && h-ysize>ysize/2)
                     {  z[ino].fmax = max(z[ino].fmax, ww[k]);
                        z[ino].fmin = min(z[ino].fmin, ww[k]);
                     }else
                     if(j<wp.minres && h-ysize<wp.minres)
                     {  z[ino].wmax = max(z[ino].wmax, ww[k]);
                        z[ino].wmin = min(z[ino].wmin, ww[k]);
                     }
                     k++;
                 }
              }

          }else               // 0=reverse
          {   
              k = 0;
              if(h<ysize)
              {  n = xsize;
                 for(i=0; i<n; i++) 
                 {   coef = z[ino].wavelet[h][i]; 
                     if(between(i, wp.imin, wp.imax) && between(coef, wp.cmin, wp.cmax) &&
                        !between(coef, wp.ignmin, wp.ignmax)) ww[k] = coef;
                     else ww[k] = 0;
                     k++;
                 }
              }else
              {  n = ysize;
                 for(j=0; j<ysize; j++) 
                 {   coef = z[ino].wavelet[j][h-ysize];
                     if(between(j, wp.imin, wp.imax) && between(coef, wp.cmin, wp.cmax) &&
                       !between(coef, wp.ignmin, wp.ignmax)) ww[k] = coef;
                     else ww[k] = 0;
                     k++;
                 }
              }

              for(nt=wp.minres*2; nt<=n; nt*=2)  wavelet_transform(wave, ww, nt, 0);

              k = 0;
              if(h<ysize)
                  for(i=0; i<xsize; i++) z[ino].wavelet[h][i] = ww[k++];                  
              else
              {   for(j=0; j<ysize; j++)
                  {   
                      if(wp.want_gray_offset)
                          value = min((int)g.maxvalue[bpp], 
                                  max(0, wp.gray_offset + cint(ww[k]*frac)));
                      else
                          value = min((int)g.maxvalue[bpp], max(0,cint(ww[k])));
                      putpixelbytes(z[ino].image[f][j]+bytesperpixel*(h-ysize), 
                          value, 1, bpp, 1); 
                      z[ino].fmax = max(z[ino].fmax, ww[k]);
                      z[ino].fmin = min(z[ino].fmin, ww[k]);
                      k++;
                  }
              }

          }
   }
   if(wp.direction==1) z[ino].waveletstate++; else  z[ino].waveletstate--;
   if(z[ino].waveletstate) z[ino].fftdisplay = WAVE; else z[ino].fftdisplay = NONE;
   delete[] ww;
}  


//--------------------------------------------------------------------------//
// wavelet_transform - called by wavelet_pyramidal_calc                     //
//--------------------------------------------------------------------------//
void wavelet_transform(Wavelet &wave, double *aa, int n, int direction)
{
   double ai, ai1, *ww;
   int i,ii,jf,jr,k,n1,ni,nj,n2,nmod;
   ww = (double*)malloc(n*sizeof(double));
   if(ww==NULL) printf("Oh nooooooooo it was null!!!!!!!!\nAaaaaaaah!!!!!\nI'm Melllltinnnnnng!");
   for(k=0;k<n;k++) ww[k]=0.0;
   nmod = wave.size * n;
   n1 = n-1;
   n2 = n/2;
   if(direction == 1)   // forward
   {   for(ii=0, i=0; i<n; i+=2, ii++)
       {   ni = i + nmod + wave.ioffset;
           nj = i + nmod + wave.joffset;
           for(k=0; k<wave.size; k++)
           {   jf = n1 & (1 + ni + k);
               jr = n1 & (1 + nj + k);
               ww[ii] += wave.filter[0][k] * aa[jf];
               ww[ii+n2] += wave.hifilter[0][k] * aa[jr];
           }
       }
   }else               // reverse
   {   for(ii=0, i=0; i<n; i+=2, ii++)
       {   ai = aa[ii];
           ai1 = aa[ii+n2];
           ni = i + nmod + wave.ioffset;
           nj = i + nmod + wave.joffset;
           for(k=0; k<wave.size; k++)
           {   jf = n1 & (1 + ni + k);
               jr = n1 & (1 + nj + k);
               ww[jf] += wave.filter[1][k] * ai;
               ww[jr] += wave.hifilter[1][k] * ai1;
           }
      }
   }
   for(k=0;k<n;k++) aa[k] = ww[k];
   free(ww);
   ww = NULL;
}



//--------------------------------------------------------------------------//
//  strip_extraneous_dashes                                                 //
//  get rid of any '-' not preceded by a whitespace.                        //
//--------------------------------------------------------------------------//
void strip_extraneous_dashes(char *s)
{
   int k, number=0;
   for(k=0;k<(int)strlen(s);k++)
   {   if(number && s[k]=='-') s[k]=' ';
       if(isdigit(s[k])) number=1;
       if(isspace(s[k])) number=0;
   }   
}


//--------------------------------------------------------------------//
// parse_string                                                       //
// input: a single command with arguments  e.g. "load test 1 1 2 3".  //
// Each word k is placed in buf[k][80]. buf[0][] is the command, the  //
// rest are arguments separated by spaces.                            //
// The no. of arguments, not counting buf[0], is placed in `noofargs'.//
// `startarg' is optional and is 0 by default.                        //
//--------------------------------------------------------------------//
void parse_string(char *cmd, char **buf, int &noofargs, int startarg, int maxlen)
{   
    int count=0;
    char word[128];
    noofargs = 0;
    do
    {    word[0]=0;
         parse_word(cmd, word, 80, count);
         if(strlen(word)==0) break;
         strncpy(buf[startarg + noofargs++], word, maxlen);
         if(count >= (int)strlen(cmd)) break;
    }while(1);
}


//--------------------------------------------------------------------//
// parse_word - returns first word demarcated by a space in `buffer'  //
// 'count' must be initialized to 0 by calling routine before first   //
// call (keeps track of buffer position).                             //
//--------------------------------------------------------------------//
void parse_word(char *buffer, char *word, int maxlen, int &count)
{
#ifdef DIGITAL
    char tempstring[1024];  
#else
    char tempstring[maxlen+1];  
#endif
    int k=count, pos=0;
    while(buffer[k])
    {   
         if(buffer[k]!=' ' && buffer[k]!='\n' ) tempstring[pos++]=buffer[k];
         else if(pos) break;
         tempstring[pos]=0;
         k++;
         if(buffer[k]=='\n') break;
    }
    count = k;    
    strcpy(word, tempstring);
}    



//--------------------------------------------------------------------------//
//  smallest                                                                //
//--------------------------------------------------------------------------//
double smallest(int ino)
{
   int i, j;
   double d=0 ,v;
   if(!z[ino].waveletexists) return 0.0;
   for(j=0; j<z[ino].ysize; j++)
   for(i=0; i<z[ino].xsize; i++)
   {   v = z[ino].wavelet[j][i]; 
       if(v < d) d = v;
   } 
   return d;
}

//--------------------------------------------------------------------------//
//  largest                                                                 //
//--------------------------------------------------------------------------//
double largest(int ino)
{
   int i, j;
   double d=0 ,v;
   if(!z[ino].waveletexists) return 0.0;
   for(j=0; j<z[ino].ysize; j++)
   for(i=0; i<z[ino].xsize; i++)
   {   v = z[ino].wavelet[j][i]; 
       if(v > d) d = v;
   } 
   return d;
}


//--------------------------------------------------------------------------//
// setwaveletpixel                                                          //
//--------------------------------------------------------------------------//
void setwaveletpixel(int x, int y, uint color, int ino)
{
  double value=0;
  x -= z[ino].xpos;
  y -= z[ino].ypos;
  if(ino>=0 && z[ino].waveletexists && x>=0 && y>=0 &&
        x<z[ino].xsize && y<z[ino].ysize && 
        z[ino].fftdisplay==WAVE)
  {   if(x<z[ino].wavelet_xminres && y<z[ino].wavelet_yminres && z[ino].wfac!=0.0)
           value = (color - z[ino].gray[3]) / z[ino].wfac + z[ino].gray[5];
      else if(z[ino].ffac!=0.0)
           value = (color - z[ino].gray[3]) / z[ino].ffac + z[ino].gray[1];
      z[ino].wavelet[y][x] = value;
  }
}


//--------------------------------------------------------------------------//
//  show_wavelet_grid                                                       //
//--------------------------------------------------------------------------//
void show_wavelet_grid(int ino, int transform_type)
{
   transform_type = transform_type;
   int i, j;
   int xs = z[ino].xpos;
   int ys = z[ino].ypos;
   int xe = xs + z[ino].xsize - 1;
   int ye = ys + z[ino].ysize - 1;
   g.inmenu++;
   for(j=z[ino].ysize-1; j>=z[ino].wavelet_yminres/2; j/=2)
   for(i=z[ino].xsize-1; i>=z[ino].wavelet_xminres/2; i/=2)
   {    line(i+xs, j+ys, i+xs, ys, 255, SET);
        line(i+xs, j+ys, xs, j+ys, 255, SET);
   }
   line(xs, ys, xs, ye, 255, SET);
   line(xs, ys, xe, ys, 255, SET);
   g.inmenu--;
}


//--------------------------------------------------------------------------//
// write_image_wavelet                                                      //
//--------------------------------------------------------------------------//
int write_image_wavelet(Wavelet &wave, Wavelet_params &wp)
{
  FILE *fp;
  if(memorylessthan(16384)){ message(g.nomemory,ERROR); return(NOMEM); }
  int i,j;
  static char waveletfile[FILENAMELENGTH]="wavelet.ascii";
  
  int status=OK;
  int size;
  int ino = wp.ino;

  if(!z[ino].waveletexists)
  {   message("No wavelet exists for this image",ERROR);
      return(BADPARAMETERS);
  }
  message("Filename:",waveletfile,PROMPT,FILENAMELENGTH-1);
  if(cant_overwrite_file(waveletfile)) return(ABORT);
  if ((fp=fopen(waveletfile, "wt")) == NULL)
  {   error_message(waveletfile, errno);
      return(CANTCREATEFILE);
  }

  fprintf(fp,"Wavelet transform of %s\n",basefilename(z[ino].name));  
  fprintf(fp,"xsize %d\n", z[ino].xsize);  
  fprintf(fp,"ysize %d\n", z[ino].ysize);  
  fprintf(fp,"bits/pixel %d\n", z[ino].bpp);  
  fprintf(fp,"format %s\n", wp.format);  
  fprintf(fp,"levels %d\n", wp.levels);  
  fprintf(fp,"xminres %d\n", z[ino].wavelet_xminres);  
  fprintf(fp,"xminres %d\n", z[ino].wavelet_yminres);  
  fprintf(fp,"wavelet %s\n", wave.filename);  
  for(j=0;j<z[ino].ysize;j++)
  {   for(i=0;i<z[ino].xsize;i++)
         fprintf(fp,"%g ",z[ino].wavelet[j][i]);
      fprintf(fp,"\n");  
  }

  fclose(fp);
  size = filesize(waveletfile);               // Check to ensure not an empty file
  if(size==0) status=ZEROLENGTH; 
  if(access(waveletfile, F_OK)) status=ERROR; // access = 0 if file exists
  if(status==OK) message("Wavelet coordinates saved");
  else  message("Ack!!! Error saving wavelet!!!",ERROR);
  return(status);
}


//--------------------------------------------------------------------------//
// read_image_wavelet                                                       //
//--------------------------------------------------------------------------//
int read_image_wavelet(int noofargs, char **arg, Wavelet &wave, Wavelet_params &wp)
{
   static char *filename=NULL;
   double ww;
   char temp1[256];
   char temp2[256];
   char temp3[256];
   int bpp=8, xminres, yminres;
   FILE *fp;
   ////  Leave this allocated
   if(filename==NULL){ filename = new char[FILENAMELENGTH]; filename[0]=0; }
   char oldfilename[FILENAMELENGTH]="none";
   if(strlen(filename)) strcpy(oldfilename, filename);
   int i, j, status=OK, xsize, ysize;
   
   if(noofargs == 0) strcpy(filename, getfilename(oldfilename));
   else  strcpy(filename, arg[1]);
      
   if((fp=fopen(filename, "rt")) == NULL)
   {   error_message(filename, errno);
       return(ERROR);
   }
   
   fscanf(fp,"%7s %9s %2s",temp1, temp2, temp3);// 'wavelet transform of'
   if(strcmp(temp1,"Wavelet") != SAME){ message("Invalid wavelet transform file",ERROR); return ERROR; }
   fscanf(fp,"%256s",temp3);                    // [filename]
   fscanf(fp,"%256s %d",temp1, &xsize);         // xsize 12345
   fscanf(fp,"%256s %d",temp1, &ysize);         // ysize 12345   
   fscanf(fp,"%256s %d",temp1, &bpp);           // bits/pixel   
   fscanf(fp,"%256s %s",temp1, wp.format);      // format name
   fscanf(fp,"%256s %d",temp1, &wp.levels);     // no. of levels
   fscanf(fp,"%256s %d",temp1, &xminres);       // x minimum resolution
   fscanf(fp,"%256s %d",temp1, &yminres);       // y minimum resolution
   fscanf(fp,"%256s %s",temp1, wave.filename);  // wavelet [waveletfilename]   
   
   if(newimage(0,0,xsize,ysize,bpp,GRAY,1,g.want_shell,PERM,1,
      g.window_border,0)!=OK) return(NOMEM); 
   allocate_image_wavelet(ci, xsize, ysize);
   z[ci].waveletstate = 1;
   z[ci].wavelet_xminres = xminres;
   z[ci].wavelet_yminres = yminres;
   z[ci].fmin = (int)g.maxvalue[bpp];
   z[ci].fmax = 0;
   z[ci].wmin = (int)g.maxvalue[bpp];
   z[ci].wmax = 0;
   z[ci].fftdisplay = WAVE;

   setimagetitle(ci, temp3);
   for(j=0;j<ysize;j++)
   for(i=0;i<xsize;i++)
   {    if(!fscanf(fp, "%lg", &ww)){ status=ERROR; break; }
        z[ci].wavelet[j][i] = ww; 
        if(j>ysize/2 && i>xsize/2)
        {   z[ci].fmax = max(z[ci].fmax, ww);
            z[ci].fmin = min(z[ci].fmin, ww);
        }else
        if(j<z[ci].wavelet_yminres && i<z[ci].wavelet_xminres)
        {   z[ci].wmax = max(z[ci].wmax, ww);
            z[ci].wmin = min(z[ci].wmin, ww);
        }
   }
   if(status==OK)
   {  switchto(ci);
      repair(ci);
   }else message("Error reading wavelets",ERROR);
   fclose(fp);
   return status;
}

